/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>   	/* standard stuff */
#include <string.h>   	/* multiple string routines */
#include <sys/types.h>	/* for opendir() ao */
#include <sys/stat.h> 	/* for open() */
#include <dirent.h>   	/* for opendir() ao */
#include <errno.h>			/* error handling functions */
#include <fcntl.h>    	/* open() ao */
#include <unistd.h>			/* lseek() */
#include <stdio.h>			/* sprintf */

#include "structs.h"
#include "world.h"
#include "config.h"
#include "handy.h"
#include "global_vars.h"
#include "syntaxhl.h"
#include "dialog.h"

char *get_home_dir();
/* check for existance, if not create */
void check_dir(char *dir);
void create_dir(char *dir);
void check_file(char *file);
void create_file(char *file);

/* void check_gtkrc(); */
void set_color(GdkColor *c, char *cstring, int r, int g, int b);
void parse_main_rc();

char *get_value(char *file, char *option);
int   get_line(int fd); /* line is put in global `curvalue' (char *) */

char *strip_space(char *s);

void put_value(char *file, char *option, char *value);
void put_beginning(char *file);

void put_color(char *file, char *name, int r, int g, int b);

void put_externals(char *worldrc);
void get_externals(char *file);

int secure_create;
char *curvalue;

void put_keypad_keys(char *worldrc);
void put_default_keypad_keys(char *worldrc);
void get_keypad_keys(char *worldrc);

/* add more options when they will be avalible */
void  gM_config_init(int argc, char *argv[]) {
  int i;
  char *s;

  config_dir = NULL;
  term_color = 1;
  secure_create = 0; /* used when creating the world files */
  curvalue = g_malloc(256);   /* used for getting options */
  keypad_keys = g_malloc(16 * sizeof(char *));
  
  for(i=1;i<argc;i++) {
    if(argv[i][0] == '-') {
      if(strcasecmp(argv[i], "-h") == 0 || strcasecmp(argv[i], "--help") == 0) {
        printf(_("gMOO, a gtk+ based graphical MOO client\n"
          		 "\n"
               "Usage: gMOO [Option...]\n"
               "\n"
               "Where Options include:\n"
               "  -h, --help                  Show this help message (DUH :))\n"
               "  -v, --version               Show the version of gMOO\n"
               "  -d <dir>, --directory <dir> Set the config dir, defaults to ~/.gMOO\n"
               "\n"
               "And of cource all the GTK+ options (not listed here)\n"
               "\n"));
        exit(0);
      } else if(strcasecmp(argv[i], "-v") == 0 || strcasecmp(argv[i], "--version") == 0) {
        printf(_("gMOO version %s\n"), VERSION);
        exit(0);
      } else if(strcasecmp(argv[i], "-d") == 0 || strcasecmp(argv[i], "--directory") == 0) {
        if(i > argc - 2) {
          printf("gMOO: Error: no value for the %s option\n", argv[i]);
        } else {
          i++;
          config_dir = argv[i];
        }
/*    } else if(blabla ) { */
      } else
        printf("gMOO: Error: Unknown option: %s\n",argv[i]);
    } else
    printf("gMOO: Error: Unknown argument: %s\n",argv[i]);
  }

  if(config_dir == NULL)
    config_dir = strcatnew(get_home_dir(), "/.gMOO");
  world_dir = strcatnew(config_dir, "/Worlds");
  lock_file_name = strcatnew(config_dir, "/lock");

  check_dir(config_dir);
  check_dir(world_dir);
  parse_main_rc();

  s = strcatnew(config_dir, "/Logs");
  check_dir(s);
  g_free(s);
#if 0 /* macros */
  s = strcatnew(config_dir, "/Macros");
  check_dir(s);
  g_free(s);
#endif
  s = strcatnew(config_dir, "/Pixmaps");
  check_dir(s);
  g_free(s);

#ifdef TCL
  s = strcatnew(config_dir, "/Scripts");
  check_dir(s);
  g_free(s);

  s = strcatnew(config_dir, "/Scripts/Client");
  check_dir(s);
  g_free(s);

  s = strcatnew(config_dir, "/Scripts/World");
  check_dir(s);
  g_free(s);
#endif
}

void gM_config_create_new_world(gM_newworld *nw) {
  char *s;
  char p[15];
  
  s = g_malloc(strlen(world_dir) + strlen(nw->name) + 2);
  strcpy(s, world_dir);
  strcat(s, "/");
  strcat(s, nw->name);

  if(exist_file(s) > 0) {
    gM_dialog_show(_("Error:"), _("That world already exist,  \nThe file has NOT been written.  "
                             "\n\nIf you want to edit that world,  "
                             "\nopen it and change the properties  "), 1);
    g_free(s);
    return;
  }
  secure_create = 1;
  create_file(s);
  secure_create = 0;
  
  put_beginning(s);

  put_value(s, HOSTNAME,    nw->host);
  sprintf(p, "%d",          nw->port);
  put_value(s, PORT,        p);
  put_value(s, USERNAME,    nw->user);
  put_value(s, PASSWORD,    nw->pass);
  put_value(s, AUTOCONNECT, nw->autoconnect > 0 ? "1" : "0");
  
  g_free(s);
}

void gM_config_recreate_world(gM_newworld *nw) {
  char *s;
  char p[15];
  int i;
  
  s = g_malloc(strlen(world_dir) + strlen(nw->name) + 2);
  strcpy(s, world_dir);
  strcat(s, "/");
  strcat(s, nw->name);

  i = remove(s);
  if(i < 0) {
    printf(_("gMOO: Error: Could not remove file: \"%s\"\n"), strerror(errno));
    exit(errno);
  }
  secure_create = 1;
  create_file(s);
  secure_create = 0;
  
  put_beginning(s);

  put_value(s, HOSTNAME,    nw->host);
  sprintf(p, "%d",          nw->port);
  put_value(s, PORT,        p);
  put_value(s, USERNAME,    nw->user);
  put_value(s, PASSWORD,    nw->pass);
  put_value(s, AUTOCONNECT, nw->autoconnect > 0 ? "1" : "0");
  
  g_free(s);
}

void gM_config_save_world(world *w) {
  char *s;
  char p[15];
  int i;
  
  s = strcat3new(world_dir, "/", w->name);

  i = remove(s);
  if(i < 0) {
    printf(_("gMOO: Error: Could not remove file: \"%s\"\n"), strerror(errno));
    exit(errno);
  }
  secure_create = 1;
  create_file(s);
  secure_create = 0;
  
  put_beginning(s);

  put_value(s, HOSTNAME,     w->host);
  sprintf(p, "%d",           w->port);
  put_value(s, PORT,         p);
  put_value(s, USERNAME,     w->user);
  put_value(s, PASSWORD,     w->password);
  put_value(s, AUTOCONNECT,  w->autoconnect > 0 ? "1" : "0");
  
  g_free(s);
}

GList *gM_config_get_world_names() {
  GList *list;
  GtkWidget *listitem;
  DIR *dir;
  struct dirent *ent;
  
  list = g_list_alloc();
  check_dir(world_dir);
  dir = opendir(world_dir);
  ent = readdir(dir);
  while(ent != NULL) {
    if(ent->d_name[0] != '.') {
/*printf("-> %s\n",ent->d_name);*/
      listitem = gtk_list_item_new_with_label(ent->d_name);
      gtk_widget_show(listitem);
      g_list_append(list, listitem);
    }
    ent = readdir(dir);
  }

  return list;
}

world *gM_config_read_world(char *worldname) {
  world *w;
  char *filename;
  char *tmp;
  
  filename = g_malloc(sizeof(char) * (strlen(world_dir) + strlen(worldname) + 2));
  strcpy(filename, world_dir);
  strcat(filename, "/");
  strcat(filename, worldname);

  w = g_malloc(sizeof(world));
  w->user           = NULL;
  w->password       = NULL;
  w->autoconnect    = 0;

  w->name =	strnew(worldname);
  w->host = get_value(filename, HOSTNAME);
  if(w->host == NULL) {
    g_free(w);
    g_free(filename);
    return NULL;
  }
  tmp = get_value(filename, PORT);
  if(tmp == NULL) {
    g_free(w->host);
    g_free(w);
    g_free(filename);
    return NULL;
  } else {
    w->port = atoi(tmp);
    g_free(tmp);
  }
  w->user = get_value(filename, USERNAME);
  if(w->user != NULL && strcmp(w->user, EMPTY_STR) == 0) {
    g_free(w->user);
    w->user = NULL;
  }
  w->password = get_value(filename, PASSWORD);
  if(w->password != NULL && strcmp(w->password, EMPTY_STR) == 0) {
    g_free(w->password);
    w->password = NULL;
  }
  tmp = get_value(filename, AUTOCONNECT);
  if(tmp == NULL) { /* not fount, an old config file ... */
    w->autoconnect = w->user == NULL ? 0 : 1;
  } else {
    w->autoconnect = atoi(tmp);
    g_free(tmp);
  }
  g_free(filename);
  return w;
}


int gM_config_remove_world_file(char *file) {
  char *s;
  int i;
  
  s = g_malloc(strlen(world_dir) + strlen(file) + 2);
  strcpy(s, world_dir);
  strcat(s, "/");
  strcat(s, file);

  i = remove(s);
  g_free(s);
  if(i < 0) {
    printf(_("gMOO: Error: Could not remove file: \"%s\"\n"), strerror(errno));
    return -1;
  }
  return 1;
}

void gM_config_rewrite_main_rc() {
  char data[] = 
    "#\n"
    "# gMOO Main configuration file, please do not edit by hand\n"
    "#\n"
    "#\n";
  char *worldrc = strcatnew(config_dir, "/gMOOrc");
  int i, j;
  char tmp[3];	

  i = open(worldrc, O_RDWR | O_CREAT | O_TRUNC, 00644);
  if(i > -1) {

	j = write(i, data, strlen(data));
    if(j > -1) {
      close(i);
      put_value(worldrc, NORMALFONT, normalfontstr);
      put_value(worldrc, BOLDFONT,   boldfontstr);
      put_color(worldrc, CBLACK	 , colors[BLACK].red   / 0xFF, 
                colors[BLACK].green   / 0xFF, colors[BLACK].blue / 0xFF);
      put_color(worldrc, CWHITE  , colors[WHITE].red   / 0xFF,
                colors[WHITE].green   / 0xFF, colors[WHITE].blue / 0xFF);
      put_color(worldrc, CRED    , colors[RED].red     / 0xFF, colors[RED].green     / 0xFF, colors[RED].blue / 0xFF);
      put_color(worldrc, CGREEN  , colors[GREEN].red   / 0xFF, colors[GREEN].green   / 0xFF,
                colors[GREEN].blue / 0xFF);
      put_color(worldrc, CBLUE   , colors[BLUE].red    / 0xFF, colors[BLUE].green    / 0xFF, colors[BLUE].blue / 0xFF);
      put_color(worldrc, CCYAN   , colors[CYAN].red    / 0xFF, colors[CYAN].green    / 0xFF, colors[CYAN].blue / 0xFF);
      put_color(worldrc, CMAGENTA, colors[MAGENTA].red / 0xFF, colors[MAGENTA].green / 0xFF,
                colors[MAGENTA].blue / 0xFF);
      put_color(worldrc, CYELLOW , colors[YELLOW].red  / 0xFF, colors[YELLOW].green  / 0xFF,
                colors[YELLOW].blue / 0xFF);
      put_color(worldrc, FGCOLOR , colors[DEFAULTFG].red  / 0xFF, colors[DEFAULTFG].green  / 0xFF,
                colors[DEFAULTFG].blue / 0xFF);
      put_color(worldrc, BGCOLOR , colors[DEFAULTBG].red  / 0xFF, colors[DEFAULTBG].green  / 0xFF,
                colors[DEFAULTBG].blue / 0xFF);

      put_color(worldrc, CSTRING, editcolors[HL_STRING - HL_STRING].red   / 0xFF, 
                                  editcolors[HL_STRING - HL_STRING].green / 0xFF,
                                  editcolors[HL_STRING - HL_STRING].blue  / 0xFF);
      put_color(worldrc, CNUMBER, editcolors[HL_NUMBER - HL_STRING].red   / 0xFF, 
                                  editcolors[HL_NUMBER - HL_STRING].green / 0xFF,
                                  editcolors[HL_NUMBER - HL_STRING].blue  / 0xFF);
      put_color(worldrc, COBJNUM, editcolors[HL_OBJNUM - HL_STRING].red   / 0xFF, 
                                  editcolors[HL_OBJNUM - HL_STRING].green / 0xFF,
                                  editcolors[HL_OBJNUM - HL_STRING].blue  / 0xFF);
      put_color(worldrc, CDOLLAR, editcolors[HL_DOLLAR - HL_STRING].red   / 0xFF, 
                                  editcolors[HL_DOLLAR - HL_STRING].green / 0xFF,
                                  editcolors[HL_DOLLAR - HL_STRING].blue  / 0xFF);
      put_color(worldrc, CVARIAB, editcolors[HL_VARIAB - HL_STRING].red   / 0xFF, 
                                  editcolors[HL_VARIAB - HL_STRING].green / 0xFF,
                                  editcolors[HL_VARIAB - HL_STRING].blue  / 0xFF);
      put_color(worldrc, CKEYWRD, editcolors[HL_KEYWRD - HL_STRING].red   / 0xFF, 
                                  editcolors[HL_KEYWRD - HL_STRING].green / 0xFF,
                                  editcolors[HL_KEYWRD - HL_STRING].blue  / 0xFF);
      put_color(worldrc, CSYMBOL, editcolors[HL_SYMBOL - HL_STRING].red   / 0xFF, 
                                  editcolors[HL_SYMBOL - HL_STRING].green / 0xFF,
                                  editcolors[HL_SYMBOL - HL_STRING].blue  / 0xFF);
      put_color(worldrc, CCOMMEN, editcolors[HL_COMMEN - HL_STRING].red   / 0xFF, 
                                  editcolors[HL_COMMEN - HL_STRING].green / 0xFF,
                                  editcolors[HL_COMMEN - HL_STRING].blue  / 0xFF);

      sprintf(tmp, "%d", logtype);
      put_value(worldrc, LOGTYPE, tmp);
      put_value(worldrc, HISTORYTYPE, use_hist < 1 ? "0" : (use_hist > 1 ? "2" : "1"));
      put_value(worldrc, USEAUTOIDENT, use_ident < 1 ? "0" : "1");
      put_value(worldrc, USEAUTODETECT, use_auto_detect < 1 ? "0" : "1");
      put_value(worldrc, CLOCKTYPE, clocktype < 1 ? "0" : (clocktype > 1 ? "2" : "1"));
      put_value(worldrc, SHOW_TOOLBAR, show_toolbar < 1 ? "0" : "1");
      put_value(worldrc, RECONNECT, reconnect < 1 ? "0" : "1");
      put_value(worldrc, POPUPCON, popupcon < 1 ? "0" : "1");
      put_value(worldrc, SAVESEND, savesend < 1 ? "0" : "1");
      sprintf(tmp, "%d", show_own);
      put_value(worldrc, SHOW_OWN, tmp);
      sprintf(tmp, "%d", beep_type);
      put_value(worldrc, BEEP_TYPE, tmp);
      put_externals(worldrc);
      put_value(worldrc, SAVE_WINDOW_POS, save_window_pos != NULL ?
                                          save_window_pos :
                                          "NotSaved");
      put_value(worldrc, SAVE_WINDOW_SIZE, save_window_size != NULL ?
                                          save_window_size :
                                          "640,480");
      sprintf(tmp, "%d", userlistsorttype);
      put_value(worldrc, USERLISTSORTTYPE, tmp);
      sprintf(tmp, "%d", homo_tabs);
      put_value(worldrc, HOMO_TABS, tmp);
      sprintf(tmp, "%d", short_tab_names);
      put_value(worldrc, SHORTTABNAMES, tmp);
#if 0 /* That damned GtkText widget is acting up again */
      sprintf(tmp, "%d", buffer_len);
      put_value(worldrc, BUFFER_LEN, tmp);
#endif
      sprintf(tmp, "%d", use_keypad_keys);
      put_value(worldrc, USE_KEYPAD_KEYS, tmp);
      put_keypad_keys(worldrc);
      sprintf(tmp, "%d", wordwrap);
      put_value(worldrc, WORDWRAP, tmp);
      sprintf(tmp, "%d", definputstyle);
      put_value(worldrc, DEFINPUTSTYLE, tmp);
      sprintf(tmp, "%d", showsep);
      put_value(worldrc, SHOWSEP, tmp);
#ifdef TCL    
      sprintf(tmp, "%d", tcl_debug);
      put_value(worldrc, TCL_DEBUG, tmp);
      sprintf(tmp, "%d", tcl_debug_on_stdout);
      put_value(worldrc, TCL_DEBUG_ON_STDOUT, tmp);
#endif /* TCL */
      sprintf(tmp, "%d", newmultiline);
      put_value(worldrc, MULTILINE, tmp);
      sprintf(tmp, "%d", multilinesize);
      put_value(worldrc, MULTILINESIZE, tmp);
      g_free(worldrc);
      return; /* success */
    } 
    printf(_("gMOO: Error: Writing in \"%s\" failed -> %s\n"), worldrc, strerror(errno));
    exit(errno);
  }
  printf(_("gMOO: Error: Creating \"%s\" failed -> %s\n"), worldrc, strerror(errno));

 exit(errno);
}


/* lock file stuff */
void gM_config_lock_open() {
  int i;

  if(exist_file(lock_file_name) > 0)
    return;
 
  i = open(lock_file_name, O_RDWR | O_CREAT, 00600);
  if(i > -1) {
    close(i);
    return; /* success */
  }
  printf(_("gMOO: Error: Creating \"%s\" failed -> %s\n"), lock_file_name, strerror(errno));
  exit(errno);
}

void gM_config_lock_close(){
  int fd;
  int i,j;
  char s[100];
  
  fd = open(lock_file_name, O_RDONLY);
  if(fd < 0) /* file doesn't exist, this is tecnally an error ... */
    return;
  i = read(fd, s, 99);
  if(i != 0 || i == 99)
    return;
  for(j = 0; j < i; j++) /* check of there are only empty lines */
    if(s[j] != 10)
      return;
  i = remove(lock_file_name);
  if(i == -1)
    printf(_("gMOO: Error: Removeing lockfile \"%s\" failed: %s\n"), lock_file_name, strerror(errno));
}

int gM_config_lock_check(char *worldname){
  int ret = 0;
  int fd;
  char s[2];
  int i;
  int j = 0;;

  worldname = strcatnew(worldname, "\n");
  fd = open(lock_file_name, O_RDONLY);
  if(fd < 0) {
    printf(_("gMOO: Error: Opening lockfile \"%s\" failed: %s\n"), lock_file_name, strerror(errno));
    return -1;
  }
  lseek(fd, 0, SEEK_SET);
  i = read(fd, s, 1);
  while(i == 1) {
    ret++;
    if(s[0] == worldname[j])
      j++;
    else
      j = 0;
    if(s[0] == 10 && worldname[j] == 0) {
      ret -= j;
      break;
    }
    i = read(fd, s, 1);
  }
  if(j == 0)
    ret = -1;

  g_free(worldname);
  return ret;
}

void gM_config_lock_add(char *worldname) {
  int fd;
  
  fd = open(lock_file_name, O_WRONLY | O_APPEND);
  if(fd < 0) {
    printf(_("gMOO: Error: Opening lockfile \"%s\" failed: %s\n"), lock_file_name, strerror(errno));
    return;
  }
  write(fd, worldname, strlen(worldname));
  write(fd, "\n", 1);
}

void gM_config_lock_remove(char *worldname) {
  int fd;
  char *file, *file2;
  char dummy[2];
  int len = 0;
  
  fd = open(lock_file_name, O_RDONLY);
  if(fd < 0) {
    printf(_("gMOO: Error: Opening lockfile \"%s\" failed: %s\n"), lock_file_name, strerror(errno));
    return;
  }
  lseek(fd, 0, SEEK_SET);

  len = gM_config_lock_check(worldname);
  if(len > -1) {
    file = g_malloc(len + 1);
    read(fd, file, len);
    file[len] = 0;
    lseek(fd, len + strlen(worldname) + 1, SEEK_SET);
  } else {
    close(fd);
    return;
  }
  dummy[0] = dummy[1] = 0;
  while(read(fd, dummy, 1) == 1) {
    len++;
    file2 = g_malloc(len + 1);
    strcpy(file2, file);
    strcat(file2, dummy);
    g_free(file);
    file = file2;
  }
  close(fd);
  fd = open(lock_file_name, O_WRONLY | O_TRUNC);
  write(fd, file, len);
  close(fd);
}


/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

char *get_home_dir() {
  char *s;
  
  s = getenv("HOME");
  if(s != NULL)
    return s;

 return "."; /* last resort -- maby another option ... */

}

void check_dir(char *dir) {
  DIR *dir_;
  
  dir_ = opendir(dir);
  if(dir_ != NULL) {

  closedir(dir_);
    return;
  } /* else: dir error */
  if(errno == ENOENT) {
    printf(_("gMOO: %s doesn't exist... creating\n"),dir);
    create_dir(dir);
  } else {
    printf(_("gMOO: Error: Opening dir \"%s\" -> %s\n"),dir,strerror(errno));
    exit(errno);
 }
}

void create_dir(char *dir) {
  if(mkdir(dir, 00755) > -1) {
/*    succes_arg("Creating %s", dir);*/
    return; /* success */
  }
  printf(_("gMOO: Error: Creating \"%s\" failed -> %s\n"), dir, strerror(errno));
  exit(errno);
}


void check_file(char *file) {
  if(exist_file(file) < 1) {
    printf(_("gMOO: %s doesn't exist... creating\n"),file);
    create_file(file);
  }
}


void create_file(char *file) {
  int i;
 
  i = open(file, O_RDWR | O_CREAT, (secure_create ? 00600 : 00644));
  if(i > -1) {
/*  succes_arg("Creating %s", file);*/
    put_beginning(file);
    close(i);
    return; /* success */
  }
  printf(_("gMOO: Error: Creating \"%s\" failed -> %s\n"), file, strerror(errno));
  exit(errno);
}

int exist_file(char *file) {
  int i;

  i = open(file, O_RDONLY);
  if(i > -1) {
    close(i);
    return 1;
  }
  if(errno == 2) {
    return -1;
  } else {
    printf(_("gMOO: Error: Opening file \"%s\" -> %s\n"),file,strerror(errno));
    exit(errno);
  }
}


void check_main_rc(char *worldrc) {
  char data[] = 
"#\n"
"# gMOO Main configuration file, please do not edit by hand\n"
"#\n"
"#\n";
  int i, j;

  i = open(worldrc, O_RDONLY);
  if(i > -1) {
    close(i);
    return;
  }
  if(errno == 2) {
    printf(_("gMOO: %s doesn't exist... creating defaults\n"),worldrc);
    i = open(worldrc, O_RDWR | O_CREAT, 00644);
    if(i > -1) {
  		j = write(i, data, strlen(data));
      if(j > -1) {
        close(i);
        put_value(worldrc, NORMALFONT, "-misc-fixed-medium-r-normal--13-*-*-*-c-*-iso8859-1");
        put_value(worldrc, BOLDFONT, "-misc-fixed-bold-r-normal--13-*-*-*-c-*-iso8859-1");
        put_color(worldrc, CBLACK	 , 0,    0,    0);
        put_color(worldrc, CWHITE  , 0x100, 0x100, 0x100);
        put_color(worldrc, CRED    , 0x100, 0,    0);
        put_color(worldrc, CGREEN  , 0,    0x100, 0);
        put_color(worldrc, CBLUE   , 0,    0,    0x100);
        put_color(worldrc, CCYAN   , 0,    0x100, 0x100);
        put_color(worldrc, CMAGENTA, 0x100, 0,    0x100);
        put_color(worldrc, CYELLOW , 0x100, 0x100, 0);
        put_color(worldrc, FGCOLOR , 0x100, 0x100, 0x100);
        put_color(worldrc, BGCOLOR , 0, 0, 0);

        put_color(worldrc, CSTRING, 256, 000,   0);
        put_color(worldrc, CNUMBER, 000, 256,   0);
        put_color(worldrc, COBJNUM, 256,  48, 254);
        put_color(worldrc, CDOLLAR, 256, 256,   0);
        put_color(worldrc, CVARIAB, 256, 257, 257);
        put_color(worldrc, CKEYWRD,   0, 256, 256);
        put_color(worldrc, CSYMBOL,  53, 137, 256);
        put_color(worldrc, CCOMMEN, 190, 109,  37);

        put_value(worldrc, HISTORYTYPE,  "2");
        put_value(worldrc, USEAUTOIDENT, "1");
        put_value(worldrc, USEAUTODETECT, "1");
        put_value(worldrc, LOGTYPE, "12"); /* (logtime | logsaveinput) */
        put_value(worldrc, CLOCKTYPE, "1"); /* clocksecond */
        put_value(worldrc, SHOW_TOOLBAR, "1");
        put_value(worldrc, EXT_MAIL, "rxvt\x1-e\x1\"pine\"\x1&");
        put_value(worldrc, EXT_NEWS, "(null)");
        put_value(worldrc, RECONNECT, "1");
        put_value(worldrc, POPUPCON, "1");
        put_value(worldrc, SAVESEND, "1");
        put_value(worldrc, SHOW_OWN, "-1");
#if 0 /* That damned GtkText widget is acting up again */
        put_value(worldrc, BUFFER_LEN, "500");
#endif /* 0 */
        put_value(worldrc, BEEP_TYPE, "4"); /* beep_type_speaker */
        put_value(worldrc, SAVE_WINDOW_POS, "NotSaved");
        put_value(worldrc, SAVE_WINDOW_SIZE, "640,480");
        put_value(worldrc, USERLISTSORTTYPE, "0");
        put_value(worldrc, SHORTTABNAMES, "0");
        put_value(worldrc, HOMO_TABS, "0");
#ifdef TCL
        put_value(worldrc, TCL_DEBUG, "0");
        put_value(worldrc, TCL_DEBUG_ON_STDOUT, "1");
#endif /* TCL */
        put_value(worldrc, MULTILINE, "0");
        put_value(worldrc, MULTILINESIZE, "50");
        put_default_keypad_keys(worldrc);
        put_value(worldrc, USE_KEYPAD_KEYS, "1");
        put_value(worldrc, WORDWRAP, "0");
        put_value(worldrc, DEFINPUTSTYLE, "0");
        put_value(worldrc, SHOWSEP, "1");
        return; /* success */
      } 
      printf(_("gMOO: Error: Writing in \"%s\" failed -> %s\n"), worldrc, strerror(errno));
      exit(errno);
    }
    printf(_("gMOO: Error: Creating \"%s\" failed -> %s\n"), worldrc, strerror(errno));
    exit(errno);
  } else {
    printf(_("gMOO: Error: Opening file \"%s\" -> %s\n"), worldrc, strerror(errno));
    exit(errno);
  }
}

void set_color(GdkColor *c, char *cstring, int r, int g, int b) {
  char s[4];

  c->pixel = 0;
  if(cstring == NULL || strlen(cstring) != 9) { /* erro in loading from the config file */
    c->red =   r * 0xFF;
    c->green = g * 0xFF;
    c->blue =  b * 0xFF;
    return;
  }
  s[0] = cstring[0];
  s[1] = cstring[1];
  s[2] = cstring[2];
  s[3] = 0;
  c->red = atoi(s) * 0xFF;

  s[0] = cstring[3];
  s[1] = cstring[4];
  s[2] = cstring[5];
  c->green = atoi(s) * 0xFF;

  s[0] = cstring[6];
  s[1] = cstring[7];
  s[2] = cstring[8];
  c->blue = atoi(s) * 0xFF;
}


void parse_main_rc() {
  char *file;
  char *s;
  file = strcatnew(config_dir, "/gMOOrc");
  check_main_rc(file);
  
  normalfontstr = get_value(file, NORMALFONT);
  if(normalfontstr == NULL)
    normalfontstr = strnew("-misc-fixed-medium-r-normal--13-*-*-*-c-*-iso8859-1");
  boldfontstr   = get_value(file, BOLDFONT);
  if(boldfontstr == NULL)
    boldfontstr = strnew("-misc-fixed-bold-r-normal--13-*-*-*-c-*-iso8859-1");

  set_color(&colors[BLACK],   get_value(file, CBLACK),   0,     0,     0);
  set_color(&colors[WHITE],   get_value(file, CWHITE),   0x100, 0x100, 0x100);
  set_color(&colors[RED],     get_value(file, CRED),     0x100, 0,     0);
  set_color(&colors[GREEN],   get_value(file, CGREEN),   0,     0x100, 0);
  set_color(&colors[BLUE],    get_value(file, CBLUE),    0,     0,     0x100);
  set_color(&colors[CYAN],    get_value(file, CCYAN) ,   0,     0x100, 0x100);
  set_color(&colors[MAGENTA], get_value(file, CMAGENTA), 0x100, 0,     0x100);
  set_color(&colors[YELLOW],  get_value(file, CYELLOW),  0x100, 0x100, 0);
  set_color(&colors[DEFAULTFG],  get_value(file, FGCOLOR),  0x100, 0x100, 0x100);
  set_color(&colors[DEFAULTBG],  get_value(file, BGCOLOR),  0, 0, 0);

  set_color(&editcolors[HL_STRING - HL_STRING], get_value(file, CSTRING), 256, 000,   0);
  set_color(&editcolors[HL_NUMBER - HL_STRING], get_value(file, CNUMBER), 000, 256,   0);
  set_color(&editcolors[HL_OBJNUM - HL_STRING], get_value(file, COBJNUM), 256,  48, 254);
  set_color(&editcolors[HL_DOLLAR - HL_STRING], get_value(file, CDOLLAR), 256, 256,   0);
  set_color(&editcolors[HL_VARIAB - HL_STRING], get_value(file, CVARIAB), 256, 257, 257);
  set_color(&editcolors[HL_KEYWRD - HL_STRING], get_value(file, CKEYWRD),   0, 256, 256);
  set_color(&editcolors[HL_SYMBOL - HL_STRING], get_value(file, CSYMBOL),  53, 137, 256);
  set_color(&editcolors[HL_COMMEN - HL_STRING], get_value(file, CCOMMEN), 190, 109,  37);

  s = get_value(file, HISTORYTYPE);
  if(s == NULL)
    use_hist = 1; /* use history, but no selective history */
  else {
    use_hist = atoi(s);
    g_free(s);
  }
  s = get_value(file, USEAUTOIDENT);
  if(s == NULL)
    use_ident = 1; /* use use auto-ident */
  else {
    use_ident = atoi(s);
    g_free(s);
  }
  s = get_value(file, USEAUTODETECT);
  if(s == NULL)
    use_auto_detect = 1;
  else {
    use_auto_detect = atoi(s);
    g_free(s);
  }
  s = get_value(file, LOGTYPE);
  if(s == NULL)
    logtype = 12;
  else {
    logtype = atoi(s);
    g_free(s);
  }
  s = get_value(file, CLOCKTYPE);
  if(s == NULL)
    clocktype = 1;
  else {
    clocktype = atoi(s);
    g_free(s);
  }
  s = get_value(file, SHOW_TOOLBAR);
  if(s == NULL)
    show_toolbar = 1;
  else {
    show_toolbar = atoi(s);
    g_free(s);
  }
  s = get_value(file, RECONNECT);
  if(s == NULL)
    reconnect = 1;
  else {
    reconnect = atoi(s);
    g_free(s);
  }
  s = get_value(file, SAVESEND);
  if(s == NULL)
    savesend = 1;
  else {
    savesend = atoi(s);
    g_free(s);
  }
  s = get_value(file, POPUPCON);
  if(s == NULL)
    popupcon = 1;
  else {
    popupcon = atoi(s);
    g_free(s);
  }
  s = get_value(file, SHOW_OWN);
  if(s == NULL)
    show_own = -1;
  else {
    show_own = atoi(s);
    g_free(s);
  }
  s = get_value(file, MULTILINE);
  if(s == NULL)
    multiline = 0;
  else {
    multiline = atoi(s);
    g_free(s);
  } newmultiline = multiline;
  s = get_value(file, MULTILINESIZE);
  if(s == NULL)
    multilinesize = 50;
  else {
    multilinesize = atoi(s);
    g_free(s);
  }
  s = get_value(file, SHORTTABNAMES);
  if(s == NULL)
    short_tab_names = 0;
  else {
    short_tab_names = atoi(s);
    g_free(s);
  }
  s = get_value(file, HOMO_TABS);
  if(s == NULL)
    homo_tabs = 0;
  else {
    homo_tabs = atoi(s);
    g_free(s);
  }

  s = get_value(file, BEEP_TYPE);
  if(s == NULL) {
    beep_type = 0;
  } else {
    beep_type = atoi(s);
    g_free(s);
  }
  if(beep_type < 4) /* backwards compatibility mode */
    beep_type =
      beep_type == 0 ? beep_type_speaker :
      beep_type == 1 ? beep_type_visual  :
      beep_type_none;

  s = get_value(file, USERLISTSORTTYPE);
  if(s == NULL)
    userlistsorttype = USERLISTSORT_NAME;
  else {
    userlistsorttype = atoi(s);
    g_free(s);
  }
#if 0 /* That damned GtkText widget is acting up again */
  s = get_value(file, BUFFER_LEN);
  if(s == NULL)
    buffer_len = 500;
  else {
    buffer_len = atoi(s);
    g_free(s);
  }
#endif /* 0 */
  s = get_value(file, WORDWRAP);
  if(s == NULL)
    wordwrap = 0;
  else {
    wordwrap = atoi(s);
    g_free(s);
  }
  s = get_value(file, DEFINPUTSTYLE);
  if(s == NULL)
    definputstyle = 0;
  else {
    definputstyle = atoi(s);
    g_free(s);
  }
  s = get_value(file, SHOWSEP);
  if(s == NULL)
    showsep = 1;
  else {
    showsep = atoi(s);
    g_free(s);
  }
#ifdef TCL
  s = get_value(file, TCL_DEBUG);
  if(s == NULL)
    tcl_debug = 0;
  else {
    tcl_debug = atoi(s);
    g_free(s);
  }
  s = get_value(file, TCL_DEBUG_ON_STDOUT);
  if(s == NULL)
    tcl_debug_on_stdout = 1;
  else {
    tcl_debug_on_stdout = atoi(s);
    g_free(s);
  }
#endif /* TCL */
  get_externals(file);
  s = get_value(file, SAVE_WINDOW_POS);
  if(s == NULL || strcmp(s, "NotSaved") == 0) {
    g_free(s);
    s = NULL;
  }
  save_window_pos = s;
  s = get_value(file, SAVE_WINDOW_SIZE);
  if(s == NULL || strcmp(s, "NotSaved") == 0) {
    g_free(s);
    s = NULL;
  }
  save_window_size = s;

  s = get_value(file, USE_KEYPAD_KEYS);
  if(s == NULL) {
    use_keypad_keys = 1;
  } else {
    use_keypad_keys = atoi(s);
    g_free(s);
  }
  get_keypad_keys(file);

  g_free(file);
}


char *get_value(char *file, char *option) {
  int fd, len;
  char *p;

  fd = open(file, O_RDONLY); /* all files have been checkt, no errohandling nessescary */
  lseek(fd, 0, SEEK_SET);  /* reset file pointer */
  len = strlen(option);

  while(get_line(fd) > 0) {
  	if(strncasecmp(option, curvalue, len) == 0) {
/*printf("=> matched with option\n\n");*/
  		p = strchr(curvalue, '=');
      if(p == NULL) {
        printf(_("gMOO: \"%s\" without a value, using default\n"), option);
        close(fd);
        return NULL;
      }
      close(fd);
      p = strip_space(p + sizeof(char));
      if(p[0] == 0) { /* the options is empty */
        g_free(p);
        printf(_("gMOO: \"%s\" without a value, using default\n"), option);
        return NULL;
      }
      return p;
    }
  }
  printf(_("gMOO: No entry for \"%s\", using default\n"), option);
  close(fd);
  return NULL;
}

int get_line(int fd) {
  int i = 0, r;
  char s[2];
  
  r = read(fd, s, 1);
  if(r < 1) {
    curvalue[0] = 0;
    return -1;
  }
  while(r > 0) {
    switch(s[0]) {
    case '#': /* skip till EOL */
      if(i == 0)
        while(s[0] != '\n')
          if(read(fd, s, 1) < 0) {
            curvalue[i] = 0;
            return 1;
          }
      curvalue[i] = s[0];
      i++;
      break;
    case '\n': /* terminate line, and return */
     curvalue[i] = 0;
      return 1;
    default:
      curvalue[i] = s[0];
      i++;
    }
    r = read(fd, s, 1);
  }
  curvalue[i] = 0;
  return 1;
}



char *strip_space(char *s) {
  char *r;
  int i, j = 0;
  int len = strlen(s);
  
  r = g_malloc(len * sizeof(char));
  for(i=0; i < len + 1; i++) {
    if(s[i] != ' ') {
      r[j] = s[i];
      j++;
    }
  }
  return r;
}


void put_value(char *file, char *option, char *value) {
  int fd;
  int len;
  int i;
  char *buf;
  
  fd = open(file, O_WRONLY); /* all files have been checkt, no errohandling nessescary */
  lseek(fd, 0, SEEK_END);

  len = strlen(option) + strlen(value) + 5;
  buf = g_malloc(len + 1);
  buf[0] = 0;
  sprintf(buf, "\n%s = %s\n", option, value);

  i = write(fd, buf, len);
  if(i < 1)
    printf(_("gMOO: Error: Writing in \"%s\" failed -> %s\n"), file, strerror(errno));

  g_free(buf);
  close(fd);
}


void put_beginning(char *file) {
  int fd;
  int i;
  char s[] = "#\n# gMOO configuration file, please do not edit by hand\n#\n#\n";
  
  fd = open(file, O_WRONLY | O_TRUNC);
  i = write(fd, s, strlen(s));
  if(i < 1)
    printf(_("gMOO: Error: Writing in \"%s\" failed -> %s\n"), file, strerror(errno));
  
  close(fd);
}

void put_color(char *file, char *name, int r, int g, int b) {
  char s[15];
  int fd;
  
  fd = open(file, O_WRONLY);
  lseek(fd, 0, SEEK_END);

  write(fd, "\n", 1);
  write(fd, name, strlen(name));
  write(fd, " = ", 3);

  sprintf(s, "000%d", r);
  write(fd, s + strlen(s) - 3, 3);
  
  sprintf(s, "000%d", g);
  write(fd, s + strlen(s) - 3, 3);
  
  sprintf(s, "000%d", b);
  write(fd, s + strlen(s) - 3, 3);
  
  write(fd, "\n", 1);
  close(fd);
}

void put_externals(char *worldrc) {
  char *s;
  int i;
  
  if(ext_mail == NULL)
    put_value(worldrc, EXT_MAIL, "(null)");
  else {
    s = strnew(ext_mail);
    for(i = 0; i < strlen(s); i++)
      if(s[i] == ' ')
        s[i] = 1;
    put_value(worldrc, EXT_MAIL, s);
    g_free(s);
  }

  if(ext_news == NULL)
    put_value(worldrc, EXT_NEWS, "(null)");
  else {
    s = strnew(ext_news);
    for(i = 0; i < strlen(s); i++)
      if(s[i] == ' ')
        s[i] = 1;
    put_value(worldrc, EXT_NEWS, s);
    g_free(s);
  }
}


void get_externals(char *file) {
  int i;

  ext_mail = get_value(file, EXT_MAIL);
  if(ext_mail != NULL) {
    if(strcmp(ext_mail, "(null)") == 0) {
      g_free(ext_mail);
      ext_mail = NULL;
    } else
      for(i = 0; i < strlen(ext_mail); i++)
        if(ext_mail[i] == 1)
          ext_mail[i] = ' ';
  }
  
  ext_news = get_value(file, EXT_NEWS);
  if(ext_news != NULL) {
    if(strcmp(ext_news, "(null)") == 0) {
      g_free(ext_news);
      ext_news = NULL;
    } else
      for(i = 0; i < strlen(ext_news); i++)
        if(ext_news[i] == 1)
          ext_news[i] = ' ';
  }
}

void put_keypad_keys(char *worldrc) {
  int i;
  char name[20];
  for(i = 0; i < 16; i++) {
    sprintf(name, "KeyPadKey%d", i);
    strsub(keypad_keys[i], ' ', 1);
    put_value(worldrc, name, keypad_keys[i] == NULL ? EMPTY_STR : keypad_keys[i]);
    strsub(keypad_keys[i], 1, ' ');
  }
}

void put_default_keypad_keys(char *worldrc) {
  put_value(worldrc, "KeyPadKey0",  "west");  /* left */
  put_value(worldrc, "KeyPadKey1",  "east");  /* right */
  put_value(worldrc, "KeyPadKey2",  "north"); /* up */
  put_value(worldrc, "KeyPadKey3",  "south"); /* down */

  put_value(worldrc, "KeyPadKey4",  "look");  /* middle */

  put_value(worldrc, "KeyPadKey5",  "nw");    /* home */
  put_value(worldrc, "KeyPadKey6",  "sw");    /* end */
  put_value(worldrc, "KeyPadKey7",  "ne");    /* pgup */
  put_value(worldrc, "KeyPadKey8",  "se");    /* pgdn */

  put_value(worldrc, "KeyPadKey9",  "up");    /* minus */
  put_value(worldrc, "KeyPadKey10", "down");  /* plus */

  put_value(worldrc, "KeyPadKey11", "here"); /* mul */
  put_value(worldrc, "KeyPadKey12", "@who"); /* div */

  put_value(worldrc, "KeyPadKey13", EMPTY_STR); /* ins */
  put_value(worldrc, "KeyPadKey14", EMPTY_STR); /* del */

  put_value(worldrc, "KeyPadKey15", EMPTY_STR); /* enter */
}

void get_keypad_keys(char *worldrc) {
  int i;
  char name[20];

  for(i = 0; i < 16; i++) {
    sprintf(name, "KeyPadKey%d", i);
    keypad_keys[i] = get_value(worldrc, name);
    if(keypad_keys[i] == NULL) {
      switch(i) {
      case 0:  keypad_keys[i] = strnew("west");  break;
      case 1:  keypad_keys[i] = strnew("east");  break;
      case 2:  keypad_keys[i] = strnew("north"); break;
      case 3:  keypad_keys[i] = strnew("south"); break;
      case 4:  keypad_keys[i] = strnew("look");  break;
      case 5:  keypad_keys[i] = strnew("nw");    break;
      case 6:  keypad_keys[i] = strnew("sw");    break;
      case 7:  keypad_keys[i] = strnew("ne");    break;
      case 8:  keypad_keys[i] = strnew("se");    break;
      case 9:  keypad_keys[i] = strnew("up");    break;
      case 10: keypad_keys[i] = strnew("down");  break;
      case 11: keypad_keys[i] = strnew("here");  break;
      case 12: keypad_keys[i] = strnew("@who");  break;
      case 13: break;
      case 14: break;
      case 15: break;
      }
    } else if(strcmp(keypad_keys[i], EMPTY_STR) == 0) {
      g_free(keypad_keys[i]);
      keypad_keys[i] = NULL;
    }
    strsub(keypad_keys[i], 1, ' ');
  }
}

