/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef gM_config_h
#define gM_config_h

#include "global_vars.h"

void  gM_config_init(int argc, char *argv[]);

/* Global vars: */
char 		*config_dir;
char 		*world_dir;
int   	 term_color;

GdkColor colors[10];
GdkColor editcolors[8];
GdkFont *normalfont;
GdkFont *boldfont;
char 		*normalfontstr;
char 		*boldfontstr;
int fgcolor;
int bgcolor;

GtkStyle *tabhigh_style;
GtkStyle *tabnorm_style;

int 		 use_hist;
int 		 use_ident;
int 		 use_auto_detect;
int 	   logtype;
int 		 clocktype;
int      show_toolbar;
char *ext_mail;
char *ext_news;
int reconnect;
int popupcon;
int savesend;
int show_own;
int beep_type;
int short_tab_names;
int homo_tabs;

int    use_keypad_keys;
char **keypad_keys;

char *save_window_pos;
char *save_window_size;
#if 0 /* That damned GtkText widget is acting up again */
int buffer_len;
#endif
int userlistsorttype;
int wordwrap;
int definputstyle;
int showsep;
#ifdef TCL
int tcl_debug;
int tcl_debug_on_stdout;
#endif
int multiline, newmultiline;
int multilinesize;

/* End Global vars */

typedef struct _gM_newworld gM_newworld;
struct _gM_newworld {
  char *name;
  char *host;
  int   port;
  char *user;
  char *pass;
  int   autoconnect;
};

#define loguse        1 /* don't keep logs */
#define logappend     2 /* append to logfile (file is truncated if not set */
#define logtime       4 /* create a logfile, each start of the MOO */
#define logsaveinput  8 /* also log input given */
#define logday       16

#define clocksecond  1
#define clockminute  2

/* lower values aren't used for backward compatibility */
#define beep_type_speaker 4
#define beep_type_visual  5
#define beep_type_none    6

#define EMPTY_STR     "--[NONE]--"
#define HOSTNAME      "Hostname"
#define PORT          "Port"
#define USERNAME      "Username"
#define PASSWORD      "Password"
#define AUTOCONNECT   "AutoConnect"

#define NORMALFONT    "NormalFont"
#define BOLDFONT      "BoldFont"

#define HISTORYTYPE   "HistoryScrollbackType"
#define USEAUTOIDENT  "EditorUseAutoIdent"
#define USEAUTODETECT "EditorAutoVerbDetect"
#define LOGTYPE       "WorldLoggingType"
#define CLOCKTYPE     "ClockType"
#define SHOW_TOOLBAR  "ShowToolbar"
#define RECONNECT     "Reconnect"
#define POPUPCON      "PopUpConnectionDialog"
#define SAVESEND      "SaveSendSettings"
#define SHOW_OWN      "ShowOwnInput"
#define BEEP_TYPE     "BeepType"

#define SAVE_WINDOW_POS  "WindowPosition"
#define SAVE_WINDOW_SIZE "WindowSize"

#define FGCOLOR       "DefaultFGColor"
#define BGCOLOR       "DefaultBGColor"

#define EXT_MAIL      "EXTERNAL_MAIL"
#define EXT_NEWS      "EXTERNAL_NEWS"

#define SHORTTABNAMES	 "ShortTabNames"
#define HOMO_TABS "HomogeneousTabs"

#define USE_KEYPAD_KEYS "UseTheKeypadKeys"
#if 0 /* That damned GtkText widget is acting up again */
#define BUFFER_LEN "BufferLength"
#endif /* 0 */

#define WORDWRAP      "WordWrap"
#define DEFINPUTSTYLE "DefaultInputStyle"
#define SHOWSEP       "ShowSeperator"

#define MULTILINE     "MultilineInputBox"
#define MULTILINESIZE "MultilineSize"

#ifdef TCL
#define TCL_DEBUG "TclDebugEnabled"
#define TCL_DEBUG_ON_STDOUT "TclDebugOnStdout"
#endif /* TCL */

#define BLACK   0
#define WHITE   1
#define RED     2
#define GREEN   3
#define BLUE    4
#define CYAN    5
#define MAGENTA 6
#define YELLOW  7
#define DEFAULTFG 8
#define DEFAULTBG 9

#define CBLACK   "ColorBlack"
#define CWHITE   "ColorWhite"
#define CRED     "ColorRed"
#define CGREEN   "ColorGreen"
#define CBLUE    "ColorBlue"
#define CCYAN    "ColorCyan"
#define CMAGENTA "ColorMagenta"
#define CYELLOW  "ColorYellow"

#define CSTRING  "EditorColorString"
#define CNUMBER  "EditorColorNumber"
#define COBJNUM  "EditorColorObjectNumber"
#define CDOLLAR  "EditorColorDollar"
#define CVARIAB  "EditorColorVariables"
#define CKEYWRD  "EditorColorKeywords"
#define CSYMBOL  "EditorColorSymbols"
#define CCOMMEN  "EditorColorComments"

#define USERLISTSORTTYPE "UserlistSortType"
#define USERLISTSORT_NAME         0
#define USERLISTSORT_NAME_REVERSE 1
#define USERLISTSORT_OBJ          2
#define USERLISTSORT_OBJ_REVERSE  3


/* Keypad movement defines */
#define LEFT   0
#define RIGHT  1
#define UP     2
#define DOWN   3
#define MIDDLE 4
#define HOME   5
#define END    6
#define PGUP   7
#define PGDN   8
#define MINUS  9
#define PLUS  10
#define MUL   11
#define DIV   12
#define INS   13
#define DEL   14
#define ENTER 15


void   gM_config_create_new_world(gM_newworld *nw);
void   gM_config_recreate_world(gM_newworld *nw);
void   gM_config_save_world(world *w);

GList *gM_config_get_world_names();

world *gM_config_read_world(char *worldname);

int    gM_config_remove_world_file(char *file);

void   gM_config_rewrite_main_rc();

/* lock file stuff: */
char *lock_file_name;
void gM_config_lock_open  ();
void gM_config_lock_close ();
int  gM_config_lock_check (char *worldname);
void gM_config_lock_add   (char *worldname);
void gM_config_lock_remove(char *worldname);

/* used by triggers.c */
int  exist_file(char *file);

#endif
