/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "do_scripts.h"

#ifdef TCL

#include <gtk/gtk.h>
#include <stdlib.h>
#include <string.h>

#include "world.h"
#include "handy.h"

#define set_width_min(wid, size) \
  if((wid)->requisition.width < size) \
    gtk_widget_set_usize((wid), size, -1);


void do_C_tell(world *w, char *text, int flags) {
  int len = flags & FLAG_N ? strlen(text) : strlen(text) + 1;
  char *s = g_malloc(len + 1);
  sprintf(s, "%s%s", text, flags & FLAG_N ? "" : "\n");
  if(flags & FLAG_RAW) {
    gM_world_print_raw(w, s, len);
  } else {
    gM_world_print(w, s, len);
  }
}

void do_C_msg (world *w, char *text, int flags) {
  char *s = g_malloc(strlen(text) + 3);
  sprintf(s, "%% %s", text);
  do_C_tell(w, s, flags);
  g_free(s);
}

void do_C_send(world *w, char *text, int flags) {
  char *s = g_malloc(strlen(text) + 2);
  char *t;
  
  sprintf(s, "%s%s", text, flags & FLAG_N ? "" : "\n");
  if(strchr(s, '\n') == NULL) { /* No '\n' char found, buffer it */
    if(w->send_buffer == NULL) {
      w->send_buffer = s;
    } else {
      t = strcatnew(w->send_buffer, s);
      g_free(s);
      g_free(w->send_buffer);
      w->send_buffer = t;
    }
  } else { /* allso check if the buffer isn't empty */
    if(w->send_buffer != NULL) {
      gM_world_write(w, w->send_buffer);
      g_free(w->send_buffer);
      w->send_buffer = NULL;
    }
    gM_world_write(w, s);
    g_free(s);
  }
}

void button_cb_0(GtkWidget *button, int *but) { *but = 0; }
void button_cb_1(GtkWidget *button, int *but) { *but = 1; }
void button_cb_2(GtkWidget *button, int *but) { *but = 2; }
int  window_cb_3(GtkWidget *window, GdkEventAny *e, int *but) { *but = 3; return FALSE; }

int do_C_msgbox(char *title, char *text, int flags) {
  int but = -1;
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *pixmap;
  GtkWidget *label;
  GtkWidget *hseparator;
  GtkWidget *hbuttonbox;
  GtkWidget *button1 = NULL;
  GtkWidget *button2 = NULL;
  GtkWidget *button3 = NULL;
  int delete;

  if(flags & FLAG_ERROR) {
    pixmap = gtk_pixmap_new(msgboxpix[ICON_ERROR], msgboxmask[ICON_ERROR]);
    if(title == NULL) title = _("Tcl Error");
  } else if(flags & FLAG_QUESTION) {
    pixmap = gtk_pixmap_new(msgboxpix[ICON_QUESTION], msgboxmask[ICON_QUESTION]);
    if(title == NULL) title = _("Tcl Question");
  } else if(flags & FLAG_WARNING) {
    pixmap = gtk_pixmap_new(msgboxpix[ICON_WARNING], msgboxmask[ICON_WARNING]);
    if(title == NULL) title = _("Tcl Warning");
  } else if(flags & FLAG_INFORMATION) {
    pixmap = gtk_pixmap_new(msgboxpix[ICON_INFORMATION], msgboxmask[ICON_INFORMATION]);
    if(title == NULL) title = _("Tcl Information");
  } else {
    pixmap = NULL;
    if(title == NULL) title = _("Tcl Message");
  }
  
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), title);
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
  gtk_window_set_policy (GTK_WINDOW (window), TRUE, TRUE, FALSE);
  
  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbox), 5);
  
  if(pixmap != NULL) {
    gtk_widget_show (pixmap);
    gtk_box_pack_start (GTK_BOX (hbox), pixmap, FALSE, TRUE, 0);
  }

  label = gtk_label_new (text);
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (hbox), label, TRUE, TRUE, 0);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  
  hseparator = gtk_hseparator_new ();
  gtk_widget_show (hseparator);
  gtk_box_pack_start (GTK_BOX (vbox), hseparator, FALSE, TRUE, 0);
  
  hbuttonbox = gtk_hbutton_box_new ();
  gtk_widget_show (hbuttonbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbuttonbox, FALSE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbuttonbox), 5);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbuttonbox), 10);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbuttonbox), 0, 0);
  
  if(flags & FLAG_ABORTRETRYIGNORE) {
    button1 = gtk_button_new_with_label(_("Abort"));
    button2 = gtk_button_new_with_label(_("Retry"));
    button3 = gtk_button_new_with_label(_("Cancel"));
  } else if(flags & FLAG_OKCANCEL) {
    button1 = gtk_button_new_with_label(_("Ok"));
    button2 = gtk_button_new_with_label(_("Cancel"));
  } else if(flags & FLAG_RETRYCANCEL) {
    button1 = gtk_button_new_with_label(_("Retry"));
    button2 = gtk_button_new_with_label(_("Cancel"));
  } else if(flags & FLAG_YESNO) {
    button1 = gtk_button_new_with_label(_("Yes"));
    button2 = gtk_button_new_with_label(_("No"));
  } else if(flags & FLAG_YESNOCANCEL) {
    button1 = gtk_button_new_with_label(_("Yes"));
    button2 = gtk_button_new_with_label(_("No"));
    button3 = gtk_button_new_with_label(_("Cancel"));
  } else { /* Assume flags & FLAG_OK */
    button1 = gtk_button_new_with_label(_("Ok"));
  }

  gtk_widget_show (button1);
  gtk_container_add (GTK_CONTAINER (hbuttonbox), button1);
  gtk_signal_connect(GTK_OBJECT(button1), "clicked", button_cb_0, &but);
  GTK_WIDGET_SET_FLAGS (button1, GTK_CAN_DEFAULT);
  if(button2 != NULL) {
    gtk_widget_show (button2);
    gtk_container_add (GTK_CONTAINER (hbuttonbox), button2);
    gtk_signal_connect(GTK_OBJECT(button2), "clicked", button_cb_1, &but);
    GTK_WIDGET_SET_FLAGS (button2, GTK_CAN_DEFAULT);
  }
  if(button3 != NULL) {
    gtk_widget_show (button3);
    gtk_container_add (GTK_CONTAINER (hbuttonbox), button3);
    gtk_signal_connect(GTK_OBJECT(button3), "clicked", button_cb_2, &but);
    GTK_WIDGET_SET_FLAGS (button3, GTK_CAN_DEFAULT);
  }

  if(flags & FLAG_DEF0) {
    gtk_widget_grab_focus(button1);
    gtk_widget_grab_default(button1);
  } else if(flags & FLAG_DEF1) {
    if(button2 == NULL) {
      gtk_widget_grab_focus(button1);
      gtk_widget_grab_default(button1);
    } else {
      gtk_widget_grab_focus(button2);
      gtk_widget_grab_default(button2);
    }
  } else {/* Assume flags & FLAG_DEF2 */
    if(button3 == NULL) {
      if(button2 == NULL) {
        gtk_widget_grab_focus(button1);
        gtk_widget_grab_default(button1);
      } else {
        gtk_widget_grab_focus(button2);
        gtk_widget_grab_default(button2);
      }
    } else {
      gtk_widget_grab_focus(button3);
      gtk_widget_grab_default(button3);
    }
  }
  /* block window deletion */
  delete = gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(gtk_true), NULL);

  gtk_widget_show(window);
  set_width_min(window, 180);

  while(but < 0)
    gtk_main_iteration_do(TRUE);
  /* unblock window deletion */
  gtk_signal_disconnect(GTK_OBJECT(window), delete);

  gtk_widget_destroy(window);

  return but;
}

int do_C_input(char *title,  char *desc, char *deftext, int flags,
               int  *retint, char **retstr) {
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *vbox2;
  GtkWidget *label;
  GtkWidget *entry;
  GtkWidget *hseparator;
  GtkWidget *hbuttonbox;
  GtkWidget *button;
  int done = -1;
  int delete;

  if(title == NULL) title = _("Tcl Input");
 
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), title);
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
  gtk_window_set_policy (GTK_WINDOW (window), TRUE, TRUE, FALSE);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  
  vbox2 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox2); 
  gtk_box_pack_start (GTK_BOX (vbox), vbox2, TRUE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (vbox2), 10);

  if(desc != NULL) {
    label = gtk_label_new(desc);
    gtk_widget_show (label);
    gtk_box_pack_start (GTK_BOX (vbox2), label, TRUE, TRUE, 5);
    gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
  }

  if(flags & FLAG_INT) {
    *retint = atoi(deftext == NULL ? "0" : deftext);
    entry = gtk_spin_button_new(GTK_ADJUSTMENT(gtk_adjustment_new(*retint,
                                                                  -2147483647,
                                                                  2147483647,
                                                                  1, 10, 10)),
                                1, 0);
    gtk_widget_show (entry);
    gtk_box_pack_start (GTK_BOX (vbox2), entry, FALSE, TRUE, 0);
  } else {
    entry = gtk_entry_new ();
    gtk_widget_show (entry);
    gtk_box_pack_start (GTK_BOX (vbox2), entry, FALSE, TRUE, 0);
    if(deftext != NULL)
      gtk_entry_set_text (GTK_ENTRY (entry), deftext);
  }

  hseparator = gtk_hseparator_new ();
  gtk_widget_show (hseparator);
  gtk_box_pack_start (GTK_BOX (vbox), hseparator, FALSE, TRUE, 0);
  
  hbuttonbox = gtk_hbutton_box_new ();
  gtk_widget_show (hbuttonbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbuttonbox, FALSE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbuttonbox), 5);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbuttonbox), 10);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbuttonbox), 0, 0);

  button = gtk_button_new_with_label (_("Ok"));
  gtk_widget_show (button);
  gtk_container_add (GTK_CONTAINER (hbuttonbox), button);
  gtk_signal_connect(GTK_OBJECT(button), "clicked", button_cb_0, &done);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);

  button = gtk_button_new_with_label (_("Cancel"));
  gtk_widget_show (button);
  gtk_container_add (GTK_CONTAINER (hbuttonbox), button);
  gtk_signal_connect(GTK_OBJECT(button), "clicked", button_cb_1, &done);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(button);
  
  delete = gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(window_cb_3), &done);

  gtk_widget_show(window);
  set_width_min(window, 180);

  while(done < 0)
    gtk_main_iteration_do(TRUE);

  if(done == 0) {
    if(flags & FLAG_INT) {
      *retint = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(entry));
    } else { /* flags & TYPE_STR */
      *retstr = strnew(gtk_entry_get_text(GTK_ENTRY(entry)));
    }
    gtk_signal_disconnect(GTK_OBJECT(window), delete);
    gtk_widget_destroy(window);
    /* The window has been destroyed the the delete callback returned TRUE */
  }
  return done == 0 ? 1 : 0;
}

#endif /* TCL */
