/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify 
 *  it under the terms of the GNU General Public License as published by 
 *  the Free Software Foundation; either version 2 of the License, or 
 *  (at your option) any later version. 
 * 
 *  This program is distributed in the hope that it will be useful, 
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of 
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 *  GNU General Public License for more details. 
 * 
 *  You should have received a copy of the GNU General Public License 
 *  along with this program; if not, write to the Free Software 
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <gdk/gdkkeysyms.h>

#include <stdio.h>
#include <string.h>
#include <sys/types.h>	/* typedefs */
#include <unistd.h>			/* close() */
#include <errno.h>			/* error handling functions */
#include <fcntl.h>    	/* open() ao */
#include <unistd.h>			/* lseek() */

#include "dialog.h"
#include "structs.h"
#include "worlds.h"
#include "world.h"
#include "editor.h"
#include "handy.h"
#include "config.h"
#include "window.h"
#include "syntaxhl.h"
#include "editordialogs.h"
#include "mcp.h"
#include "world.h"

void really_close(editor *e);
int  send_text(editor *e);

void save_text(editor *e);
void open_text(editor *e);
void filesel_cancel(GtkWidget *but, void *pointer);
void filesel_ok_save(GtkWidget *but, editor *e);
void filesel_ok_open(GtkWidget *but, editor *e);
GtkWidget *filesel;

void do_enter(GtkWidget *text);
void do_backspace(GtkWidget *text);

int  editor_changed(GtkWidget *text, editor *e);
void editor_key_press(GtkWidget *text, GdkEventKey *event, editor *e);
int  editor_key_press_after(GtkWidget *text, GdkEventKey *event, editor *e);
void update_line_label(GtkWidget *taxt, GdkEvent *event, editor *e);

char *get_line_around(char *s, int pos, int *begin, int *end);
void handle_color(editor *e, char *what);

void insert_text(GtkWidget *text, char *s, int len, int *pos, editor *e);
void delete_text(GtkWidget *text, int start_pos, int end_pos, editor *e);

int  gM_editor_new(world *w, char *verbname, char *upload) {
  editor *e;
  editor **el;
  GtkWidget *tmp;
  int i;
  char *name;
  
  e = g_malloc(sizeof(editor));
  e->world = w;
  e->verbname = verbname != NULL ? strnew(verbname) : NULL;
  e->upload = upload != NULL ? strnew(upload) : NULL;
  e->has_changed = 0;
  e->curline = 1;
  e->syntax = 0;
  e->simpleedit = NULL;
  
  e->main = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(e->main);
  

  e->text = gtk_text_new (NULL, NULL);
  gtk_box_pack_start(GTK_BOX(e->main), e->text, TRUE, TRUE, 0);
  gtk_signal_connect(GTK_OBJECT(e->text), "key_press_event", GTK_SIGNAL_FUNC(editor_key_press), e);
  gtk_signal_connect_after(GTK_OBJECT(e->text), "key_press_event", GTK_SIGNAL_FUNC(editor_key_press_after), e);
  e->sig = gtk_signal_connect_after(GTK_OBJECT(e->text), "changed", GTK_SIGNAL_FUNC(editor_changed), e);
  gtk_signal_connect_after(GTK_OBJECT(e->text), "button_press_event", GTK_SIGNAL_FUNC(update_line_label), e);
  set_global_color(e->text);

  gtk_signal_connect(GTK_OBJECT(e->text), "insert_text", GTK_SIGNAL_FUNC(insert_text), e);
  gtk_signal_connect(GTK_OBJECT(e->text), "delete_text", GTK_SIGNAL_FUNC(delete_text), e);

  gtk_text_set_editable(GTK_TEXT(e->text), TRUE);
  GTK_TEXT(e->text)->default_tab_width = 2;
  gtk_text_set_word_wrap(GTK_TEXT(e->text), TRUE);
  gtk_widget_set_name(e->text, "output");
  gtk_widget_show(e->text);

  
  tmp = gtk_vscrollbar_new(GTK_ADJUSTMENT(GTK_TEXT(e->text)->vadj));
  gtk_widget_show(tmp);
  gtk_box_pack_start(GTK_BOX(e->main), tmp, FALSE, TRUE, 0);

  if(verbname != NULL) {
    name = g_malloc(21);
    limit_to_20(FOR_EDITOR, verbname, &name);
    e->label = gtk_label_new(short_tab_names > 0 ? name : verbname);
    gtk_widget_show(e->label);
    g_free(name);
  }

  if(w->editcount == 0) {
    w->edit = g_malloc(sizeof(editor *));
    w->edit[0] = e;
    w->editcount = 1;
    return 0;
  } else {
    el = g_malloc(sizeof(editor *) * w->editcount + 1);
    for(i = 0; i < w->editcount; i++) {
      el[i] = w->edit[i];
    }
    el[w->editcount] = e;
    w->editcount++;
    g_free(w->edit);
    w->edit = el;
    return w->editcount - 1;
  }
}

void gM_editor_new_mcp(world *w, mcpsimpleedit *se) {
  editor *e;
  char *name;

  e = w->edit[gM_editor_new(w, NULL, NULL)];

  name = g_malloc(21);
  limit_to_20(FOR_EDITOR, se->name, &name);
  e->label = gtk_label_new(short_tab_names > 0 ? name : se->name);
  gtk_widget_show(e->label);
  g_free(name);

  e->simpleedit = se;
  se->editor = e;

}



int  gM_editor_close(editor *e) {
  int i;
  char *s;
  
  if(e->has_changed > 0) {
    s = strcat3new(_("You have been editing "),
      					   e->verbname != NULL ? e->verbname : e->simpleedit->name,
      						 _("\nIt has not been send to the MOO(/MUD).\n"
                   "\n"
                   "Do you wish to send it now?"));
    i = gM_dialog_show_question(_("Question"), s, bYes | bNo | bCancel);
    g_free(s);
    switch(i) {
    case bYes:
      if(send_text(e) < 0)
        return -1;
    case bNo:
      break;
    case bCancel:
      return -1;
    }
  }
  really_close(e);
  return 1;
}

void really_close(editor *e) {
  int i, j;
  editor **el;

  if(e->world->editcount == 1) {
    e->world->editcount--;
    g_free(e->world->edit);
  } else {
    el = g_malloc(sizeof(editor *) * (e->world->editcount - 1));
    j = 0;
    for(i = 0; i < e->world->editcount; i++) {
      if(e->world->edit[i] != e) {
        el[j] = e->world->edit[i];
        j++;
      }
    }
    e->world->editcount--;
    g_free(e->world->edit);
    e->world->edit = el;
  }

  /* editor is free now */
  if(e->verbname != NULL) {
    g_free(e->verbname);
    g_free(e->upload);
  } else
    gM_mcp_simpleedit_free(e->simpleedit);
    
  gtk_widget_destroy(e->main);
  gM_worlds_reorder_pages();
  gtk_notebook_set_page(GTK_NOTEBOOK(notebook), e->world->page_num);
  g_free(e);
}

int send_text(editor *e) {
  char *s;
  int len;

  if(e->world->sock < 1) {
    gM_dialog_show(_("Error:"), _("Could not send text, world not connected!  "),1);
    return -1;
  }

  len = gtk_text_get_length(GTK_TEXT(e->text));
  s = gtk_editable_get_chars(GTK_EDITABLE(e->text), 0, len);

  if(e->upload != NULL) {
    gM_world_write(e->world, e->upload);
    if(e->upload[strlen(e->upload) - 1] != '\n')
      gM_world_write(e->world, "\n");
    gM_world_write(e->world, s);
    if(s[len - 1] == '\n')
      gM_world_write(e->world, ".\n");
    else
      gM_world_write(e->world, "\n.\n");
  } else {
    gM_mcp_simpleedit_upload(e->world, e->simpleedit, s, len);
  }

  gtk_notebook_set_page(GTK_NOTEBOOK(notebook), e->world->page_num);
  e->has_changed = 0;

  g_free(s);

  return 1;
}


void gM_editor_add(editor *e, char *s, int len) {
  gtk_text_insert(GTK_TEXT(e->text), normalfont, NULL, NULL, s, len);
}



void gM_editor_data_received(editor *e) {
  char *s, *t;
  int i;

  gM_window_add_editor(e);

  if(e->upload != NULL) {
    s = strrchr(e->upload, ':');
    if(s != NULL) {
      t = strrchr(e->upload, '.');
      if(t != NULL) {
        if(s > t)
          i = 1;
        else
          i = 0;
        }
      else
        i = 1;
    } else
      i = 0;
  } else {
    if(strcasecmp(e->simpleedit->type, "moo-code") == 0)
      i = 1;
    else
      i = 0;
  }
  gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(tb_editsyntax), use_auto_detect > 0 ? i : FALSE);

}



void gM_editor_handle_menu(GtkWidget *menuitem, int option) {
  int i, page_num;
  world *w;
  editor *e;


  page_num = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
  w = gM_worlds_get_world_by_editor_main(gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page_num), &i);
  e = w->edit[i];
  
  switch(option) {
  case SEND:
    if(e->has_changed < 1)
      gM_dialog_show(_("Message:"), _("There has been no change in the text...  \n\nHas not been send."), 0);
    else
      send_text(e);
    break;
  case SENDCLOSE:
    if(send_text(e) > 0)
      really_close(e);
    break;
  case CLOSE:
    really_close(e);
    break;
  case SAVE:
  	save_text(e);
    break;
  case OPEN:
  	open_text(e);
    break;
  case SYNTAX:
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tb_editsyntax),
                                GTK_CHECK_MENU_ITEM(menuitem)->active);
    break;
  case FIND:
    gM_editordialogs_find(e);
    break;
  case REPLACE:
    gM_editordialogs_replace(e);
    break;
  case AGAIN:
    gM_editordialogs_again();
    break;
  case GOTOLINE:
    gM_editordialogs_gotoline(e);
    break;
  }
}

void save_text(editor *e) {
  filesel = gtk_file_selection_new(_("Save to a local file..."));
  gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), e->verbname != NULL ? e->verbname : e->simpleedit->name);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button), "clicked", GTK_SIGNAL_FUNC(filesel_cancel), NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button), "clicked", GTK_SIGNAL_FUNC(filesel_ok_save), e);
  gtk_window_set_modal(GTK_WINDOW(filesel), TRUE);
  gtk_window_position(GTK_WINDOW(filesel), GTK_WIN_POS_MOUSE);
  gtk_widget_show(filesel);
}

void open_text(editor *e) {
  filesel = gtk_file_selection_new(_("Open a local file..."));
  gtk_file_selection_set_filename(GTK_FILE_SELECTION(filesel), e->verbname != NULL ? e->verbname : e->simpleedit->name);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->cancel_button), "clicked",  GTK_SIGNAL_FUNC(filesel_cancel), NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filesel)->ok_button), "clicked", GTK_SIGNAL_FUNC(filesel_ok_open), e);
  gtk_window_set_modal(GTK_WINDOW(filesel), TRUE);
  gtk_window_position(GTK_WINDOW(filesel), GTK_WIN_POS_MOUSE);
  gtk_widget_show(filesel);
}

void filesel_cancel(GtkWidget *but, void *pointer) {
  gtk_widget_hide(filesel);
  gtk_widget_destroy(filesel);
}

void filesel_ok_save(GtkWidget *but, editor *e) {
  char *file = gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
  int fd;
  char *s;
  int len;

  len = gtk_text_get_length(GTK_TEXT(e->text));
  s = gtk_editable_get_chars(GTK_EDITABLE(e->text), 0, len);
  
  fd = open(file, O_WRONLY | O_CREAT | O_TRUNC, 00644);
  if(fd < 1) {
    g_free(s);
    s = strcatnew(_("Error opening file:  \n"), strerror(errno));
    gM_dialog_show(_("Error:"), s, 1);
  } else {
    if(write(fd, s, len) != len) {
      g_free(s);
      s = strcatnew(_("Error writing to file:  \n"), strerror(errno));
      gM_dialog_show(_("Error:"), s, 1);
    }
    /* succes: */
    close(fd);
  }
  g_free(s);
  filesel_cancel(but, NULL);
}

#define LEN 4096
void filesel_ok_open(GtkWidget *but, editor *e) {
  char *file = gtk_file_selection_get_filename(GTK_FILE_SELECTION(filesel));
  int fd;
  char *s;
  int len;

  fd = open(file, O_RDONLY);
  if(fd < 1) {
    s = strcatnew("Error opening file:  \n", strerror(errno));
    gM_dialog_show("Error:", s, 1);
    g_free(s);
  } else {
    gtk_text_freeze(GTK_TEXT(e->text));
    lseek(fd, 0, SEEK_SET);
    s = g_malloc(sizeof(char) * LEN);
    while(1) {
      len = read(fd, s, LEN - 1);
      if(len == 0) {
        break; /* should not be able to happen */
      }
      if(s[len - 1] != 0)
        s[len] = 0; /* add terminating nul char ... */
      gtk_text_insert(GTK_TEXT(e->text), normalfont, NULL, NULL, s, len);	/* print it on the screen */
      if(len != LEN - 1)	/* check if the buffer was acedently FULL */
        break;
    }
    g_free(s);
   
    /* succes: */
    close(fd);
    gtk_text_thaw(GTK_TEXT(e->text));
  }
  filesel_cancel(but, NULL);
}

/* The syntax button callback, also used for the menu item */
void gM_editor_do_syntax(GtkWidget *but, void *pointer) {
  int on;
  editor *e;
  world *w;
  char *s;
  char *shlline;
  int i;
  int pos;
  float value;

  on = GTK_TOGGLE_BUTTON(but)->active;
  gtk_signal_handler_block(GTK_OBJECT(syntaxmenu), syntaxhlsignal);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(syntaxmenu), on);
  gtk_signal_handler_unblock(GTK_OBJECT(syntaxmenu), syntaxhlsignal);
  w = gM_worlds_get_world_by_editor_main(
    	gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), 
      gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook))), &i);
  if(w == NULL)
    return;
  e = w->edit[i];
  e->syntax = on;

  gtk_signal_handler_block(GTK_OBJECT(e->text), e->sig);

  pos = gtk_editable_get_position(GTK_EDITABLE(e->text));
  i = gtk_text_get_length(GTK_TEXT(e->text));
  value = GTK_ADJUSTMENT(GTK_TEXT(e->text)->vadj)->value;
  if(on > 0) {
    gtk_text_freeze(GTK_TEXT(e->text));
    s = gtk_editable_get_chars(GTK_EDITABLE(e->text), 0, i);
    shlline = g_malloc((i * 2) + 1);
    shl(shlline, s);
    gtk_editable_delete_text(GTK_EDITABLE(e->text), 0, i);
    handle_color(e, shlline);
    g_free(shlline);
    gtk_text_thaw(GTK_TEXT(e->text));
  } else {
    s = gtk_editable_get_chars(GTK_EDITABLE(e->text), 0, i);
    gtk_editable_delete_text(GTK_EDITABLE(e->text), 0, i);
    gtk_text_insert(GTK_TEXT(e->text), normalfont, NULL, NULL, s, i);
  }
  g_free(s);
  gtk_adjustment_set_value(GTK_ADJUSTMENT(GTK_TEXT(e->text)->vadj), value);
  gtk_editable_set_position(GTK_EDITABLE(e->text), pos);
  gtk_text_set_point(GTK_TEXT(e->text), pos);

  gtk_signal_handler_unblock(GTK_OBJECT(e->text), e->sig);
}



int editor_changed(GtkWidget *text, editor *e) {
  int pos, begin, end;
  char *line;
  char *shlline;
/*  char *shlline = g_malloc(150); */
  char *s;
  int len;

  e->has_changed = 1;
  if(e->syntax < 1)
    return FALSE;
    
  gtk_signal_handler_block(GTK_OBJECT(text), e->sig);
  
  pos = gtk_editable_get_position(GTK_EDITABLE(text));
  s = gtk_editable_get_chars(GTK_EDITABLE(text), 0, gtk_text_get_length(GTK_TEXT(text)));

  get_line_around(s, pos, &begin, &end);
  len = end - begin;
  line = g_malloc(len + 1);
  line = strncpy(line, s + begin, len);
  line[len] = 0;

/*printf("line == \"%s\" (van %d tot %d met len %d)\n", line, begin, end, len);*/
  if(len > 0) {
/*printf("hier\n");*/
    shlline = g_malloc((2 * len) + 1);
    shl(shlline, line);

    gtk_text_freeze(GTK_TEXT(text));
  	gtk_editable_delete_text(GTK_EDITABLE(text), begin, end);
    handle_color(e, shlline);
    gtk_text_thaw(GTK_TEXT(text));

    gtk_editable_set_position(GTK_EDITABLE(text), pos);
    gtk_text_set_point(GTK_TEXT(text), pos);

  	g_free(shlline);
  }

  g_free(line);
  g_free(s);
  gtk_signal_handler_unblock(GTK_OBJECT(text), e->sig);
  return TRUE;
}




void do_enter(GtkWidget *text) {
  int pos;
  char *s;
  int i;
  int count = 0;
  
  if(use_ident < 1) {
    gtk_text_insert(GTK_TEXT(text), normalfont, NULL, NULL, "\n", 1);
    return;
  }
  pos = gtk_editable_get_position(GTK_EDITABLE(text));
  s = gtk_editable_get_chars(GTK_EDITABLE(text), 0, gtk_text_get_length(GTK_TEXT(text)));

  for(i = pos != 0 ? pos - 1 : 0; i >= 0; i--)
    if(s[i] == '\n')
      break;
  for(i = i != 0 ? i + 1 : 0; i < pos; i++)
    if(s[i] == ' ')
      count++;
    else
      break;
  if(strncasecmp(s + i, "if",    2) == 0 ||
     strncasecmp(s + i, "else",  4) == 0 ||
     strncasecmp(s + i, "while", 5) == 0 ||
     strncasecmp(s + i, "for",   3) == 0 ||
     strncasecmp(s + i, "try",   3) == 0 ||
     strncasecmp(s + i, "exept", 5) == 0)
    count += 2;
  if(count == 0)
    gtk_text_insert(GTK_TEXT(text), normalfont, NULL, NULL, "\n", 1);
  else {
    g_free(s);
    s = g_malloc(count + 2);
    s[0] = '\n';
    for(i = 1; i <= count; i++) {
      s[i] = ' ';
    }
    s[count + 1] = 0;
    gtk_text_insert(GTK_TEXT(text), normalfont, NULL, NULL, s, count + 1);
  }
  g_free(s);
}

void do_backspace(GtkWidget *text) {
  int pos;
  char *s;
  int i;
  int count = 0;

  pos = gtk_editable_get_position(GTK_EDITABLE(text));
  if(pos == 0)
    return;
  s = gtk_editable_get_chars(GTK_EDITABLE(text), 0, gtk_text_get_length(GTK_TEXT(text)));
  pos--;
  if(s[pos] != ' ') {
    gtk_text_backward_delete(GTK_TEXT(text), 1);
    g_free(s);
    return;
  }
  for(i = pos; i >= 0; i--)
    if(s[i] == '\n')
      break;
    else if(s[i] == ' ')
      count++;
    else {
      gtk_text_backward_delete(GTK_TEXT(text), 1);
      g_free(s);
      return;
    }
  if(count == 0 || count % 2 != 0)
    gtk_text_backward_delete(GTK_TEXT(text), 1);
  else
    gtk_text_backward_delete(GTK_TEXT(text), 2);
  g_free(s);
}


void editor_key_press(GtkWidget *text, GdkEventKey *event, editor *e) {
  int num;
  int pos = gtk_editable_get_position(GTK_EDITABLE(text));


  if(event->keyval == GDK_Tab ||
     event->keyval == GDK_KP_Tab ||
     event->keyval == GDK_ISO_Left_Tab ||
     event->keyval == GDK_Return ||
     event->keyval == GDK_BackSpace) {
    gtk_signal_emit_stop_by_name(GTK_OBJECT(text), "key_press_event");
    gtk_text_set_point(GTK_TEXT(text), pos);
    e->has_changed = 1;
    if(event->keyval == GDK_BackSpace)
      do_backspace(text);
    else if(event->keyval == GDK_Tab || event->keyval == GDK_KP_Tab || event->keyval == GDK_ISO_Left_Tab) {
      num = g_list_length(GTK_NOTEBOOK(notebook)->children);
      if(event->state & 4) { /* CTR-tab */
        if(e->page_num != num - 1)
          gtk_notebook_next_page(GTK_NOTEBOOK(notebook));
        else
          gtk_notebook_set_page(GTK_NOTEBOOK(notebook), 0);
      } else if(event->state & 1) { /* (CTRL) SHIFT-tab */
        if(e->page_num != 0)
          gtk_notebook_prev_page(GTK_NOTEBOOK(notebook));
        else
          gtk_notebook_set_page(GTK_NOTEBOOK(notebook), num - 1);
      } else /* chr(7) to "  " replacement */
        gtk_text_insert(GTK_TEXT(text), normalfont, NULL, NULL, "  ", 2);
    } else if(event->keyval == GDK_Return )
      do_enter(text);
  } else if(event->keyval == GDK_Left  && event->state & 4) {
    num = g_list_length(GTK_NOTEBOOK(notebook)->children);
    if(e->page_num != 0)
      gtk_notebook_prev_page(GTK_NOTEBOOK(notebook));
    else
      gtk_notebook_set_page(GTK_NOTEBOOK(notebook), num - 1);
  } else if(event->keyval == GDK_Right && event->state & 4) {
    num = g_list_length(GTK_NOTEBOOK(notebook)->children);
    if(e->page_num != num - 1)
      gtk_notebook_next_page(GTK_NOTEBOOK(notebook));
    else
      gtk_notebook_set_page(GTK_NOTEBOOK(notebook), 0);
  } else if(event->state == 4 && event->keyval >= GDK_1 && event->keyval <= GDK_9) {
    gtk_notebook_set_page(GTK_NOTEBOOK(notebook), event->keyval - GDK_1);
  }
}

int editor_key_press_after(GtkWidget *text, GdkEventKey *event, editor *e) {
  update_line_label(text, NULL, e);

  if(event->state == 4 &&
     event->keyval != GDK_Left &&
     event->keyval != GDK_Right &&
     event->keyval != GDK_Up &&
     event->keyval != GDK_Down &&
     event->keyval != GDK_Page_Up &&
     event->keyval != GDK_Page_Down)
    return FALSE;
  return TRUE;
}

void update_line_label(GtkWidget *text, GdkEvent *event, editor *e) {
  char *s = gtk_editable_get_chars(GTK_EDITABLE(text), 0, gtk_text_get_length(GTK_TEXT(text)));
  int pos = gtk_editable_get_position(GTK_EDITABLE(text));
  int i, count = 0;

  if(pos != 0)
    for(i = pos - 1; i > -1; i--)
      if(s[i] == '\n')
        count++;
  g_free(s);
  if(count == 0)
    s = strnew(_("Line : 1"));
  else {
    s = g_malloc(15);
    sprintf(s, _("Line : %d"), count + 1);
  }
  gtk_label_set_text(GTK_LABEL(tb_linenum), s);
  g_free(s);
}



char *get_line_around(char *s, int pos, int *begin, int *end) {
  int i = pos;
  char *ret;
  while(s[i] != 10 && s[i] != 0)
    i++;
  *end =  i; /* s[i] != 0 ? i + 1 : i; */

  i = pos > 0 ? pos - 1 : 0;
  while(s[i] != 10 && i != 0)
    i--;
  *begin = i != 0 ? i + 1 : i;
  ret = g_malloc(*end - *begin + 2);
  strncpy(ret, s + *begin, *end - *begin);
  ret[*end - *begin] = 0;
  return ret;
}



void handle_color(editor *e, char *what) {
  int i;
  int j = 0;
  char *buf = g_malloc(sizeof(char) * strlen(what));


  for(i = 0; what[i] != 0; i++) {
    buf[j] = what[i];
/*printf("buf[%d] == '%c' (%d)\n", j, buf[j], buf[j]);*/
    if(buf[j] >= HL_STRING && buf[j] <= HL_SYMBOL) {
      buf[j] = 0;
/*printf("inserting \"%s\"with color %d\n", buf, w->edit->curcolor);*/
      gtk_text_insert(GTK_TEXT(e->text), normalfont, &editcolors[e->curcolor], NULL, buf, -1);
      j = 0;
    }
    if(what[i] == 0) {
      gtk_text_insert(GTK_TEXT(e->text), normalfont, &editcolors[e->curcolor], NULL, buf, -1);
/*printf("BUF == \"%s\"\n", buf);*/
      g_free(buf);
      return;
    } else if(what[i] >= HL_STRING && what[i] <= HL_COMMEN) {
      e->curcolor = what[i] - HL_STRING; /* enure a [0..7] range */
    } else {
      j++;
    }
  }
 
  buf[j] = 0;
  gtk_text_insert(GTK_TEXT(e->text), normalfont, &editcolors[e->curcolor], NULL, buf, -1);
/*printf("BUF(last) == \"%s\"\n", buf);*/
  g_free(buf);
}



void insert_text(GtkWidget *text, char *s, int len, int *pos, editor *e) {
/*  printf("Inserting \"%s\", with len = %d and *pos == %d\n", s, len, *pos);*/
}

void delete_text(GtkWidget *text, int start_pos, int end_pos, editor *e) {
/*  printf("deleting from %d to %d\n", start_pos, end_pos); */
}
 
 
 
