/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include "structs.h"
#include "help.h"
#include "handy.h"
#include "dialog.h"
#include "config.h"

#define OK     0
#define CANCEL 1
#define HELP   2

#define FIND    1
#define REPLACE 2

#define CURRENT 0
#define BEGIN   1

void destroy_cb(GtkWidget *w, void *pointer);
void but_cb(GtkWidget *butwidget, int but);
void cb_cb(GtkWidget *checkbox, void *pointer);
void rb_cb(GtkWidget *radio, int value);
int find_text(char *what, int *begin, int *end);
void select_text(int begin, int end);

void destroy_cb2(GtkWidget *w, void *pointer);
void but_cb2(GtkWidget *butwidget, int but);
void replace_text(int begin, int *end, char *s);

void destroy_cb3(GtkWidget *w, void *pointer);
void but_cb3(GtkWidget *butwidget, int but);

GtkWidget *dialog_find    = NULL;
GtkWidget *dialog_replace = NULL;
GtkWidget *dialog_goto    = NULL;

GtkWidget *findent;
GtkWidget *replaceent;
int use_case = FALSE;
int use_begin = FALSE;
editor *curedit;
GtkWidget *spinbutton;
int max_spin = 0;

int   lastcommand = 0;
char *lastfind    = NULL;
char *lastreplace = NULL;

void gM_editordialogs_find(editor *e) {
  GtkWidget *vbox;
  GtkWidget *vbox2;
  GtkWidget *hbox;
  GtkWidget *hbox2;
  GtkWidget *hbb;
  GtkWidget *tmp;
  GtkWidget *tmp2;
  GSList *rgroup_group = NULL;
 
  if(dialog_find != NULL) /* only one find dialog open at the time */
    return;

  dialog_find = gtk_window_new (GTK_WINDOW_DIALOG);
  gtk_window_set_title (GTK_WINDOW (dialog_find), _("Find:"));
  gtk_window_position (GTK_WINDOW (dialog_find), GTK_WIN_POS_MOUSE);
  gtk_window_set_policy (GTK_WINDOW (dialog_find), FALSE, FALSE, TRUE);
  gtk_signal_connect(GTK_OBJECT(dialog_find), "destroy", GTK_SIGNAL_FUNC(destroy_cb), NULL);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (dialog_find), vbox);

  vbox2 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox2);
  gtk_box_pack_start (GTK_BOX (vbox), vbox2, TRUE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (vbox2), 6);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox2), hbox, FALSE, TRUE, 0);  

  tmp = gtk_label_new (_("Text to Find: "));
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (hbox), tmp, FALSE, TRUE, 0);
  gtk_label_set_justify (GTK_LABEL (tmp), GTK_JUSTIFY_RIGHT);
  gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);

  findent = gtk_entry_new();
  if(lastfind != NULL) {
    gtk_entry_set_text(GTK_ENTRY(findent), lastfind);
    gtk_editable_select_region(GTK_EDITABLE(findent), 0, strlen(lastfind));
  }
  gtk_widget_show(findent);
  gtk_box_pack_start(GTK_BOX(hbox), findent, TRUE, TRUE, 0);
  gtk_widget_grab_focus(findent);

  hbox2 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox2);
  gtk_box_pack_start (GTK_BOX (vbox2), hbox2, FALSE, TRUE, 0);

  tmp = gtk_label_new (_("Start position: "));
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (hbox2), tmp, TRUE, TRUE, 0);

  tmp2 = gtk_radio_button_new_with_label (rgroup_group, _("current location"));
  rgroup_group = gtk_radio_button_group (GTK_RADIO_BUTTON (tmp2));
  gtk_widget_show (tmp2);
  gtk_box_pack_start (GTK_BOX (hbox2), tmp2, TRUE, TRUE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp2), use_begin == CURRENT ? TRUE : FALSE);

  tmp = gtk_radio_button_new_with_label (rgroup_group, _("beginning"));
  rgroup_group = gtk_radio_button_group (GTK_RADIO_BUTTON (tmp));
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (hbox2), tmp, TRUE, TRUE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), use_begin == BEGIN ? TRUE : FALSE);

  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(rb_cb), (void *) BEGIN);
  gtk_signal_connect(GTK_OBJECT(tmp2), "clicked", GTK_SIGNAL_FUNC(rb_cb), (void *) CURRENT);

  tmp = gtk_check_button_new_with_label (_("Case sensitive"));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), use_case);
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (vbox2), tmp, FALSE, TRUE, 0);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(cb_cb), NULL);
                                                     
  tmp = gtk_hseparator_new ();
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (vbox), tmp, FALSE, TRUE, 5);

  hbb = gtk_hbutton_box_new ();
  gtk_widget_show (hbb);
  gtk_box_pack_start (GTK_BOX (vbox), hbb, FALSE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbb), 5);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbb), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbb), 10);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbb), 0, 0);

  tmp = gtk_button_new_with_label ("Ok");
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb), (void *) OK);

  tmp = gtk_button_new_with_label ("Cancel");
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);   
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb), (void *) CANCEL);
  
  tmp = gtk_button_new_with_label ("Help");
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb), (void *) HELP);

  gtk_widget_show(dialog_find);
  
  curedit = e;
}

void destroy_cb(GtkWidget *w, void *pointer) {
  dialog_find = NULL;
}

void but_cb(GtkWidget *butwidget, int but) {
  char *s;
  int begin;
  int end;
  
  switch(but) {
  case OK:
    s = gtk_entry_get_text(GTK_ENTRY(findent));
    lastcommand = FIND;
    if(lastfind != NULL)
      g_free(lastfind);
    lastfind = strnew(s);
    if(s[0] != 0) { /* don't want to add empty strings */
      if(find_text(s, &begin, &end))
        select_text(begin, end);
      else {
        gtk_widget_hide(dialog_find);
        gtk_widget_destroy(dialog_find);
        dialog_find = NULL;
        gM_dialog_show(_("Message"), _("String NOT found  "), 1);
        break;
      }
    } else {
      gM_dialog_show(_("Huh?"), _("Find WHAT ??  "), 1);
      lastcommand = 0;
    }
  case CANCEL:
    gtk_widget_hide(dialog_find);
    gtk_widget_destroy(dialog_find);
    break;
  case HELP:
    gM_help_show_find();
  }
}

void cb_cb(GtkWidget *checkbox, void *pointer) {
  use_case = GTK_TOGGLE_BUTTON(checkbox)->active;
}

void rb_cb(GtkWidget *radio, int value) {
  use_begin = value;
}

int find_text(char *what, int *begin, int *end) {
  char *tmp;
  char *where;
  char *s;
  int len;
  int i = 0;

  tmp = gtk_editable_get_chars(GTK_EDITABLE(curedit->text), 0, gtk_text_get_length(GTK_TEXT(curedit->text)));
  if(use_begin == CURRENT) {
    if(GTK_EDITABLE(curedit->text)->selection_start_pos ==
            GTK_EDITABLE(curedit->text)->selection_end_pos) {
      i = gtk_editable_get_position(GTK_EDITABLE(curedit->text));
    } else {
      i = GTK_EDITABLE(curedit->text)->selection_end_pos;
    }
    where = tmp + i;
  } else
    where = tmp;
  len = strlen(what);

  s = strchr(where, what[0]);
  while(s != NULL) {
    if((use_case && strncasecmp(s, what, len) == 0) || strncmp(s, what, len) == 0) {
  		*begin = s - where + i;
      *end = *begin + len;
      g_free(tmp);
      return TRUE;
    }
  	s = s + 1;
    s = strchr(s, what[0]);
  }
  g_free(tmp);
  return FALSE;
}

void select_text(int begin, int end) {
  gtk_editable_set_position(GTK_EDITABLE(curedit->text), begin);
  gtk_editable_select_region(GTK_EDITABLE(curedit->text),
                             begin, end);
}


void gM_editordialogs_replace(editor *e) {
  GtkWidget *vbox;
  GtkWidget *vbox2;
  GtkWidget *hbox2;
  GtkWidget *hbb;
  GtkWidget *tmp;
  GtkWidget *tmp2;
  GtkWidget *table;
  GSList *rgroup_group = NULL;
 
  if(dialog_replace != NULL) /* only one find dialog open at the time */
    return;

  dialog_replace = gtk_window_new (GTK_WINDOW_DIALOG);
  gtk_window_set_title (GTK_WINDOW (dialog_replace), _("Replace:"));
  gtk_window_position (GTK_WINDOW (dialog_replace), GTK_WIN_POS_MOUSE);
  gtk_window_set_policy (GTK_WINDOW (dialog_replace), FALSE, FALSE, TRUE);
  gtk_signal_connect(GTK_OBJECT(dialog_replace), "destroy", GTK_SIGNAL_FUNC(destroy_cb2), NULL);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (dialog_replace), vbox);

  vbox2 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox2);
  gtk_box_pack_start (GTK_BOX (vbox), vbox2, TRUE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (vbox2), 6);

  table = gtk_table_new (2, 2, FALSE);
  gtk_widget_show (table);
  gtk_box_pack_start (GTK_BOX (vbox2), table, FALSE, TRUE, 0);

  tmp = gtk_label_new (_("New Text: "));
  gtk_widget_show (tmp);
  gtk_table_attach (GTK_TABLE (table), tmp, 0, 1, 1, 2,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);

  tmp = gtk_label_new (_("Text to Find: "));
  gtk_widget_show (tmp);
  gtk_table_attach (GTK_TABLE (table), tmp, 0, 1, 0, 1,
                    (GtkAttachOptions) 0, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);

  replaceent = gtk_entry_new ();
  if(lastreplace != NULL) {
    gtk_entry_set_text(GTK_ENTRY(replaceent), lastreplace);
    gtk_editable_select_region(GTK_EDITABLE(replaceent), 0, strlen(lastreplace));
  }
  gtk_widget_show (replaceent);
  gtk_table_attach (GTK_TABLE (table), replaceent, 1, 2, 1, 2,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);

  findent = gtk_entry_new ();
  if(lastfind != NULL) {
    gtk_entry_set_text(GTK_ENTRY(findent), lastfind);
    gtk_editable_select_region(GTK_EDITABLE(findent), 0, strlen(lastfind));
  }
  gtk_widget_show (findent);
  gtk_table_attach (GTK_TABLE (table), findent, 1, 2, 0, 1,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_widget_grab_focus(findent);
 

  hbox2 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox2);
  gtk_box_pack_start (GTK_BOX (vbox2), hbox2, FALSE, TRUE, 0);

  tmp = gtk_label_new (_("Start position: "));
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (hbox2), tmp, TRUE, TRUE, 0);

  tmp2 = gtk_radio_button_new_with_label (rgroup_group, _("current location"));
  rgroup_group = gtk_radio_button_group (GTK_RADIO_BUTTON (tmp2));
  gtk_widget_show (tmp2);
  gtk_box_pack_start (GTK_BOX (hbox2), tmp2, TRUE, TRUE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp2), use_begin == CURRENT ? TRUE : FALSE);

  tmp = gtk_radio_button_new_with_label (rgroup_group, _("beginning"));
  rgroup_group = gtk_radio_button_group (GTK_RADIO_BUTTON (tmp));
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (hbox2), tmp, TRUE, TRUE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), use_begin == BEGIN ? TRUE : FALSE);

  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(rb_cb), (void *) BEGIN);
  gtk_signal_connect(GTK_OBJECT(tmp2), "clicked", GTK_SIGNAL_FUNC(rb_cb), (void *) CURRENT);

  tmp = gtk_check_button_new_with_label (_("Case sensitive"));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), use_case);
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (vbox2), tmp, FALSE, TRUE, 0);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(cb_cb), NULL);
                                                     
  tmp = gtk_hseparator_new ();
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (vbox), tmp, FALSE, TRUE, 5);

  hbb = gtk_hbutton_box_new ();
  gtk_widget_show (hbb);
  gtk_box_pack_start (GTK_BOX (vbox), hbb, FALSE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbb), 5);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbb), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbb), 10);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbb), 0, 0);

  tmp = gtk_button_new_with_label (_("Ok"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb2), (void *) OK);

  tmp = gtk_button_new_with_label (_("Cancel"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);   
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb2), (void *) CANCEL);
  
  tmp = gtk_button_new_with_label (_("Help"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb2), (void *) HELP);

  gtk_widget_show(dialog_replace);
  
  curedit = e;
}

void destroy_cb2(GtkWidget *w, void *pointer) {
  dialog_replace = NULL;
}

void but_cb2(GtkWidget *butwidget, int but) {
  char *s, *t;
  int begin;
  int end;
  
  switch(but) {
  case OK:
    lastcommand = REPLACE;

    s = gtk_entry_get_text(GTK_ENTRY(findent));
    if(lastfind != NULL)
      g_free(lastfind);
    lastfind = strnew(s);

    t = gtk_entry_get_text(GTK_ENTRY(replaceent));
    if(lastreplace != NULL)
      g_free(lastreplace);
    lastreplace = strnew(t);

    if(t[0] == 0 && gM_dialog_show_question(_("Conformation"), _("Really replace by an empty string?"),
                       bYes | bNo) == bNo) {
      lastcommand = 0;
      break;
    }
    if(s[0] != 0) { /* don't want to find empty strings */
      if(find_text(s, &begin, &end)) {
        replace_text(begin, &end, t);
        select_text(begin, end);
      } else {
        gtk_widget_hide(dialog_replace);
        gtk_widget_destroy(dialog_replace);
        dialog_find = NULL;
        gM_dialog_show(_("Message"), _("String NOT found  \nNoting replaced"), 1);
        break;
      }
    } else {
      gM_dialog_show(_("Huh?"), _("Find WHAT ??  "), 1);
      lastcommand = 0;
    }
  case CANCEL:
    gtk_widget_hide(dialog_replace);
    gtk_widget_destroy(dialog_replace);
    break;
  case HELP:
    gM_help_show_replace();
  }
}

void replace_text(int begin, int *end, char *s) {
  gtk_signal_handler_block(GTK_OBJECT(curedit->text), curedit->sig);
  gtk_text_freeze(GTK_TEXT(curedit->text));

  gtk_editable_delete_text(GTK_EDITABLE(curedit->text), begin, *end);
  gtk_text_set_point(GTK_TEXT(curedit->text), begin);
  *end = begin + strlen(s);
  gtk_text_insert(GTK_TEXT(curedit->text), normalfont, NULL, NULL, s, *end - begin);

  gtk_text_thaw(GTK_TEXT(curedit->text));
  gtk_signal_handler_unblock(GTK_OBJECT(curedit->text), curedit->sig);
  curedit->has_changed = 1;
}

void gM_editordialogs_again() {
  int begin;
  int end;
  int i;

  switch(lastcommand) {
  case 0:
    gM_dialog_show(_("Message"), _("Do what ???  "), 1);
    break;
  case FIND:
    if(find_text(lastfind, &begin, &end)) {
      select_text(begin, end);
    } else 
      gM_dialog_show(_("Message"), _("Nothing found  "), 1);
    break;
  case REPLACE:
    if(find_text(lastfind, &begin, &end)) {
      select_text(begin, end);
      i = gM_dialog_show_question(_("Question:"), _("Replace this text ?"), bYes | bNo | bAll);
      if(i == bNo)
        break;
      gtk_editable_set_position(GTK_EDITABLE(curedit->text), begin);
      replace_text(begin, &end, lastreplace);
      if(i == bAll) {
        i = use_begin;
        use_begin = CURRENT;
        gtk_editable_set_position(GTK_EDITABLE(curedit->text), begin);
        while(find_text(lastfind, &begin, &end)) {
          replace_text(begin, &end, lastreplace);
          gtk_editable_set_position(GTK_EDITABLE(curedit->text), end);
        }
        use_begin = i;
      }
      select_text(begin, end);
    } else 
      gM_dialog_show(_("Message"), _("Nothing found to replace  "), 1);
    break;
  }
}

void gM_editordialogs_gotoline(editor *e) {
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *tmp;
  GtkWidget *hbb;
  int i;
  char *s;
 
  if(dialog_goto != NULL) /* only one find dialog open at the time */
    return;

  curedit = e;
  s = gtk_editable_get_chars(GTK_EDITABLE(curedit->text), 0, gtk_text_get_length(GTK_TEXT(curedit->text)));
  max_spin = 0;
  for(i = 0; s[i] != 0; i++)
    if(s[i] == 10)
      max_spin++;
  g_free(s);

  dialog_goto = gtk_window_new (GTK_WINDOW_DIALOG);
  gtk_window_set_title (GTK_WINDOW (dialog_goto), _("Goto line:"));
  gtk_window_position (GTK_WINDOW (dialog_goto), GTK_WIN_POS_MOUSE);
  gtk_window_set_policy (GTK_WINDOW (dialog_goto), FALSE, FALSE, TRUE);
  gtk_signal_connect(GTK_OBJECT(dialog_goto), "destroy", GTK_SIGNAL_FUNC(destroy_cb3), NULL);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (dialog_goto), vbox);


  hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, TRUE, 0);

  tmp = gtk_label_new (_("Enter Line number: "));
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (hbox), tmp, TRUE, TRUE, 0);

  spinbutton = gtk_spin_button_new (GTK_ADJUSTMENT(gtk_adjustment_new(1, 1, max_spin, 1, 10, 10)), 1, 0);
  gtk_widget_show (spinbutton);
  gtk_box_pack_start (GTK_BOX (hbox), spinbutton, TRUE, TRUE, 0);
  gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(spinbutton), TRUE);
  gtk_spin_button_set_wrap(GTK_SPIN_BUTTON(spinbutton), FALSE);
  gtk_widget_grab_focus(spinbutton);
                                                       
  tmp = gtk_hseparator_new ();
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (vbox), tmp, FALSE, TRUE, 5);

  hbb = gtk_hbutton_box_new ();
  gtk_widget_show (hbb);
  gtk_box_pack_start (GTK_BOX (vbox), hbb, FALSE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbb), 5);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbb), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbb), 10);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbb), 0, 0);

  tmp = gtk_button_new_with_label (_("Ok"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb3), (void *) OK);

  tmp = gtk_button_new_with_label (_("Cancel"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);   
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb3), (void *) CANCEL);
  
  tmp = gtk_button_new_with_label (_("Help"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb3), (void *) HELP);

  gtk_widget_show(dialog_goto);
}

void destroy_cb3(GtkWidget *w, void *pointer) {
  dialog_goto = NULL;
}

void but_cb3(GtkWidget *butwidget, int but) {
  int i, j, k;
  char *s;
  
  switch(but) {
  case OK:
    i = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton));
  	i--;
    s = gtk_editable_get_chars(GTK_EDITABLE(curedit->text), 0, gtk_text_get_length(GTK_TEXT(curedit->text)));
    k = 0;
    for(j = 0; k < i; j++) {
      if(s[j] == 10)
        k++;
    }
    g_free(s);
    gtk_editable_set_position(GTK_EDITABLE(curedit->text), j);
  case CANCEL:
    gtk_widget_hide(dialog_goto);
    gtk_widget_destroy(dialog_goto);
    break;
  case HELP:
    gM_help_show_goto();
  }
}
