/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "structs.h"
#include "fromval.h"
#include "handy.h"

/* Uncomment for standaline testing:
#define STANDALONE
*/
var    *do_parse(char *s, int *pos);
#define skip_spaces(s) while(s[0] == ' ') (s)++;

#ifdef STANDALONE
#define strnew(str) strcpy(g_malloc(strlen(str) + 1), str)
#endif

var *str_to_var(char *s) {
  int pos = 0;
  return do_parse(s, &pos);
}

char *var_to_str(var *v) {
  char s[20];
  int i;
  var *v2;
  char *ret, *tmp;
  
  switch(v->type) {
  case TYPE_EMPTY:
    return strnew("");
  case TYPE_INT:
    sprintf(s, "%d", v->i);
    return strnew(s);
  case TYPE_FLOAT:
    sprintf(s, "%f", v->f);
    return strnew(s);
  case TYPE_STRING:
    ret = g_malloc(v->strlen + 3);
    sprintf(ret, "\"%s\"", v->str);
    return ret;
  case TYPE_OBJ:
    sprintf(s, "#%d", v->obj);
    return strnew(s);
  case TYPE_ERROR:
    return strnew(v->error);
  case TYPE_LIST:
    v2 = v->list;
    ret = strnew("{");
    for(i = 0; i < v->listlen; i++) {
      tmp = var_to_str(v2);
      ret = g_realloc(ret, strlen(ret) + strlen(tmp) + 3);
      if(i != 0)
        strcat(ret, ", ");
      strcat(ret, tmp);
      g_free(tmp);
      v2 = v2->nextlistitem;
    }
    ret = g_realloc(ret, strlen(ret) + 2);
    strcat(ret, "}");
    return ret;
  }
  /* error, this shouldn't happen */
  return NULL;
}

/******************************
 * The basic parsing function *
 ******************************/

var *do_parse(char *s, int *pos) {
  var *ret = var_malloc();
  var *vartmp;
  char *tmp;
  int i;
  int j;
  int k;

  skip_spaces(s);

  if((s[0] >= '0' && s[0] <= '9') || /* TYPE_INT or TYPE_FLOAT */
     s[0] == '-') {
    tmp = g_malloc(strlen(s) + 1);
    i = j = 0;
    if(s[0] == '-') {
      s++;
      j += 2;
    }
    while((s[i] >= '0' && s[i] <= '9') || s[i] == '.') {
      tmp[i] = s[i];
      if(s[i] == '.') /* a float is found */
        j += 1;
      i++;
    }
    tmp[i] = 0;       /* terminate tmp */
    if(j == 1 || j == 3) {           /* if float */
      ret->type = TYPE_FLOAT;
      ret->f    = atof(tmp);
      if(j == 3) {
        ret->f *= -1;
        i++;
      }
    } else if(j == 0 || j == 2){
      ret->type = TYPE_INT;
      ret->i    = atoi(tmp);
      if(j == 2) {
        ret->i *= -1;
        i++;
      }
    }
    g_free(tmp);
    (*pos) += i;
  } else if(s[0] == '{') {         /* TYPE_LIST */
  	i = 1; /* number of chars skipped */
    j = k = 0; /* number of elements, total skipped chars */
    ret->type = TYPE_LIST;
    vartmp = ret;
    while(1) {
      s += i;
      k += i;
      i = 0;
      while(s[0] == ' ') { /* skip spaces */
        (s)++;
        k++;
      }
      if(s[0] == '}') {
        break;
      } else if(s[0] == ',') {
        s++;
        k++;
      }
      while(s[0] == ' ') { /* skip spaces */
        (s)++;
        k++;
      }
      j++; /* the listitem that will be read */
      if(s[i] != 0 && s[i] != 10) {
        if(j == 1) {
          vartmp->list = do_parse(s + i, &i);  /* recursif function call */
          vartmp = vartmp->list;
        } else {
          vartmp->nextlistitem = do_parse(s + i, &i);  /* recusrif function call */
          vartmp = vartmp->nextlistitem;
        }
      } else {
        j--;
        break;
      }
    }
    ret->listlen = j;  /* done */
    (*pos) += k + 1;
  } else if(s[0] == '\"') {        /* TYPE_STR */
    tmp = g_malloc(strlen(s) + 1);
    i = 1;
    j = 0;
    while(1) {
      if(s[i] == '\\' && (s[i + 1] == '\\' || s[i + 1] == '\"')) {
        tmp[j] = s[i + 1];
        i++;
      } else if(s[i] == '\"') {
        tmp[j] = 0;
        i++; 
        break;
      } else if(s[i] == 0) { /* major error, unterminatred string ! */
        tmp[j] = 0;
        break;
      } else
        tmp[j] = s[i];
      i++;
      j++;
    }
    ret->type = TYPE_STRING;
    ret->str = strnew(tmp);
    ret->strlen = j;
    g_free(tmp);
    (*pos) += i;
  } else if(s[0] == '#') {         /* TYPE_OBJ */
    tmp = g_malloc(strlen(s) + 1);
    i = j = 1;
    if(s[1] == '-') { /* negative object number */
      i = j = 2;
    }
    while(s[i] >= '0' && s[i] <= '9') {
      tmp[i - j] = s[i];
      i++;
    }
    tmp[i - j] = 0;       /* terminate tmp */
    ret->type = TYPE_OBJ;
    ret->obj  = atoi(tmp) * (j == 2 ? -1 : 1);
    g_free(tmp);
    (*pos) += i;
  } else {                         /* must be an TYPE_ERROR */
    tmp = g_malloc(strlen(s) + 1);
    i = 0;
    while(s[i] != ',' &&
          s[i] != ' ' &&
          s[i] != '}') {
      tmp[i] = s[i];
      i++;
    }
    tmp[i] = 0;       /* terminate tmp */
    ret->type = TYPE_ERROR;
    ret->error = strnew(tmp);
    g_free(tmp);
    (*pos) += i;
  }
  return ret;
}

/**************
 * util funcs *
 **************/

var *var_malloc() {
  var *v = g_malloc(sizeof(var));
  v->type = TYPE_EMPTY;
  v->str = NULL;
  v->list = NULL;
  v->nextlistitem = NULL;
  return v;
}

void var_free(var *v) {
  var *v2;
  int i, len;
  
  if(v == NULL)
    return;
  if(v->type == TYPE_LIST) {
    /* free the list first */
    v2 = v;
    len = v->listlen;
    v = v->list;
    g_free(v2);
    for(i = 0; i < len; i++) {
      v2 = v;
      v = v->nextlistitem;
      var_free(v2);
    }
    return; /* the list is completely free()ed */
  }
  if(v->type == TYPE_STRING && v->strlen != 0)
    g_free(v->str);
  g_free(v);
}




void gM_var_print(var *v) {
  char *s = var_to_str(v);
  printf("%s", s);
  g_free(s);
#if 0
  int i, j;
  switch(v->type) {
  case TYPE_INT:
    printf("%d", v->i);
    break;
  case TYPE_FLOAT:
    printf("%f", v->f);
    break;
  case TYPE_STRING:
    printf("\"%s\"", v->str);
    break;
  case TYPE_OBJ:
    printf("#%d", v->obj);
    break;
  case TYPE_LIST:
    printf("{");
    j = v->listlen;
    v = v->list;
    for(i = 0; i < j; i++) {
      gM_var_print(v);
      printf("%s", i == j - 1 ? "" : ", ");
      v = v->nextlistitem;
    }
    printf("}");
    break;
  case TYPE_ERROR:
    printf("%s", v->error);
    break;
  }
#endif
}

#ifdef STANDALONE
int main() {
  var *v;
  char *s = "{1, #0, {#-100, {\"1\",2,#3}, \"\\\"hoi\\12 != 3\\\"\", -1}, #12, \"hoi\", {-12.345}}";
  v = str_to_var(s);

  printf("\n%s =>\n", s);
  gM_var_print(v);
  printf("\n");
  
  return 1;
}
#endif
