/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <time.h>
#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>
#include <glib.h>

#include "handy.h"

int parse_yes_or_no(char *what) {

  if(!what)
    return -1;
  if(strcasecmp(what, "yes") == 0 ||
     strcasecmp(what, "true") == 0 ||
     strcasecmp(what, "on") == 0 ||
     strcasecmp(what, "1") == 0 ||
     strcasecmp(what, "one") == 0 ||
     strcasecmp(what, "ja") == 0)
    return 1;
  else if(strcasecmp(what, "no") == 0 ||
    			strcasecmp(what, "false") == 0 ||
          strcasecmp(what, "off") == 0 ||
          strcasecmp(what, "0") == 0 ||
          strcasecmp(what, "none") == 0 ||
          strcasecmp(what, "nee") == 0)
    return 0;
  else
    return -1;
}

char *get_time() {
  time_t t;
  static char s[9] = "";
  
  t = time(NULL);
  strncpy(s, ctime(&t) + 11, 8);
  s[8] = 0;
  return s;
}

char *get_date() {
  time_t t;
  struct tm *timep;
  static char s[11] = "";
  
  t = time(NULL);
  timep = localtime(&t);
  sprintf(s, "%d-%s%d-%s%d", 1900 + timep->tm_year,
            timep->tm_mon + 1  < 10 ? "0" : "", timep->tm_mon + 1, 
            timep->tm_mday < 10 ? "0" : "", timep->tm_mday);
  return s;
}


char *strtolower(char *string, int len) {
  int i;
  char *ret = g_malloc(len + 1);
  
  for(i = 0; i < len; i++) {
    ret[i] = tolower(string[i]);
  }
  ret[len] = 0;

  return ret;
}

char *strrcasechr(char *s, char c) {
  char *ret = NULL;
  int i = 0;
  for(; s[i] != 0; i++) {
    if(s[i] == c ||
      (s[i] >= 'A' && s[i] <= 'Z' && s[i] == c - 'a' + 'A') ||
      (s[i] >= 'a' && s[i] <= 'z' && s[i] == c + 'a' - 'A'))
      ret = s + i;
  }
  return ret;
}

void strsub(char *s, char replace, char by) {
  int i;

  if(s == NULL)
    return;
  for(i = 0; s[i] != 0; i++)
    if(s[i] == replace)
      s[i] = by;
}


void limit_to_20(int for_what, char *from, char **to) {
  char *pro = NULL;
  int prelen = 0, prolen = 0;

/*printf("Startiung conversion\n");*/
  (*to)[0] = 0;
  if(strlen(from) > 20) {
    if(for_what == FOR_EDITOR) {
      pro = strrchr(from, ':');
      if(pro == NULL)
        for_what = FOR_WORLD;
      else {
        pro++;
        prolen = strlen(pro);
        prelen = pro - from - 1;
        strncpy((*to), from, 20);
        (*to)[prelen > 20 ? 20 : prelen] = 0;
/*printf("----------123456789|123456789|123456789|123456789|123456789|\n");
printf("  *to == \"%s\"\n",*to);
printf("  pro == \"%s\"\n",pro);
printf("1\n");*/
        if(20 - prolen - 4 > 0) {
/*printf("2\n");*/
          if(20 - prolen - 1  < prelen) {
/*printf("3 (starting \"...\" at pos: 20 - prolen - 4 == %d\n",20-prolen-4);*/
            strcpy((*to) + 20 - prolen - 4, "...");
/*  					(*to)[16 - prolen] = '.';
            (*to)[17 - prolen] = '.';
  					(*to)[18 - prolen] = '.';
  					(*to)[19 - prolen] = 0;*/
          }
/*else printf("4\n");*/
        } else {
/*printf("5\n");*/
          strcpy((*to), "...");
        }
/*printf("6\n");*/
        strcat((*to), ":");
        strncat((*to), pro, 20 - strlen((*to)) < prolen ? 20 - strlen((*to)) - 3 : prolen);
        if(strlen((*to)) == 17) {
          strcat((*to), "...");
        }
      }
    }
    if(for_what == FOR_WORLD) {
      strncpy((*to), from, 17);
      ((*to))[17] = 0;
      strcat((*to), "...");
    }
  } else
    strcpy((*to), from);
/*printf("   \"%s\"(%d)\n-> \"%s\"(%d)\n\n", from,strlen(from), (*to),strlen((*to)));*/
}

char *strconcatnew(int argc, ...) {
  char *ret = NULL;
  int i;
  va_list argv;
  char *s;
  int len = 1;

  va_start(argv, argc);
  for(i = 0; i < argc; i++) {
    s = va_arg(argv, char *);
    len += strlen(s);
    ret = g_realloc(ret, len);
    sprintf(ret, "%s%s", i != 0 ? ret : "", s);
  }
  va_end(argv);
 
  return ret;
}

void strappend(char **str, char c) {
  int len;

  if((*str) == NULL) {
    (*str) = g_malloc(2);
    (*str)[0] = c;
    (*str)[1] = 0;
    return;
  }
  len = strlen((*str));
  (*str) = g_realloc((*str), len + 2);
  (*str)[len] = c;
  (*str)[len + 1] = 0;
}

int strendswith(char *s, char *s2, int cases) {
  int i;
  int j;
  if(s[0] == 0) return 0;
  if(s2[0] == 0) return 1;
  for(i = 0; s[i]  != 0; i++);
  for(j = 0; s2[j] != 0; j++);
  if(j > i) return 0;
  i--; j--;
  for(; j > -1 ; j--) {
    if(cases) {
      if(strncmp(s2 + j, s + i, 1) != 0)
        return 0;
    } else { /* not case sensitive */
      if(strncasecmp(s2 + j, s + i, 1) != 0)
        return 0;
    }
    i--;
  }
  return 1;
}

char *strcasestr(char *haystack, char *needle) {
  int len = strlen(needle);
  for(; haystack[0] != 0 && strncasecmp(haystack, needle, len) != 0;
        haystack++);
  return haystack[0] == 0 ? NULL : haystack;
}

char *skip_esc_code(char *s) {
  if(s[0] != 27 || s[1] != '[') return s;
  s += 2;
  for(; (s[0] >= '0' && s[0] <= '9') || s[0] == ';'; s++);
  return s + 1;
}

#if 0 /* Old version */
char *strip_chars(char *s, int len, int *j) {
  int i;
  char *ret = g_malloc(len + 1);
  *j = 0;
  
  for(i = 0; s[i] != 0; i++) {
    if(s[i] == 27) {
      i++;
      if(s[i] == '[') {
        for(i++; (s[i] >= '0' && s[i] <= '9') || s[i] == ';'; i++) {
          if(s[i] == 0)
            break;
        }
      }
    } else if(s[i] > 27) {
      ret[*j] = s[i];
      (*j)++;
    }
  }
  ret[*j] = 0;
#ifdef DEBUG
  if(*j > len) {
    printf("***************************\n***************************\n"
           "MAJOR ERROR IN handy.c:strip_chars() ... the len wasn't "
           "large enough for the string, buffer overflow has occured!"
           "Crash is iminent !!!\n"
           "***************************\n***************************\n"
           "Forcing crash (I hope) to enable gdb debug !!\n";
    strcpy(NULL, "THIS OUGHT TO CAUSE A SIGSEV");
  }
#endif
  return ret;
}
#endif /* 0 */


char *strip_chars(char *s, int len, int *j) {
  int i;
  char *ret = g_malloc(len + 1);
  *j = 0;
  
  for(i = 0; i < len; i++) {
    if(s[i] >= ' ') {
      ret[*j] = s[i];
      (*j)++;
    } else if(s[i] == 27) {
      i++;
      if(s[i] == '[') {
        i++;
        while((s[i] >= '0' && s[i] <= '9') || s[i] == ';') {
          if(s[i] == 0)
            goto endloop;
          i++;
        }
      }
      if(s[i] == 0)
        break;
    }
  }
  endloop:
  ret[*j] = 0;
#ifdef DEBUG
  if(*j > len) {
    printf("***************************\n***************************\n"
           "MAJOR ERROR IN handy.c:strip_chars() ... the len wasn't "
           "large enough for the string, buffer overflow has occured!"
           "Crash is iminent !!!\n"
           "***************************\n***************************\n"
           "Forcing crash (I hope) to enable gdb debug !!\n";
    strcpy(NULL, "THIS OUGHT TO CAUSE A SIGSEV");
  }
#endif
  return ret;
}

char *get_1st_from(char *s) {
  char *ret = g_malloc(strlen(s) + 1);
  int i;
  for(i = 0; 1; i++) {
    ret[i] = s[i];
    if(s[i] == 0)
      return ret;
    if(s[i] == ' ') {
      ret[i] = 0;
      return ret;
    }
  }
}

char *get_point_to_2nd(char *line) {
  for(; 1; line++) {
    if(line[0] == ' ') {
      line++;
      return line;
    } else if(line[0] == 0)
      return line;
  }
  /* line only added to please compilers */
  return line;
}

char *format_for_tcl(char *command, char *ID, char *line) {
  char *ret = g_malloc(strlen(command) + strlen(ID) + 2 * strlen(line) + 5);
  char *p   = get_point_to_2nd(line);
  int i,j;
  sprintf(ret, "%s %s ", command, ID);
  i = p - line;
  j = strlen(ret);
  for(; line[i] != 0; i++) {
    if(line[i] == '\"' ||
       line[i] == '$' ||
       line[i] == '{' ||
       line[i] == '}' ||
       line[i] == '\\' ||
       line[i] == '[' ||
       line[i] == ']') {
      ret[j] = '\\';
      j++;
    }
    ret[j] = line[i];
    j++;
  }
  ret[j] = 0;
  return ret;
}

char *quote_for_tcl(char *line) {
  char *ret = g_malloc(2 * strlen(line) + 5);
  int i,j;
  sprintf(ret, "\"");
  i = 0;
  j = 1;
  for(; line[i] != 0; i++) {
    if(line[i] == '{' ||
       line[i] == '}' ||
       line[i] == '\\' ||
       line[i] == '\"' ||
       line[i] == '[' ||
       line[i] == ']') {
      ret[j] = '\\';
      j++;
    }
    ret[j] = line[i];
    j++;
  }
  ret[j] = '\"';
  ret[j+1] = 0;
  return ret;
}

#if 0 /* testing */
int main() {
  
  printf("=> %d\n", strendswith("Hallo", "lo", 1));
  printf("=> %d\n", strendswith("Hallo", "llfdgfggf", 1));
  printf("=> %d\n", strendswith("Hallo", "LLO", 1));
  printf("=> %d\n", strendswith("Hallo", "aLlO", 0));


  return 0;
}
#endif /* testing */
