/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <gtk/gtk.h> 

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>

#include "structs.h"
#include "help.h"
#include "global_vars.h"
#include "config.h"
#include "window.h"
#include "handy.h"

#define max(a,b) (a > b ? a : b)

void destroy_dialog(GtkWidget *but, GtkWidget *dialog) {
  gtk_widget_destroy(dialog);
}

void gM_help_dialog(char *title, char *data, int maxwidth) {
  GtkWidget *dialog;
  GtkWidget *tmp;
  GtkWidget *text;
  GtkWidget *sb;
  dialog = gtk_dialog_new ();
  gtk_window_set_title(GTK_WINDOW(dialog), title);
  gtk_window_set_policy(GTK_WINDOW(dialog), TRUE, TRUE, FALSE);
  gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
  gtk_widget_show(GTK_DIALOG(dialog)->vbox);
  gtk_widget_show(GTK_DIALOG(dialog)->action_area);
  gtk_container_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 10);
  tmp = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(tmp);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), tmp, TRUE, TRUE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(tmp), 5);
  text = gtk_text_new(NULL, NULL);
  gtk_widget_show(text);
  gtk_box_pack_start(GTK_BOX(tmp), text, TRUE, TRUE, 0);
  gtk_text_insert(GTK_TEXT(text), normalfont, NULL, NULL, data, -1);
  gtk_widget_set_usize(text, gdk_char_width(normalfont, 'm') * maxwidth + 15, 300);
  GTK_WIDGET_UNSET_FLAGS(text, GTK_CAN_FOCUS);
  GTK_WIDGET_UNSET_FLAGS(text, GTK_CAN_DEFAULT);
  sb = gtk_vscrollbar_new(GTK_TEXT(text)->vadj);
  gtk_widget_show(sb);
  gtk_box_pack_start(GTK_BOX(tmp), sb, FALSE, TRUE, 0);
  GTK_WIDGET_UNSET_FLAGS(sb, GTK_CAN_FOCUS);
  GTK_WIDGET_UNSET_FLAGS(sb, GTK_CAN_DEFAULT);
  tmp = gtk_button_new_with_label("   OK   ");
  GTK_WIDGET_SET_FLAGS(tmp, GTK_CAN_DEFAULT);
  GTK_WIDGET_SET_FLAGS(tmp, GTK_HAS_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(destroy_dialog), dialog);
  gtk_widget_show(tmp);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), tmp, TRUE, TRUE, 0);
  gtk_widget_grab_focus(tmp);
  gtk_widget_show(dialog);
}

void show_help_dialog(char *about, char *name) {
#define BUFFER_LEN 100
  GtkWidget *dialog;
  GtkWidget *tmp;
  GtkWidget *text;
  GtkWidget *sb;
  
  char *title;
  char *tmp_s = NULL;
  char *filename;
  char buf[BUFFER_LEN + 1];
  int maxwidth = 0;
  int i;
  FILE *f;
  
  title = strcatnew(_("Help: "), about);
  
  dialog = gtk_dialog_new ();
  gtk_window_set_title(GTK_WINDOW(dialog), title);
  gtk_window_set_policy(GTK_WINDOW(dialog), TRUE, TRUE, FALSE);
  gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
  
  g_free(title);

  gtk_widget_show(GTK_DIALOG(dialog)->vbox);
  gtk_widget_show(GTK_DIALOG(dialog)->action_area);
  gtk_container_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 10);

  tmp = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(tmp);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), tmp, TRUE, TRUE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(tmp), 5);

  text = gtk_text_new(NULL, NULL);
  gtk_widget_show(text);
  gtk_box_pack_start(GTK_BOX(tmp), text, TRUE, TRUE, 0);

  tmp_s = getenv("LANG");
  if(tmp_s == NULL)
    tmp_s = "C";
  for(i = 0; i < strlen(tmp_s); i++)
    tmp_s[i] = toupper(tmp_s[i]);
  filename = strconcatnew(5, SHAREDIR, "/gMOO/Help/", tmp_s, "/", name);
  f = fopen(filename, "r");
  if(f == NULL) {
    g_free(filename);
    tmp_s = "C";
    filename = strconcatnew(5, SHAREDIR, "/gMOO/Help/", tmp_s, "/", name);
    f = fopen(filename, "r");
  }
  if(f != NULL) {
    while(fgets(buf, BUFFER_LEN, f) != NULL) {
      gtk_text_insert(GTK_TEXT(text), normalfont, NULL, NULL, buf, -1);
      maxwidth = maxwidth < strlen(buf) ? strlen(buf) : maxwidth;
    }
    fclose(f);
  } else {
    gtk_text_insert(GTK_TEXT(text), normalfont, NULL, NULL,
      _("Error opening help file:\n  "), -1);
    gtk_text_insert(GTK_TEXT(text), normalfont, NULL, NULL,
      strerror(errno), -1);
    gtk_text_insert(GTK_TEXT(text), normalfont, NULL, NULL,
      "\n\n-> ", -1);
    gtk_text_insert(GTK_TEXT(text), normalfont, NULL, NULL,
      filename, -1);
      
    maxwidth = 50;
  }
  g_free(filename);
  

  gtk_widget_set_usize(text, gdk_char_width(normalfont, 'm') * (maxwidth + 1), 300);
  GTK_WIDGET_UNSET_FLAGS(text, GTK_CAN_FOCUS);
  GTK_WIDGET_UNSET_FLAGS(text, GTK_CAN_DEFAULT);
  
  sb = gtk_vscrollbar_new(GTK_TEXT(text)->vadj);
  gtk_widget_show(sb);
  gtk_box_pack_start(GTK_BOX(tmp), sb, FALSE, TRUE, 0);
  GTK_WIDGET_UNSET_FLAGS(sb, GTK_CAN_FOCUS);
  GTK_WIDGET_UNSET_FLAGS(sb, GTK_CAN_DEFAULT);
  
  tmp = gtk_button_new_with_label("   OK   ");
  GTK_WIDGET_SET_FLAGS(tmp, GTK_CAN_DEFAULT);
  GTK_WIDGET_SET_FLAGS(tmp, GTK_HAS_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(destroy_dialog), dialog);
  gtk_widget_show(tmp);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), tmp, TRUE, TRUE, 0);
  gtk_widget_grab_focus(tmp);
  
  gtk_widget_show(dialog);
}

void gM_help_show_about() {
  gM_help_dialog("About: gMOO", 
                 "gMOO version "VERSION":\n"
                 "\n"
                 "gMOO is a to be flexible MOO\n"
                 "and MUD client. (More MOO then\n"
                 "MUD though)\n"
                 "\n"
                 "Author / Thanks to:\n"
                 "  Gert \"Gs2\" Scholten\n"
                 "  - Main programmer\n"
                 "  David \"Vello\" de Kloet\n"
                 "  - creator of the syntax highlighter\n"
                 "  Paul \"Magmemion\" van Tilburg\n"
                 "  - main beta tester\n"
                 "  - debian package maintainer\n"
                 "  - biggest fan :)\n"
                 "\n"
                 "\n"
                 "It is still in a development\n"
                 "version, but features are coming\n"
                 "along nicely :)"
                 , 42);
}

void gM_help_show_about_world(world *w) {
  char text[100];
  mcpicon *icon;
  int i,len;
  GtkWidget *dialog;
  GtkWidget *tmp;
  GtkWidget *textw;
  GtkWidget *sb;

  dialog = gtk_dialog_new ();
  gtk_window_set_title(GTK_WINDOW(dialog), "About current world:");
  gtk_window_set_policy(GTK_WINDOW(dialog), TRUE, TRUE, FALSE);
  gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
  gtk_widget_set_usize(dialog, 450, 450);

  gtk_widget_show(GTK_DIALOG(dialog)->vbox);
  gtk_widget_show(GTK_DIALOG(dialog)->action_area);
  gtk_container_border_width(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 10);

  tmp = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(tmp);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), tmp, TRUE, TRUE, 0);

  textw = gtk_text_new(NULL, NULL);
  gtk_widget_show(textw);
  gtk_box_pack_start(GTK_BOX(tmp), textw, TRUE, TRUE, 0);
  
  sb = gtk_vscrollbar_new(GTK_TEXT(textw)->vadj);
  gtk_widget_show(sb);
  gtk_box_pack_start(GTK_BOX(tmp), sb, FALSE, TRUE, 0);
  
  tmp = gtk_button_new_with_label("   OK   ");
  GTK_WIDGET_SET_FLAGS(tmp, GTK_CAN_DEFAULT);
  GTK_WIDGET_SET_FLAGS(tmp, GTK_HAS_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(destroy_dialog), dialog);
  gtk_widget_show(tmp);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), tmp, TRUE, TRUE, 0);
  gtk_widget_grab_focus(tmp);
  

  if(w == NULL)
     w = gM_window_get_current_editor_world();
  
  if(w == NULL) {
    gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, "No world has been opened", -1);
  } else {
    /* World data */
    sprintf(text, ":%d", w->port);
    gtk_text_insert(GTK_TEXT(textw), boldfont, NULL, NULL,     "World data:", -1);
    gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, "\n Worldname : ", -1);
    gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL,                w->name, -1);
    gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, "\n Location  : ", -1);
    gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL,                w->host, -1);
    gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL,                text, -1);
    gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, "\n Username  : ", -1);
    gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, w->user != NULL ? w->user : "unknown", -1);
    gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, "\n Connected : ", -1);
    if(w->sock > 0)
      gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[GREEN], NULL, "Yes\n", -1);
    else {
      gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[RED], NULL, "No\n", -1);
      gtk_widget_set_usize(textw, 350, 200);
      gtk_widget_show(dialog);
      return;
    }

    /* Editor data */
    gtk_text_insert(GTK_TEXT(textw), boldfont, NULL, NULL,     "\nEditor data:\n", -1);
    gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL,   " Editors open: ", -1);
    if(w->editcount == 0) {
      gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[RED], NULL, "None\n", -1);
    } else {
      sprintf(text, "%d\n", w->editcount);
      gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, text, -1);
      for(i = 0; i < w->editcount; i++) {
        sprintf(text, "%d", i + 1);
      gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, " Editor ", -1);
      gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, text, -1);
      gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, " is editing \"", -1);
      gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL,
                      w->edit[i]->simpleedit == NULL ? 
                      w->edit[i]->verbname : w->edit[i]->simpleedit->name, -1);
      gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, "\"", -1);
      gtk_text_insert(GTK_TEXT(textw), boldfont, NULL, NULL,
                      w->edit[i]->has_changed > 0 ? " (changed)" : "", -1);
      gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, "\n", -1);
      }
    }
    
    /* MCP data */
    gtk_text_insert(GTK_TEXT(textw), boldfont, NULL, NULL,     "\nMCP data:\n", -1);
    gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL,   " MCP support in this MOO      : ", -1);
    if(w->session == NULL) {
      gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[RED], NULL, "No\n", -1);
    } else {
      gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[GREEN], NULL, "Yes ", -1);
      sprintf(text, "(version %d.%d)\n", w->session->mcpversion[0], w->session->mcpversion[1]);
      gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, text, -1);

      gtk_text_insert(GTK_TEXT(textw), boldfont, NULL, NULL, " MOO supports:\n", -1);

      gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, " o mcp-negotiate              : ", -1);
      if(w->session->negotiateversion != NULL) {
        sprintf(text, "(version %d.%d)\n", w->session->negotiateversion[0], w->session->negotiateversion[1]);
        gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[GREEN], NULL,   "Yes ", -1);
        gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, text, -1);
      } else
        gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[RED], NULL, "No\n", -1);

      gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, " o dns-org-mud-moo-simpleedit : ", -1);
      if(w->session->simpleeditversion != NULL) {
        sprintf(text, "(version %d.%d)\n", w->session->simpleeditversion[0], w->session->simpleeditversion[1]);
        gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[GREEN], NULL, "Yes ", -1);
        gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, text, -1);
      } else
        gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[RED], NULL, "No\n", -1);

      gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, " o dns-nl-vgmoo-client        : ", -1);
      if(w->session->vgmoo_clientversion != NULL) {
        sprintf(text, "(version %d.%d)\n", w->session->vgmoo_clientversion[0], w->session->vgmoo_clientversion[1]);
        gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[GREEN], NULL, "Yes ", -1);
        gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, text, -1);
      } else
        gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[RED], NULL, "No\n", -1);

      gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, " o dns-nl-vgmoo-userlist      : ", -1);
      if(w->session->userlistversion != NULL) {
        sprintf(text, "(version %d.%d)\n", w->session->userlistversion[0], w->session->userlistversion[1]);
        gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[GREEN], NULL, "Yes ", -1);
        gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, text, -1);
        if(w->session->userlist->iconcount > 0) {
          sprintf(text, "%d", w->session->userlist->iconcount);
          gtk_text_insert(GTK_TEXT(textw), boldfont, NULL, NULL, "   MOO requested ", -1);
          gtk_text_insert(GTK_TEXT(textw), boldfont, NULL, NULL, text, -1);
          gtk_text_insert(GTK_TEXT(textw), boldfont, NULL, NULL, " icons:\n", -1);
          icon = w->session->userlist->icons;
          len = 0;
          for(i = 0; i < w->session->userlist->iconcount; i++) {
            len = max(len, strlen(icon->name));
            icon = icon->next;
          }
          icon = w->session->userlist->icons;
          for(i = 0; i < w->session->userlist->iconcount; i++) {
            gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, "   - ", -1);
            sprintf(text, "%s                                           ", icon->name);
            text[len + 1] = 0;
            gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, text, -1);
            gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, ": ", -1);
            if(icon->pixmap == NULL)
              gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[RED], NULL, "Not found\n", -1);
            else
              gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[GREEN], NULL, "Found\n", -1);
            icon = icon->next;
          }
        } else
          gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL,"   MOO requested no icons\n", -1);
        gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, "   Your object number is: ", -1);
        if(w->session->userlist->player_obj < 0) {
          gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[RED], NULL, "not (yet) specified\n", -1);
        } else {
          sprintf(text, "#%d\n", w->session->userlist->player_obj);
          gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[GREEN], NULL, text, -1);
        }
        if(w->session->userlist->friendscount > 0) {
          sprintf(text, "%d", w->session->userlist->friendscount);
          gtk_text_insert(GTK_TEXT(textw), boldfont, NULL, NULL, "   MOO specified ", -1);
          gtk_text_insert(GTK_TEXT(textw), boldfont, NULL, NULL, text, -1);
          gtk_text_insert(GTK_TEXT(textw), boldfont, NULL, NULL, " friends:\n", -1);
          sprintf(text, "#%d", w->session->userlist->friends[0]);
          gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, "   - ", -1);
          gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, text, -1);
          for(i = 1; i < w->session->userlist->friendscount; i++) {
            sprintf(text, "#%d", w->session->userlist->friends[i]);
            gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, ", ", -1);
            gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, text, -1);
          }
          gtk_text_insert(GTK_TEXT(textw), normalfont, NULL, NULL, "\n", -1);
        } else
          gtk_text_insert(GTK_TEXT(textw), normalfont, &colors[BLUE], NULL, "   MOO specified no friends\n", -1);
      } else
        gtk_text_insert(GTK_TEXT(textw), boldfont, &colors[RED], NULL, "No\n", -1);
    }
  }

  gtk_widget_set_usize(textw, 350, 200);
  gtk_widget_show(dialog);
}


void gM_help_show_index() {
  show_help_dialog(_("Index"), "index.help");
}

void gM_help_show_new_and_properties() {
  show_help_dialog(_("New/Properties"), "properties.help");
}

void gM_help_show_open() {
  show_help_dialog(_("Open"), "open.help");
}

void gM_help_show_pref_history() {
  show_help_dialog(_("Preferences: History"), "pref_hist.help");
}
void gM_help_show_pref_logs() {
  show_help_dialog(_("Preferences: Logs"), "pref_logs.help");
}
void gM_help_show_pref_beeptype() {
  show_help_dialog(_("Preferences: General"), "pref_general.help");
}

void gM_help_show_pref_visuals() {
  show_help_dialog(_("Preferences: Interface"), "pref_interface.help");
}
void gM_help_show_pref_window() {
  show_help_dialog(_("Preferences: Window"), "pref_window.help");
}
void gM_help_show_pref_worldhandling() {
  show_help_dialog(_("Preferences: World handling"), "pref_world.help");
}
void gM_help_show_pref_tabs() {
  show_help_dialog(_("Preferences: Tabs"), "pref_tabs.help");
}

void gM_help_show_pref_keypad() {
  show_help_dialog(_("Preferences: Keypad"), "pref_keypad.help");
}


void gM_help_show_pref_fonts() {
  show_help_dialog(_("Preferences: Display"), "pref_dfonts.help");
}
void gM_help_show_pref_dcolors() {
  show_help_dialog(_("Preferences: Display"), "pref_dcolors.help");
}

void gM_help_show_pref_autoindent() {
  show_help_dialog(_("Preferences: Editor"), "pref_editor.help");
}
void gM_help_show_pref_ecolors() {
  show_help_dialog(_("Preferences: Editor"), "pref_ecolor.help");
}

void gM_help_show_pref_tcl() {
#ifdef TCL
  show_help_dialog(_("Preferences: Tcl"), "pref_tcl.help");
#else
  show_help_dialog(_("Preferences: Tcl"), "pref_notcl.help");
#endif
}

void gM_help_show_pref_externals() {
  show_help_dialog(_("Preferences: Externals"), "pref_externals.help");
}


void gM_help_show_find() {
  show_help_dialog(_("Editor: Find"), "edit_find.help");
}

void gM_help_show_replace() {
  show_help_dialog(_("Editor: Replace"), "edit_replace.help");
}

void gM_help_show_goto() {
  show_help_dialog(_("Editor: Goto"), "edit_goto.help");
}

void gM_help_show_send() {
  show_help_dialog(_("World: Send"), "world_send.help");
}

void gM_help_show_config_info() {
  char *text = g_malloc(3 * strlen(PREFIX) + 3 * strlen(config_dir) + 512);
  sprintf(text,  "gMOO version "VERSION":\n"
                 "\n"
                 "gMOO is configured to be installed in:\n"
                 " "PREFIX"/bin/gMOO\n\n"
                 
                 "gMOO will store user specific config data in:\n"
                 " %s\n\n"
                 
                 "gMOO will search these dirs in following\n"
                 "order to find requested userlist pixmaps:\n"
                 " %s/Pixmaps\n"
                 " "SHAREDIR"/gMOO/Pixmaps\n\n"
                 
                 "gMOO will get the help files from:\n"
                 " "SHAREDIR"/gMOO/Help/$LANG/\n\n"
                 
                 "%s%s%s"
                 , config_dir, config_dir,
#ifdef TCL
                 "gMOO will run the tcl scrips found in these\n"
                 "dirs in the following order, to let the user\n"
                 "overwrite macros:\n"
                 " "SHAREDIR"/gMOO/Scripts\n"
                 " ",config_dir,"/Scripts\n"
#else
                 "gMOO won't look for scripts, since Tcl was\n"
                 "disabled/not found\n","",""
#endif
                 );

  gM_help_dialog("Help: Configuration info", text, 50);
  g_free(text);
}

void gM_help_show_keyboard() {
  show_help_dialog(_("Keyboard"), "keyboard.help");
}

void gM_help_show_connect() {
  show_help_dialog(_("Connect dialog"), "connect.help");
}

void gM_help_show_triggers() {
  show_help_dialog(_("Triggers & highlites"), "triggers.help");
}


