/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include "handy.h"
#include "structs.h"
#include "history.h"
#include "config.h"

#define text_get_text(text) \
gtk_editable_get_chars(GTK_EDITABLE(text), 0, gtk_text_get_length(text))
#define text_set_text(text, s) \
gtk_editable_delete_text(GTK_EDITABLE(text), 0, gtk_text_get_length(text));\
gtk_text_insert(text, NULL, NULL, NULL, s, -1)

void gM_history_add(world *w) {
  char *s;
  if(use_hist < 1)
    return;

  if(multiline > 0)
    s = text_get_text(GTK_TEXT(w->input));
  else
    s = gtk_entry_get_text(GTK_ENTRY(w->input));
  if(s[0] == 0) /* don't add empty strings */
    return;
  if(w->hist == NULL) { /* add basic struct */
    w->hist = g_malloc(sizeof(history));
    w->hist->next = w->hist->prev = NULL;
    w->hist->s = NULL;
  } else {
    while(w->hist->next != NULL) {
      w->hist = w->hist->next;
    }
    if(strcmp(w->hist->prev->s, s) == 0)
      return;
  }
  w->hist->next = g_malloc(sizeof(history));
  w->hist->next->prev = w->hist;
  w->hist->next->next = NULL;
  w->hist->next->s = NULL;
  w->hist->s = multiline > 0 ? s : strnew(s);
  w->hist = w->hist->next;
}

void gM_history_up(world *w) {
  history *h;
  
  if(use_hist < 1 || w->hist == NULL || w->hist->prev == NULL)
    return;
  if(w->hist->s == NULL) {
    if(w->selection != NULL)
      g_free(w->selection);
    w->selection = multiline > 0 ? text_get_text(GTK_TEXT(w->input)) :
                   strnew(gtk_entry_get_text(GTK_ENTRY(w->input)));
  }
  if(use_hist < 2 || w->selection == NULL) /* normal hist */
    w->hist = w->hist->prev;
  else { /* selective history */
    for(h = w->hist->prev; h != NULL; h = h->prev)
      if(strncasecmp(h->s, w->selection, strlen(w->selection)) == 0)
        break;
    if(h != NULL)
      w->hist = h;
  }
  if(multiline > 0) {
    text_set_text(GTK_TEXT(w->input), w->hist->s == NULL ? w->selection : w->hist->s);
  } else
    gtk_entry_set_text(GTK_ENTRY(w->input), w->hist->s == NULL ? w->selection : w->hist->s);
}

void gM_history_down(world *w) {
  history *h;

  
  if(use_hist < 1 || w->hist == NULL || w->hist->next == NULL)
    return;
  if(use_hist < 2 || w->selection == NULL) /* normal hist */
    w->hist = w->hist->next;
  else { /* selective history */
    for(h = w->hist->next; h != NULL; h = h->next)
      if(h->s == NULL || strncasecmp(h->s, w->selection, strlen(w->selection)) == 0)
        break;
    if(h != NULL)
      w->hist = h;
  }
  if(multiline > 0) {
    text_set_text(GTK_TEXT(w->input), w->hist->s == NULL ? w->selection : w->hist->s);
  } else
    gtk_entry_set_text(GTK_ENTRY(w->input), w->hist->s == NULL ? w->selection : w->hist->s);
}

void gM_history_godown(world *w) {

  if(multiline > 0) {
    text_set_text(GTK_TEXT(w->input), "");
  } else
    gtk_entry_set_text(GTK_ENTRY(w->input), "");
  if(w->hist != NULL)
    for(; w->hist->next != NULL; w->hist = w->hist->next);
}

void gM_history_purge(world *w) {
  history *h;

  gM_history_godown(w);
  h = w->hist;
  w->hist = NULL;

  if(w->selection != NULL)
    g_free(w->selection);

  if(h == NULL)
    return;

  while(h->prev != NULL) {
    h = h->prev;
    if(h->next->s != NULL)
      g_free(h->next->s);
    g_free(h->next);
  }
  if(h->s != NULL)
    g_free(h->s);
  g_free(h);
}

