/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
 
#include "structs.h"
#include "global_vars.h"
#include "editor.h"
#include "handy.h"
#include "mcp.h"
#include "world.h"
#include "userlist.h"
#include "config.h"
#include "window.h"

/* UserListSession defines */
#define ulsinit   0
#define ulsicons  1
#define ulsfields 2
#define ulsupdate 3

#define AUTHKEY_LENGTH 5
  
int mcpmin[2]          = {2, 1};
int mcpmax[2]          = {2, 1};
int negotiatemin[2]    = {1, 0};
int negotiatemax[2]    = {2, 0};
int simpleeditmin[2]   = {1, 0};
int simpleeditmax[2]   = {1, 0};
int userlistmin[2]     = {1, 0};
int userlistmax[2]     = {1, 0};
int vgmoo_clientmin[2] = {1, 0};
int vgmoo_clientmax[2] = {1, 0};

/* MAIN handlers */
void mcp_negotiate_init(world *w, char **words, int count);
void mcp_negotiate_send_all(world *w);
void mcp_negotiate_handle(world *w, char **words, int count);

void mcp_simpleedit_handle(world *w, char **words, int count);
void mcp_simpleedit_add(world *w, char *reference, char *name, char *type, char *datatag);
void mcp_simpleedit_add_text(world *w, mcpsimpleedit *se, char *text);
void mcp_simpleedit_finished(world *w, mcpsimpleedit *se);

void mcp_vgmoo_userlist_handle(world *w, char **words, int count);
mcpuserlist *mcp_vgmoo_userlist_add(mcpsession *session, char *data, int type);

void mcp_vgmoo_client_handle(world *w, char **words, int count);

mcpmultiline *mcp_multiline_new(mcpsession *session, char *datatag, char *nametag);
void mcp_multiline_continue(mcpsession *session, char *line, int len);
void mcp_multiline_end(world *w, char *line, int len);
      
/* UTILS */
char **get_words_from(char *line, int len, int *count);
void   extract_version_from(char *word, int *version);

int   *get_best_version(int *min1, int *max1, int *min2, int *max2); /* returns NULL if versions don't match */
int    compare_version(int *ver1, int *ver2);  /* TRUE if ver1 >= ver2 */
int   *min_version(int *ver1, int *ver2);      /* returns new malloc()ed minimum version */
/* create a new malloc()ed key (or datatag) */
char *new_authkey();
/* fille datatag and text with new malloc()ed data, or NULL on an error */
void get_data_from_multiline(char *line, int len, char **datatag, char **nametag, char **text);
void handle_multiline_text(mcpsession *session, mcpmultiline *ml, char *text);

/* MAIN handlers */
void gM_mcp_do_out_of_band(world *w, char *line, int len) {
  char **words;
  int count;
  int i;

  if(len < 2)
    return;
  if(strncasecmp(line, "* ", 2) == 0) {
    mcp_multiline_continue(w->session, line, len);
    return;
  } else if(strncasecmp(line, ": ", 2) == 0) {
    mcp_multiline_end(w, line, len);
    return;
  }
  
  words = get_words_from(line, len, &count);
  if(words == NULL || count == 0) {
    return;
  } else if(count % 2 == 1) {
    mcp_negotiate_init(w, words, count);
  } else if(w->session != NULL && strcmp(w->session->authkey, words[1]) == 0) {
    if(strncasecmp(words[0], "mcp-negotiate", 13) == 0)
      mcp_negotiate_handle(w, words, count);
    else if(strncasecmp(words[0], "dns-org-mud-moo-simpleedit", 26) == 0)
      mcp_simpleedit_handle(w, words, count);
    else if(strncasecmp(words[0], "dns-nl-vgmoo-userlist", 21) == 0)
      mcp_vgmoo_userlist_handle(w, words, count);
    else if(strncasecmp(words[0], "dns-nl-vgmoo-client", 19) == 0)
      mcp_vgmoo_client_handle(w, words, count);
  }
  for(i = 0; i < count; i++) {
    g_free(words[i]);
  }
  g_free(words);
}

void gM_mcp_close(world *w) {
  mcpsimpleedit *se, *setmp;
  mcpmultiline *ml, *mltmp;
  if(w->session == NULL)
    return;
  g_free(w->session->authkey);
  g_free(w->session->mcpversion);
  g_free(w->session->negotiateversion);
  if(w->session->simpleeditversion != NULL)
    g_free(w->session->simpleeditversion);
  if(w->session->userlistversion != NULL)
    g_free(w->session->userlistversion);

  if(w->session->simpleedit != NULL) {
    for(se = w->session->simpleedit; se != NULL; se = setmp) {
      setmp = se->next;
      gM_mcp_simpleedit_free(se);
    }
  }
  if(w->session->multiline != NULL) {
    for(ml = w->session->multiline; ml != NULL; ml = mltmp) {
      mltmp = ml->next;
      gM_mcp_multiline_free(ml);
    }
  }
  if(w->session->userlist != NULL) {
    gM_userlist_free(w->session->userlist);
    gtk_clist_clear(GTK_CLIST(w->userlist));
  }
  g_free(w->session);
  w->session = NULL;
}


void mcp_negotiate_init(world *w, char **words, int count) {
  int i;
  int *min;
  int *max;
  int *version = NULL;

  if(strcasecmp(words[0], "mcp") != 0 /* error, a non authkey cmd that isn't the first line */
     || w->session != NULL) {         /* wow, someone is playing ticks, initiating MCP twice */
    return;
  }
  min = g_malloc(2 * sizeof(int));
  max = g_malloc(2 * sizeof(int));

  min[0] = -1;
  max[0] = -1;

  for(i = 1; i < count; i += 2) {
    if(strcasecmp(words[i], "version:") == 0) {
      extract_version_from(words[i + 1], min);
    } else if(strcasecmp(words[i], "to:") == 0) {
      extract_version_from(words[i + 1], max);
    }
  }
  if(min[0] != -1 || max[0] != -1) { /* check if args where found */
    version = get_best_version(min, max, mcpmin, mcpmax);
    if(version != NULL) { /* versions match */
    	w->session = g_malloc(sizeof(mcpsession));
      w->session->authkey = new_authkey();
    	w->session->mcpversion = version;
      w->session->negotiateversion = NULL;
      w->session->negotiate_done = 0;
      w->session->multiline = NULL;
      w->session->simpleedit = NULL;
      w->session->simpleeditversion = NULL;
      w->session->userlist = NULL;
      w->session->userlistversion = NULL;
      w->session->vgmoo_clientversion = NULL;
      w->session->world = w;
      mcp_negotiate_send_all(w);
    }
  }
  g_free(min);
  g_free(max);
}


void mcp_negotiate_send_all(world *w) {
  char s[70 + 20 + 20]; /* strlen + authlen + packagenamelen */
  sprintf(s, "#$#mcp authentication-key: %s version: %d.%d to: %d.%d\n",
             w->session->authkey, mcpmin[0], mcpmin[1], mcpmax[0], mcpmax[1]);
  gM_world_write(w, s);

  sprintf(s, "#$#mcp-negotiate-can %s package: "
             "mcp-negotiate min-version: %d.%d max-version: %d.%d\n",
             w->session->authkey,
             negotiatemin[0], negotiatemin[1], negotiatemax[0], negotiatemin[1]);
  gM_world_write(w, s);

  sprintf(s, "#$#mcp-negotiate-can %s package: dns-org-mud-moo-simpleedit "
             "min-version: %d.%d max-version: %d.%d\n",
             w->session->authkey,
             simpleeditmin[0], simpleeditmin[1], simpleeditmax[0], simpleeditmax[1]);
  gM_world_write(w, s);

  sprintf(s, "#$#mcp-negotiate-can %s package: dns-nl-vgmoo-userlist "
             "min-version: %d.%d max-version: %d.%d\n",
             w->session->authkey,
             userlistmin[0], userlistmin[1], userlistmax[0], userlistmax[1]);
  gM_world_write(w, s);

  sprintf(s, "#$#mcp-negotiate-can %s package: dns-nl-vgmoo-client "
             "min-version: %d.%d max-version: %d.%d\n",
             w->session->authkey,
             vgmoo_clientmin[0], vgmoo_clientmin[1], vgmoo_clientmax[0], vgmoo_clientmax[1]);
  gM_world_write(w, s);

  /* check and send other packages */
  sprintf(s, "#$#mcp-negotiate-end %s\n", w->session->authkey);
  gM_world_write(w, s);
}

void mcp_negotiate_handle(world *w, char **words, int count) {
  int i;
  char *pack = NULL;
  int  *packmin;
  int  *packmax;

  if(w->session->negotiate_done > 0)
    return;
  if(strcasecmp(words[0], "mcp-negotiate-can") == 0) {
    packmin = g_malloc(2 * sizeof(int)); packmin[0] = -1;
    packmax = g_malloc(2 * sizeof(int)); packmax[0] = -1;
    for(i = 2; i < count; i += 2) {
      if(strcasecmp(words[i], "package:") == 0) {
        pack = words[i + 1];
      } else if(strcasecmp(words[i], "min-version:") == 0) {
        extract_version_from(words[i + 1], packmin);
      } else if(strcasecmp(words[i], "max-version:") == 0) {
        extract_version_from(words[i + 1], packmax);
      }
    }
    if(packmin[0] != -1 && packmax[0] != -1 && pack != NULL) {
      if(strcasecmp(pack, "mcp-negotiate") == 0) { /* the mcp-negotiate package */
        w->session->negotiateversion = get_best_version(packmin, packmax, negotiatemin, negotiatemax);
/*printf("Package mcp-negotiate has been added\n"); */
      } else if(strcasecmp(pack, "dns-org-mud-moo-simpleedit") == 0) { /* the simpleedit package */
        w->session->simpleeditversion = get_best_version(packmin, packmax, simpleeditmin, simpleeditmax);
/*printf("Package dns-org-mud-moo-simpleedit has been added\n"); */
      } else if(strcasecmp(pack, "dns-nl-vgmoo-userlist") == 0) { /* the userlist package */
        w->session->userlistversion = get_best_version(packmin, packmax, userlistmin, userlistmax);
      } else if(strcasecmp(pack, "dns-nl-vgmoo-client") == 0) {
        w->session->vgmoo_clientversion = get_best_version(packmin, packmax, vgmoo_clientmin, vgmoo_clientmax);
        gM_mcp_vgmoo_client_info(w->session);
      } /* else check for other packages */
/*else printf("Package %s not added ... don't support it (yet)\n", pack); */
    }
    g_free(packmin);
    g_free(packmax);
  } else if(strcasecmp(words[0], "mcp-negotiate-end") == 0) {
    w->session->negotiate_done = 1;
  }
}

/* SIMPLEEDIT package: */
void mcp_simpleedit_handle(world *w, char **words, int count) {
  char *reference = NULL;
  char *name = NULL;
  char *type = NULL;
  char *datatag = NULL;
  int has_content = 0;
  int i;

  if(w->session->simpleeditversion == NULL) /* packageversions didn't match */
    return;
    
  if(strcasecmp(words[0], "dns-org-mud-moo-simpleedit-content") == 0) {
    for(i = 2; i < count; i += 2) {
      if(strcasecmp(words[i], "reference:") == 0) {
        reference = words[i + 1];
      } else if(strcasecmp(words[i], "name:") == 0) {
        name = words[i + 1];
      } else if(strcasecmp(words[i], "type:") == 0) {
        type =  words[i + 1];
      } else if(strcasecmp(words[i], "content*:") == 0) {
        has_content = 1;
      } else if(strcasecmp(words[i], "_data-tag:") == 0) {
        datatag = words[i + 1];
      }
    }
    if(reference == NULL || name == NULL || type == NULL || has_content == 0 || datatag == NULL)
      return; /* error occured, missing params */
    mcp_simpleedit_add(w, reference, name, type, datatag);
  }
}

void mcp_simpleedit_add(world *w, char *reference, char *name, char *type, char *datatag) {
  mcpsimpleedit *se;
  mcpmultiline *ml;

  if(w->session->simpleedit == NULL) {
    w->session->simpleedit =  g_malloc(sizeof(mcpsimpleedit));
    se = w->session->simpleedit;
  } else {
    for(se = w->session->simpleedit; se->next != NULL; se = se->next);
    se->next = g_malloc(sizeof(mcpsimpleedit));
    se = se->next;
  }
  se->reference = strnew(reference);
  se->name = strnew(name);
  se->type = strnew(type);
  se->next = NULL;
  gM_editor_new_mcp(w, se);

  ml = mcp_multiline_new(w->session, datatag, "content");

  se->ml = ml;
}

void gM_mcp_simpleedit_upload(world *w, mcpsimpleedit *se, char *text, int len) {
  char *tmp;
  char *line;
  char *key;
  int i;
  int j = 0;

  key = new_authkey();
  line = g_malloc(strlen(w->session->authkey) + 
                strlen(se->reference) + 
                strlen(se->type) +
                strlen(key) + 100);
  sprintf(line, "#$#dns-org-mud-moo-simpleedit-set %s "
                "reference: \"%s\" "
                "type: \"%s\" "
                "content*: \"\" "
                "_data-tag: %s \n",
                w->session->authkey,
                se->reference,
                se->type,
                key);
  gM_world_write(w, line);
  g_free(line);

  tmp = g_malloc(len + 1);
  for(i = 0; i < len; i++) {
    tmp[j] = text[i];
    if(text[i] == '\n') {
      tmp[j] = 0;
      line = g_malloc(strlen(key) + strlen(tmp) + 20);
      sprintf(line, "#$#* %s content: %s\n", key, tmp);
      gM_world_write(w, line);
      g_free(line);
      j = -1;
    }
    j++;
  }
  if(j != 0) {
    tmp[j] = 0;
    line = g_malloc(strlen(key) + strlen(tmp) + 20);
    sprintf(line, "#$#* %s content: %s\n", key, tmp);
    gM_world_write(w, line);
    g_free(line);
  }
  g_free(tmp);
  line = g_malloc(strlen(key) + 10);
  sprintf(line, "#$#: %s\n", key);
  gM_world_write(w, line);

  g_free(line);
  g_free(key);
}

void mcp_simpleedit_add_text(world *w, mcpsimpleedit *se, char *text) {
  gM_editor_add(se->editor, text, strlen(text));
}

void mcp_simpleedit_finished(world *w, mcpsimpleedit *se) {
  gM_editor_data_received(se->editor);

  se->ml = NULL;
}

/* VGMoo userlist package */
void mcp_vgmoo_userlist_handle(world *w, char **words, int count) {
  int icons = 0;
  int fields = 0;
  int d = 0;
  char *datatag = NULL;
  int i;

  if(w->session->userlistversion == NULL) /* packageversions didn't match */
    return;
  if(strcasecmp(words[0], "dns-nl-vgmoo-userlist") == 0) {
    for(i = 2; i < count; i += 2) {
      if(strcasecmp(words[i], "icons*:") == 0) {
        icons = 1;
      } else if(strcasecmp(words[i], "fields*:") == 0) {
        fields = 1;
      } else if(strcasecmp(words[i], "d*:") == 0) {
        d = 1;
      } else if(strcasecmp(words[i], "_data-tag:") == 0) {
        datatag = words[i + 1];
      }
    }
    if(icons == 0 || fields == 0 || d == 0 || datatag == NULL) {
      return; /* error occured, missing params */
    }
    mcp_vgmoo_userlist_add(w->session, NULL, ulsinit);
    w->session->userlist->iconline = mcp_multiline_new(w->session, datatag, "icons");
    w->session->userlist->fieldline = mcp_multiline_new(w->session, datatag, "fields");
    w->session->userlist->listline = mcp_multiline_new(w->session, datatag, "d");
    /* Show the userlist: */
    gtk_notebook_set_page(GTK_NOTEBOOK(notebook), w->page_num);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(userlistmenu), TRUE);
  } else if(strcasecmp(words[0], "dns-nl-vgmoo-userlist-you") == 0) {
    for(i = 2; i < count; i += 2) {
      if(strcasecmp(words[i], "nr:") == 0) {
        mcp_vgmoo_userlist_add(w->session, NULL, ulsinit);
        gM_userlist_player_obj(w->session->userlist, words[i + 1]);
        break;
      }
    }
  } else if(strcasecmp(words[0], "dns-nl-vgmoo-userlist-friends") == 0) {
    for(i = 2; i < count; i += 2) {
      if(strcasecmp(words[i], "friends:") == 0) {
        mcp_vgmoo_userlist_add(w->session, NULL, ulsinit);
        gM_userlist_friends(w->session->userlist, words[i + 1]);
        break;
      }
    }
  } else if(strcasecmp(words[0], "dns-nl-vgmoo-userlist-menu") == 0) {
    for(i = 2; i < count; i += 2) {
      if(strcasecmp(words[i], "menu:") == 0) {
        mcp_vgmoo_userlist_add(w->session, NULL, ulsinit);
        gM_userlist_add_menu(w->session->userlist, words[i + 1]);
        break;
      }
    }
  }
}

mcpuserlist *mcp_vgmoo_userlist_add(mcpsession *session, char *data, int type) {
  mcpuserlist *ul;
  
  if(type == ulsinit) {
    if(session->userlist != NULL) {
  		return NULL;
    }
    ul = session->userlist = g_malloc(sizeof(mcpuserlist));
    ul->list = NULL;
    ul->listlen = 0;
    ul->icons = NULL;
    ul->iconcount = 0;
    ul->fields = NULL;
    ul->fieldcount = 0;
    ul->player_obj = -1;
    ul->friendscount = 0;
    ul->friends = NULL;
    ul->menucount = 0;
    ul->menus = NULL;
    return ul;
  } else if(type == ulsicons) {
    gM_userlist_do_icons(session->userlist, data);
  } else if(type == ulsfields) {
    gM_userlist_do_fields(session->userlist, data);
  } else if(type == ulsupdate) {
    gM_userlist_do_users(session->world, data);
  }
  return NULL;
}

/* VGMoo-client package */
void mcp_vgmoo_client_handle(world *w, char **words, int count) {
  /* This functions does nothing, since there coulnd't come a
   * vgmoo-client command from the MOO (one way package).
   * That might change in the future though */
}

void gM_mcp_vgmoo_client_info(mcpsession *session) {
  char text[82 + AUTHKEY_LENGTH + strlen(VERSION)];

  if(session->vgmoo_clientversion == NULL)
    return;
  sprintf(text, "#$#dns-nl-vgmoo-client-info %s "
                "name: gMOO "
                "text-version: \"%s\" "
                "internal-version: 0\n",
                session->authkey,
                VERSION);
  gM_world_write(session->world, text);
}

void gM_mcp_vgmoo_client_screensize(mcpsession *session) {
  char text[80 + AUTHKEY_LENGTH];
  int c,r;

  if(session->vgmoo_clientversion == NULL)
    return;

  c = (session->world->output->allocation.width - 15) / gdk_char_width(normalfont, 'm');
  r = (session->world->output->allocation.height - 15) / gdk_char_height(normalfont, 'J');
  if(c != session->world->outputsize[0] || r != session->world->outputsize[1]) {
    session->world->outputsize[0] = c;
    session->world->outputsize[1] = r;
    sprintf(text, "#$#dns-nl-vgmoo-client-screensize %s "
                  "cols: %d "
                  "rows: %d \n",
                  session->authkey,
                  c,r);
    gM_world_write(session->world, text);
  }
}


/* MULTILINE stuff: */
mcpmultiline *mcp_multiline_new(mcpsession *session, char *datatag, char *nametag) {
  mcpmultiline *ml;

  if(session->multiline == NULL) {
    session->multiline = g_malloc(sizeof(mcpmultiline));
    ml = session->multiline;
  } else {
    for(ml = session->multiline; ml->next != NULL; ml = ml->next);
    ml->next = g_malloc(sizeof(mcpmultiline));
    ml = ml->next;
  }
  ml->datatag = strnew(datatag);
  ml->nametag = strnew(nametag);
  ml->next = NULL;
  
  return ml;
}

void mcp_multiline_continue(mcpsession *session, char *line, int len) {
  char *datatag;
  char *nametag;
  char *text;
  mcpmultiline *ml;

  if(session == NULL || session->multiline == NULL) /* multiline CAN'T be right */
    return;

  get_data_from_multiline(line, len, &datatag, &nametag, &text);
  if(datatag == NULL || nametag == NULL || text == NULL)
    return;
/*printf("Multiline received with datatag == >>%s<< and %s: >>%s<<\n", datatag, nametag, text);*/
  for(ml = session->multiline; ml != NULL; ml = ml->next) {
    if(strcmp(datatag, ml->datatag) == 0 && strcasecmp(nametag, ml->nametag) == 0) {
      handle_multiline_text(session, ml, text);
      break;
    }
  }
  g_free(nametag);
  g_free(datatag); /* error, the datatag wasn't found in the structs */
  g_free(text);
}


void mcp_multiline_end(world *w, char *line, int len) {
  char *datatag;
  int i, j;
  mcpmultiline *ml = NULL, *tmp = NULL;
  mcpsimpleedit *se = NULL, *setmp = NULL;

  datatag = g_malloc(len + 1);
  for(i = 1; line[i] == ' '; i++);
  for(j = 0; line[i] != ' ' && line[i] != '\n'; i++) {
    datatag[j] = line[i];
    j++;
  }
  datatag[j] = 0;

  for(ml = w->session->multiline; ml != NULL; ml = ml->next) {
    if(strcasecmp(ml->datatag, datatag) == 0) {
      if(w->session->multiline != ml) {
        for(tmp = w->session->multiline; tmp->next != ml; tmp = tmp->next);
        tmp->next = ml->next; /* removed ml from the list */
      } else
        w->session->multiline = ml->next; /* removed ml from the list */
      break;
    }
  }
  /* Now we find the package containing this multiline */
  for(se = w->session->simpleedit; se != NULL; se = se->next) {
    if(se->ml == ml) {
      if(w->session->simpleedit != se) {
        for(setmp = w->session->simpleedit; setmp->next != se; setmp = setmp->next);
        setmp->next = se->next; /* removed se from the list */
      } else
        w->session->simpleedit = se->next; /* removed se from the list */
      break;
    }
  }
  if(se != NULL) {
/*    gM_editor_new_mcp(w, se); */
    mcp_simpleedit_finished(w, se);
  }/*    do this to check for other packages */
/* currently, this is an error: */

	gM_mcp_multiline_free(ml);
  
  g_free(datatag);
}



void gM_mcp_multiline_free(mcpmultiline *ml) {
/*
printf("\e[33mfreeing multiline with:\n datatag == %s\n nametag == %s\e[0m\n",
ml->datatag,ml->nametag);
*/
  g_free(ml->datatag);
  g_free(ml->nametag);
  g_free(ml);
}

void gM_mcp_simpleedit_free(mcpsimpleedit *se) {
  if(se->ml != NULL)
    gM_mcp_multiline_free(se->ml);
  g_free(se->reference);
  g_free(se->name);
  g_free(se->type);
  g_free(se);
}


/*********************************

  UTILS

 *********************************/

char **get_words_from(char *line, int len, int *count) {
  char **tmp;
  char **ret;
  char *s;
  int i, j;

  tmp = g_malloc((len + 1) * sizeof(char *));
  s = g_malloc(len + 1);
  j = 0; *count = 0;
  for(i = 0; i < len; i++) {
    if(line[i] == ' '	|| i == len - 1) {
      if(i == len - 1 && line[i] != ' ') {
        s[j] = line[i];
        j++;
      }
      s[j] = 0;
      tmp[*count] = s;
      s = g_malloc(len - j + 1);
      j = 0;
      (*count)++;
      while(line[i + 1] == ' ')
        i++;
    } else if(line[i] == '\"') {
      if(j != 0) { /* error, int the line */
        for(j = 0; j < *count; j++)
          g_free(tmp[j]);
        g_free(s);
/*printf("ERROR: \" encountert wasn't first char of a word (j == %d)\n", j);*/
        return NULL;
      }
      i++;
      while(line[i] != 0) { /* line[i] should never be 0 */
        if(line[i] == '\\') {
          i++;
          switch(line[i]) {
          case '\\':
            s[j] = '\\';
            break;
          case '\"':      
           s[j] = '\"';   
            break;
          default: /* ERROR */
/*printf("ERROR: wrong \\ at pos %d\n", i);*/
             for(j = 0; j < *count; j++)
               g_free(tmp[j]);
             g_free(s);
             return NULL;
          }
        } else if(line[i] == '"') {
          break;
        } else {
          s[j] = line[i];
        }
        j++;
        i++;
      }
      if(line[i] == 0) {/* ERROR */
/*printf("ERROR: unterminated string\n");*/
         for(j = 0; j < *count; j++)
           g_free(tmp[j]);
         g_free(s);
         return NULL;
      }
      /* now add s with the next loop */
    } else {
      s[j] = line[i];
      j++;
    }
  }  /* tmp -> ret */
  ret = g_malloc((*count) * sizeof(char *));
  for(i = 0; i < *count; i++) {
    ret[i] = tmp[i];
  }
  g_free(tmp);
  return ret;
}


void extract_version_from(char *word, int *version) {
  int len = strlen(word);
  char *tmp = g_malloc(len + 1);
  int i, j = 0;
  int minor = 0;
  
  for(i = 0; i < len; i++) {
    if(word[i] == '.' && minor == 0) {
      minor = 1;
      tmp[j] = 0;
      version[0] = atoi(tmp);
      j = 0;
    } else if(word[i] >= '0' && word[i] <= '9') {
      tmp[j] = word[i];
      j++;
    } else { /* error, not a version number */
      g_free(tmp);
      version[0] = -1;
      version[1] = -1;
      return;
    }
  }
  tmp[j] = 0;
  version[1] = atoi(tmp);
  g_free(tmp);
/*printf("extract_version_from \"%s\" -> {%d, %d}\n", word, version[0], version[1]);*/
}


int *get_best_version(int *min1, int *max1, int *min2, int *max2) {
  if(compare_version(max2, min1) && compare_version(max1, min2))
    return min_version(max1, max2);
  else
     return NULL;
}


/* TRUE if ver1 >= ver2 */
int compare_version(int *ver1, int *ver2) {
  if(ver1[0] > ver2[0])
    return TRUE;
  else if(ver1[0] == ver2[0] && ver1[1] >= ver2[1])
    return TRUE;
  return FALSE;
}

/* returns new g_malloc()ed minimum version */
int *min_version(int *ver1, int *ver2) {
  int *r = g_malloc(2 * sizeof(int));
  
  if(ver1[0] > ver2[0] || (ver1[0] == ver2[0] && ver1[1] >= ver2[1])) {
    r[0] = ver1[0];
    r[1] = ver1[1];
  } else {
    r[0] = ver2[0];
    r[1] = ver2[1];
  }
  return r;
}

char *new_authkey() {
/* Mo old func, now replaced ...
  char *s = malloc(15);
  srandom((int) time(NULL));
  sprintf(s, "%d", random());
  return s;
*/
  char *s = g_malloc(AUTHKEY_LENGTH + 1);
  int i;
  
  for(i = 0; i < AUTHKEY_LENGTH; i++) {
    s[i] = (rand() % 94) + '!';
    if(strchr("\"\\!:$",s[i]))
      i--;
  }
  s[i] = 0;
  return s;
}

/* Capi's C++ key generator routine
CHString CHCoder::StGenString(int len)
{
	int c;
	CHString Key;
	while(len-- > 0)
	{
		c = (rand() % 94)+ (int)'!';
		if(strchr("\"\\!:$",c))
			len++;
		else
			Key += (char)c;
	}
	return Key;
}
*/

void get_data_from_multiline(char *line, int len, char **datatag, char **nametag, char **text) {
  int i, j;
  char *tmp;
  
  for(i = 1; line[i] == ' '; i++);
  for(j = i; line[j] != ' ' && line[j] != 0; j++);
  *datatag = g_malloc(j - i + 1);
  strncpy(*datatag, line + i, j - i);
  (*datatag)[j - i] = 0;
  for(i = j; line[i] == ' '; i++);

  tmp = g_malloc(len - i + 1);
  for(j = 0; line[i] != ' '; i++) {  /*  && line[j] != 0 */
    tmp[j] = line[i];
    j++;
  }
  tmp[j - 1] = 0;
  if(line[i - 1] != ':') { /* error */
    g_free(*datatag);
    g_free(tmp);
    *datatag = *nametag = *text = NULL;
    return;
  }
  *nametag = strnew(tmp);
  g_free(tmp);
  i++;
  *text = strnew(line + i);
/*printf("{datatag, nametag, text} == {\"%s\", \"%s\", >>%s<<}\n", *datatag, *nametag, *text);*/
}


void handle_multiline_text(mcpsession *session, mcpmultiline *ml, char *text) {
  mcpsimpleedit *se = NULL;
  mcpuserlist *ul = NULL;
/*printf("adding line %d to a multiline: >>%s<<\n", ml->len + 1, text);*/
  
  if(session->simpleedit != NULL) {
    for(se = session->simpleedit; se != NULL; se = se->next)
      if(se->ml == ml)
        break;
  } else if(session->userlist != NULL) {
    if(session->userlist->listline == ml ||
       session->userlist->iconline == ml ||
       session->userlist->fieldline == ml)
      ul = session->userlist;
  } /* else if( check for other multiline packages */

  /* no handle it: */
  if(se != NULL) {
    mcp_simpleedit_add_text(session->world, se, text);
  }  else if(ul != NULL) {
    if(ul->listline == ml)
      mcp_vgmoo_userlist_add(session, text, ulsupdate);
    else if(ul->iconline == ml)
      mcp_vgmoo_userlist_add(session, text, ulsicons);
    else if(ul->fieldline == ml)
      mcp_vgmoo_userlist_add(session, text, ulsfields);
  }
}


