/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <stdio.h>

#include "structs.h"
#include "preferences.h"
#include "config.h"
#include "syntaxhl.h"
#include "world.h"
#include "worlds.h"
#include "window.h"
#include "handy.h"
#include "help.h"
#include "global_vars.h"

/* global vars */
GtkWidget *dialog = NULL;
GtkWidget *tree;
GtkWidget *titlelabel;
GtkWidget *frame;
GtkWidget *current = NULL;
GtkWidget *butOk, *butApply, *butHelp;
GtkWidget    *section[18];
GtkCTreeNode *sectionnode[18];
char *sectionnames[18];

/* the widgets with the (changed) data: */
GtkWidget *g_h_no, *g_h_simple, *g_h_selective;
GtkWidget *g_l_use, *g_l_append, *g_l_day, *g_l_time, *g_l_saveinput;
GtkWidget *g_b_speaker, *g_b_visual;
GtkWidget *i_v_toolbar, *i_v_seperator, *i_v_definputstyle, *i_v_wordwrap,
          *i_v_dont, *i_v_sec, *i_v_min, *i_v_multi;
GtkWidget *i_w_position, *i_w_size;
GtkWidget *i_h_recon, *i_h_showown, *i_h_color, *i_h_colors[8], *i_h_popup, *i_h_savesend;
GtkWidget *i_t_homo, *i_t_short;
GtkWidget *k_enable, *k_entry[16];
GtkWidget *d_f_normal, *d_f_bold;
GdkColor   d_c_colors[10];
GtkWidget *d_c_preview;
GtkWidget *e_a_enable, *e_a_auto;
GdkColor   e_c_colors[8];
GtkWidget *e_c_preview;
#ifdef TCL
GtkWidget *t_enable, *t_onstdout;
#endif
GtkWidget *x_mail, *x_news;
/* #defines 
  ===========
   0. +-General
   1. | |-History
   2. | |-Logs
   3. | `-Beep type
   4. +-Interface
   5. | |-Visuals
   6. | |-Window
   7. | |-World handling
   8. | `-Tabs
   9. |-Keypad
  10. +-Display
  11. | |-Fonts
  12. | `-Colors
  13. +-Editor
  14. | |-Auto indent
  15. | `-Colors
  16. |-Tcl
  17. `-Externals
  ===========
*/
#define GENERAL        0
#define HISTORY        1
#define LOGS           2
#define BEEPTYPE       3
#define INTERFACE      4
#define VISUALS        5
#define WINDOW         6
#define WORLDHANDLING  7
#define TABS           8
#define KEYPAD         9
#define DISPLAY       10
#define FONTS         11
#define DCOLORS       12
#define EDITOR        13
#define AUTOINDENT    14
#define ECOLORS       15
#define MTCL          16
#define EXTERNALS     17

#define OK    0
#define APPLY 1
#define CLOSE 2
#define HELP  3

/* functions */
void button_callback(GtkWidget *wid, int i);

void something_changed(GtkWidget *dummy, void *dummy2);

void do_apply();
void do_close();
void do_help();

void select_row(GtkWidget *ctree, GtkCTreeNode *rnode, int col, void *pointer);
void tree_insert_all();

void create_widgets();

void update_previews();

void gM_preferences() {
  GtkWidget *vbox;
  GtkWidget *hpaned;
  GtkWidget *scrollw;
  GtkWidget *vbox2;
  GtkWidget *tmp;
  GtkWidget *hbb;
  int        i;

  if(dialog != NULL) return;

  sectionnames[ 0] = _("");
  sectionnames[ 1] =   _("General: History");
  sectionnames[ 2] =   _("General: Logs");
  sectionnames[ 3] =   _("General: Beep type");
  sectionnames[ 4] = _("");
  sectionnames[ 5] =   _("Interface: Visuals");
  sectionnames[ 6] =   _("Interface: Window");
  sectionnames[ 7] =   _("Interface: World handling");
  sectionnames[ 8] =   _("Interface: Tabs");
  sectionnames[ 9] =   _("Keypad");
  sectionnames[10] = _("");
  sectionnames[11] =   _("Display: Fonts");
  sectionnames[12] =   _("Display: Display colors");
  sectionnames[13] = _("");
  sectionnames[14] =   _("Editor: Auto indent");
  sectionnames[15] =   _("Editor: Colors");
  sectionnames[16] = _("Tcl");
  sectionnames[17] = _("Externals");

  dialog = gtk_window_new (GTK_WINDOW_DIALOG);
  gtk_window_set_default_size(GTK_WINDOW(dialog), 600, 350);
  gtk_window_set_title (GTK_WINDOW (dialog), _("Preferences"));
  gtk_window_set_policy (GTK_WINDOW (dialog), TRUE, TRUE, FALSE);
  gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(dialog), "delete_event", GTK_SIGNAL_FUNC(gtk_true), NULL);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (dialog), vbox);

  hpaned = gtk_hpaned_new ();
  gtk_widget_show (hpaned);
  gtk_box_pack_start (GTK_BOX (vbox), hpaned, TRUE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hpaned), 5);

  scrollw = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrollw);
  gtk_container_add (GTK_CONTAINER (hpaned), scrollw);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrollw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  tree = gtk_ctree_new (1, 0);
  gtk_widget_set_usize(tree, 130, -1);
  gtk_widget_show (tree);
  gtk_container_add (GTK_CONTAINER (scrollw), tree);
  gtk_clist_set_column_width (GTK_CLIST (tree), 0, 80);
  gtk_clist_column_titles_hide (GTK_CLIST (tree));
  tree_insert_all();
  gtk_signal_connect(GTK_OBJECT(tree), "tree_select_row", GTK_SIGNAL_FUNC(select_row), NULL);

  vbox2 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox2);
  gtk_container_add (GTK_CONTAINER (hpaned), vbox2);

  titlelabel = gtk_label_new ("");
  gtk_widget_show (titlelabel);
  gtk_box_pack_start (GTK_BOX (vbox2), titlelabel, FALSE, TRUE, 0);
  gtk_misc_set_alignment (GTK_MISC (titlelabel), 0, 0.5);

  frame = gtk_frame_new (NULL);
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (vbox2), frame, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);

  tmp = gtk_hseparator_new ();
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (vbox), tmp, FALSE, FALSE, 0);

  hbb = gtk_hbutton_box_new ();
  gtk_widget_show (hbb);
  gtk_box_pack_start (GTK_BOX (vbox), hbb, FALSE, FALSE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbb), 5);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbb), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbb), 10);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbb), 0, 0);

  butOk = gtk_button_new_with_label (_("Ok"));
  gtk_widget_show (butOk);
  gtk_container_add (GTK_CONTAINER (hbb), butOk);
  GTK_WIDGET_SET_FLAGS (butOk, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(butOk), "clicked", GTK_SIGNAL_FUNC(button_callback), (void *) OK);

  butApply = gtk_button_new_with_label (_("Apply"));
  gtk_widget_show (butApply);
  gtk_container_add (GTK_CONTAINER (hbb), butApply);
  GTK_WIDGET_SET_FLAGS (butApply, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(butApply), "clicked", GTK_SIGNAL_FUNC(button_callback), (void *) APPLY);

  tmp = gtk_button_new_with_label (_("Close"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(button_callback), (void *) CLOSE);

  butHelp = gtk_button_new_with_label (_("Help"));
  gtk_widget_show (butHelp);
  gtk_container_add (GTK_CONTAINER (hbb), butHelp);
  GTK_WIDGET_SET_FLAGS (butHelp, GTK_CAN_DEFAULT);
  gtk_widget_grab_focus (butHelp);
  gtk_widget_grab_default (butHelp);
  gtk_signal_connect(GTK_OBJECT(butHelp), "clicked", GTK_SIGNAL_FUNC(button_callback), (void *) HELP);

  gtk_widget_set_sensitive(butOk,    FALSE);
  gtk_widget_set_sensitive(butApply, FALSE);
  gtk_widget_set_sensitive(butHelp,  FALSE);

  /* duplicate the colors */
  for(i = 0; i < 10; i++) {
    d_c_colors[i].pixel = 0;
    d_c_colors[i].red   = colors[i].red;
    d_c_colors[i].green = colors[i].green;
    d_c_colors[i].blue  = colors[i].blue;
  }
  for(i = 0; i < 8; i++) {
    e_c_colors[i].pixel = 0;
    e_c_colors[i].red   = editcolors[i].red;
    e_c_colors[i].green = editcolors[i].green;
    e_c_colors[i].blue  = editcolors[i].blue;
  }

  /* create all the sections */
  create_widgets();
  
  update_previews();

  gtk_widget_show(dialog);
}

void button_callback(GtkWidget *wid, int i) {
  switch(i) {
  case OK:    do_apply();
              do_close(); break;
  case APPLY: do_apply(); break;
  case CLOSE: do_close(); break;
  case HELP:  do_help();  break;
  }
}

void something_changed(GtkWidget *dummy, void *dummy2) {
  gtk_widget_set_sensitive(butOk,    TRUE);
  gtk_widget_set_sensitive(butApply, TRUE);
}


#define active(toggle)    (GTK_TOGGLE_BUTTON(toggle)->active)
#define set_str_value(var, val)           \
  if(var != NULL)                         \
    g_free(var);                            \
  var = val != NULL ? strnew(val) : NULL
#define set_str_value2(var, val)          \
  if(var != NULL)                         \
    g_free(var);                            \
  var = val[0] != 0 ? strnew(val) : NULL
#define set_font(font, fontstr, value) \
    g_free(fontstr);                     \
    fontstr = strnew(gtk_entry_get_text(GTK_ENTRY(value))); \
    font = gdk_font_load(fontstr)
int get_show_own() {
  int i;
  if(!active(i_h_showown))
    return -1;
  for(i = 0; i < 8; i++)
    if(GTK_OPTION_MENU(i_h_color)->menu_item == i_h_colors[i])
      return i;
  return -1;
}
void set_keypad_values() {
  int i;
  for(i = 0; i < 16; i++) {
    if(keypad_keys[i] != NULL)
      g_free(keypad_keys[i]);
    keypad_keys[i] = strnew(gtk_entry_get_text(GTK_ENTRY(k_entry[i])));
    if(keypad_keys[i][0] == 0) {
      g_free(keypad_keys[i]);
      keypad_keys[i] = NULL;
    }
  }
}
void set_colors() {
  int i;
  for(i = 0; i < 10; i++) {
    colors[i].pixel = 0;
    colors[i].red   = d_c_colors[i].red;
    colors[i].green = d_c_colors[i].green;
    colors[i].blue  = d_c_colors[i].blue;
    if(i < 8) {
      editcolors[i].pixel = 0;
      editcolors[i].red   = e_c_colors[i].red;
      editcolors[i].green = e_c_colors[i].green;
      editcolors[i].blue  = e_c_colors[i].blue;
    }
  }
}
void do_apply() {
/* storing the widget data in the global vars */
  use_hist        = active(g_h_no)     ? 0 :
                    active(g_h_simple) ? 1 :
                                  2;
  logtype         = (active(g_l_use)       ? loguse       : 0) |
                    (active(g_l_append)    ? logappend    : 0) |
                    (active(g_l_day)       ? logday       : 0) |
                    (active(g_l_time)      ? logtime      : 0) |
                    (active(g_l_saveinput) ? logsaveinput : 0);

  beep_type       = (active(g_b_speaker) && active(g_b_visual)) ?
                      beep_type_speaker + beep_type_visual  :
                    active(g_b_speaker) ? beep_type_speaker :
                    active(g_b_visual)  ? beep_type_visual  :
                                    beep_type_none; 
  show_toolbar    = active(i_v_toolbar);
  newmultiline    = active(i_v_multi);
  showsep         = active(i_v_seperator);
  definputstyle   = active(i_v_definputstyle);
  wordwrap        = active(i_v_wordwrap);
  clocktype       = active(i_v_dont) ? 0 :
                    active(i_v_sec)  ? clocksecond :
                                       clockminute;
  save_window_pos = active(i_w_position) ? strnew("") : NULL;
  save_window_size= active(i_w_size)     ? strnew("") : NULL;
  reconnect       = active(i_h_recon);
  popupcon        = active(i_h_popup);
  savesend        = active(i_h_savesend);
  show_own        = get_show_own();
  homo_tabs       = active(i_t_homo);
  short_tab_names = active(i_t_short);
  use_keypad_keys = active(k_enable);
  set_keypad_values();
  set_font(normalfont, normalfontstr, d_f_normal);
  set_font(boldfont,   boldfontstr,   d_f_bold);
  set_colors();
  use_ident       = active(e_a_enable);
  use_auto_detect = active(e_a_auto);
#ifdef TCL
  tcl_debug       = active(t_enable);
  tcl_debug_on_stdout = active(t_onstdout);
#endif
  set_str_value2(ext_mail, gtk_entry_get_text(GTK_ENTRY(x_mail)));
  set_str_value2(ext_news, gtk_entry_get_text(GTK_ENTRY(x_news)));
  
  /* update the real time updatable stuff */
  gtk_widget_set_sensitive(toolbarhandle, show_toolbar);
  gtk_notebook_set_homogeneous_tabs(GTK_NOTEBOOK(notebook), homo_tabs);
  gM_worlds_set_tab_size();
  gM_worlds_show_sep();
  gM_worlds_set_input_color();
  gM_worlds_set_wordwrap(wordwrap);
  gM_worlds_update_clock();
  gM_worlds_update_default_color();

  gM_config_rewrite_main_rc();

/* set the widgets back to unsensitive */
  gtk_widget_set_sensitive(butOk,    FALSE);
  gtk_widget_set_sensitive(butApply, FALSE);
}

void do_close() {
  int i;

  for(i = 0; i < 18; i++) {
    gtk_widget_destroy(section[i]);
  }
  gtk_widget_destroy(dialog);
  dialog = NULL;
  current = NULL;
}

void do_help() {
  int i;
  for(i = 0; i < 18; i++)	if(section[i] == current) break;
  switch(i) {
  case 1:  gM_help_show_pref_history();       break;
  case 2:  gM_help_show_pref_logs();          break;
  case 3:  gM_help_show_pref_beeptype();      break;
  case 5:  gM_help_show_pref_visuals();       break;
  case 6:  gM_help_show_pref_window();        break;
  case 7:  gM_help_show_pref_worldhandling(); break;
  case 8:  gM_help_show_pref_tabs();          break;
  case 9:  gM_help_show_pref_keypad();        break;
  case 11: gM_help_show_pref_fonts();         break;
  case 12: gM_help_show_pref_dcolors();       break;
  case 14: gM_help_show_pref_autoindent();    break;
  case 15: gM_help_show_pref_ecolors();       break;
  case 16: gM_help_show_pref_tcl();           break;
  case 17: gM_help_show_pref_externals();     break;
  }
}

void select_row(GtkWidget *ctree, GtkCTreeNode *rnode, int col, void *pointer) {
  int row;
  for(row = 0; rnode != sectionnode[row]; row++);
  
  if(current != NULL) {
    gtk_widget_hide(current);
    (GTK_OBJECT(current)->ref_count)++;
    gtk_container_remove(GTK_CONTAINER(frame), current);
  }
  current = section[row];
  gtk_container_add(GTK_CONTAINER(frame), current);
  gtk_widget_show(current);

  gtk_label_set_text(GTK_LABEL(titlelabel), sectionnames[row]);

  gtk_widget_set_sensitive(butHelp, sectionnames[row][0] != 0);
}


GtkCTreeNode *ctree_insert_node(char *name, GtkCTreeNode *parent) {
  char *text[1];
  text[0] = name;
  return gtk_ctree_insert_node(GTK_CTREE(tree), parent, NULL, text, 0,
                        NULL, NULL, NULL, NULL, FALSE, TRUE);
}

void tree_insert_all() {
  sectionnode[0]    = ctree_insert_node(_("General"),      NULL);
    sectionnode[1]  = ctree_insert_node(_("History"),        sectionnode[0]);
    sectionnode[2]  = ctree_insert_node(_("Logs"),           sectionnode[0]);
    sectionnode[3]  = ctree_insert_node(_("Beep type"),      sectionnode[0]);
  sectionnode[4]    = ctree_insert_node(_("Interface"),    NULL);
    sectionnode[5]  = ctree_insert_node(_("Visuals"),        sectionnode[4]);
    sectionnode[6]  = ctree_insert_node(_("Window"),         sectionnode[4]);
    sectionnode[7]  = ctree_insert_node(_("World handling"), sectionnode[4]);
    sectionnode[8]  = ctree_insert_node(_("Tabs"),           sectionnode[4]);
  sectionnode[9]    = ctree_insert_node(_("Keypad"),       NULL);
  sectionnode[10]   = ctree_insert_node(_("Display"),      NULL);
    sectionnode[11] = ctree_insert_node(_("Fonts"),          sectionnode[10]);
    sectionnode[12] = ctree_insert_node(_("Colors"),         sectionnode[10]);
  sectionnode[13]   = ctree_insert_node(_("Editor"),       NULL);
    sectionnode[14] = ctree_insert_node(_("Auto indent"),    sectionnode[13]);
    sectionnode[15] = ctree_insert_node(_("Colors"),         sectionnode[13]);
  sectionnode[16]   = ctree_insert_node(_("Tcl"),          NULL);
  sectionnode[17]   = ctree_insert_node(_("Externals"),    NULL);
}

/* callbacks for the create_widget function (toggling and stuff) */
void log_set_sensitive(GtkWidget *check, void *pointer) {
  gtk_widget_set_sensitive(g_l_append, GTK_TOGGLE_BUTTON(check)->active);
  gtk_widget_set_sensitive(g_l_day, GTK_TOGGLE_BUTTON(check)->active);
  gtk_widget_set_sensitive(g_l_time, GTK_TOGGLE_BUTTON(check)->active);
  gtk_widget_set_sensitive(g_l_saveinput, GTK_TOGGLE_BUTTON(check)->active);
}
void color_sensitive(GtkWidget *check, void *pointer) {
  gtk_widget_set_sensitive(i_h_color, GTK_TOGGLE_BUTTON(check)->active);
}
void keypad_set_sensitive(GtkWidget *check, void *pointer) {
  int i;
  for(i = 0; i < 16; i++)
    gtk_widget_set_sensitive(k_entry[i], GTK_TOGGLE_BUTTON(check)->active);
}
void destroy_widget(GtkWidget *dummy, GtkWidget *to_destroy) {
  gtk_widget_hide(to_destroy);
  gtk_widget_destroy(to_destroy);
}
void handle_fs_ok(GtkWidget *ok_but, GtkWidget *entry) {
  GtkWidget *fsd = gtk_widget_get_toplevel(ok_but);
  gtk_entry_set_text(GTK_ENTRY(entry),
    gtk_font_selection_dialog_get_font_name(
      GTK_FONT_SELECTION_DIALOG(fsd)));
  destroy_widget(ok_but, fsd);
}
void font_but_cb(GtkWidget *but, GtkWidget *entry) {
  GtkWidget *fsd;
  fsd = gtk_font_selection_dialog_new(entry == d_f_normal ?
                                      _("Select normal font") :
                                      _("Select bold font"));
  gtk_window_set_modal(GTK_WINDOW(fsd), TRUE);
  gtk_window_position(GTK_WINDOW(fsd), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(GTK_FONT_SELECTION_DIALOG(fsd)->cancel_button),
                     "clicked", GTK_SIGNAL_FUNC(destroy_widget), fsd);
  gtk_signal_connect(GTK_OBJECT(GTK_FONT_SELECTION_DIALOG(fsd)->ok_button),
                     "clicked", GTK_SIGNAL_FUNC(handle_fs_ok), entry);

  gtk_widget_show(fsd);
}
char *get_title_for(GdkColor *color) {
  int i;
  for(i = 0; i != 10 && color != &d_c_colors[i]; i++);
  if(i == 10) { /* it is an editor color */
    for(i = 0; color != &e_c_colors[i]; i++);
    i += HL_STRING;
  }
  switch(i) {
  case BLACK     : return _("Select a color for BLACK");
  case WHITE     : return _("Select a color for WHITE");
  case RED       : return _("Select a color for RED");
  case GREEN     : return _("Select a color for GREEN");
  case BLUE      : return _("Select a color for BLUE");
  case CYAN      : return _("Select a color for CYAN");
  case MAGENTA   : return _("Select a color for MAGENTA");
  case YELLOW    : return _("Select a color for YELLOW");
  case DEFAULTFG : return _("Select a default foreground color");
  case DEFAULTBG : return _("Select a default background color");
  case HL_STRING : return _("Select a color for STRINGS");
  case HL_NUMBER : return _("Select a color for NUMBERS");
  case HL_OBJNUM : return _("Select a color for OBJECTS");
  case HL_DOLLAR : return _("Select a color for $ PROPERTIES");
  case HL_VARIAB : return _("Select a color for VARIABLES");
  case HL_KEYWRD : return _("Select a color for KEYWORDS");
  case HL_SYMBOL : return _("Select a color for SYMBOLS");
  case HL_COMMEN : return _("Select a color for KEYWORDS");
  }
  return _("-error-");
}
void handle_cs_ok(GtkWidget *ok_but, GdkColor *color) {
  GtkWidget *csd = gtk_widget_get_toplevel(ok_but);
  gdouble d[3];
  int i;

  gtk_color_selection_get_color(
    GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(csd)->colorsel), d);

  for(i = 0; i != 10 && color != &d_c_colors[i]; i++);
  if(i == 10) { /* it is an editor color */
    for(i = 0; color != &e_c_colors[i]; i++);
    e_c_colors[i].pixel = 0;
    e_c_colors[i].red   = d[0] * 0xFFFF;
    e_c_colors[i].green = d[1] * 0xFFFF;
    e_c_colors[i].blue  = d[2] * 0xFFFF;
  } else {
    d_c_colors[i].pixel = 0;
    d_c_colors[i].red   = d[0] * 0xFFFF;
    d_c_colors[i].green = d[1] * 0xFFFF;
    d_c_colors[i].blue  = d[2] * 0xFFFF;
  }
  destroy_widget(ok_but, csd);
  something_changed(ok_but, NULL);
  update_previews();
}
void color_but_cb(GtkWidget *but, GdkColor *color) {
  double d[3] = {(double) color->red   / 0x10000,
                 (double) color->green / 0x10000,
                 (double) color->blue  / 0x10000 };
  char  *s    = get_title_for(color);
  GtkWidget *csd;
  
  csd = gtk_color_selection_dialog_new(s);
  gtk_window_position(GTK_WINDOW(csd), GTK_WIN_POS_MOUSE);
  gtk_window_set_modal(GTK_WINDOW(csd), TRUE);
  gtk_widget_hide(GTK_COLOR_SELECTION_DIALOG(csd)->help_button);
  gtk_signal_connect(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(csd)->cancel_button), "clicked",
                     GTK_SIGNAL_FUNC(destroy_widget), csd);
  gtk_signal_connect(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(csd)->ok_button), "clicked", 
                     GTK_SIGNAL_FUNC(handle_cs_ok), color);
  gtk_color_selection_set_color(
    GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(csd)->colorsel), d);

  gtk_widget_show(csd);
}
#ifdef TCL
void onstdout_sensitive(GtkWidget *check, void *pointer) {
  gtk_widget_set_sensitive(t_onstdout, GTK_TOGGLE_BUTTON(check)->active);
}
#endif
void multi_sensetive(GtkWidget *check, void *pointer) {
  gtk_widget_set_sensitive(i_v_seperator,     !GTK_TOGGLE_BUTTON(check)->active);
}
/* macros for the create_widget function */
#define add_def_signal(widget, signal) \
  gtk_signal_connect(GTK_OBJECT(widget), signal, \
  GTK_SIGNAL_FUNC(something_changed), NULL)
#define radio_but(name, widget, group, box, on)                  \
  widget = gtk_radio_button_new_with_label(group, name);         \
  radiogroup = gtk_radio_button_group(GTK_RADIO_BUTTON(widget)); \
  gtk_widget_show(widget);                                       \
  gtk_box_pack_start(GTK_BOX(box), widget, FALSE, TRUE, 0);      \
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), (on))
#define check_but(name, widget, box, on)                         \
  widget = gtk_check_button_new_with_label(name);                \
  gtk_widget_show(widget);                                       \
  gtk_box_pack_start(GTK_BOX(box), widget, FALSE, TRUE, 0);      \
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), (on)); \
  add_def_signal(widget, "clicked")
#define new_menu(name, widget, menu)             \
  widget = gtk_menu_item_new_with_label(name);   \
  gtk_widget_show(widget);                       \
  gtk_container_add(GTK_CONTAINER(menu), widget)
#define keypad_entry(num, x1, x2, y1, y2, name)       \
  k_entry[num] = gtk_entry_new();                       \
  gtk_widget_show(k_entry[num]);                        \
  gtk_table_attach_defaults(GTK_TABLE(table), k_entry[num], x1, x2, y1, y2); \
  gtk_entry_set_text(GTK_ENTRY(k_entry[num]),	keypad_keys[num] != NULL ? keypad_keys[num] : "");      \
  tmp = gtk_label_new(name);                          \
  gtk_widget_show(tmp);                               \
  gtk_table_attach_defaults(GTK_TABLE(table), tmp, x1 - 1, x1, y1, y2); \
  gtk_misc_set_alignment(GTK_MISC(tmp), 1.0, 0.5);    \
  add_def_signal(k_entry[num], "changed")
#define color_but(name, num)                                       \
  tmp = gtk_button_new_with_label(name);                           \
  gtk_widget_show (tmp);                                           \
  gtk_container_add (GTK_CONTAINER (vbb), tmp);                    \
  if(num >= HL_STRING)                                             \
    gtk_signal_connect(GTK_OBJECT(tmp), "clicked",                 \
      GTK_SIGNAL_FUNC(color_but_cb), &e_c_colors[num - HL_STRING]);\
  else                                                             \
    gtk_signal_connect(GTK_OBJECT(tmp), "clicked",                 \
      GTK_SIGNAL_FUNC(color_but_cb), &d_c_colors[num])

/* This collossal function creates al the widgets incl signals for
   all the sections */
void create_widgets() {
  GtkWidget *menu;
  GSList *radiogroup;
  GtkWidget *tmp;
  GtkWidget *table;
  GtkWidget *frame;
  GtkWidget *vbb;

/**** General section: ****/
  	section[GENERAL] = gtk_label_new(""); /* `fake' widget */
/**** General: history ****/
    section[HISTORY]  = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[HISTORY]), 10);
    radiogroup = NULL;
    radio_but(_("No history"),        g_h_no,        radiogroup, section[HISTORY], use_hist  < 1);
    radio_but(_("Simple history"),    g_h_simple,    radiogroup, section[HISTORY], use_hist == 1);
    radio_but(_("Selective history"), g_h_selective, radiogroup, section[HISTORY], use_hist  > 1);
    add_def_signal(g_h_no,        "clicked");
    add_def_signal(g_h_simple,    "clicked");
    add_def_signal(g_h_selective, "clicked");
/**** General: logs ****/
    section[LOGS]     = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[LOGS]), 10);
    radiogroup = NULL;
    check_but(_("Use logs"), g_l_use, section[LOGS], logtype & loguse);
    radio_but(_("Append to a single file"), g_l_append, radiogroup, section[LOGS], logtype & logappend);
    radio_but(_("New log per day"),         g_l_day,    radiogroup, section[LOGS], logtype & logday);
    radio_but(_("New log per connection"),  g_l_time,   radiogroup, section[LOGS], logtype & logtime);
    check_but(_("Log user input"),          g_l_saveinput, section[LOGS], logtype & logsaveinput);
    gtk_signal_connect(GTK_OBJECT(g_l_use), "clicked", GTK_SIGNAL_FUNC(log_set_sensitive), NULL);
    log_set_sensitive(g_l_use, NULL);
    add_def_signal(g_l_append,    "clicked");
    add_def_signal(g_l_day,       "clicked");
    add_def_signal(g_l_time,      "clicked");
    add_def_signal(g_l_saveinput, "clicked");
/**** General: beep type ****/
    section[BEEPTYPE] = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[BEEPTYPE]), 10);
    check_but(_("Beep to the speaker"), g_b_speaker, section[BEEPTYPE], beep_type == beep_type_speaker ||
                                                                     beep_type == beep_type_speaker +
                                                                     beep_type_visual);
    check_but(_("Visual beep"),         g_b_visual,  section[BEEPTYPE], beep_type == beep_type_visual ||
                                                                     beep_type == beep_type_speaker +
                                                                     beep_type_visual);
/**** Interface ****/
  	section[INTERFACE] = gtk_label_new(""); /* `fake' widget */
/**** Interface: Visuals ****/
    section[VISUALS] = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[VISUALS]), 10);
    check_but(_("Show toolbar"),   i_v_toolbar,   section[VISUALS], show_toolbar);
    check_but(_("Multiline input bar   - NOTE: You'll have to restart gMOO to see this :("),
              i_v_multi, section[VISUALS], newmultiline);
    check_but(_("Show seperator"), i_v_seperator, section[VISUALS], showsep);
    check_but(_("Default input style"), i_v_definputstyle, section[VISUALS], definputstyle);
    check_but(_("Word wrap"),   i_v_wordwrap,   section[VISUALS], wordwrap);
    radiogroup = NULL;
    radio_but(_("Don't show counter"),        i_v_dont, radiogroup, section[VISUALS], clocktype == 0);
    radio_but(_("Update counter in seconds"), i_v_sec,  radiogroup, section[VISUALS], clocktype == clocksecond);
    radio_but(_("Update counter in minutes"), i_v_min,  radiogroup, section[VISUALS], clocktype == clockminute);
    add_def_signal(i_v_dont, "clicked");
    add_def_signal(i_v_sec,  "clicked");
    add_def_signal(i_v_min,  "clicked");
    gtk_signal_connect(GTK_OBJECT(i_v_multi), "clicked", GTK_SIGNAL_FUNC(multi_sensetive), NULL);
    multi_sensetive(i_v_multi, NULL);
/**** Interface: Window ****/
    section[WINDOW] = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[WINDOW]), 10);
    check_but(_("Save position on exit"), i_w_position, section[WINDOW], save_window_pos  != NULL);
    check_but(_("Save size on exit"),     i_w_size,     section[WINDOW], save_window_size != NULL);
/**** Interface: World handling ****/
    section[WORLDHANDLING] = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[WORLDHANDLING]), 10);
    check_but(_("Reconnect on disconnect"), i_h_recon,   section[WORLDHANDLING], reconnect);
    check_but(_("Pop up connection dialog, when not autoconnecting"), i_h_popup, section[WORLDHANDLING], popupcon);
    check_but(_("Save settings in send dialog"), i_h_savesend, section[WORLDHANDLING], savesend);
    check_but(_("Show own input"),          i_h_showown, section[WORLDHANDLING], show_own > -1);
    menu = gtk_menu_new();
      gtk_widget_show(menu);
    i_h_color = gtk_option_menu_new();
      gtk_widget_show(i_h_color);
      gtk_box_pack_start(GTK_BOX(section[WORLDHANDLING]), i_h_color, FALSE, TRUE, 0);
      gtk_option_menu_set_menu(GTK_OPTION_MENU(i_h_color), menu);
    new_menu(_("Black"),   i_h_colors[BLACK],   menu);
    new_menu(_("White"),   i_h_colors[WHITE],   menu);
    new_menu(_("Red"),     i_h_colors[RED],     menu);
    new_menu(_("Green"),   i_h_colors[GREEN],   menu);
    new_menu(_("Blue"),    i_h_colors[BLUE],    menu);
    new_menu(_("Cyan"),    i_h_colors[CYAN],    menu);
    new_menu(_("Magenta"), i_h_colors[MAGENTA], menu);
    new_menu(_("Yellow"),  i_h_colors[YELLOW],  menu);
    gtk_option_menu_set_history(GTK_OPTION_MENU(i_h_color), show_own >= 0 ? show_own : YELLOW);
    gtk_signal_connect_after(GTK_OBJECT(i_h_showown), "clicked", GTK_SIGNAL_FUNC(color_sensitive), NULL);
    color_sensitive(i_h_showown, NULL);
/**** Interface: Tabs ****/
    section[TABS] = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[TABS]), 10);
    check_but(_("Homogeneous tabs"),               i_t_homo,  section[TABS], homo_tabs);
    check_but(_("Set the length of the "
              "name on a tab to a maximum "
              "of 20 chars"), i_t_short, section[TABS], short_tab_names);
/**** Keypad ****/
  	section[KEYPAD] = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[KEYPAD]), 10);
    gtk_widget_set_usize(section[KEYPAD], 300, -1);
    check_but(_("Enable keypad aliasses"), k_enable, section[KEYPAD], use_keypad_keys);
    frame = gtk_frame_new(_("Keypad bindings"));
      gtk_container_set_border_width(GTK_CONTAINER(section[KEYPAD]), 5);
      gtk_widget_show(frame);
      gtk_box_pack_start(GTK_BOX(section[KEYPAD]), frame, TRUE, TRUE, 0);
    table = gtk_table_new(5, 8, TRUE);
      gtk_container_set_border_width(GTK_CONTAINER(table), 5);
      gtk_widget_show(table);
      gtk_container_add(GTK_CONTAINER(frame), table);
    keypad_entry(HOME,   1, 2, 1, 2, _("Home "));
    keypad_entry(LEFT,   1, 2, 2, 3, _("Left "));
    keypad_entry(END,    1, 2, 3, 4, _("End "));
    keypad_entry(INS,    1, 4, 4, 5, _("Ins "));
    keypad_entry(DIV,    3, 4, 0, 1, _("/ "));
    keypad_entry(UP,     3, 4, 1, 2, _("Up "));
    keypad_entry(MIDDLE, 3, 4, 2, 3, _("5 "));
    keypad_entry(DOWN,   3, 4, 3, 4, _("Down "));
    keypad_entry(MUL,    5, 6, 0, 1, _("* "));
    keypad_entry(PGUP,   5, 6, 1, 2, _("PgUp "));
    keypad_entry(RIGHT,  5, 6, 2, 3, _("Right "));
    keypad_entry(PGDN,   5, 6, 3, 4, _("PgDn "));
    keypad_entry(DEL,    5, 6, 4, 5, _("Del "));
    keypad_entry(MINUS,  7, 8, 0, 1, _("- "));
    keypad_entry(PLUS,   7, 8, 1, 3, _("+ "));
    keypad_entry(ENTER,  7, 8, 3, 5, _("Enter ")); 
    gtk_signal_connect(GTK_OBJECT(k_enable), "clicked", GTK_SIGNAL_FUNC(keypad_set_sensitive), NULL);
    keypad_set_sensitive(k_enable, NULL);
/**** Display ****/
  	section[DISPLAY] = gtk_label_new(""); /* fake widget */
/**** Display: Fonts ****/
    section[FONTS] = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[FONTS]), 10);
    table = gtk_table_new (2, 3, FALSE);
      gtk_widget_show (table);
      gtk_box_pack_start(GTK_BOX(section[FONTS]), table, FALSE, TRUE, 0);
      gtk_table_set_row_spacings (GTK_TABLE (table), 5);
      gtk_table_set_col_spacings (GTK_TABLE (table), 5);
    tmp = gtk_label_new (_("Normal font "));
      gtk_widget_show (tmp);
      gtk_table_attach(GTK_TABLE (table), tmp, 0, 1, 0, 1,
                       (GtkAttachOptions) GTK_FILL,
                       (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
      gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);
    tmp = gtk_label_new (_("Bold font "));
      gtk_widget_show (tmp);
      gtk_table_attach(GTK_TABLE (table), tmp, 0, 1, 1, 2,
                       (GtkAttachOptions) GTK_FILL,
                       (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
      gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);
    d_f_normal = gtk_entry_new ();
      gtk_widget_show (d_f_normal);
      gtk_table_attach(GTK_TABLE (table), d_f_normal, 1, 2, 0, 1,
                      (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
                      (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
      gtk_entry_set_text(GTK_ENTRY(d_f_normal), normalfontstr);
      add_def_signal(d_f_normal, "changed");
      gtk_editable_set_editable(GTK_EDITABLE(d_f_normal), FALSE);
    d_f_bold = gtk_entry_new ();
      gtk_widget_show (d_f_bold);
      gtk_table_attach(GTK_TABLE (table), d_f_bold, 1, 2, 1, 2,
                      (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
                      (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
      gtk_entry_set_text(GTK_ENTRY(d_f_bold), boldfontstr);
      add_def_signal(d_f_bold, "changed");
      gtk_editable_set_editable(GTK_EDITABLE(d_f_bold), FALSE);
    tmp = gtk_button_new_with_label (_(" Select "));
      gtk_widget_show (tmp);
      gtk_table_attach(GTK_TABLE (table), tmp, 2, 3, 0, 1,
                      (GtkAttachOptions) GTK_FILL,
                      (GtkAttachOptions) GTK_EXPAND, 0, 0);
      gtk_signal_connect(GTK_OBJECT(tmp), "clicked",
                         GTK_SIGNAL_FUNC(font_but_cb), d_f_normal);
    tmp = gtk_button_new_with_label (_(" Select "));
      gtk_widget_show (tmp);
      gtk_table_attach(GTK_TABLE (table), tmp, 2, 3, 1, 2,
                      (GtkAttachOptions) GTK_FILL,
                      (GtkAttachOptions) GTK_EXPAND, 0, 0);
      gtk_signal_connect(GTK_OBJECT(tmp), "clicked",
                         GTK_SIGNAL_FUNC(font_but_cb), d_f_bold);
/**** Display: Colors ****/
    section[DCOLORS] = gtk_hbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[DCOLORS]), 10);
    vbb = gtk_vbutton_box_new ();
      gtk_widget_show (vbb);
      gtk_box_pack_start(GTK_BOX(section[DCOLORS]), vbb, FALSE, TRUE, 0);
      gtk_button_box_set_layout (GTK_BUTTON_BOX (vbb), GTK_BUTTONBOX_START);
      gtk_button_box_set_spacing (GTK_BUTTON_BOX (vbb), 0);
      gtk_button_box_set_child_size (GTK_BUTTON_BOX (vbb), 0, 0);
    color_but(_("Normal FG"), DEFAULTFG);
    color_but(_("Normal BG"), DEFAULTBG);
    tmp = gtk_hseparator_new();
      gtk_widget_show(tmp);
      gtk_container_add (GTK_CONTAINER (vbb), tmp);
    color_but(_("Black")  , BLACK);
    color_but(_("Red")    , RED);
    color_but(_("Green")  , GREEN);
    color_but(_("Yellow") , YELLOW);
    color_but(_("Blue")   , BLUE);
    color_but(_("Magenta"), MAGENTA);
    color_but(_("Cyan")   , CYAN);
    color_but(_("White")  , WHITE);
    frame = gtk_frame_new(_("Example"));
      gtk_widget_show(frame);
      gtk_box_pack_start(GTK_BOX(section[DCOLORS]), frame, TRUE, TRUE, 0);
    tmp = gtk_vbox_new(FALSE, 0);
      gtk_container_set_border_width(GTK_CONTAINER(tmp), 5);
      gtk_widget_show(tmp);
      gtk_container_add(GTK_CONTAINER(frame), tmp);
    d_c_preview = gtk_text_new(NULL, NULL);
      gtk_widget_show(d_c_preview);
      gtk_container_add(GTK_CONTAINER(tmp), d_c_preview);
/**** Editor ****/
  	section[EDITOR] = gtk_label_new(""); /* fake widget */
/**** Editor: Auto indent ****/
    section[AUTOINDENT] = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[AUTOINDENT]), 10);
    check_but(_("Enable auto indent"), e_a_enable, section[AUTOINDENT], use_ident);
    check_but(_("Detect if text is a verb, and then enable the syntax highlighter"),
              e_a_auto, section[AUTOINDENT], use_auto_detect);
/**** Editor: Auto indent ****/
    section[ECOLORS] = gtk_hbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[ECOLORS]), 10);
    vbb = gtk_vbutton_box_new ();
      gtk_widget_show (vbb);
      gtk_box_pack_start(GTK_BOX(section[ECOLORS]), vbb, FALSE, TRUE, 0);
      gtk_button_box_set_layout (GTK_BUTTON_BOX (vbb), GTK_BUTTONBOX_START);
      gtk_button_box_set_spacing (GTK_BUTTON_BOX (vbb), 0);
      gtk_button_box_set_child_size (GTK_BUTTON_BOX (vbb), 0, 0);
    color_but(_("Strings"),      HL_STRING);
    color_but(_("Numbers"),      HL_NUMBER);
    color_but(_("Objects"),      HL_OBJNUM);
    color_but(_("$ properties"), HL_DOLLAR);
    color_but(_("Variables"),    HL_VARIAB);
    color_but(_("Keywords"),     HL_KEYWRD);
    color_but(_("Symbols"),      HL_SYMBOL);
    color_but(_("Comments"),     HL_COMMEN);
    frame = gtk_frame_new(_("Example"));
      gtk_widget_show(frame);
      gtk_box_pack_start(GTK_BOX(section[ECOLORS]), frame, TRUE, TRUE, 0);
    tmp = gtk_vbox_new(FALSE, 0);
      gtk_container_set_border_width(GTK_CONTAINER(tmp), 5);
      gtk_widget_show(tmp);
      gtk_container_add(GTK_CONTAINER(frame), tmp);
    e_c_preview = gtk_text_new(NULL, NULL);
      gtk_widget_show(e_c_preview);
      gtk_container_add(GTK_CONTAINER(tmp), e_c_preview);
/**** Tcl ****/
#ifdef TCL
  	section[MTCL] = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[MTCL]), 10);
    check_but(_("Enable tcl debug"), t_enable, section[MTCL], tcl_debug);
    check_but(_("Tcl debug and errors on stdout"),
              t_onstdout, section[MTCL], tcl_debug_on_stdout);
    gtk_signal_connect(GTK_OBJECT(t_enable), "clicked",
                       GTK_SIGNAL_FUNC(onstdout_sensitive), NULL);
    onstdout_sensitive(t_enable, NULL);
#else /* ndef TCL */
    section[MTCL] = gtk_label_new(_("Tcl scripting not compiled in to gMOO."));
#endif
/**** Externals ****/
  	section[EXTERNALS] = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(section[EXTERNALS]), 10);
    table = gtk_table_new (3, 2, FALSE);
      gtk_widget_show (table);
      gtk_box_pack_start(GTK_BOX(section[EXTERNALS]), table, FALSE, TRUE, 0);
      gtk_table_set_row_spacings (GTK_TABLE (table), 5);
      gtk_table_set_col_spacings (GTK_TABLE (table), 5);
    tmp = gtk_label_new (_("Mailreader "));
      gtk_widget_show (tmp);
      gtk_table_attach(GTK_TABLE (table), tmp, 0, 1, 0, 1,
                       (GtkAttachOptions) GTK_FILL,
                       (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
      gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);
    tmp = gtk_label_new (_("Newsreader "));
      gtk_widget_show (tmp);
      gtk_table_attach(GTK_TABLE (table), tmp, 0, 1, 1, 2,
                       (GtkAttachOptions) GTK_FILL,
                       (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
      gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);
    x_mail = gtk_entry_new ();
      gtk_widget_show (x_mail);
      gtk_table_attach(GTK_TABLE (table), x_mail, 1, 2, 0, 1,
                      (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
                      (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
      gtk_entry_set_text(GTK_ENTRY(x_mail), ext_mail != NULL ? ext_mail : "");
      add_def_signal(x_mail, "changed");
    x_news = gtk_entry_new ();
      gtk_widget_show (x_news);
      gtk_table_attach(GTK_TABLE (table), x_news, 1, 2, 1, 2,
                      (GtkAttachOptions) GTK_EXPAND | GTK_FILL,
                      (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
      gtk_entry_set_text(GTK_ENTRY(x_news), ext_news != NULL ? ext_news : "");
      add_def_signal(x_news, "changed");
    tmp = gtk_label_new (_("NOTE: Read the help (press the help button)"));
      gtk_widget_show (tmp);
      gtk_table_attach_defaults(GTK_TABLE (table), tmp, 0, 2, 2, 3);
      gtk_misc_set_alignment (GTK_MISC (tmp), 0.0, 0.5);
}

#define i(text, color1, color2, tekst) \
  gtk_text_insert(GTK_TEXT(text), nf, color1,color2, tekst, -1);
void update_previews() {
  GdkFont  *nf, *bf;
  GtkStyle *style;
  int i;
char text[] =
"\"This is an example verb\";\n"
"lis = {#11, 99, \"String\"};\n"
"for i in (lis)\n"
"  if(typeof(i) ==STR)\n"
"    #2:tell(i);\n"
"  elseif(typeof(i) == NUM)\n"
"    nofify(#2, tostr(i));\n"
"  else\n"
"    t:tell($string_utils:nn(i));\n"
"  endif\n"
"  #2:tell(\"This was number \", tostr(i in lis));\n"
"endif\n"
"\"as you can see,\";\n"
"\"completely useless :)\";\n";
  char *s;
  int j = 0;
  char *buf;
  int curcolor = 0;
  
  nf = gdk_font_load(gtk_entry_get_text(GTK_ENTRY(d_f_normal)));
  bf = gdk_font_load(gtk_entry_get_text(GTK_ENTRY(d_f_bold)));

  style = gtk_style_copy(gtk_widget_get_style(d_c_preview));
  memcpy(&style->fg[GTK_STATE_NORMAL],   &d_c_colors[DEFAULTFG], sizeof(GdkColor));
  memcpy(&style->text[GTK_STATE_NORMAL], &d_c_colors[DEFAULTFG], sizeof(GdkColor));
  memcpy(&style->bg[GTK_STATE_NORMAL],   &d_c_colors[DEFAULTBG], sizeof(GdkColor));
  memcpy(&style->base[GTK_STATE_NORMAL], &d_c_colors[DEFAULTBG], sizeof(GdkColor));
  style->font = nf;

  gtk_widget_set_style(d_c_preview, style);
  gtk_widget_set_style(e_c_preview, style);


  gtk_text_freeze(GTK_TEXT(d_c_preview));
  
  gtk_editable_delete_text(GTK_EDITABLE(d_c_preview), 0, gtk_text_get_length(GTK_TEXT(d_c_preview))); 

  gtk_text_insert(GTK_TEXT(d_c_preview), nf, NULL, NULL, _("This is text in the normal font\n"), -1);
  gtk_text_insert(GTK_TEXT(d_c_preview), bf, NULL, NULL, _("This is text in the bold font\n"),   -1);

  gtk_text_insert(GTK_TEXT(d_c_preview), nf, NULL, NULL,
    "\nForeground  Background\n----------------------\n", -1);
  i(d_c_preview, &d_c_colors[BLACK], NULL, "     BLACK  ");
  i(d_c_preview, NULL, &d_c_colors[BLACK], "BLACK     \n");
  i(d_c_preview, &d_c_colors[RED], NULL,   "       RED  ");
  i(d_c_preview, NULL, &d_c_colors[RED],   "RED       \n");
  i(d_c_preview, &d_c_colors[GREEN], NULL, "     GREEN  ");
  i(d_c_preview, NULL, &d_c_colors[GREEN], "GREEN     \n");
  i(d_c_preview, &d_c_colors[YELLOW], NULL,"    YELLOW  ");
  i(d_c_preview, NULL, &d_c_colors[YELLOW],"YELLOW    \n");
  i(d_c_preview, &d_c_colors[BLUE], NULL,  "      BLUE  ");
  i(d_c_preview, NULL, &d_c_colors[BLUE],  "BLUE      \n");
  i(d_c_preview, &d_c_colors[MAGENTA], NULL,"   MAGENTA  ");
  i(d_c_preview, NULL, &d_c_colors[MAGENTA],"MAGENTA   \n");
  i(d_c_preview, &d_c_colors[CYAN], NULL,  "      CYAN  ");
  i(d_c_preview, NULL, &d_c_colors[CYAN],  "CYAN      \n");
  i(d_c_preview, &d_c_colors[WHITE], NULL, "     WHITE  ");
  i(d_c_preview, NULL, &d_c_colors[WHITE], "WHITE     \n");

  gtk_text_thaw(GTK_TEXT(d_c_preview));





  s = shl(g_malloc((strlen(text) * 2) + 2), text);
  buf = g_malloc(strlen(s) + 1);

  gtk_text_freeze(GTK_TEXT(e_c_preview));
  gtk_editable_delete_text(GTK_EDITABLE(e_c_preview), 0, gtk_text_get_length(GTK_TEXT(e_c_preview))); 

  for(i = 0; s[i] != 0; i++) {
    buf[j] = s[i];
    if(buf[j] >= HL_STRING && buf[j] <= HL_SYMBOL) {
      buf[j] = 0;

      gtk_text_insert(GTK_TEXT(e_c_preview), nf, &e_c_colors[curcolor], NULL, buf, -1);
      j = 0;
    }
    if(s[i] == 0) {
      gtk_text_insert(GTK_TEXT(e_c_preview), nf, &e_c_colors[curcolor], NULL, buf, -1);
      g_free(s);
      gtk_text_thaw(GTK_TEXT(e_c_preview));
      g_free(buf);
      return;
    } else if(s[i] >= HL_STRING && s[i] <= HL_COMMEN) {
      curcolor = s[i] - HL_STRING;
    } else {
      j++;
    }
  }
  buf[j] = 0;
  gtk_text_insert(GTK_TEXT(e_c_preview), nf, &e_c_colors[curcolor], NULL, buf, -1);
  g_free(buf);

  gtk_text_thaw(GTK_TEXT(e_c_preview));

  g_free(s);

  gtk_style_unref(style);
}

