/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "structs.h"

#ifdef TCL /* skip the entire C file if no Tcl support */

#include <gtk/gtk.h>

#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>

#include "handy.h"
#include "structs.h"
#include "config.h"
#include "worlds.h"
#include "world.h"
#include "do_scripts.h"
#include "dialog.h"

/* Macro struct: */
typedef struct _macro macro;
struct _macro {
  char *name;
  char *command;
  char *desc;
  unsigned int global : 1;

  macro *next;
};
macro *macros = NULL;

void  eval_files(Tcl_Interp *with, char *where);

int C_msg     (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);
int C_tell    (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);
int C_send    (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);
int C_msgbox  (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);
int C_input   (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);

int C_debugmsg(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);

int C_bgerror (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);

int C_register_macro(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);
int C_registered_macros(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);

int C_info_world(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);

void gM_script_init(char *appname) {
  char *name;

  Tcl_FindExecutable(appname);
  Tcl_SetServiceMode(TCL_SERVICE_ALL);

  master_interp = Tcl_CreateInterp();
  Tcl_Init(master_interp);
  Tcl_Preserve(master_interp);

  gtk_timeout_add(50, (GtkFunction) gM_script_idle_func, NULL);

  Tcl_CreateCommand(master_interp, "msgbox",   C_msgbox,   NULL, NULL);
  Tcl_CreateCommand(master_interp, "input",    C_input,    NULL, NULL);
  Tcl_CreateCommand(master_interp, "tell",     C_tell,     NULL, NULL);
  Tcl_CreateCommand(master_interp, "msg",      C_msg,      NULL, NULL);
  Tcl_CreateCommand(master_interp, "send",     C_send,     NULL, NULL);
  Tcl_CreateCommand(master_interp, "debugmsg", C_debugmsg, NULL, NULL);
  Tcl_CreateCommand(master_interp, "bgerror",  C_bgerror,  NULL, NULL);

  Tcl_CreateCommand(master_interp, "register.macro", C_register_macro,  NULL, NULL);
  Tcl_CreateCommand(master_interp, "registered.macros", C_registered_macros, NULL, NULL);

  Tcl_CreateCommand(master_interp, "info.world", C_info_world, NULL, NULL);

  name = strcatnew(PREFIX, "/share/gMOO/Scripts/Client");
  eval_files(master_interp, name);
  g_free(name);

  name = strcatnew(config_dir, "/Scripts/Client");
  eval_files(master_interp, name);
  g_free(name);
}

void gM_script_exit() {
  Tcl_Release(master_interp);
  Tcl_DeleteInterp(master_interp);
}

void gM_script_init_for(world *w) {
  char *name;

  w->tcl_name = strcatnew("wrld", w->name);
  strsub(w->tcl_name, ' ', '_'); /* remove the fucking spaces, or else
                                  * that create slave command fucks up */
  w->interp = Tcl_CreateSlave(master_interp, w->tcl_name, 1);
  Tcl_Init(w->interp);
  Tcl_Preserve(w->interp);

  Tcl_CreateCommand(w->interp, "msg",      C_msg,      NULL, NULL);
  Tcl_CreateCommand(w->interp, "tell",     C_tell,     NULL, NULL);
  Tcl_CreateCommand(w->interp, "send",     C_send,     NULL, NULL);
  Tcl_CreateCommand(w->interp, "msgbox",   C_msgbox,   NULL, NULL);
  Tcl_CreateCommand(w->interp, "input",    C_input,    NULL, NULL);
  Tcl_CreateCommand(w->interp, "debugmsg", C_debugmsg, NULL, NULL);
  Tcl_CreateCommand(w->interp, "bgerror",  C_bgerror,  NULL, NULL);

  Tcl_CreateCommand(w->interp, "register.macro",  C_register_macro,  NULL, NULL);
  Tcl_CreateCommand(w->interp, "registered.macros", C_registered_macros, NULL, NULL);

  Tcl_CreateCommand(w->interp, "info.world", C_info_world, NULL, NULL);

  name = strcatnew(PREFIX, "/share/gMOO/Scripts/World");
  eval_files(w->interp, name);
  g_free(name);

  name = strcatnew(config_dir, "/Scripts/World");
  eval_files(w->interp, name);
  g_free(name);
}

int gM_script_idle_func(void *pointer) {
  Tcl_ServiceAll();
  
  return TRUE;
}

void gM_script_do_eval(world *w, char *s) {
  gM_world_print(w, _("\e[36mTcl: Evalling: "), -1);
  gM_world_print(w, s, -1);
  gM_world_print(w, "\e[0m\n", -1);
  if(Tcl_Eval(w->interp, s) == TCL_OK) {
    gM_world_print(w, "\e[36m=> TCL_OK\e[0m\n", -1);
  } else {
    gM_world_print(w, "\e[36m=> TCL_ERROR: ", -1);
    gM_world_print(w, Tcl_GetStringResult(w->interp), -1);
    gM_world_print(w, "\e[0m\n", -1);
  }
}

void gM_script_eval_script(world *w, char *script) {
  if(Tcl_Eval(w->interp, script) != TCL_OK && tcl_debug) {
    gM_world_print(w, "\e[31m% TCL_ERROR: ", -1);
    gM_world_print(w, Tcl_GetStringResult(w->interp), -1);
    gM_world_print(w, "\e[0m\n", -1);
  }
}

void gM_script_do_macro(world *w, char *line) {
  macro *m;
  char *s;
  line++; /* skip the first '/' char */
  s = get_1st_from(line);
  
  for(m = macros; m != NULL; m = m->next) {
    if(strcasecmp(m->name, s) == 0) { /* Found ! */
      g_free(s);
      s = format_for_tcl(m->command, m->global ? w->tcl_name : "", line);
      if(Tcl_Eval(m->global ? master_interp : w->interp, s) != TCL_OK && tcl_debug) {
        gM_world_print(w, "\e[31m% TCL_ERROR: ", -1);
        gM_world_print(w, Tcl_GetStringResult(m->global ? master_interp : w->interp), -1);
        gM_world_print(w, "\e[0m\n", -1);
      }
      g_free(s);
      return;
    }
  }
  gM_world_print(w, _("\e[31m% Error, macroname \""), -1);
  gM_world_print(w, s, -1);
  gM_world_print(w, _("\" not found\e[0m\n"), -1);
  g_free(s);
}

/*********************************
 *
 * local funcs
 *
 *********************************/

void eval_files(Tcl_Interp *with, char *where) {
  DIR *dir;
  struct dirent *ent;
  char *tmp;
  int i;

  dir = opendir(where);
  if(dir != NULL) {
    ent = readdir(dir);
    while(ent != NULL) {
      if(ent->d_name[0] != '.') { /* skip hidden files, `..' and `.' */
        tmp = strcat3new(where, "/", ent->d_name);
        i = Tcl_EvalFile(with, tmp);
        if(i != TCL_OK)
          printf(_("Tcl Error evalling file %s: \"%s\"\n"), tmp, Tcl_GetStringResult(with));
        g_free(tmp);
      }
      ent = readdir(dir);
    }
  }
}

/*********************************
 *
 * C additions to the standard Tcl lib
 *
 *********************************/
char *__str__;

#define set_result(i, s)       Tcl_SetResult   (i, s, NULL)
#define set_int_result(i, num) Tcl_SetObjResult(i, Tcl_NewIntObj(num))
#define E_ARGS(i, proc) \
  __str__ = strconcatnew(2, proc, _(": Incorrect number of arguments")); \
  Tcl_SetResult(i, __str__, (Tcl_FreeProc *)free); \
  return TCL_ERROR
#define E_WORLDNF(i, proc) \
  __str__ = strconcatnew(2, proc, _(": World not found in opened worlds")); \
  Tcl_SetResult(i, __str__, (Tcl_FreeProc *)free); \
  return TCL_ERROR
#define E_SYNTAX(i, proc) \
  __str__ = strconcatnew(2, proc, _("Syntax error")); \
  Tcl_SetResult(i, __str__, (Tcl_FreeProc *)free); \
  return TCL_ERROR
#define E_OPTION(i, proc) \
  __str__ = strconcatnew(2, proc, _("Wrong option")); \
  Tcl_SetResult(i, __str__, (Tcl_FreeProc *)free); \
  return TCL_ERROR

#define set_obj_var(i, var, val) \
  Tcl_ObjSetVar2(i, Tcl_NewStringObj(var, -1), NULL, val, 0)
#define set_string_var(i, var, val) \
  set_obj_var(i, var, Tcl_NewStringObj(val, -1))
#define set_int_var(i, var, val) \
  set_obj_var(i, var, Tcl_NewIntObj(val))


int C_msg(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  world *w;
  int i;
  int flags = 0;

  if(interp == master_interp) {
    if(argc < 3) { E_ARGS(interp, "msg"); }
    w = gM_worlds_get_by_tcl_id(argv[argc - 2]);
  } else {
    if(argc < 2) { E_ARGS(interp, "msg"); }
    w = gM_worlds_get_by_interp(interp);
  }

  if(w == NULL) { E_WORLDNF(interp, "msg"); }

  for(i = 1; i < argc - (interp == master_interp ? 2 : 1); i++) {
    if(argv[i][0] == '-') {
      if(strcmp(argv[i], "-raw") == 0) {
        flags |= FLAG_RAW;
      } else if(strcmp(argv[i], "-n") == 0) {
        flags |= FLAG_N;
      } else { /* error, invalid value */
        E_OPTION(interp, "msg");
      }
    } else {
      E_SYNTAX(interp, "msg");
    }
  }
  do_C_msg(w, argv[argc - 1], flags);
  return TCL_OK;
}

int C_tell(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  world *w;
  int i;
  int flags = 0;

  if(interp == master_interp) {
    if(argc < 3) { E_ARGS(interp, "tell"); }
    w = gM_worlds_get_by_tcl_id(argv[argc - 2]);
  } else {
    if(argc < 2) { E_ARGS(interp, "tell"); }
    w = gM_worlds_get_by_interp(interp);
  }
  if(w == NULL) { E_WORLDNF(interp, "tell"); }

  for(i = 1; i < argc - (interp == master_interp ? 2 : 1); i++) {
    if(argv[i][0] == '-') {
      if(strcmp(argv[i], "-raw") == 0) {
        flags |= FLAG_RAW;
      } else if(strcmp(argv[i], "-n") == 0) {
        flags |= FLAG_N;
      } else { /* error, invalid value */
        E_OPTION(interp, "tell");
      }
    } else {
      E_SYNTAX(interp, "tell");
    }
  }
  do_C_tell(w, argv[argc - 1], flags);
  return TCL_OK;
}

int C_send(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  world *w;
  int i;
  int flags = 0;

  if(interp == master_interp) {
    if(argc < 3) { E_ARGS(interp, "send"); }
    w = gM_worlds_get_by_tcl_id(argv[argc - 2]);
  } else {
    if(argc < 2) { E_ARGS(interp, "send"); }
    w = gM_worlds_get_by_interp(interp);
  }

  if(w == NULL) { E_WORLDNF(interp, "send"); }

  for(i = 1; i < argc - (interp == master_interp ? 2 : 1); i++) {
    if(argv[i][0] == '-') {
      if(strcmp(argv[i], "-raw") == 0) {      /* ignored */
      } else if(strcmp(argv[i], "-n") == 0) {
        flags |= FLAG_N;
      } else { /* error, invalid value */
        E_OPTION(interp, "send");
      }
    } else {
      E_SYNTAX(interp, "send");
    }
  }
  do_C_send(w, argv[argc - 1], flags);
  return TCL_OK;
}

int C_msgbox(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  int i;
  int flags = 0;
  char *title = NULL;
  int tmp;

  if(argc < 2) { E_ARGS(interp, "msgbox"); }

  for(i = 1; i < argc - 1; i++) {
    if(argv[i][0] == '-') {
      if(strcmp(argv[i], "--") == 0) {
        break;
      } else if(strcmp(argv[i], "-icon") == 0) {
        if((i++) > argc - 2) { E_ARGS(interp, "msgbox"); }
        if(strcmp(argv[i], "stop") == 0 || strcmp(argv[i], "error") == 0)
          flags |= FLAG_ERROR;
        else if(strcmp(argv[i], "question") == 0)
          flags |= FLAG_QUESTION;
        else if(strcmp(argv[i], "warning") == 0 || strcmp(argv[i], "exclamation") == 0)
          flags |= FLAG_WARNING;
        else if(strcmp(argv[i], "information") == 0 || strcmp(argv[i], "asterix") == 0)
          flags |= FLAG_INFORMATION;
        else {
          set_result(interp, _("Wrong option value"));
          return TCL_ERROR;
        }
      } else if(strcmp(argv[i], "-buttons") == 0) {
        if((i++) > argc - 2) { E_ARGS(interp, "msgbox"); }
        if(strcmp(argv[i], "abortretryignore") == 0)
          flags |= FLAG_ABORTRETRYIGNORE;
        else if(strcmp(argv[i], "ok") == 0)
          flags |= FLAG_OK;
        else if(strcmp(argv[i], "okcancel") == 0)
          flags |= FLAG_OKCANCEL;
        else if(strcmp(argv[i], "retrycancel") == 0)
          flags |= FLAG_RETRYCANCEL;
        else if(strcmp(argv[i], "yesno") == 0)
          flags |= FLAG_YESNO;
        else if(strcmp(argv[i], "yesnocancel") == 0)
          flags |= FLAG_YESNOCANCEL;
        else {
          set_result(interp, _("Wrong option value"));
          return TCL_ERROR;
        }
      } else if(strcmp(argv[i], "-stdbutton") == 0) {
        if((i++) > argc - 2) { E_ARGS(interp, "msgbox"); }
        tmp = atoi(argv[i]);
        switch(tmp) {
        case 0: flags |= FLAG_DEF0; break;
        case 1: flags |= FLAG_DEF1; break;
        case 2: flags |= FLAG_DEF2; break;
        default: set_result(interp, _("Wrong option value")); return TCL_ERROR;
        }
      } else if(strcmp(argv[i], "-title") == 0) {
        if((i++) > argc - 2) { E_ARGS(interp, "msgbox"); }
        title = argv[i];
      } else {
        E_OPTION(interp, "msgbox");
      }
    } else { /* error */
      E_SYNTAX(interp, "msgbox");
    }
  }
  if(i != argc - 1) { E_ARGS(interp, "msgbox"); }
  set_int_result(interp, do_C_msgbox(title, argv[i], flags));
  return TCL_OK;
}

int C_input (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  int i;
  int flags = 0;
  char *title   = NULL;
  char *desc    = NULL;
  char *deftext = NULL;
  char *var;
  int   retint;
  char *retstr;

  if(argc < 2) { E_ARGS(interp, "input"); }

  for(i = 1; i < argc - 1; i++) {
    if(argv[i][0] == '-') {
      if(strcmp(argv[i], "--") == 0) {
        break;
      } else if(strcmp(argv[i], "-type") == 0) {
        if((i++) > argc - 2) { E_ARGS(interp, "input"); }
        if(strcmp(argv[i], "int") == 0 || strcmp(argv[i], "integer") == 0)
          flags |= FLAG_INT;
        else if(strcmp(argv[i], "string") == 0)
          flags |= FLAG_STRING;
        else {
          set_result(interp, _("Wrong option value"));
          return TCL_ERROR;
        }
      } else if(strcmp(argv[i], "-title") == 0) {
        if((i++) > argc - 2) { E_ARGS(interp, "input"); }
        title = argv[i];
      } else if(strcmp(argv[i], "-desc") == 0 || strcmp(argv[i], "-description") == 0) {
        if((i++) > argc - 2) { E_ARGS(interp, "input"); }
        desc = argv[i];
      } else {
        E_OPTION(interp, "input");
      }
    } else { /* error */
      if(i == argc - 2) /* This one is correct */
        break;
      E_SYNTAX(interp, "input");
    }
  }
  if(i == argc - 2) { /* Check if the [defaulttext] exists */
    var = argv[i];
    deftext = argv[i + 1];
  } else if(i == argc - 1) {
    var = argv[i];
  } else {
    E_ARGS(interp, "input");
  }
  i = do_C_input(title, desc, deftext, flags, &retint, &retstr);
  set_int_result(interp, i);
  if(i == 1) {
    if(flags & FLAG_INT) {
      set_int_var(interp, var, retint);
    } else { /* rettype == FLAG_STRING */
      set_string_var(interp, var, retstr);
      g_free(retstr);
    }
  }
  return TCL_OK;
}

int C_debugmsg(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  int i;
  char *s;
  world *w;

  if(tcl_debug < 1) {
    set_result(interp, _("Debug mode disabled"));
    return TCL_OK;
  }

  if(argc < 2 || argc > 4) { E_ARGS(interp, "debugmsg"); }
  for(i = 1; i < argc - 1; i++) {
    if(argv[i][0] == '-') {
      if(strcmp(argv[i], "-raw") == 0) {      /* ignored */
      } else if(strcmp(argv[i], "-n") == 0) { /* ignored */
      } else { /* error, invalid value */
        E_OPTION(interp, "debugmsg");
      }
    } else {
      E_SYNTAX(interp, "debugmsg");
    }
  }

  if(tcl_debug_on_stdout > 0)
    printf(_("\e[36mTcl debug: %s\e[0m\n"), argv[i]);
  else {
    w = gM_worlds_get_by_interp(interp);
    if(w != NULL) {
      s = strconcatnew(3, _("\e[36m% Tcl debug: "), argv[i], "\e[0m\n");
      gM_world_print(w, s, -1);
    } else {
      s = strconcatnew(2, _("Tcl debug: "), argv[i]);
      gM_dialog_show(_("Tcl Debug: "), s, 0);
    }
    g_free(s);
  }

  return TCL_OK;
}

int C_bgerror (ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  world *w;
  char *name;
  char *error;
  char *s;
  
  w = gM_worlds_get_by_interp(interp);
  name = w == NULL ? strnew(_("master interpreter")) :
                     strcatnew(_("interpreter for "), w->name);
  error = Tcl_Concat(argc - 1, argv + 1);

  if(tcl_debug_on_stdout > 0)
    printf(_("\e[31mTcl bgerror in the %s: %s\e[0m\n"), name, error);
  else {
    if(w != NULL) {
      s = strconcatnew(5, _("\e[31m% Tcl bgerror in the "), name, ": ",
                          error,  "\e[0m\n");
      gM_world_print(w, s, -1);
    } else {
      s = strconcatnew(4, _("Tcl bgerror in the "), name, ": ",  error);
      gM_dialog_show(_("Tcl Error: "), s, 0);
    }
    g_free(s);
  }

  g_free(name);
  Tcl_Free(error);
  
  return TCL_OK;
}

int C_register_macro(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  int i;
  macro *m;

  if(argc != 4) {
    E_ARGS(interp, "register.macro");
  }
  if(argv[1][0] == '/' && argv[1][1] == 0) {
    set_result(interp, _("register.macro: Error, macro '/' is a reserved macro"));
    return TCL_ERROR;
  }
  /* Check for spaces in the macro name */
  for(i = 0; argv[1][i] != 0; i++) {
    if(argv[1][i] == ' ') {
      set_result(interp, _("register.macro: The macro name containes a space"));
      return TCL_ERROR;
    }
  }

  if(macros == NULL) {
    m = macros = g_malloc(sizeof(macro));
  } else {
    for(m = macros; m != NULL; m = m->next) {
      if(strcasecmp(argv[1], m->name) == 0) { /* replaceing that macro */
        g_free(m->name);
        g_free(m->command);
        g_free(m->desc);
        goto do_set;
      }
    }
    for(m = macros; m->next != NULL; m = m->next);
    m->next = g_malloc(sizeof(macro));
    m = m->next;
  }
  m->next = NULL;

  do_set:
  m->name    = strnew(argv[1]);
  m->command = strnew(argv[2]);
  m->desc    = strnew(argv[3]);
  m->global  = interp == master_interp;
  
  return TCL_OK;
}

int C_registered_macros(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  int global = -1;
  macro *m;
  Tcl_Obj *list;
  
  if(argc != 1 && argc != 2) {
    E_ARGS(interp, "registered.macros");
  }
  if(argc == 2) {
    if(strcmp(argv[1], "-client") == 0)
      global = 1;
    else if(strcmp(argv[1], "-world") == 0)
      global = 0;
  }
  if(macros == NULL) {
    set_result(interp, "");
    return TCL_OK;
  }
  list = Tcl_NewListObj(0, NULL);

  for(m = macros; m != NULL; m = m->next) {
    if(global == -1 || global == m->global) { /* add it */
      Tcl_ListObjAppendElement(interp, list, Tcl_NewStringObj(m->name, -1));
      Tcl_ListObjAppendElement(interp, list, Tcl_NewStringObj(m->command, -1));
      Tcl_ListObjAppendElement(interp, list, Tcl_NewStringObj(m->desc, -1));
    }
  }
  Tcl_SetObjResult(interp, list);
  return TCL_OK;
}

int C_info_world(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) {
  world *w;
  if(interp == master_interp) {
    if(argc != 3) { E_ARGS(interp, "info.world"); }
    w = gM_worlds_get_by_tcl_id(argv[argc - 2]);
  } else {
    if(argc != 2) { E_ARGS(interp, "info.world"); }
    w = gM_worlds_get_by_interp(interp);
  }
  if(w == NULL) { E_WORLDNF(interp, "info.world"); }
  
  if(strcmp(argv[argc - 1], "screenwidth")         == 0) {
    set_int_result(interp, (w->output->allocation.width - 15) /
                            gdk_char_width(normalfont, 'm'));
  } else if(strcmp(argv[argc - 1], "screenheigth") == 0) {
    set_int_result(interp, (w->output->allocation.height - 15) /
                            gdk_char_height(normalfont, 'J'));
  } else if(strcmp(argv[argc - 1], "name")         == 0) {
    set_result(interp, w->name);
  } else if(strcmp(argv[argc - 1], "hostname")     == 0) {
    set_result(interp, w->host);
  } else if(strcmp(argv[argc - 1], "port")         == 0) {
    set_int_result(interp, w->port);
  } else if(strcmp(argv[argc - 1], "connected")    == 0) {
    set_int_result(interp, w->sock > -1);
  }
  return TCL_OK;
}

#endif /* TCL */
 
