/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef gM_types_h
#define gM_types_h

/* global includes */
#include <gtk/gtk.h>
#include <time.h>
#include <regex.h>

#include "global_vars.h"

/* struct defenitions */
typedef struct _mcpsession   mcpsession;
typedef struct _mcpmultiline mcpmultiline;
typedef struct _mcpsimpledit mcpsimpleedit;
typedef struct _mcpuserlist  mcpuserlist;
typedef struct _mcpuser      mcpuser;
typedef struct _mcpicon      mcpicon;

typedef struct _history history;
typedef struct _editor  editor;
typedef struct _world   world;

typedef struct _var     var;

typedef struct _trigger trigger;
typedef struct _trigger_data trigger_data;
#ifdef TCL
#include <tcl.h>
#include "script.h"
#endif


/* struct implementations */
struct _mcpuserlist {
  mcpuser *list;
  int listlen;
  mcpmultiline *listline;
  
  mcpicon *icons;
  int      iconcount;
  mcpmultiline *iconline;
  
  char **fields;
  int    fieldcount;
  mcpmultiline *fieldline;
  
  int player_obj;
  
  int  friendscount;
  int *friends;
  
  int menucount;
  char ***menus;
};


struct _mcpuser {
  char *name;
  int object;
  int icon;

  int cloaked;
  int away;
  int idle;

  char **args;  /* used to store the other arguments that the MOO       */
  int    argc;  /* might send along with the #obj,name,icon information */
  
  mcpuser *next;
  int pos;
  mcpuser *prev;
  /* add other supported things here: */
};

struct _mcpicon {
  char *name;
  GdkPixmap *pixmap;
  GdkBitmap *mask;
  
  mcpicon *next;
};


struct _mcpsession {
  world *world;
  char *authkey;

  int *mcpversion;       /* {major, minor} */
  
  int *negotiateversion; /* {major, minor} */
  int negotiate_done;

  mcpmultiline *multiline;

  /* structs here */
  mcpsimpleedit *simpleedit;
  int *simpleeditversion; /* {major, minor} */
  
  mcpuserlist *userlist;
  int *userlistversion;   /* {major, minor} */
  
  int *vgmoo_clientversion; /* {major, minor} */
};


struct _mcpsimpledit {
  mcpsimpleedit *next;
  editor *editor;
  
  char *reference;
  char *name;
  char *type;
  mcpmultiline *ml;
};


struct _mcpmultiline {
  mcpmultiline *next;

  char *datatag;
  char *nametag;
};


struct _history {
  history *next;
  history *prev;
  char *s;
};


struct _editor {
  world *world;
  
  char *verbname;		/* name of the verb(property) beiing edited */
  char *upload;			/* command to upload */
  
  int page_num;			/* the number in the notebook */

  int has_changed;      /* check of the editor needs saving */
  int curline;
  
  int syntax;  /* wether or not, the highlighter is on */

  int curcolor;         /* curent color for the syntaxhl parser */
  int sig;              /* signal ID */
  
  mcpsimpleedit *simpleedit;   /* MCP compatibility */
  
  GtkWidget *main;	/* vbox as the notebookpages child */
  GtkWidget *text;	/* GtkText to enter the text in */
  GtkWidget *syntaxHL; /* button for syntaxhl toggle */
  GtkWidget *label;	/* label to be used on the notebook */
};


struct _world {
  char *name;				/* Name of the world */
  char *host;				/* Name of the host */
  int   port;				/* the port */
  
  char *user;				/* username */
  char *password;		/* the password */
  
  int autoconnect;  /* if the user wants to autoconnect */
  
  int   sock;				/* socket to write to / read from */
  int   gsock;			/* int that gdk_input_add returns */
  
  world *prev;
  world *next;
  
  /* history pointer */
  history *hist;
  char *selection;
  
  /* externel checking */
  int get_for_editor;
  editor **edit;
  int editcount;

  /* color */
  int fgcolor;
  int bgcolor;
  GdkFont *curfont;
  
  /* logging: */
  int log;
  
  /* macros */
#if 0 /* macros */
  macro *mac;
#endif
  /* indicate of the world is locked */
  int locked;
  
  /* MCP session strucht */
  mcpsession *session;
  
  /* userlist (if enables) toggle (shown or hidden) */
  int userlist_on;
  
  /* use for discarding output using the PREFIX and SUFFIX */
  int ignore;
  
  /* gtk widget stuff */
  int page_num;		/* nummer of the notebook */

  /* thingie for linefeeds*/
  int insert_10;
  int lines_pending;
#if 0 /* That damned GtkText widget is acting up again */
  int buffer_len;    /* number of lines in the buffer ... */
#endif  
  /* Resize check (GTK BUG) ? */
  int outputsize[2];
  
  /* Partial ANSI escape buffer for escape sequences that get
   * terminated across ppp packets */
  char *ansibuffer;

  /* flag is set when some wierd code is received, that set the
   * local echo off ... don't know what other flags can come like
   * this */
  int force_no_echo;

  /* The connection counter */
  time_t connect_time; /* time the world connected */
  
  /* Really reconnect (protection for a bad-passwd-recon-loop */
  int really_reconnect;
  
  /* Tcl related stuff */
#ifdef TCL
  Tcl_Interp *interp;
  char *tcl_name;     /* That damned Tcl won't accept a ' ' char in the
                       * name for the slave interpreter :(
                       * The name is ``strsub("wrld"+w->name," ","_");''
                       */
  char *send_buffer;  /* Buffer send commands till a \n arrives, so it
                       * can't interfere with the #$# mcp commands
                       */
#endif
  /* Trigger & highlite stuff */
  int  highlitecount; /* may be uninitialized, highlites != NULL is used */
  int *highlites;

  int  runcount;
  trigger_data **runs; /* may be uninitialized, w->runs > 0 is used */
  
  char *oob_buffer;    /* MCP Out Of Band Buffering */

  /* Widgets for the `send' dialog */
  GtkWidget *senddialog; /* GtkWindow */
  GtkWidget *sendfirst;  /* GtkEntry */
  GtkWidget *sendtext;   /* GtkText */
  GtkWidget *sendlast;   /* GtkEntry */
  
  /* Widgets for the `find' dialog        */
  GtkWidget *finddialog; /* GtkDialog   */
  GtkWidget *findentry;  /* GtkEntry    */
  GtkWidget *findcase;   /* GtkCheckBox */
  GtkWidget *findok;     /* GtkButton   */
  int        findpos;    /* int (duh)   */
  
  /* widgets for the `conect' dialog */
  GtkWidget *conname;
  GtkWidget *conpasswd;

  /* Widgets for displaying the information */
  GtkWidget *label; /* for ther notebook */
  GtkWidget *main; /* vbox of vpaned */
    GtkWidget *hpaned;
      GtkWidget *output_scroll;
      	GtkWidget *output;
      GtkWidget *userlist_scroll;
        GtkWidget *userlist;
      GtkWidget *seperator;
    GtkWidget *ihbox;
      GtkWidget *input;
#if 0
      GtkWidget *clock;    /* label to diplay in */
#endif
};


struct _var {
  char  type;
  
  int   i;
  double f;
  int   strlen;
  char *str;
  int   obj;
  char *error;
  var  *list;
  
  int   listlen;
  var  *nextlistitem;
};


struct _trigger {
  /* name */
  char        *name;
  /* matching */
  unsigned int enabled        : 1;
  char        *pattern;
  unsigned int casesensitive  : 1;
  unsigned int checkremaining : 1;
  unsigned int chance         : 7; /* [0..100] */
  unsigned int match          : 3; /* [0..4] */
  /* color */
  unsigned int color          : 1;
  unsigned int onlymatched    : 1;
  unsigned int fgcolor        : 1;  /* use this color (or not) */
  unsigned int fgcolorr       : 16; /* [0..FFFF] */
  unsigned int fgcolorg       : 16; /* [0..FFFF] */
  unsigned int fgcolorb       : 16; /* [0..FFFF] */
  unsigned int bgcolor        : 1;  /* use this color (or not) */
  unsigned int bgcolorr       : 16; /* [0..FFFF] */
  unsigned int bgcolorg       : 16; /* [0..FFFF] */
  unsigned int bgcolorb       : 16; /* [0..FFFF] */
  /* script */
  unsigned int run            : 1;
  unsigned int istcl          : 1;
  char        *script;
  /* Sound */
  unsigned int beep           : 1;
  unsigned int play           : 1;
  char        *playfile;

  regex_t      preg;
  unsigned int freepreg       : 1;
  trigger     *prev;
  trigger     *next;
};

struct _trigger_data {
  trigger *t;
  int begin;
  int end;
  char *matched;
  char *line;
};

#endif
