/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>
#include <regex.h>

#include "structs.h"
#include "triggers.h"
#include "handy.h"
#include "config.h"
#include "world.h"
#include "help.h"
#include "global_vars.h"
#ifdef TCL
#include "script.h"
#endif

#define mSTART 0
#define mEND   1
#define mCONT  2
#define mWILD  3
#define mREGX  4

trigger *triggers = NULL;


#define do_loop()                \
  for(i = 0; i < len; i++) {     \
    if(t->pattern[i] == '^' ||   \
       t->pattern[i] == '.' ||   \
       t->pattern[i] == '[' ||   \
       t->pattern[i] == '$' ||   \
       t->pattern[i] == '*' ||   \
       t->pattern[i] == '\\') {  \
      s[j] = '\\';               \
      j++;                       \
    }                            \
    s[j] = t->pattern[i];        \
    j++;                         \
  }
/* Formats the pattern for regex use */
char *format_pattern(trigger *t) {
  char *s;
  int i;
  int j;
  int len;
/* have to quote these chars: ^ . [ $ * \  */
  switch(t->match) {
  case mSTART:
    len = strlen(t->pattern);
    s = g_malloc(2 * len + 3);
    s[0] = '^'; /* regex beginning of line */
    j = 1;
    do_loop()
    s[j] = 0;
    break;
  case mEND:
    len = strlen(t->pattern);
    s = g_malloc(2 * len + 3);
    j = 0;
    do_loop()
    s[j] = '$';
    s[j + 1] = 0;
    break;
  case mCONT:
    len = strlen(t->pattern);
    s = g_malloc(2 * len + 2);
    j = 0;
    do_loop()
    s[j] = 0;
    break;
  case mWILD:
    len = strlen(t->pattern);
    s = g_malloc(2 * len + 1);
    j = 0;
    for(i = 0; i < len; i++) {
      if(t->pattern[i] == '^' ||
         t->pattern[i] == '.' ||
         t->pattern[i] == '[' ||
         t->pattern[i] == '$' ||
         t->pattern[i] == '*' ||
         t->pattern[i] == '\\') {
        s[j] = t->pattern[i] == '*' ? '.' : '\\';
        j++;
      }
      s[j] = t->pattern[i];
      j++;
    }
    s[j] = 0;
    break;
  default: /*case mREGX: */
    s = strnew(t->pattern);
    break;
  }
  return s;
}


#define read_str(var)       \
  var = NULL;               \
  while(1) {                \
    i = read(fd, &buf, 1);  \
    if(i != 1) {  \
      stop_loading(t);      \
      return;               \
    }                       \
    strappend(&(var), buf); \
    if(buf == 0)            \
      break;                \
  }

#define read_int(var) \
  i = read(fd, &(buf), 1); \
  if(i != 1) {  \
    stop_loading(t);      \
    return;               \
  } else (var) = buf;

#define read_color(var) \
  i = read(fd, &(buf2), sizeof(int)); \
  if(i != sizeof(int)) {  \
    stop_loading(t);      \
    return;               \
  } else (var) = buf2;

void stop_loading(trigger *t) {
  if(t->name != NULL)
    g_free(t->name);
  if(t->script != NULL)
    g_free(t->script);
  if(t->prev == NULL)
    triggers = NULL;
  else
    t->prev->next = NULL;
  g_free(t);
}

void gM_triggers_load() {
  char *fname = strconcatnew(3, config_dir, "/", "triggers");
  int fd;
  trigger *t;
  char buf;
  unsigned int buf2;
  int i;
  char *match;

  if(exist_file(fname) < 1) {
    printf(_("gMOO: %s doesn't exist... creating\n"), fname);
    fd = open(fname, O_RDWR | O_CREAT, 00644);
    if(fd > -1) {
      close(fd);
      return;
    }
    printf(_("gMOO: Error: Creating \"%s\" failed -> %s\n"), fname, strerror(errno));
    g_free(fname);
    exit(errno);
  }
  fd = open(fname, O_RDONLY);
  g_free(fname);

  i = read(fd, &buf, 1); /* version char (if written) */

  t = triggers = g_malloc(sizeof(trigger));
  t->next = t->prev = NULL;
  t->name = t->pattern = t->script = t->playfile = NULL;
  if(i != 1) { /* error, empty file */
    g_free(triggers);
    triggers = NULL;
    close(fd);
    return;
  } /* Ok, lets start readig the data structs */
  read(fd, &buf, 1); /* the first seperator */
  while(1) {
    read_str(t->name)
    read_int(t->enabled)
    read_str(t->pattern)
    read_int(t->casesensitive)
    read_int(t->checkremaining)
    read_int(t->chance)
    read_int(t->match)
    read_int(t->color)
    read_int(t->onlymatched)
    read_int(t->fgcolor)
    read_color(t->fgcolorr)
    read_color(t->fgcolorg)
    read_color(t->fgcolorb)
    read_int(t->bgcolor)
    read_color(t->bgcolorr)
    read_color(t->bgcolorg)
    read_color(t->bgcolorb)
    read_int(t->run)
    read_int(t->istcl)
    read_str(t->script)
    read_int(t->beep)
    read_int(t->play)
    read_str(t->playfile)

    /* compile the regex pattern buffer */
    match = format_pattern(t);
    i = REG_EXTENDED;
    if(!t->casesensitive)
      i |= REG_ICASE;
    i = regcomp(&t->preg, match, i);
    g_free(match);
    t->freepreg = (i == 0);
    /* t is ``full'' */
  	if(read(fd, &buf, 1) == 1) { /* The seperator for the next trigger has been read */
      t->next = g_malloc(sizeof(trigger));
      t->next->prev = t;
      t = t->next;
      t->next = NULL;
      t->name = t->pattern = t->script = t->playfile = NULL;
    } else
      return;
  }
}
#define write_str(var) \
  write(fd, var, strlen(var) + 1)
#define write_int(var) \
  buf = (var);         \
  write(fd, &buf, 1)
#define write_color(var) \
  buf2 = (var);         \
  write(fd, &buf2, sizeof(int))
void gM_triggers_store() {
  char *fname = strconcatnew(3, config_dir, "/", "triggers");
  int fd;
  trigger *t;
  char buf;
  unsigned int buf2;

  if(! exist_file(fname)) {
    printf(_("gMOO: %s doesn't exist... creating\n"),fname);
    fd = open(fname, O_WRONLY | O_APPEND | O_TRUNC | O_CREAT, 00644);
    if(fd < 0) {
      printf(_("gMOO: Error: Creating \"%s\" failed -> %s\n"), fname, strerror(errno));
      g_free(fname);
      exit(errno);
    }
  } else
    fd = open(fname, O_WRONLY | O_APPEND | O_TRUNC);
  g_free(fname);
  
  buf = 0;
  write(fd, &buf, 1); /* version number */
  /* write all the structs */
  for(t = triggers; t != NULL; t = t->next) {
    buf = '@';          /* seperator */
    write(fd, &buf, 1); /* I use an @ for easy reconition in the hex editor */
    write_str(t->name);
    write_int(t->enabled);
    write_str(t->pattern);
    write_int(t->casesensitive);
    write_int(t->checkremaining);
    write_int(t->chance);
    write_int(t->match);
    write_int(t->color);
    write_int(t->onlymatched);
    write_int(t->fgcolor);
    write_color(t->fgcolorr);
    write_color(t->fgcolorg);
    write_color(t->fgcolorb);
    write_int(t->bgcolor);
    write_color(t->bgcolorr);
    write_color(t->bgcolorg);
    write_color(t->bgcolorb);
    write_int(t->run);
    write_int(t->istcl);
    write_str(t->script);
    write_int(t->beep);
    write_int(t->play);
    write_str(t->playfile);
  }
}

trigger *get_trigger(int num) {
  int i = 0;
  trigger *t = triggers;
  for(; i < num; i++) t = t->next;
  return t;
}

void gM_triggers_get_color(int trig,
                           gushort *fgr, gushort *fgg, gushort *fgb,
                           gushort *bgr, gushort *bgg, gushort *bgb,
                           int *flags) {
  trigger *t = get_trigger(trig);
  *fgr = t->fgcolorr;
  *fgg = t->fgcolorg;
  *fgb = t->fgcolorb;
  *bgr = t->bgcolorr;
  *bgg = t->bgcolorg;
  *bgb = t->bgcolorb;
  *flags = 0;
  *flags |= t->fgcolor ? T_F_FG : 0;
  *flags |= t->bgcolor ? T_F_BG : 0;
}


/******************************************************************
 ******************************************************************
 ******************************************************************
 ******************************************************************
                           Widget stuff
 ******************************************************************
 ******************************************************************
 ******************************************************************
 ******************************************************************/
GtkWidget *window = NULL;

GtkWidget *h_clist;
GtkWidget *h_name;
GtkWidget *h_enabled;
GtkWidget *h_pattern;
GtkWidget *h_casesensitive;
GtkWidget *h_checkremaining;
GtkObject *h_chance_adj;
GtkWidget *h_chance;
GtkWidget *h_match;
GtkWidget *h_match_item[5];
GtkWidget *h_color;
GtkWidget *h_onlymatched;
GtkWidget *h_fgcolor_c;
GtkWidget *h_fgcolor;
GtkWidget *h_bgcolor_c;
GtkWidget *h_bgcolor;
GtkWidget *h_preview;
GtkWidget *h_run;
#ifdef TCL
GtkWidget *h_istcl;
#endif
GtkWidget *h_beep;
GtkWidget *h_play;
GtkWidget *h_playfile;

GtkWidget *b_playfile;
GtkWidget *b_up;
GtkWidget *b_down;
GtkWidget *b_new;
GtkWidget *b_delete;
GtkWidget *b_edit;

int signals[
#ifdef TCL
15
#else
14
#endif
];

void set_sensitives();
void load_entries();
void select_entry_row(GtkCList *clist, int row);
void update_values();
void new_trigger();
void delete_trigger();
void handlers_block(int yes);
void update_preview();
void row_up();
void row_down();
void get_set_script();

#define active(widget) (GTK_TOGGLE_BUTTON(widget)->active)
#define focus() (GTK_CLIST(h_clist)->focus_row)
int triggercount() {
  int i = 0;
  trigger *t;
  for(t = triggers; t != NULL; t = t->next) i++;
  return i;
}
int get_trigger_pos(trigger *trig) {
  int i = 0;
  trigger *t = triggers;
  for(; t != trig; t = t->next) i++;
  return i;
}
int match_history() {
  int i;
  for(i = 0; h_match_item[i] != GTK_OPTION_MENU(h_match)->menu_item; i++);
  return i;
}
void destroy_wid(GtkWidget *but, GtkWidget *destroy) {
  gtk_widget_hide(destroy);
  gtk_widget_destroy(destroy);
}
void select_color_ok(GtkWidget *wid, GtkWidget *but) {
  GtkWidget *csd = gtk_widget_get_toplevel(wid);
  gdouble d[3];
  trigger *t = get_trigger(focus());

  gtk_color_selection_get_color(
    GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(csd)->colorsel), d);
  if(but == h_fgcolor) {
    t->fgcolorr = d[0] * 0xFFFF;
    t->fgcolorg = d[1] * 0xFFFF;
    t->fgcolorb = d[2] * 0xFFFF;
  } else {
    t->bgcolorr = d[0] * 0xFFFF;
    t->bgcolorg = d[1] * 0xFFFF;
    t->bgcolorb = d[2] * 0xFFFF;
  }
  destroy_wid(NULL, csd);
  update_preview();
}
void select_color(GtkWidget *but) {
  trigger *t = get_trigger(focus());
  double d[3] = {(double) (but == h_fgcolor ? t->fgcolorr : t->bgcolorr) / 0xFFFF,
                 (double) (but == h_fgcolor ? t->fgcolorg : t->bgcolorg) / 0xFFFF,
                 (double) (but == h_fgcolor ? t->fgcolorb : t->bgcolorb) / 0xFFFF};
  char  *s    = (but == h_fgcolor ? _("Select a foreground color")
                                  : _("Select a background color"));
  GtkWidget *csd;
  
  csd = gtk_color_selection_dialog_new(s);
  gtk_window_position(GTK_WINDOW(csd), GTK_WIN_POS_MOUSE);
  gtk_window_set_modal(GTK_WINDOW(csd), TRUE);
  gtk_widget_hide(GTK_COLOR_SELECTION_DIALOG(csd)->help_button);
  gtk_signal_connect(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(csd)->cancel_button), "clicked",
                     GTK_SIGNAL_FUNC(destroy_wid), csd);
  gtk_signal_connect(GTK_OBJECT(GTK_COLOR_SELECTION_DIALOG(csd)->ok_button), "clicked", 
                     GTK_SIGNAL_FUNC(select_color_ok), but);
  gtk_color_selection_set_color(
    GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(csd)->colorsel), d);

  gtk_widget_show(csd);
}
void _select_entry_row(GtkWidget *w, int row) {
  gtk_clist_select_row(GTK_CLIST(w), row, 0);
}
void close_window() {
  gtk_widget_destroy(window);
  window = NULL;
  gM_triggers_store();
}

#define add_update_signal(widget, signal) \
  gtk_signal_connect(GTK_OBJECT(widget), signal, GTK_SIGNAL_FUNC(update_values), NULL);
#define make_menu_item(widget, text) \
  widget = gtk_menu_item_new_with_label(text);\
  gtk_widget_show(widget);\
  gtk_menu_append(GTK_MENU(menu), widget);\
  add_update_signal(widget, "activate")

void gM_triggers_show_dialog() {
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *frame;
  GtkWidget *vbox2;
  GtkWidget *scrollw;
  GtkWidget *table;
  GtkWidget *tmp;
  GtkWidget *vbox3;
  GtkWidget *hbb;
  GtkWidget *menu;
  GtkWidget *frame2;
  GtkWidget *vbox4;

  if(window != NULL)
    return;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), _("Triggers & Highlites"));
  gtk_window_set_policy (GTK_WINDOW (window), FALSE, TRUE, FALSE);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC(close_window), NULL);
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  hbox = gtk_hbox_new (FALSE, 10);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbox), 5);

  frame = gtk_frame_new (_("Entries"));
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (hbox), frame, FALSE, TRUE, 0);
//  gtk_widget_set_usize (frame, 130, -2);

  vbox2 = gtk_vbox_new (FALSE, 5);
  gtk_widget_show (vbox2);
  gtk_container_add (GTK_CONTAINER (frame), vbox2);
  gtk_container_border_width (GTK_CONTAINER (vbox2), 5);

  scrollw = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrollw);
  gtk_box_pack_start (GTK_BOX (vbox2), scrollw, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrollw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  h_clist = gtk_clist_new (1);
  gtk_widget_show (h_clist);
  gtk_container_add (GTK_CONTAINER (scrollw), h_clist);
  gtk_clist_set_column_width (GTK_CLIST (h_clist), 0, 80);
  gtk_clist_column_titles_hide (GTK_CLIST (h_clist));
  gtk_signal_connect_after(GTK_OBJECT(h_clist), "select_row", GTK_SIGNAL_FUNC(select_entry_row), NULL);

  table = gtk_table_new (2, 3, FALSE);
  gtk_widget_show (table);
  gtk_box_pack_start (GTK_BOX (vbox2), table, FALSE, TRUE, 0);

  b_up = gtk_button_new_with_label (_("Up"));
  gtk_widget_show (b_up);
  gtk_table_attach (GTK_TABLE (table), b_up, 0, 1, 0, 1,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_signal_connect(GTK_OBJECT(b_up), "clicked", GTK_SIGNAL_FUNC(row_up), NULL);

  b_down = gtk_button_new_with_label (_("Down"));
  gtk_widget_show (b_down);
  gtk_table_attach (GTK_TABLE (table), b_down, 0, 1, 1, 2,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_signal_connect(GTK_OBJECT(b_down), "clicked", GTK_SIGNAL_FUNC(row_down), NULL);

  b_new = gtk_button_new_with_label (_("New"));
  gtk_widget_show (b_new);
  gtk_table_attach (GTK_TABLE (table), b_new, 2, 3, 0, 1,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_signal_connect(GTK_OBJECT(b_new), "clicked", GTK_SIGNAL_FUNC(new_trigger), NULL);

  b_delete = gtk_button_new_with_label (_("Delete"));
  gtk_widget_show (b_delete);
  gtk_table_attach (GTK_TABLE (table), b_delete, 2, 3, 1, 2,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_signal_connect(GTK_OBJECT(b_delete), "clicked", GTK_SIGNAL_FUNC(delete_trigger), NULL);

  tmp = gtk_vseparator_new ();
  gtk_widget_show (tmp);
  gtk_table_attach (GTK_TABLE (table), tmp, 1, 2, 0, 2,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 10, 0);

  vbox2 = gtk_vbox_new (FALSE, 5);
  gtk_widget_show (vbox2);
  gtk_box_pack_start (GTK_BOX (hbox), vbox2, TRUE, TRUE, 0);

  frame = gtk_frame_new (_("Name"));
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (vbox2), frame, FALSE, TRUE, 0);

  vbox3 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox3);
  gtk_container_add (GTK_CONTAINER (frame), vbox3);
  gtk_container_border_width (GTK_CONTAINER (vbox3), 4);

  h_name = gtk_entry_new ();
  gtk_widget_show (h_name);
  gtk_box_pack_start (GTK_BOX (vbox3), h_name, TRUE, TRUE, 0);
  signals[0] = add_update_signal(h_name, "changed");

  frame = gtk_frame_new (_("Matching"));
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (vbox2), frame, TRUE, TRUE, 0);

  table = gtk_table_new (3, 5, FALSE);
  gtk_widget_show (table);
  gtk_container_add (GTK_CONTAINER (frame), table);
  gtk_container_border_width (GTK_CONTAINER (table), 5);

  h_enabled = gtk_check_button_new_with_label (_("Enabled"));
  gtk_widget_show (h_enabled);
  gtk_table_attach (GTK_TABLE (table), h_enabled, 0, 1, 2, 3,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[1] = add_update_signal(h_enabled, "clicked");

  tmp = gtk_label_new (_("Pattern "));
  gtk_widget_show (tmp);
  gtk_table_attach (GTK_TABLE (table), tmp, 0, 1, 1, 2,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);

  tmp = gtk_label_new (_("Lines that "));
  gtk_widget_show (tmp);
  gtk_table_attach (GTK_TABLE (table), tmp, 0, 1, 0, 1,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);

  h_pattern = gtk_entry_new ();
  gtk_widget_show (h_pattern);
  gtk_table_attach (GTK_TABLE (table), h_pattern, 1, 5, 1, 2,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[2] = add_update_signal(h_pattern, "changed");

  h_casesensitive = gtk_check_button_new_with_label (_("Case sensitive"));
  gtk_widget_show (h_casesensitive);
  gtk_table_attach (GTK_TABLE (table), h_casesensitive, 1, 2, 2, 3,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[3] = add_update_signal(h_casesensitive, "clicked");

  h_checkremaining = gtk_check_button_new_with_label (_("Check remaining entries"));
  gtk_widget_show (h_checkremaining);
  gtk_table_attach (GTK_TABLE (table), h_checkremaining, 2, 5, 2, 3,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[4] = add_update_signal(h_checkremaining, "clicked");

  h_chance_adj = gtk_adjustment_new (100, 0, 100, 1, 10, 10);
  h_chance = gtk_spin_button_new (GTK_ADJUSTMENT (h_chance_adj), 1, 0);
  gtk_widget_show (h_chance);
  gtk_table_attach (GTK_TABLE (table), h_chance, 3, 4, 0, 1,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[5] = add_update_signal(h_chance, "changed");

  tmp = gtk_label_new ("%");
  gtk_widget_show (tmp);
  gtk_table_attach (GTK_TABLE (table), tmp, 4, 5, 0, 1,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);

  tmp = gtk_label_new (_("Chance "));
  gtk_widget_show (tmp);
  gtk_table_attach (GTK_TABLE (table), tmp, 2, 3, 0, 1,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_label_set_justify (GTK_LABEL (tmp), GTK_JUSTIFY_LEFT);
  gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);

  h_match = gtk_option_menu_new ();
  gtk_widget_show (h_match);
  gtk_table_attach (GTK_TABLE (table), h_match, 1, 2, 0, 1,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  menu = gtk_menu_new ();
  make_menu_item(h_match_item[0], _("start with"));
  make_menu_item(h_match_item[1], _("end with"));
  make_menu_item(h_match_item[2], _("contain"));
  make_menu_item(h_match_item[3], _("wildcard match"));
  make_menu_item(h_match_item[4], _("regex match"));
  gtk_option_menu_set_menu (GTK_OPTION_MENU (h_match), menu);

  frame = gtk_frame_new (_("Color"));
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (vbox2), frame, TRUE, TRUE, 0);

  table = gtk_table_new (3, 3, FALSE);
  gtk_widget_show (table);
  gtk_container_add (GTK_CONTAINER (frame), table);
  gtk_container_border_width (GTK_CONTAINER (table), 5);

  h_color = gtk_check_button_new_with_label (_("Color the text"));
  gtk_widget_show (h_color);
  gtk_table_attach (GTK_TABLE (table), h_color, 0, 1, 0, 1,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[6] = add_update_signal(h_color, "clicked");
  gtk_signal_connect(GTK_OBJECT(h_color), "clicked", GTK_SIGNAL_FUNC(update_preview), NULL);

  h_onlymatched = gtk_check_button_new_with_label (_("Only matched words"));
  gtk_widget_show (h_onlymatched);
  gtk_table_attach (GTK_TABLE (table), h_onlymatched, 0, 1, 1, 2,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[7] = add_update_signal(h_onlymatched, "clicked");
  gtk_signal_connect(GTK_OBJECT(h_onlymatched), "clicked", GTK_SIGNAL_FUNC(update_preview), NULL);


  h_fgcolor_c = gtk_check_button_new_with_label(_("Foreground color "));
  gtk_widget_show (h_fgcolor_c);
  gtk_table_attach (GTK_TABLE (table), h_fgcolor_c, 1, 2, 0, 1,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[8] = add_update_signal(h_fgcolor_c, "clicked");
  gtk_signal_connect(GTK_OBJECT(h_fgcolor_c), "clicked", GTK_SIGNAL_FUNC(update_preview), NULL);

  h_bgcolor_c = gtk_check_button_new_with_label (_("Background color "));
  gtk_widget_show (h_bgcolor_c);
  gtk_table_attach (GTK_TABLE (table), h_bgcolor_c, 1, 2, 1, 2,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[9] = add_update_signal(h_bgcolor_c, "clicked");
  gtk_signal_connect(GTK_OBJECT(h_bgcolor_c), "clicked", GTK_SIGNAL_FUNC(update_preview), NULL);


  h_fgcolor = gtk_button_new_with_label(" ... ");
  gtk_widget_show (h_fgcolor);
  gtk_table_attach (GTK_TABLE (table), h_fgcolor, 2, 3, 0, 1,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_signal_connect(GTK_OBJECT(h_fgcolor), "clicked", GTK_SIGNAL_FUNC(select_color), NULL);

  h_bgcolor = gtk_button_new_with_label(" ... ");
  gtk_widget_show (h_bgcolor);
  gtk_table_attach (GTK_TABLE (table), h_bgcolor, 2, 3, 1, 2,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_signal_connect(GTK_OBJECT(h_bgcolor), "clicked", GTK_SIGNAL_FUNC(select_color), NULL);

  
  frame2 = gtk_frame_new(NULL);
  gtk_widget_show (frame2);
  gtk_table_attach_defaults(GTK_TABLE(table), frame2, 0, 3, 2, 3);
  gtk_frame_set_shadow_type(GTK_FRAME(frame2), GTK_SHADOW_IN);

  vbox4 = gtk_vbox_new (FALSE, 5);
  gtk_widget_show (vbox4);
  gtk_container_add (GTK_CONTAINER (frame2), vbox4);
  gtk_container_border_width (GTK_CONTAINER (vbox4), 2);

  h_preview = gtk_text_new(NULL, NULL);
  gtk_widget_show(h_preview);
  gtk_container_add(GTK_CONTAINER(vbox4), h_preview);
  gtk_widget_set_usize(h_preview, -1, gdk_char_height(normalfont, 'J') + 10);
  set_global_color(h_preview);
  GTK_WIDGET_UNSET_FLAGS(h_preview, GTK_CAN_FOCUS);

  frame = gtk_frame_new (_("Script"));
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (vbox2), frame, TRUE, TRUE, 0);

  table = gtk_table_new (2, 2, FALSE);
  gtk_widget_show (table);
  gtk_container_add (GTK_CONTAINER (frame), table);
  gtk_container_border_width (GTK_CONTAINER (table), 5);
  gtk_table_set_col_spacings (GTK_TABLE (table), 5);

  h_run = gtk_check_button_new_with_label (_("Run script"));
  gtk_widget_show (h_run);
  gtk_table_attach (GTK_TABLE (table), h_run, 0, 2, 0, 1,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[10] = add_update_signal(h_run, "clicked");
#ifdef TCL
  h_istcl = gtk_check_button_new_with_label (_("Script is Tcl"));
  gtk_widget_show (h_istcl);
  gtk_table_attach (GTK_TABLE (table), h_istcl, 1, 2, 1, 2,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[14] = add_update_signal(h_istcl, "clicked");
#endif
  b_edit = gtk_button_new_with_label (_("  Edit script  "));
  gtk_widget_show (b_edit);
  gtk_table_attach (GTK_TABLE (table), b_edit, 0, 1, 1, 2,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_signal_connect(GTK_OBJECT(b_edit), "clicked", GTK_SIGNAL_FUNC(get_set_script), NULL);


  frame = gtk_frame_new (_("Sound"));
  gtk_widget_show (frame);
  gtk_box_pack_start (GTK_BOX (vbox2), frame, TRUE, TRUE, 0);

  table = gtk_table_new (2, 3, FALSE);
  gtk_widget_show (table);
  gtk_container_add (GTK_CONTAINER (frame), table);
  gtk_container_border_width (GTK_CONTAINER (table), 5);
  gtk_table_set_col_spacings (GTK_TABLE (table), 5);

  h_beep = gtk_check_button_new_with_label (_("Beep the speaker"));
  gtk_widget_show (h_beep);
  gtk_table_attach (GTK_TABLE (table), h_beep, 0, 3, 0, 1,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[11] = add_update_signal(h_beep, "clicked");

  h_play = gtk_check_button_new_with_label (_("Play sound file"));
  gtk_widget_show (h_play);
  gtk_table_attach (GTK_TABLE (table), h_play, 0, 1, 1, 2,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[12] = add_update_signal(h_play, "clicked");

  h_playfile = gtk_entry_new();
  gtk_widget_show (h_playfile);
  gtk_table_attach (GTK_TABLE (table), h_playfile, 1, 2, 1, 2,
                    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  signals[13] = add_update_signal(h_playfile, "changed");

  b_playfile = gtk_button_new_with_label (" ... ");
  gtk_widget_show (b_playfile);
  gtk_table_attach (GTK_TABLE (table), b_playfile, 2, 3, 1, 2,
                    (GtkAttachOptions) GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  /* TODO: (sound) add the callback */

  tmp = gtk_hseparator_new ();
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (vbox), tmp, FALSE, TRUE, 0);

  hbb = gtk_hbutton_box_new ();
  gtk_widget_show (hbb);
  gtk_box_pack_start (GTK_BOX (vbox), hbb, FALSE, FALSE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbb), 5);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbb), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbb), 10);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbb), 0, 0);

  tmp = gtk_button_new_with_label (_("Ok"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);
  gtk_signal_connect_after(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(close_window), NULL);

  tmp = gtk_button_new_with_label (_("Help"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);
  gtk_widget_grab_focus (tmp);
  gtk_widget_grab_default (tmp);
  gtk_signal_connect_after(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(gM_help_show_triggers), NULL);

  load_entries();

  gtk_widget_show(window);
}

void set_sensitives() {
  trigger *t;
  int i;

  if(triggers == NULL) { /* no triggers */
    gtk_widget_set_sensitive(h_name, FALSE);
    gtk_widget_set_sensitive(h_enabled, FALSE);
    gtk_widget_set_sensitive(h_pattern, FALSE);
    gtk_widget_set_sensitive(h_casesensitive, FALSE);
    gtk_widget_set_sensitive(h_checkremaining, FALSE);
    gtk_widget_set_sensitive(h_chance, FALSE);
    gtk_widget_set_sensitive(h_match, FALSE);
    gtk_widget_set_sensitive(h_color, FALSE);
    gtk_widget_set_sensitive(h_onlymatched, FALSE);
    gtk_widget_set_sensitive(h_fgcolor, FALSE);
    gtk_widget_set_sensitive(h_fgcolor_c, FALSE);
    gtk_widget_set_sensitive(h_bgcolor, FALSE);
    gtk_widget_set_sensitive(h_bgcolor_c, FALSE);
    gtk_widget_set_sensitive(h_run, FALSE);
#ifdef TCL
    gtk_widget_set_sensitive(h_istcl, FALSE);
#endif
    gtk_widget_set_sensitive(h_beep, FALSE);
    gtk_widget_set_sensitive(h_play, FALSE);
    gtk_widget_set_sensitive(h_playfile, FALSE);

    gtk_widget_set_sensitive(b_playfile, FALSE);
    gtk_widget_set_sensitive(b_up, FALSE);
    gtk_widget_set_sensitive(b_down, FALSE);
    gtk_widget_set_sensitive(b_delete, FALSE);
    gtk_widget_set_sensitive(b_edit, FALSE);
    return;
  }
  t = get_trigger(focus());
  gtk_widget_set_sensitive(b_up, focus() == 0 ? FALSE : TRUE);
  gtk_widget_set_sensitive(b_down, focus() == triggercount() - 1 ? FALSE : TRUE);
  gtk_widget_set_sensitive(b_delete,         TRUE);
  gtk_widget_set_sensitive(b_edit,           t->enabled && t->run);
  gtk_widget_set_sensitive(b_playfile,       FALSE);

  gtk_widget_set_sensitive(h_name,           t->enabled);
  gtk_widget_set_sensitive(h_enabled,        TRUE);
  gtk_widget_set_sensitive(h_pattern,        t->enabled);
  gtk_widget_set_sensitive(h_casesensitive,  t->enabled);
  gtk_widget_set_sensitive(h_checkremaining, t->enabled);
  gtk_widget_set_sensitive(h_chance,         t->enabled);
  gtk_widget_set_sensitive(h_match,          t->enabled);
  for(i = 0; i < 5; i++)
   gtk_widget_set_sensitive(h_match_item[i], TRUE);
  gtk_widget_set_sensitive(h_color,          t->enabled);
  gtk_widget_set_sensitive(h_onlymatched,    t->enabled);
  gtk_widget_set_sensitive(h_fgcolor_c,      t->enabled && t->color);
  gtk_widget_set_sensitive(h_fgcolor,        t->enabled && t->color && t->fgcolor);
  gtk_widget_set_sensitive(h_bgcolor_c,      t->enabled && t->color);
  gtk_widget_set_sensitive(h_bgcolor,        t->enabled && t->color && t->bgcolor);
  gtk_widget_set_sensitive(h_run,            t->enabled);
#ifdef TCL
  gtk_widget_set_sensitive(h_istcl,          t->enabled && t->run);
#endif
  gtk_widget_set_sensitive(h_beep,           t->enabled);
  gtk_widget_set_sensitive(h_play,           FALSE);
  gtk_widget_set_sensitive(h_playfile,       FALSE);
}

void load_entries() {
  trigger *t;
  char *text[1];
  
  if(triggers == NULL) {
    set_sensitives();
    return;
  }
  
  for(t = triggers; t != NULL; t = t->next) {
    text[0] = t->name;
    gtk_clist_append(GTK_CLIST(h_clist), text);
  }
  gtk_clist_select_row(GTK_CLIST(h_clist), 0, 0);
}

void select_entry_row(GtkCList *clist, int row) {
  trigger *t = get_trigger(row);

  handlers_block(TRUE);

  set_sensitives();

  gtk_entry_set_text(GTK_ENTRY(h_name), t->name);
  gtk_option_menu_set_history(GTK_OPTION_MENU(h_match), t->match);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(h_chance), (float) t->chance);
  gtk_entry_set_text(GTK_ENTRY(h_pattern), t->pattern);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(h_enabled), t->enabled);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(h_casesensitive), t->casesensitive);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(h_checkremaining), t->checkremaining);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(h_color), t->color);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(h_onlymatched), t->onlymatched);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(h_fgcolor_c), t->fgcolor);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(h_bgcolor_c), t->bgcolor);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(h_run), t->run);
#ifdef TCL
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(h_istcl), t->istcl);
#endif
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(h_beep), t->beep);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(h_play), t->play);
  gtk_entry_set_text(GTK_ENTRY(h_playfile), t->playfile);

  update_preview();

  handlers_block(FALSE);
}

void update_values() {
  trigger *t = get_trigger(focus());
  char *match;
  int   i;

  g_free(t->name);
  g_free(t->pattern);
  g_free(t->playfile);

  t->name           = strnew(gtk_entry_get_text(GTK_ENTRY(h_name)));
  t->enabled        = active(h_enabled);
  t->pattern        = strnew(gtk_entry_get_text(GTK_ENTRY(h_pattern)));
  t->casesensitive  = active(h_casesensitive);
  t->checkremaining = active(h_checkremaining);
  t->chance         = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(h_chance));
  t->match          = match_history();
  t->color          = active(h_color);
  t->onlymatched    = active(h_onlymatched);
  t->fgcolor        = active(h_fgcolor_c);
  t->bgcolor        = active(h_bgcolor_c);
  t->run            = active(h_run);
#ifdef TCL
  t->istcl          = active(h_istcl);
#endif
  t->beep           = active(h_beep);
  t->play           = active(h_play);
  t->playfile       = strnew(gtk_entry_get_text(GTK_ENTRY(h_playfile)));

  gtk_clist_set_text(GTK_CLIST(h_clist), focus(), 0, t->name);

  if(t->freepreg)
    regfree(&t->preg);
  match = format_pattern(t);
  i = REG_EXTENDED;
  if(!t->casesensitive)
    i |= REG_ICASE;
  i = regcomp(&t->preg, match, i);
  g_free(match);
  t->freepreg = (i == 0);
    
  set_sensitives();
}

void new_trigger() {
  trigger *t;
  char *text[1];
  
  if(triggers == NULL) {
    t = triggers = g_malloc(sizeof(trigger));
    t->next = t->prev = NULL;
  } else {
    t = triggers;
    triggers        = g_malloc(sizeof(trigger));
    t->prev         = triggers;
    triggers->next  = t;
    triggers->prev  = NULL;
    t               = triggers;
  }
  t->name           = strnew(_("New trigger"));
  t->enabled        = 1;
  t->pattern        = strnew("");
  t->casesensitive  = 0;
  t->checkremaining = 0;
  t->chance         = 100;
  t->match          = mCONT;
  t->color          = 0;
  t->onlymatched    = 1;
  t->fgcolor        = 1;
  t->fgcolorr       = 0;
  t->fgcolorg       = 0;
  t->fgcolorb       = 0;
  t->bgcolor        = 1;
  t->bgcolorr       = 64879;
  t->bgcolorg       = 0xFFFF;
  t->bgcolorb       = 20315;
  t->run            = 0;
  t->istcl          = 0;
  t->script         = strnew("");
  t->beep           = 0;
  t->play           = 0;
  t->playfile       = strnew("");
  t->freepreg       = 0;
  text[0] = t->name;
  gtk_clist_insert(GTK_CLIST(h_clist), 0, text);
  GTK_CLIST(h_clist)->focus_row = 0;
  gtk_clist_select_row(GTK_CLIST(h_clist), 0, 0);
}

void delete_trigger() {
  int row = focus();
  trigger *t = get_trigger(row);
  
  if(t == triggers) {
    triggers = t->next;
  }
  if(t->next != NULL)
    t->next->prev = t->prev;
  if(t->prev != NULL)
    t->prev->next = t->next;
  g_free(t->name);
  g_free(t->pattern);
  g_free(t->script);
  g_free(t->playfile);
  if(t->freepreg)
    regfree(&t->preg);
  g_free(t);
  gtk_clist_remove(GTK_CLIST(h_clist), row);
  if(triggers == NULL) {
    set_sensitives();
    update_preview();
  } else
    gtk_clist_select_row(GTK_CLIST(h_clist), row < triggercount() ? row : triggercount() - 1, 0);
    
}

void handlers_block(int yes) {
  if(yes) {
    gtk_signal_handler_block(GTK_OBJECT(h_name), signals[0]);
    gtk_signal_handler_block(GTK_OBJECT(h_enabled), signals[1]);
    gtk_signal_handler_block(GTK_OBJECT(h_pattern), signals[2]);
    gtk_signal_handler_block(GTK_OBJECT(h_casesensitive), signals[3]);
    gtk_signal_handler_block(GTK_OBJECT(h_checkremaining), signals[4]);
    gtk_signal_handler_block(GTK_OBJECT(h_chance), signals[5]);
    gtk_signal_handler_block(GTK_OBJECT(h_color), signals[6]);
    gtk_signal_handler_block(GTK_OBJECT(h_onlymatched), signals[7]);
    gtk_signal_handler_block(GTK_OBJECT(h_fgcolor_c), signals[8]);
    gtk_signal_handler_block(GTK_OBJECT(h_bgcolor_c), signals[9]);
    gtk_signal_handler_block(GTK_OBJECT(h_run), signals[10]);
#ifdef TCL
    gtk_signal_handler_block(GTK_OBJECT(h_istcl), signals[14]);
#endif
    gtk_signal_handler_block(GTK_OBJECT(h_beep), signals[11]);
    gtk_signal_handler_block(GTK_OBJECT(h_play), signals[12]);
    gtk_signal_handler_block(GTK_OBJECT(h_playfile), signals[13]);
  } else {
    gtk_signal_handler_unblock(GTK_OBJECT(h_name), signals[0]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_enabled), signals[1]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_pattern), signals[2]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_casesensitive), signals[3]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_checkremaining), signals[4]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_chance), signals[5]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_color), signals[6]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_onlymatched), signals[7]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_fgcolor_c), signals[8]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_bgcolor_c), signals[9]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_run), signals[10]);
#ifdef TCL
    gtk_signal_handler_unblock(GTK_OBJECT(h_istcl), signals[14]);
#endif
    gtk_signal_handler_unblock(GTK_OBJECT(h_beep), signals[11]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_play), signals[12]);
    gtk_signal_handler_unblock(GTK_OBJECT(h_playfile), signals[13]);
  }
}

void update_preview() {
  GdkColor fg, bg;
  trigger *t;

  gtk_editable_delete_text(GTK_EDITABLE(h_preview), 0, gtk_text_get_length(GTK_TEXT(h_preview)));
  if(triggers == NULL) {
    return;
  }
  t = get_trigger(focus());
  fg.pixel = 0;
  fg.red   = t->fgcolorr;
  fg.green = t->fgcolorg;
  fg.blue  = t->fgcolorb;
  bg.pixel = 0;
  bg.red   = t->bgcolorr;
  bg.green = t->bgcolorg;
  bg.blue  = t->bgcolorb;

  if(t->color) {
    if(t->onlymatched) {
      gtk_text_insert(GTK_TEXT(h_preview), NULL, NULL, NULL,
                      "That ", -1);
      gtk_text_insert(GTK_TEXT(h_preview), NULL, t->fgcolor ? &fg : NULL,
                                                       t->bgcolor ? &bg : NULL,
                            "quick brown fox", -1);
      gtk_text_insert(GTK_TEXT(h_preview), NULL, NULL, NULL,
                                         " seems to jump a lot  :)", -1);
    } else {
      gtk_text_insert(GTK_TEXT(h_preview), NULL, t->fgcolor ? &fg : NULL,
                                                       t->bgcolor ? &bg : NULL,
                      "That quick brown fox seems to jump a lot  :)", -1);
    }
  } else {
    gtk_text_insert(GTK_TEXT(h_preview), NULL, NULL, NULL,
                    "That quick brown fox seems to jump a lot  :)", -1);
  }
}

void row_up() {
  int row = focus();
  trigger *t = get_trigger(row);
  
  if(triggers == t->prev) { /* switch the 1st 2 */
    triggers = t;
  }
  /* remove from list */
  t->prev->next = t->next;
  if(t->next != NULL)
    t->next->prev = t->prev;
  /* insert again */
  t->next = t->prev;
  t->prev = t->next->prev;
  t->next->prev = t;
  if(t->prev != NULL)
    t->prev->next = t;
  gtk_clist_swap_rows(GTK_CLIST(h_clist), row, row - 1);
  GTK_CLIST(h_clist)->focus_row = row - 1;
  set_sensitives();
}

void row_down() {
  int row = focus();
  trigger *t = get_trigger(row);
  
  if(triggers == t) { /* switch the 1st 2 */
    triggers = t->next;
  }
  /* remove from list */
  if(t->prev != NULL)
    t->prev->next = t->next;
  t->next->prev = t->prev;
  /* insert again */
  t->prev = t->next;
  t->next = t->prev->next;
  t->prev->next = t;
  if(t->next != NULL)
    t->next->prev = t;
  gtk_clist_swap_rows(GTK_CLIST(h_clist), row, row + 1);
  GTK_CLIST(h_clist)->focus_row = row + 1;
  set_sensitives();
}

void set_0(GtkWidget *dummy, int *i) { *i = 0; }
void set_1(GtkWidget *dummy, int *i) { *i = 1; }

void get_set_script() {
  GtkWidget *se_window;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *text;
  GtkWidget *tmp;
  GtkWidget *hbb;
  
  trigger *t = get_trigger(focus());
  int br;

  GtkStyle style;

  se_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_object_set_data (GTK_OBJECT (se_window), "se_window", se_window);
  gtk_window_set_title (GTK_WINDOW (se_window), _("Edit script"));
  gtk_window_position (GTK_WINDOW (se_window), GTK_WIN_POS_MOUSE);
  gtk_window_set_policy (GTK_WINDOW (se_window), TRUE, TRUE, FALSE);
  gtk_window_set_modal(GTK_WINDOW(se_window), TRUE);
  gtk_widget_set_usize(se_window, 400, -1);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_object_set_data (GTK_OBJECT (se_window), "vbox", vbox);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (se_window), vbox);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_object_set_data (GTK_OBJECT (se_window), "hbox", hbox);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbox), 5);

  text = gtk_text_new (NULL, NULL);
  gtk_object_set_data (GTK_OBJECT (se_window), "text", text);
  gtk_widget_show (text);
  gtk_box_pack_start (GTK_BOX (hbox), text, TRUE, TRUE, 0);
  gtk_editable_set_editable(GTK_EDITABLE(text), TRUE);

  memcpy(&style, text->style, sizeof(GtkStyle));
  style.font = normalfont;
  gtk_widget_set_style(text, &style);


  tmp = gtk_vscrollbar_new (GTK_ADJUSTMENT (gtk_adjustment_new (0, 0, 0, 0, 0, 0)));
  gtk_object_set_data (GTK_OBJECT (se_window), "tmp", tmp);
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (hbox), tmp, FALSE, TRUE, 0);

  tmp = gtk_hseparator_new ();
  gtk_object_set_data (GTK_OBJECT (se_window), "tmp", tmp);
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (vbox), tmp, FALSE, TRUE, 5);

  hbb = gtk_hbutton_box_new ();
  gtk_object_set_data (GTK_OBJECT (se_window), "hbb", hbb);
  gtk_widget_show (hbb);
  gtk_box_pack_start (GTK_BOX (vbox), hbb, FALSE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbb), 5);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbb), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbb), 10);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbb), 0, 0);

  tmp = gtk_button_new_with_label (_("Ok"));
  gtk_object_set_data (GTK_OBJECT (se_window), "tmp", tmp);
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(set_0), &br);

  tmp = gtk_button_new_with_label (_("Cancel"));
  gtk_object_set_data (GTK_OBJECT (se_window), "tmp", tmp);
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);
  gtk_widget_grab_default (tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(set_1), &br);

  gtk_widget_grab_focus(text);

  gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, t->script, -1);

  gtk_widget_show(se_window);

  br = -1;

  while(br < 0)
    gtk_main_iteration_do(TRUE);

  if(br == 0) { /* Ok pressed */
    g_free(t->script);
    t->script = gtk_editable_get_chars(GTK_EDITABLE(text), 0, gtk_text_get_length(GTK_TEXT(text)));
  }
  
  gtk_widget_hide(se_window);
  gtk_widget_destroy(se_window);
}


/******************************************************************
 ******************************************************************
 ******************************************************************
 ******************************************************************
                           Parsing stuff
 ******************************************************************
 ******************************************************************
 ******************************************************************
 ******************************************************************/

/* returns TRUE on match, then begin is set to the 1st char, and
   end is set to the last + 1 */
int   matches (world *w, trigger *t, char *line, int len,
              int *count, int **begin, int **end);
/* Inser the T_START and T_END char in the string */
void do_color(world *w, char *line, int len, trigger *t,
  	           int count, int *begin, int *end);
/* Schedule the script for running */
void  do_add_run  (world *w, trigger *t, int begin, int end, char *line);

/* Play the sound file */
void  do_play (world *w, trigger *t);

int gM_triggers_check(world *w, char *line, int len) {
  trigger *t;
  int count, *begin, *end;
  int ret = FALSE;

  if(triggers == NULL || w == NULL || len == 0)
    return FALSE;

  for(t = triggers; t != NULL; t = t->next) {
    if(t->enabled) {
      if(matches(w, t, line, len, &count, &begin, &end)) {
#ifdef DEBUG
printf("*** Line matched %d time(s) to a trigger\n", count);
#endif
        if(t->color) {
          if(!(t->onlymatched)) {
            if(count != 1) {
              count = 1;
              g_free(begin);
              g_free(end);
              begin = g_malloc(sizeof(int));
              end   = g_malloc(sizeof(int));
            }
            begin[0] = 0;
            end[0] = len;
          }
          do_color(w, line, len, t, count, begin, end);
        }
        if(t->run)
          do_add_run(w, t, begin[0], end[0], line);
        if(t->beep)
          gdk_beep();
        if(t->play)
          do_play(w, t);
        g_free(begin);
        g_free(end);
        ret = TRUE;
        if(!(t->checkremaining))
          break;
      } /* matches */
    } /* enabled */
  } /* for-loop */
  return ret;
}

int matches (world *w, trigger *t, char *line, int len,
              int *count, int **begin, int **end) {
  regmatch_t *matched;
  int i,j,prefix;

  if(!t->freepreg) /* There was an error compiling it ... */
    return FALSE;


  matched = g_malloc(sizeof(regmatch_t) * len); /* there can't be more matches */
  i = regexec(&t->preg, line, len, matched, 0);

  if(i == 0) { /* match */
    j = 0;
    prefix = 0;
    *begin = g_malloc(sizeof(int) * len);
    *end   = g_malloc(sizeof(int) * len);
    do {
      for(i = 0; matched[i].rm_so != -1; i++);
      for(i--; i > -1; i--) {
        (*begin)[j] = prefix + matched[i].rm_so;
        (*end)[j]   = prefix + matched[i].rm_eo;
        j++;
      }
      prefix = (*end)[j - 1];
    } while(regexec(&t->preg, line + prefix, len - prefix, matched, REG_NOTBOL) == 0);
    *count = j;
    g_free(matched);
    return TRUE;
  }
  g_free(matched);
  return FALSE;

}

void do_add_color(world *w, int trig, int begin, int end) {
  int i;

  for(i = begin; i < end; i++)
    if(w->highlites[i] == -1)
      w->highlites[i] = trig;
}

void do_color(world *w, char *line, int len, trigger *t,
  	           int count, int *begin, int *end) {
  int i;
  int trig = get_trigger_pos(t);

  if(w->highlites == NULL) {
    w->highlites = malloc(sizeof(int) * len);
    for(i = 0; i < len; i++)
      w->highlites[i] = -1;
    w->highlitecount = len;
  }
  for(i = 0; i < count; i++) {
    do_add_color(w, trig, begin[i], end[i]);
  }
}

void do_add_run(world *w, trigger *t, int begin, int end, char *line) {
  trigger_data **tmp;
  int i;
#ifdef TCL
  char *s;
#endif
  w->runcount++;
  tmp = w->runs;
  w->runs = g_malloc(sizeof(trigger_data *) * w->runcount);
  for(i = 0; i < w->runcount - 1; i++) {
    w->runs[i] = tmp[i];
  }

  w->runs[i] = g_malloc(sizeof(trigger_data));
  w->runs[i]->t = t;
#ifdef TCL
  w->runs[i]->begin = begin;
  w->runs[i]->end = end;
  w->runs[i]->line = quote_for_tcl(line);
  s = strnew(line + begin);
  s[end - begin] = 0;
  w->runs[i]->matched = quote_for_tcl(s);
  g_free(s);
#endif
}

void gM_triggers_run(world *w, trigger_data *tdata) {
  int len;
  trigger *t = tdata->t;
#ifdef TCL
  char *script;
  char *tmp;
  char *tmp2;
  static int ID = 0;
#endif

  if(t->script[0] == 0) /* pretty empty eh :) */
    return;
#ifdef TCL
  if(t->istcl) {
    tmp = g_malloc(40);
    sprintf(tmp, "%d %d ", tdata->begin, tdata->end);
    tmp2 = g_malloc(20);
    sprintf(tmp2, "_%d", ID);
    script = strconcatnew(13, "proc _trigger_",
                             w->tcl_name,
                             tmp2,
                             " {begin end matched line} {\n",
                             t->script,
                             "\n}\n _trigger_",
                             w->tcl_name,
                             tmp2,
                             " ",
                             tmp,
                             tdata->matched,
                             " ",
                             tdata->line
                             );
    strsub(script, '\n', ';');
    gM_script_eval_script(w, script);
    g_free(script);
    g_free(tmp);
    g_free(tmp2);
    ID++;
    return;
  }
#endif
  len = strlen(t->script);
  gM_world_write(w, t->script);
  if(t->script[len - 1] != 10)
  gM_world_write(w, "\n");
}

void do_play (world *w, trigger *t) {
  printf("//TODO: Add sound support and play the soundfile\n");
}

