/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "structs.h"
#include "window.h"
#include "userlist.h"
#include "handy.h"
#include "fromval.h"
#include "mcp.h"
#include "config.h"
#include "world.h"
#include "worlds.h"
#include "global_vars.h"

#define LIST_INIT    0  /*  =  */
#define LIST_ADD     1  /*  +  */
#define LIST_REMOVE  2  /*  -  */
#define LIST_IDLE    3  /*  <  */
#define LIST_UNIDLE  4  /*  >  */
#define LIST_CLOAK   5  /*  (  */
#define LIST_UNCLOAK 6  /*  )  */
#define LIST_AWAY    7  /*  [  */
#define LIST_UNAWAY  8  /*  ]  */
#define LIST_UPDATE  9  /*  *  */

struct _usermenudata {
  world *world;
  mcpuserlist *ul;
  GtkWidget *menuwidget;
  int row;
};
struct _usermenudata usermenudata = {NULL, NULL, NULL, -1};

/* UTILS */
void clist_add_user(world *w, GtkWidget *clist, mcpuser *user, int pos);
void resize_cb(GtkWidget *userlist, GtkAllocation *allocation, world *w);
int  resize_col_cb(GtkWidget *userlist, GtkAllocation *allocation, world *w);

void set_type(int *type);

void do_users(world *w, var *v, int type);
void do_user(world *w, var *v, int type);

void list_clear(world *w);
void user_add(world *w, var *v);
void do_user_add(world *w, mcpuser *u);
void user_remove(world *w, var *v);
void user_idle(world *w, var *v);
void do_user_add_idle(world *w, mcpuser *u);
void user_unidle(world *w, var *v);
void user_cloak(world *w, var *v);
void user_uncloak(world *w, var *v);
void user_away(world *w, var *v);
void user_unaway(world *w, var *v);
void user_update(world *w, var *v);

mcpuser *mcpuser_new(var *v);
void     mcpuser_free(mcpuser *u);
mcpuser *mcpuser_find(mcpuserlist *ul, var *v);
void mcpuser_remove(mcpuserlist *ul, mcpuser *u);

void mcpicons_free(mcpicon *icon);
mcpicon *mcpicon_new(char *name);
GdkPixmap *get_userlistpix(world *w, mcpuser *u);
GdkBitmap *get_userlistmask(world *w, mcpuser *u);
void set_userlistpixmap(world *w, GtkWidget *clist, mcpuser *u);

void userlistmenus_new(mcpuserlist *ul, var *v);
void userlistmenus_free(mcpuserlist *ul);

void button_pressed(GtkCList *clist, GdkEventButton *e, world *w);
void menu_pressed(GtkMenuItem *mitem, int i);

char *parse_menu_string(mcpuserlist *ul, int pos, char *text);

/* returns:
 * -1 if u1 must be inserted befor u2
 *  1 if u1 must be added after u2
 */
int compare_users(mcpuser *u1, mcpuser *u2);

void userlisttitleclick(GtkWidget *columnbutton, int column, world *thisw);


void gM_userlist_init(world *w) {
  char *titles[] = {_("I"), _("Name"), _("Obj #")};

  w->userlist_scroll = gtk_scrolled_window_new(NULL, NULL);
  gtk_widget_show(w->userlist_scroll);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(w->userlist_scroll), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  gtk_paned_pack2(GTK_PANED(w->hpaned), w->userlist_scroll, FALSE, FALSE);
  GTK_WIDGET_UNSET_FLAGS(w->userlist_scroll, GTK_CAN_FOCUS);

  w->userlist = gtk_clist_new_with_titles(3, titles);
  gtk_widget_show(w->userlist);
  GTK_WIDGET_UNSET_FLAGS(w->userlist, GTK_CAN_FOCUS);
  gtk_container_add(GTK_CONTAINER(w->userlist_scroll), w->userlist);
  
  /* Icon column */
  gtk_clist_set_column_width(GTK_CLIST(w->userlist),       0, 16);
  gtk_clist_column_title_passive(GTK_CLIST(w->userlist),   0);
  /* Name column */
  gtk_clist_set_column_auto_resize(GTK_CLIST(w->userlist), 1, TRUE);
  /* Object culumn */
  gtk_clist_set_column_auto_resize(GTK_CLIST(w->userlist), 2, TRUE);

  gtk_signal_connect(GTK_OBJECT(w->userlist), "click_column", userlisttitleclick, w);
  gtk_signal_connect(GTK_OBJECT(w->userlist), "button_press_event", button_pressed, w);

  gtk_clist_set_row_height(GTK_CLIST(w->userlist), 16);

  w->userlist_on = 0;
}

void gM_userlist_show(world *w) {
  gtk_widget_reparent(w->output_scroll, w->hpaned);

  GTK_PANED(w->hpaned)->child1_shrink = TRUE;
  GTK_PANED(w->hpaned)->child1_resize = TRUE;
  GTK_PANED(w->hpaned)->child2_shrink = FALSE;
  GTK_PANED(w->hpaned)->child2_resize = FALSE;

  w->userlist_on = 1;
  
  if(multiline > 0) {
    gtk_paned_pack1(GTK_PANED(w->main), w->hpaned, TRUE, TRUE);
    gtk_widget_unref(w->hpaned);
  }
  gtk_widget_show(w->hpaned);
}

void gM_userlist_hide(world *w) {
  w->userlist_on = 0;

  if(multiline > 0) {
    gtk_widget_ref(w->hpaned);
    gtk_container_remove(GTK_CONTAINER(w->main), w->hpaned);
    gtk_widget_ref(w->output_scroll);
    gtk_container_remove(GTK_CONTAINER(w->hpaned), w->output_scroll);
    gtk_paned_pack1(GTK_PANED(w->main), w->output_scroll, TRUE, TRUE);
  	gtk_widget_unref(w->output_scroll);
  } else {
    gtk_widget_hide(w->hpaned);
    gtk_widget_reparent(w->output_scroll, w->main);
  }
}


void gM_userlist_toggle(GtkWidget *ulmenu, void *pointer) {
  world *w = gM_window_get_current_world();

  if(w == NULL)
    return;
  gtk_signal_handler_block(GTK_OBJECT(userlistmenu), userlistsignal);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(userlistmenu), GTK_TOGGLE_BUTTON(ulmenu)->active);
  gtk_signal_handler_unblock(GTK_OBJECT(userlistmenu), userlistsignal);

  if(GTK_TOGGLE_BUTTON(ulmenu)->active > 0) {
    gM_userlist_show(w);
  } else {
    gM_userlist_hide(w);
  }
}

void gM_userlist_player_obj(mcpuserlist *ul, char *data) {
  var *v;
  
  if(ul == NULL)
    return;
  v = str_to_var(data);
  if(v == NULL)
    return;
  if(v->type == TYPE_OBJ) {
    ul->player_obj = v->obj;
/*printf("\e[32m-> Succes in determining player object: #%d\e[0m\n", ul->player_obj);*/
  }
  var_free(v);
}

void gM_userlist_friends(mcpuserlist *ul, char *data) {
  var *v, *v2;
  int i,j;
  
  v = str_to_var(data);
  if(v == NULL)
    return;
  if(v->type == TYPE_OBJ) {
    if(ul->friendscount > 0)
      g_free(ul->friends);
    ul->friendscount = 1;
    ul->friends = g_malloc(sizeof(int));
    ul->friends[0] = v->obj;
  } else if(v->type == TYPE_LIST) {
    if(ul->friendscount != 0)
      g_free(ul->friends);
    ul->friendscount = v->listlen;
    if(ul->friendscount > 0) {
      ul->friends = g_malloc(sizeof(int) * ul->friendscount);
      v2 = v->list;
      j = 0;
      for(i = 0; i < ul->friendscount; i++) {
        if(v2->type == TYPE_OBJ) {
          ul->friends[j] = v2->obj;
          j++;
        } else
          ul->friendscount--;
        v2 = v2->nextlistitem;
      }
    }
  }
/*
for(i=0;i<ul->friendscount;i++)
printf("\e[32m   Found a friend: #%d\e[0m\n", ul->friends[i]); */
  var_free(v);
}

void gM_userlist_add_menu(mcpuserlist *ul, char *data) {
  var *v, *v2;
  int i;
/*printf("Kijken naar de menu dingus...\n");*/
  v = str_to_var(data);
  if(v == NULL)
    return;
  if(v->type == TYPE_LIST && v->listlen > 0) {
/*printf("Okee, het is een list, klopt die: ");*/
    v2 = v->list;
    for(i = 0; i < v->listlen; i++) {
      if( !(v2->type == TYPE_LIST && v2->listlen == 2 &&
            v2->list->type == TYPE_STRING &&
            v2->list->nextlistitem->type == TYPE_STRING) &&
          !(v2->type == TYPE_INT && v2->i == 0)) {
/*printf("NEE\n");*/
        var_free(v);
        return;
      }
    }
/*printf("JA\n");*/
    userlistmenus_new(ul, v);
  }
/*else printf("Het is geen list\n");*/
  var_free(v);
}

/***********************************
 *
 * GLOBAL PARSING AND ADDING STUFF
 *
 ***********************************/
void gM_userlist_free(mcpuserlist *ul) {
  int i;
  mcpuser *u = NULL, *u2 = NULL;
  
  if(ul == NULL)
    return;
  for(u = ul->list; u != NULL; u = u->next) {
    u2 = u->prev;
    g_free(u->name);
    if(u2 != NULL)
      g_free(u2);
  }
  if(u2 != NULL)
    g_free(u2);
  if(u != NULL)
    g_free(u);
  mcpicons_free(ul->icons);
  for(i = 0; i < ul->fieldcount; i++)
    g_free(ul->fields[i]);
  if(ul->fields != NULL)
    g_free(ul->fields);
  if(ul->friendscount > 0)
    g_free(ul->friends);
  userlistmenus_free(ul);
}

void gM_userlist_do_icons(mcpuserlist *ul, char *data) {
  var *v;
  int i;
  mcpicon *icon;
/*printf("Adding icons line to the userlist\n");*/
  
  if(data == NULL || ul->icons != NULL)
    return;
  v = str_to_var(data);
  if(v == NULL) {
/*printf("\e[31m** Error in parsing >>%s<<\e[0m\n", data);*/
  	return;
  }
/*
printf("-> ");
gM_var_print(v);
printf("\n");
*/
  if(v->type == TYPE_LIST)
    ul->iconcount = v->listlen;
  else if(v->type == TYPE_STRING)
    ul->iconcount = 1;
  else { /* error */
/*printf("\e[31m** Error, not a list or a string value\e[0m\n");*/
    var_free(v);
    return;
  }
  if(ul->iconcount != 0) {/* empty list -> ignore it */
    if(v->type == TYPE_STRING)
      ul->icons = mcpicon_new(v->str);
    else {
      v = v->list;
      icon = ul->icons = mcpicon_new(v->str);
      v = v->nextlistitem;
      for(i = 1; i < ul->iconcount; i++) {
        if(v->type == TYPE_STRING) {
          icon->next = mcpicon_new(v->str);
          icon = icon->next;
        } else
          ul->iconcount--;
        v = v->nextlistitem;
      }
    }
/*printf("\e[32m-> %d icons found (and added ?)\e[0m\n", ul->iconcount);*/
  }
/*else printf("\e[31m** Error, empty list not valid\e[0m\n");*/
  var_free(v);
/*printf("\e[32mAdding complete\e[0m\n\n");*/
}

void gM_userlist_do_fields(mcpuserlist *ul, char *data) {
  var *v;
  int i;
/*printf("Adding fields line to the userlist\n");*/
  
  if(data == NULL || ul->fields != NULL) {
/*if(data == NULL) printf("\e[31m** No data in the fields line\e[0m\n");*/
/*if(ul->fields != NULL) printf("\e[31m** ul->fields != NULL\e[0m\n");*/
    return;
  }
  v = str_to_var(data);
  if(v == NULL) {
/*printf("\e[31m** Error in parsing >>%s<<\e[0m\n", data);*/
  	return;
  }
/*
printf("-> ");
gM_var_print(v);
printf("\n");
*/
  if(v->type == TYPE_LIST)
    ul->fieldcount = v->listlen;
  else if(v->type == TYPE_STRING)
    ul->fieldcount = 1;
  else { /* error */
/*printf("\e[31m** Error, not a list or a string value\e[0m\n");*/
    var_free(v);
    return;
  }
  if(ul->fieldcount != 0) {/* empty list -> ignore it */
    ul->fields = g_malloc(sizeof(char *) * ul->fieldcount);
    if(v->type == TYPE_STRING) {
      ul->fields[0] = strcpy(g_malloc(v->strlen), v->str);
/*printf("   Field detected: >>%s<<\n", v->str);*/
    } else {
      v = v->list;
      for(i = 0; i < ul->fieldcount; i++) {
        ul->fields[i] = strcpy(g_malloc(v->strlen), v->str);
/*printf("   Field detected: >>%s<<\n", v->str);*/
        v = v->nextlistitem;
      }
    }
/*printf("\e[32m-> Succes, %d fields registrated\e[0m\n", ul->fieldcount);*/
  }
/*else printf("\e[31m** Error, empty list not valid\e[0m\n");*/
  var_free(v);
/*printf("\n");*/
}

void gM_userlist_do_users(world *w, char *data) {
  var *v;
  int type;
/*printf("Updating/adding/removing (a) user(s) [from/to] the userlist\n");*/
  
  if(data == NULL)
    return;

  if(w->session->userlist->iconcount < 1) {
/*printf("\e[31m** Error: no iconlist has been defined\e[0m\n");*/
  	return;
  } else if(w->session->userlist->fieldcount < 1) {
/*printf("\e[31m** Error: no fields have been defined\e[0m\n");*/
  	return;
  }

  while(data[0] == ' ')
    data++;
  type = data[0];
  data++;
  while(data[0] == ' ')
    data++;

  v = str_to_var(data);
  if(v == NULL || v->type == TYPE_EMPTY) {
/*printf("\e[31m** Error in parsing >>%s<<\e[0m\n", data);*/
  	return;
  }

  set_type(&type);
/*
printf("-> ");
gM_var_print(v);
printf("\n");
*/
  if(v->type == TYPE_LIST) {
    do_users(w, v, type);
  }
/*else printf("\e[31m** Error, value isn't a list\e[0m\n");*/

  var_free(v);
/*printf("\n");*/
}

/****************
 *
 *  Utils
 ****************/
 
void clist_add_user(world *w, GtkWidget *clist, mcpuser *user, int pos) {
  char *empty[]  = {"", "", ""};
  char s[20];
/*printf("   Adding user \"%s\" to the list\n", user->name);*/

  gtk_clist_insert(GTK_CLIST(clist), pos, empty);
  set_userlistpixmap(w, clist, user);
  gtk_clist_set_text(GTK_CLIST(clist), pos, 1, user->name);
  sprintf(s, "%d", user->object);
  gtk_clist_set_text(GTK_CLIST(clist), pos, 2, s);
}

void resize_cb(GtkWidget *userlist, GtkAllocation *allocation, world *w) {
  int delta;
  GtkCList *clist = GTK_CLIST(userlist);
/*printf("resized to %dx%d+%d+%d\n", allocation->width, allocation->height, allocation->x, allocation->y);*/

  delta = clist->column[2].area.width - clist->column[2].width;
  if (delta > 0){

      gtk_clist_set_column_width (clist, 1, clist->column[1].width + delta);
      return;
  }

}

int resize_col_cb(GtkWidget *userlist, GtkAllocation *allocation, world *w) {
return FALSE;
}



void set_type(int *type) {
  switch(*type) {
  case '=': 
/*printf("-> *type == LIST_INIT\n");*/
    *type = LIST_INIT;
    break;
  case '+':
/*printf("-> *type == LIST_ADD\n");*/
    *type = LIST_ADD;
    break;
  case '-':
/*printf("-> *type == LIST_REMOVE\n");*/
    *type = LIST_REMOVE;
    break;
  case '<':
/*printf("-> *type == LIST_IDLE\n");*/
    *type = LIST_IDLE;
    break;
  case '>':
/*printf("-> *type == LIST_UNIDLE\n");*/
    *type = LIST_UNIDLE;
    break;
  case '(':
/*printf("-> *type == LIST_CLOAK\n");*/
    *type = LIST_CLOAK;
    break;
  case ')':
/*printf("-> *type == LIST_UNCLOAK\n");*/
    *type = LIST_UNCLOAK;
    break;
  case '[':
/*printf("-> *type == LIST_AWAY\n");*/
    *type = LIST_AWAY;
    break;
  case ']':
/*printf("-> *type == LIST_UNAWAY\n");*/
    *type = LIST_UNAWAY;
    break;
  case '*':
/*printf("-> *type == LIST_UPDATE\n");*/
    *type = LIST_UPDATE;
    break;
  default:
/*printf("\e[31m** Error in getting *type, '%c' is not supported\e[0m\n", *type);*/
    *type = -1;
    break;
  }
}

void do_users(world *w, var *v, int type) {
  var *v2;
  int i;


  if((type == LIST_ADD || type == LIST_UPDATE) &&
    v->list->type != TYPE_LIST)
    do_user(w, v, type);
  else {
    if(type == LIST_INIT) {
      list_clear(w);
      type = LIST_ADD;
    }
    v2 = v->list;
    for(i = 0; i < v->listlen; i++) {
      do_user(w, v2, type);
      v2 = v2->nextlistitem;
    }
  }
}

void do_user(world *w, var *v, int type) {
  switch(type) {
  case LIST_ADD:     user_add(w, v);     break;
  case LIST_REMOVE:  user_remove(w, v);  break;
  case LIST_IDLE:    user_idle(w, v);    break;
  case LIST_UNIDLE:  user_unidle(w, v);  break;
  case LIST_CLOAK:   user_cloak(w, v);   break;
  case LIST_UNCLOAK: user_uncloak(w, v); break;
  case LIST_AWAY:    user_away(w, v);    break;
  case LIST_UNAWAY:  user_unaway(w, v);  break;
  case LIST_UPDATE:  user_update(w, v);  break;
/*default: printf("HUH: wierd event, type == %d\n", type);*/
  }
}



void list_clear(world *w) {
  mcpuser *u;

/*printf("   Handler: clearing list\n");*/
  u = w->session->userlist->list;
  while(u != NULL) {
    g_free(u->name);
    if(u->next == NULL) {
      g_free(u);
      break;
    }
    u = u->next;
    g_free(u->prev);
  }
  w->session->userlist->list = NULL;
  w->session->userlist->listlen = 0;
  
  gtk_clist_clear(GTK_CLIST(w->userlist));
}

void user_add(world *w, var *v) {
  mcpuser *u;
  
/*printf("   Handler: adding user to the list (rewritten)\n");*/
  if(v->type != TYPE_LIST ||
     v->listlen < 3 ||
     v->list->type != TYPE_OBJ ||
     v->list->nextlistitem->type != TYPE_STRING ||
     v->list->nextlistitem->nextlistitem->type != TYPE_INT) {
/*printf("\e[31m** Error in the user definition\e[0m\n");*/
  	return;
  }
  u = mcpuser_new(v);
  do_user_add(w, u);
  w->session->userlist->listlen++;
  clist_add_user(w, w->userlist, u, u->pos);
}  

void do_user_add(world *w, mcpuser *u) {
  mcpuser *u2;

  if(w->session->userlist->listlen == 0) {
    w->session->userlist->list = u;
    u->pos = 0;
  } else {
    for(u2 = w->session->userlist->list; u2 != NULL; u2 = u2->next) {
      if(u2->idle == 1 ||           /* user u must be added befor the idle user */
      compare_users(u, u2) < 0){/* user (u) must be added befor u2 */

/*printf("Adding user \"%s\" BEFORE user \"%s\"\n", u->name, u2->name);*/
        u->next = u2;
        u->prev = u2->prev;
        if(u->next != NULL)
          u->next->prev = u;
        if(u->prev != NULL) {
          u->prev->next = u;
          for(u2  = u->prev; u2 != NULL; u2 = u2->prev) /* set the .pos for user (u) */
            if(u2->cloaked == 0) {
              u->pos = u2->pos + 1;
              break;
            } else if(u2->prev == NULL) {
              u->pos = 0;
              break;
            }
        } else {
          w->session->userlist->list = u;
          u->pos = 0;
        }
        for(u2 = u->next; u2 != NULL; u2 = u2->next)
          if(u2->cloaked == 0)
            u2->pos++;
        break;
      } else if(u2->next == NULL) {         /* user must be added after the last on*/
/*printf("Adding user \"%s\" AFTER the last one (\"%s\")\n", u->name, u2->name);*/
        u->prev = u2;
        u2->next = u;
        u->next = NULL;
        for(; u2 != NULL; u2 = u2->prev) /* set the .pos for user (u) */
          if(u2->cloaked == 0) {
            u->pos = u2->pos + 1;
            break;
          } else if(u2->prev == NULL) {
            u->pos = 0;
            break;
          }
        break;
      }
    }
  }
}

void user_remove(world *w, var *v) {
  mcpuser *u, *u2;

/*printf("   Handler: removing user (rewritten)\n");*/
  u = mcpuser_find(w->session->userlist, v);
  if(u == NULL)
    return;
  mcpuser_remove(w->session->userlist, u);
  /* user removed, now update the userlist and the rest of the users'pos */
  gtk_clist_remove(GTK_CLIST(w->userlist), u->pos);

  if(u->cloaked != 0) {
    for(u2 = u->next; u2 != NULL; u2 = u2->next) {
      if(u2->cloaked == 0)
        u2->pos--;
    }
  }
  w->session->userlist->listlen--;
  mcpuser_free(u);
}

void user_idle(world *w, var *v) {
  mcpuser *u;
  int oldpos;
  
/*printf("   Handler: idling user (rewritten)\n");*/
  u = mcpuser_find(w->session->userlist, v);
  if(u == NULL)
    return;
/*printf("  \e[33mIdling user \"%s\"\e[0m\n", u->name);*/

  oldpos = u->pos;
  mcpuser_remove(w->session->userlist, u);
  u->idle = 1;

  do_user_add_idle(w, u);

  if(u->pos != oldpos)
    gtk_clist_row_move(GTK_CLIST(w->userlist), oldpos, u->pos);
  set_userlistpixmap(w, w->userlist, u);
}

void do_user_add_idle(world *w, mcpuser *u) {
  mcpuser *u2;

  if(w->session->userlist->listlen == 0) {
    w->session->userlist->list = u;
    u->pos = 0;
  } else {
    for(u2 = w->session->userlist->list; u2 != NULL; u2 = u2->next) {
      if(u2->idle == 1 &&           /* user u must be added befor the idle user */
      compare_users(u, u2) < 0){/* user (u) must be added befor u2 */

/*printf("Adding IDLE user \"%s\" BEFORE user \"%s\"\n", u->name, u2->name);*/
        u->next = u2;
        u->prev = u2->prev;
        if(u->next != NULL)
          u->next->prev = u;
        if(u->prev != NULL) {
          u->prev->next = u;
          for(u2  = u->prev; u2 != NULL; u2 = u2->prev) /* set the .pos for user (u) */
            if(u2->cloaked == 0) {
              u->pos = u2->pos + 1;
              break;
            } else if(u2->prev == NULL) {
              u->pos = 0;
              break;
            }
        } else {
          w->session->userlist->list = u;
          u->pos = 0;
        }
        for(u2 = u->next; u2 != NULL; u2 = u2->next)
          if(u2->cloaked == 0)
            u2->pos++;
        break;
      } else if(u2->next == NULL) {         /* user must be added after the last one */
/*printf("Adding IDLE user \"%s\" AFTER the last one (\"%s\")\n", u->name, u2->name);*/
        u->prev = u2;
        u2->next = u;
        u->next = NULL;
        for(; u2 != NULL; u2 = u2->prev) /* set the .pos for user (u) */
          if(u2->cloaked == 0) {
            u->pos = u2->pos + 1;
            break;
          } else if(u2->prev == NULL) {
            u->pos = 0;
            break;
          }
        break;
      }
    }
  }
}

void user_unidle(world *w, var *v) {
  mcpuser *u;
  int oldpos;
  
/*printf("   Handler: unidling user (rewritten)\n");*/
  u = mcpuser_find(w->session->userlist, v);
  if(u == NULL)
    return;
/*printf("  \e[33mUunidling user \"%s\"\e[0m\n", u->name);*/

  oldpos = u->pos;
  mcpuser_remove(w->session->userlist, u);
  u->idle = 0;

  do_user_add(w, u);
  if(u->pos != oldpos)
    gtk_clist_row_move(GTK_CLIST(w->userlist), oldpos, u->pos);
  set_userlistpixmap(w, w->userlist, u);
}

void user_cloak(world *w, var *v) {
  mcpuser *u;

/*printf("   Handler: cloaking user (rewritten)\n");*/
  u = mcpuser_find(w->session->userlist, v);
  if(u == NULL)
    return;
/*printf("  \e[33mCloaking user \"%s\"\e[0m\n", u->name);*/

  gtk_clist_remove(GTK_CLIST(w->userlist), u->pos);
  u->pos = -1;
  u->cloaked = 1;
  for(u = u->next; u != NULL; u = u->next)
    if(u->cloaked == 0)
      u->pos--;
}

void user_uncloak(world *w, var *v) {
  mcpuser *u, *u2;

/*printf("   Handler: decloaking user (rewritten)\n");*/
  u = mcpuser_find(w->session->userlist, v);
  if(u == NULL)
    return;
/*printf("  \e[33mDecloaking user \"%s\"\e[0m\n", u->name);*/
  for(u2 = u->prev; u2 != NULL; u2 = u2->prev)
    if(u2->cloaked == 0) {
      u->pos = u2->pos + 1;
      break;
    }
  if(u2 == NULL)
    u->pos = 0;
  for(u2 = u->next; u2 != NULL; u2 = u2->next)
    if(u2->cloaked == 0)
      u2->pos++;
  u->cloaked = 0;
  clist_add_user(w, w->userlist, u, u->pos);
}

void user_away(world *w, var *v) {
  mcpuser *u;

/*printf("   Handler: awaying user (rewritten)\n");*/
  u = mcpuser_find(w->session->userlist, v);
  if(u == NULL)
    return;
/*printf("  \e[33mAwaying user \"%s\"\e[0m\n", u->name);*/
  u->away = 1;
  set_userlistpixmap(w, w->userlist, u);
}

void user_unaway(world *w, var *v) {
  mcpuser *u;

/*printf("   Handler: awaying user (rewritten)\n");*/
  u = mcpuser_find(w->session->userlist, v);
  if(u == NULL)
    return;
/*printf("  \e[33mAwaying user \"%s\"\e[0m\n", u->name);*/
  u->away = 0;
  set_userlistpixmap(w, w->userlist, u);
}

void user_update(world *w, var *v) {
  mcpuser *u, *new;
  char s[20];
  
  if(v->type != TYPE_LIST ||
     v->listlen < 3 ||
     v->list->type != TYPE_OBJ ||
     v->list->nextlistitem->type != TYPE_STRING ||
     v->list->nextlistitem->nextlistitem->type != TYPE_INT) {
/*printf("\e[31m** Error in the user definition\e[0m\n");*/
  	return;
  }
  
  u = mcpuser_find(w->session->userlist, v->list);
  if(u == NULL)
    return;

  new = mcpuser_new(v);
  new->next = u->next;
  new->prev = u->prev;
  if(u->prev != NULL)
    u->prev->next = new;
  if(u->next != NULL)
    u->next->prev = new;
  new->cloaked = u->cloaked;
  new->idle = u->idle;
  new->away = u->away;
  new->pos = u->pos;
  mcpuser_free(u);
  
  set_userlistpixmap(w, w->userlist, new);
  gtk_clist_set_text(GTK_CLIST(w->userlist), new->pos, 1, new->name);
  sprintf(s, "%d", new->object);
  gtk_clist_set_text(GTK_CLIST(w->userlist), new->pos, 2, s);
}


mcpuser *mcpuser_new(var *v) {
  mcpuser *u;
  int i;
  var *v2;
  
  u = g_malloc(sizeof(mcpuser));
  u->object  =  v->list->obj;
  u->name    =  strcpy(g_malloc(v->list->nextlistitem->strlen + 1), 
                       v->list->nextlistitem->str);
  u->icon    =  v->list->nextlistitem->nextlistitem->i;
  u->cloaked =  0;
  u->away    =  0;
  u->idle    =  0;
  u->next    =  NULL;
  u->pos     = -1;
  u->prev    =  NULL;
  if(v->listlen == 3) {
    u->argc = 0;
    u->args = NULL;
  } else {
    u->argc = v->listlen - 3;
    v2 = v->list->nextlistitem->nextlistitem->nextlistitem;
    u->args = g_malloc(u->argc * sizeof(char *));
    for(i = 0; i < u->argc; i++) {
      u->args[i] = var_to_str(v2);
      v2 = v2->nextlistitem;
    }
  }
  return u;
}

void mcpuser_free(mcpuser *u) {
/*printf("  Free()ing user \"%s\"\n", u->name);*/
  if(u == NULL)
    return;
  if(u->name != NULL)
    g_free(u->name);
  g_free(u);
}

mcpuser *mcpuser_find(mcpuserlist *ul, var *v) {
  mcpuser *u = NULL;
  if(v->type != TYPE_OBJ) {
/*printf("\e[31m** Error, the to be removed user isn't an object (v->type == %d)\e[0m\n",v->type);*/
  	return NULL;
  }
  if(ul->listlen > 0)
    for(u = ul->list; u != NULL; u = u->next) {
      if(u->object == v->obj)
        break;
    }
/*else printf("\e[31m** Error, het is een lege list waar in wordt gezocht\e[0m\n");*/
/*if(u == NULL)printf("\e[31m** Error, no user found\e[0m\n");*/
  return u;
}

void mcpuser_remove(mcpuserlist *ul, mcpuser *u) {
  mcpuser *u2;
  
  if(u->next != NULL)
    u->next->prev = u->prev;
  if(u->prev != NULL)
    u->prev->next = u->next;
  else
    ul->list = u->next;
  for(u2 = u->next; u2 != NULL; u2 = u2->next)
    if(u2->cloaked == 0)
      u2->pos--;
}

void mcpicons_free(mcpicon *icon) {
  mcpicon *tmp = icon;
  
  while(tmp != NULL) {
    icon = tmp;
    tmp = tmp->next;
    g_free(icon->name);
    /* this _should_ free the pixmap, if I'm correct
     * however, it gives a lot of errors, so I'll comment them out ...
    gdk_pixmap_unref(icon->pixmap);
    gdk_bitmap_unref(icon->mask);
     */
    g_free(icon);
  }
}

mcpicon *mcpicon_new(char *name) {
  mcpicon *icon;
  char *filename;
  int len = strlen(name);

  name = strtolower(name, len);
  filename = g_malloc(14 + len + strlen(config_dir));
  sprintf(filename, "%s/Pixmaps/%s.xpm", config_dir, name);

  icon = g_malloc(sizeof(mcpicon));
  icon->name = strnew(name);
  icon->pixmap = NULL;
  icon->mask = NULL;
  icon->next = NULL;

/*printf("  Request for icon >>%s<<\n", name);*/
  icon->pixmap = gdk_pixmap_create_from_xpm(
                 main_window->window,
                 &icon->mask,
                 &main_window->style->bg[GTK_STATE_NORMAL],
                 filename);

  if(icon->pixmap == NULL) { /* Icon not found in userdit, let's check the
                              * global config dir, maby it's there */
    g_free(filename);
    filename = g_malloc(26 + len + strlen(PREFIX));
    sprintf(filename, PREFIX"/share/gMOO/Pixmaps/%s.xpm", name);
    icon->pixmap = gdk_pixmap_create_from_xpm(
                   main_window->window,
                   &icon->mask,
                   &main_window->style->bg[GTK_STATE_NORMAL],
                   filename);
  }
    
  g_free(filename);
  return icon;
}

GdkPixmap *get_userlistpix(world *w, mcpuser *u) {
  int i;
  mcpicon *icon;
  GdkPixmap *ret;
  
  if(u->idle == 1)
    if(u->away == 1)
      ret = userlistpix[IDLEAWAY];
    else
      ret = userlistpix[IDLE];
  else if(u->away)
    ret = userlistpix[AWAY];
  else if(u->icon == 0) /* No icon for this user */
    return NULL;
  else if(u->icon < 0 || u->icon > w->session->userlist->iconcount) 
    ret = userlistpix[UNKNOWN];
  else {
    icon = w->session->userlist->icons;  
    for(i = 0; i < u->icon -1; i++)
      icon = icon->next;
    ret = icon->pixmap;
  }
  return ret == NULL ? userlistpix[UNKNOWN] : ret;
}

GdkBitmap *get_userlistmask(world *w, mcpuser *u) {
  int i;
  mcpicon *icon;
  GdkBitmap *ret;
  
  if(u->idle == 1)
    if(u->away == 1)
      ret = userlistmask[IDLEAWAY];
    else
      ret = userlistmask[IDLE];
  else if(u->away)
    ret = userlistmask[AWAY];
  else if(u->icon == 0) /* No icon for this user */
    return NULL;
  else if(u->icon < 0 || u->icon > w->session->userlist->iconcount) 
    ret = userlistmask[UNKNOWN];
  else {
    icon = w->session->userlist->icons;  
    for(i = 0; i < u->icon -1; i++)
      icon = icon->next;
    ret = icon->mask;
  }
  return ret == NULL ? userlistmask[UNKNOWN] : ret;
}


void set_userlistpixmap(world *w, GtkWidget *clist, mcpuser *u) {
  GdkPixmap *pm = get_userlistpix(w, u);
  GdkBitmap *bm = get_userlistmask(w, u);

  if(pm == NULL || bm == NULL)
    gtk_clist_set_text(GTK_CLIST(clist), u->pos, 0, "");
  else
    gtk_clist_set_pixmap(GTK_CLIST(clist), u->pos, 0, pm, bm);
}

void userlistmenus_new(mcpuserlist *ul, var *v) {
  var *v2;
  int i;
  
  if(ul->menucount > 0) {
/*printf("Freeing the old userlistmenus\n");*/
  	userlistmenus_free(ul);
  }
/*printf("Creating userlistmenus:\n");*/
  ul->menucount = v->listlen;
  ul->menus = g_malloc(sizeof(char **) * v->listlen);
  v2 = v->list;
/*
gM_var_print(v);
printf("\n");
*/
  for(i = 0; i < v->listlen; i++) {
    if(v2->type == TYPE_INT) {
      ul->menus[i] = NULL;
/*printf("[%d] SEPERATOR\n", i);*/
    } else {
      ul->menus[i] = g_malloc(sizeof(char *) * 2);
      ul->menus[i][0] = strnew(v2->list->str);
      ul->menus[i][1] = strnew(v2->list->nextlistitem->str);
/*printf("[%d] \"%s\" -> \"%s\"\n", i, ul->menus[i][0], ul->menus[i][1]);*/
    }
    v2 = v2->nextlistitem;
  }
/*printf("Finished userlistmenus\n");*/
}

void userlistmenus_free(mcpuserlist *ul) {
  int i;
  
  if(ul->menucount < 1)
    return;
    
  for(i = 0; i < ul->menucount; i++) {
    if(ul->menus[i] != NULL) {
      g_free(ul->menus[i][0]);
      g_free(ul->menus[i][1]);
      g_free(ul->menus[i]);
    }
  }
  g_free(ul->menus);
}

void remove_(char *s) {
  for(; s[0] != 0; s++) {
    if(s[0] == '&')
      memmove(s, s + 1, strlen(s));
  }
}

void button_pressed(GtkCList *clist, GdkEventButton *e, world *w) {
  int row = -1;
  GtkWidget *tmp;
  char *s;
  int i;

  gtk_clist_get_selection_info(clist, e->x, e->y, &row, NULL);
  if(row < 0 ||
     e->button != 3 ||
     w->session == NULL ||
     w->session->userlist == NULL) {
    return;
  }

  if(usermenudata.menuwidget != NULL)
    gtk_widget_destroy(usermenudata.menuwidget);

  usermenudata.menuwidget = gtk_menu_new();
  usermenudata.row = row;
  usermenudata.world = w;
  usermenudata.ul = w->session->userlist;
  
  if(usermenudata.ul->menucount < 1) {
    tmp = gtk_menu_item_new_with_label(_("(empty)"));
    gtk_widget_show(tmp);
    gtk_widget_set_sensitive(tmp, FALSE);
    gtk_container_add(GTK_CONTAINER(usermenudata.menuwidget), tmp);
  } else
    for(i = 0; i < usermenudata.ul->menucount; i++) {
      if(usermenudata.ul->menus[i] == NULL) { /* seperator */
/*printf("adding SEPERATOR\n");*/
        tmp = gtk_menu_item_new();
        gtk_widget_show(tmp);
        gtk_container_add(GTK_CONTAINER(usermenudata.menuwidget), tmp);
      } else { /* menu item */
        s = parse_menu_string(usermenudata.ul, row, usermenudata.ul->menus[i][0]);
        if(s == NULL)
          s = strnew(_("-error-"));
        remove_(s);
/*printf("s == >>%s<<\n", s);*/
        tmp = gtk_menu_item_new_with_label(s);
        g_free(s);
        gtk_signal_connect(GTK_OBJECT(tmp), "activate", menu_pressed, (void *) i);
      	gtk_widget_show(tmp);
        gtk_container_add(GTK_CONTAINER(usermenudata.menuwidget), tmp);
      }
    }
/*printf("\n");*/

  gtk_menu_popup(GTK_MENU(usermenudata.menuwidget), NULL, NULL, NULL, NULL, 0, 0);

}

void menu_pressed(GtkMenuItem *mitem, int i) {
  char *s;

  gtk_widget_destroy(usermenudata.menuwidget);
  usermenudata.menuwidget = NULL;

  s = parse_menu_string(usermenudata.ul, usermenudata.row, usermenudata.ul->menus[i][1]);
  if(s == NULL) {
  	gM_world_print(usermenudata.world, _("\e[31m% Error in parsing the userlist menucommand !\e[0m\n"), -1);
  	return;
  }
/*printf("MUST SEND -> >>%s<<\n", s);*/
  gM_world_write(usermenudata.world, s);
  if(s[strlen(s) - 1] != 10)
    gM_world_write(usermenudata.world, "\n");
  g_free(s);
}


char *parse_menu_string(mcpuserlist *ul, int pos, char *s) {
  char *ret;
  mcpuser *u;
  int i,j,k,len;
  char *tmp;

  for(u = ul->list; u != NULL; u = u->next)
    if(u->pos == pos)
      break;
  if(u == NULL) {
/*printf("User at pos %d NOT found !!!\n", pos);*/
  	return NULL;
  }
  len = strlen(s);
  ret = g_malloc(5);
  j = 0;
  for(i = 0; i < len; i++) {
/*printf("s[i] == \"%s\"[%d] == '%c' (%d)\n", s,i, s[i], s[i]);*/
    if(s[i] == '$') {
      i++;
      if(s[i] == '$') {
        ret[j] = '$';
        j++;
      } else if(s[i] == 'n' || s[i] == 'N') {
        ret[j] = '\n';
        j++;
      } else if(s[i] == '(') {
        tmp = g_malloc(20); /* need is for a couple of buffer things */
        k = 0;
        ret[j] = 0;
        for(i++; s[i] != ')'; i++) {
          if(s[i] >= '0' && s[i] <= '9')
            tmp[k] = s[i];
          else { /* error */
            g_free(tmp);
            g_free(ret);
/*printf("error\n");*/
            return NULL;
          }
          k++;
        }
        tmp[k] = 0;
        k = atoi(tmp);
/*printf("Seeing for prop number %d\n", k);*/
/*printf("ret == >>%s<<\n", ret);*/
        if(k == 1) { /* object */
          ret = g_realloc(ret, j + 40);
          sprintf(ret, "%s#%d", ret, u->object);
          j = strlen(ret);
        } else if(k == 2) { /* name */
          ret = g_realloc(ret, j + 1 + strlen(u->name));
          strcat(ret, u->name);
          j = strlen(ret);
        } else if(k == 3) { /* icon */
          ret = g_realloc(ret, j + 40);
          sprintf(ret, "%s%d", ret, u->object);
          j = strlen(ret);
        } else if(k < 3 + u->argc) {/* one of the args */
          k -= 3;
          ret = g_realloc(ret, j + 2 + strlen(u->args[k]));
          strcat(ret, u->args[k]);
          j = strlen(ret);
        } else { /* error */
          g_free(tmp);
          g_free(ret);
          return NULL;
        }
        g_free(tmp);
      } else {
        g_free(ret);
        return NULL;
      }
    } else {
      ret[j] = s[i];
      ret = g_realloc(ret, (j++) + 5);
/*ret[j]=0;printf("ret is nu \"%s\"\n",ret);*/
    }
  }
  ret[j] = 0;
  return ret;
}

int compare_users(mcpuser *u1, mcpuser *u2) {
  int ret = 0;

  switch(userlistsorttype) {
  case USERLISTSORT_NAME:
    ret = strcasecmp(u1->name, u2->name) < 0 ? -1 :  1;
    break;
  case USERLISTSORT_NAME_REVERSE:
    ret = strcasecmp(u1->name, u2->name) < 0 ?  1 : -1;
    break;
  case USERLISTSORT_OBJ:
    ret = u1->object < u2->object ?  -1 :  1;
    break;
  case USERLISTSORT_OBJ_REVERSE:
    ret = u1->object < u2->object ?   1 : -1;
    break;
  }
/*printf(" Placing user %s %s user %s\n",u1->name, ret < 0 ? "before" : "after", u2->name);*/
  return ret;
}

void userlisttitleclick(GtkWidget *columnbutton, int column, world *thisw) {
  world *w;
  mcpuser *u;

  gtk_widget_grab_focus(thisw->input);
  switch(column) {
  case 1: /* name */
/*printf("Name column clicked");*/
  	if(userlistsorttype == USERLISTSORT_NAME) {
      userlistsorttype = USERLISTSORT_NAME_REVERSE;
/*printf(", set to reverse name\n");*/
    } else {
      userlistsorttype = USERLISTSORT_NAME;
/*printf(", set to name\n");*/
    }
    break;
  case 2: /* obj  */
/*printf("Obj column clicked");*/
  	if(userlistsorttype == USERLISTSORT_OBJ) {
      userlistsorttype = USERLISTSORT_OBJ_REVERSE;
/*printf(", set to reverse object\n");*/
    } else {
      userlistsorttype = USERLISTSORT_OBJ;
/*printf(", set to object\n");*/
    }
    break;
  }
/*printf("reordering lists\n");*/
  for(w = firstworld; w != NULL; w = w->next) {
    if(w->session != NULL &&
       w->session->userlist != NULL &&
       w->session->userlist->listlen > 0) {
/*printf("  Reordering for world \"%s\"\n", w->name);*/
      u = w->session->userlist->list;
      w->session->userlist->listlen = 0;
      w->session->userlist->list    = NULL;
      gtk_clist_clear(GTK_CLIST(w->userlist));
      while(1) {
        if(u->next == NULL) {
          u->prev = NULL;
          if(u->idle == 1)
            do_user_add_idle(w, u);
          else
            do_user_add(w, u);
          w->session->userlist->listlen++;
          clist_add_user(w, w->userlist, u, u->pos);
          break;
        } else {
          u = u->next;
          u->prev->next = u->prev->prev = NULL;
          if(u->prev->idle == 1)
            do_user_add_idle(w, u->prev);
          else
            do_user_add(w, u->prev);
          w->session->userlist->listlen++;
          clist_add_user(w, w->userlist, u->prev, u->prev->pos);
        }
      }
    }
  }
}



