/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "global_vars.h"
#include "structs.h"
#include "worlds.h"
#include "window.h"
#include "dialogs.h"
#include "help.h"
#include "config.h"
#include "editor.h"
#include "worlddialogs.h"
#include "userlist.h"
#include "world.h"
#include "do_scripts.h"
#include "preferences.h"
#include "triggers.h"

int gM_window_destroy();
int gM_window_delete(GtkWidget *window, GdkEvent  *event, void *data);

void gM_window_handle_menu(GtkWidget *w, int i);

void switch_page(GtkWidget *nb, GtkWidget *nbp, int page_num, void *pointer);

void handle_discon();

void update_discon_pixmap(GtkWidget *dc, void *pointer);

void toggle_userlist(GtkWidget *menuitem, void *pointer);

#define mNew 		 0
#define mOpen		 1
#define mPref		 2

#define mConDia     10
#define mDiscon	    11
#define mClose      12
#define mUpdateLine 13
#define mSend       14
#define mFind       15
#define mProp		    16
#define mCScreen    17
#define mCBuffer    18
#define mSendSel    19

#define mHelp		    20
#define mAbout      21
#define mAboutWorld 22
#define mConfigInfo 23
#define mKeyboard   24

/* gMOO Logo icon -> Will someone create a _cool_ one ? */
#include "../share/gMOO.xpm"

#include "pix/new.xpm"
#include "pix/open.xpm"
#include "pix/exit.xpm"
#include "pix/connect_on.xpm"
#include "pix/connect_off.xpm"
#include "pix/send.xpm"
#include "pix/close.xpm"
#include "pix/editsend.xpm"
#include "pix/editsendclose.xpm"
#include "pix/editsave.xpm"
#include "pix/editsyntax.xpm"
#include "pix/editfind.xpm"
#include "pix/editreplace.xpm"
#include "pix/editagain.xpm"
#include "pix/editgoto.xpm"
#include "pix/properties.xpm"
#include "pix/preferences.xpm"
#include "pix/triggers.xpm"

#include "pix/userlist.xpm" /* toolbarbutton */
#include "pix/userunknown.xpm"
#include "pix/useridle.xpm"
#include "pix/useraway.xpm"
#include "pix/useridleaway.xpm"

#ifdef TCL
#include "pix/icon_error.xpm"
#include "pix/icon_question.xpm"
#include "pix/icon_warning.xpm"
#include "pix/icon_information.xpm"
#endif /* TCL */

GtkWidget *connect_on;
GtkWidget *connect_off;

void gM_window_create() {
  GtkWidget *MenuBar;
  GtkWidget *mainvb;
	GtkWidget *tmpmenu;
	GtkWidget *tmp;
  GtkAccelGroup *ag;
  GtkWidget *hbox, *frame;

  GtkWidget *iconw;
  GdkPixmap *icon;
  GdkBitmap *mask;  
  
  int x = -1;
  int y = -1;
  int h = 480;
  int w = 640;

  char *s;

  main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect(GTK_OBJECT(main_window), "delete_event", GTK_SIGNAL_FUNC(gM_window_delete), NULL);
 	gtk_window_set_policy(GTK_WINDOW(main_window), TRUE, TRUE, FALSE);
  gtk_window_set_title(GTK_WINDOW(main_window), "gMOO-" VERSION);
  gtk_window_set_wmclass (GTK_WINDOW(main_window), "gMOO_main_window", "gMOO");

  ag = gtk_accel_group_new();
  gtk_accel_group_attach(ag, GTK_OBJECT(main_window));

	mainvb = gtk_vbox_new(FALSE, 0);
	gtk_widget_show(mainvb);
	gtk_container_add(GTK_CONTAINER(main_window), mainvb);

  tmp = gtk_handle_box_new();
  gtk_widget_show(tmp);
  gtk_handle_box_set_handle_position(GTK_HANDLE_BOX(tmp), GTK_POS_LEFT);
  gtk_handle_box_set_snap_edge(GTK_HANDLE_BOX(tmp), GTK_POS_TOP);
  gtk_box_pack_start(GTK_BOX(mainvb), tmp, FALSE, TRUE, 0);

  MenuBar = gtk_menu_bar_new();
  gtk_widget_show(MenuBar);
  gtk_container_add(GTK_CONTAINER(tmp), MenuBar);

/*** Begin Submenu Main ***/
  tmp = gtk_menu_item_new_with_label(_("Main"));
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(MenuBar), tmp);

  tmpmenu = gtk_menu_new();
  gtk_widget_show(tmpmenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(tmp), tmpmenu);
 
  tmp = gtk_tearoff_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("New..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mNew);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Open..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mOpen);
  gtk_widget_add_accelerator(tmp,
                             "activate",
                             ag,
                             GDK_o,
                             GDK_CONTROL_MASK,
                             GTK_ACCEL_VISIBLE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Preferences..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mPref);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Highlites & Triggers..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_triggers_show_dialog), NULL);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Quit"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_destroy), NULL);
  gtk_widget_add_accelerator(tmp,
                             "activate",
                             ag,
                             GDK_q,
                             GDK_CONTROL_MASK,
                             GTK_ACCEL_VISIBLE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

/*** Begin Submenu %Worldname% ***/
  worldmenu = gtk_menu_item_new_with_label("");
  gtk_container_add(GTK_CONTAINER(MenuBar), worldmenu);

  tmpmenu = gtk_menu_new();
  gtk_widget_show(tmpmenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(worldmenu), tmpmenu);

  tmp = gtk_tearoff_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Connect dialog..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mConDia);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  disconmenu = gtk_menu_item_new_with_label(_("(Dis)connect"));
  gtk_widget_show(disconmenu);
  gtk_signal_connect(GTK_OBJECT(disconmenu), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mDiscon);
  gtk_widget_add_accelerator(disconmenu,
                             "activate",
                             ag,
                             GDK_F3,
                             GDK_CONTROL_MASK,
                             GTK_ACCEL_VISIBLE);
  gtk_container_add(GTK_CONTAINER(tmpmenu), disconmenu);

  tmp = gtk_menu_item_new_with_label(_("Close"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mClose);
  gtk_widget_add_accelerator(tmp,
                             "activate",
                             ag,
                             GDK_F4,
                             GDK_CONTROL_MASK,
                             GTK_ACCEL_VISIBLE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);
  
  tmp = gtk_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);


  tmp = gtk_menu_item_new_with_label(_("Send..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mSend);
  gtk_widget_add_accelerator(tmp,
                             "activate",
                             ag,
                             GDK_s,
                             GDK_CONTROL_MASK,
                             GTK_ACCEL_VISIBLE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);
#if 0 /* This doesn't work yet, I don't know how to get a selection :( */
  tmp = gtk_menu_item_new_with_label("Send Selection");
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mSendSel);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);
#endif
  findmenu = gtk_menu_item_new_with_label(_("Find..."));
  gtk_signal_connect(GTK_OBJECT(findmenu), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mFind);
  gtk_widget_add_accelerator(findmenu,
                             "activate",
                             ag,
                             GDK_f,
                             GDK_CONTROL_MASK,
                             GTK_ACCEL_VISIBLE);
  gtk_widget_show(findmenu);
  gtk_container_add(GTK_CONTAINER(tmpmenu), findmenu);

  tmp = gtk_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Clear Screen"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mCScreen);
  gtk_widget_add_accelerator(tmp,
                             "activate",
                             ag,
                             GDK_l,
                             GDK_CONTROL_MASK,
                             GTK_ACCEL_VISIBLE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Clear Buffer"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mCBuffer);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Update linelength"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mUpdateLine);
  gtk_widget_add_accelerator(tmp,
                             "activate",
                             ag,
                             GDK_F9,
                             GDK_CONTROL_MASK,
                             GTK_ACCEL_VISIBLE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  userlistmenu = gtk_check_menu_item_new_with_label(_("Userlist"));
  gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(userlistmenu), TRUE);
  userlistsignal = gtk_signal_connect(GTK_OBJECT(userlistmenu), "activate", GTK_SIGNAL_FUNC(toggle_userlist), NULL);
  gtk_widget_add_accelerator(userlistmenu,
                             "activate",
                             ag,
                             GDK_F5,
                             GDK_CONTROL_MASK,
                             GTK_ACCEL_VISIBLE);
  gtk_widget_show(userlistmenu);
  gtk_container_add(GTK_CONTAINER(tmpmenu), userlistmenu);

  tmp = gtk_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Properties..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (void *) mProp);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

/*** Begin Submenu Editor ***/
  editmenu = gtk_menu_item_new_with_label(_("Editor"));
  gtk_container_add(GTK_CONTAINER(MenuBar), editmenu);

  tmpmenu = gtk_menu_new();
  gtk_widget_show(tmpmenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(editmenu), tmpmenu);

  tmp = gtk_tearoff_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Send"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_editor_handle_menu), (void *) SEND);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Send & Close"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_editor_handle_menu), (void *) SENDCLOSE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Close"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_editor_handle_menu), (void *) CLOSE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Open..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_editor_handle_menu), (void *) OPEN);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Save..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_editor_handle_menu), (void *) SAVE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  syntaxmenu = gtk_check_menu_item_new_with_label(_("Syntax HL"));
  syntaxhlsignal = gtk_signal_connect(GTK_OBJECT(syntaxmenu), "activate", GTK_SIGNAL_FUNC(gM_editor_handle_menu), 
                   (void *) SYNTAX);
  gtk_check_menu_item_set_show_toggle(GTK_CHECK_MENU_ITEM(syntaxmenu), TRUE);
  gtk_widget_show(syntaxmenu);
  gtk_container_add(GTK_CONTAINER(tmpmenu), syntaxmenu);

/*** Begin Submenu Search ***/
  searchmenu = gtk_menu_item_new_with_label(_("Search"));
  gtk_container_add(GTK_CONTAINER(MenuBar), searchmenu);

  tmpmenu = gtk_menu_new();
  gtk_widget_show(tmpmenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(searchmenu), tmpmenu);

  tmp = gtk_tearoff_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Find..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_editor_handle_menu), (void *) FIND);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Replace..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_editor_handle_menu), (void *) REPLACE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Again"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_editor_handle_menu), (void *) AGAIN);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Go to Line..."));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_editor_handle_menu), (void *) GOTOLINE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);
  
/*** Begin Submenu Help ***/
  tmp = gtk_menu_item_new_with_label(_("Help"));
  gtk_menu_item_right_justify(GTK_MENU_ITEM(tmp));
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(MenuBar), tmp);

  tmpmenu = gtk_menu_new();
  gtk_widget_show(tmpmenu);
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(tmp), tmpmenu);
 
  tmp = gtk_tearoff_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("About gMOO"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (char *) mAbout);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("World info"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (char *) mAboutWorld);
  gtk_widget_add_accelerator(tmp,
                             "activate",
                             ag,
                             GDK_F1,
                             GDK_CONTROL_MASK,
                             GTK_ACCEL_VISIBLE);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Configure info"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (char *) mConfigInfo);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Keyboard"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (char *) mKeyboard);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new();
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

  tmp = gtk_menu_item_new_with_label(_("Help"));
  gtk_signal_connect(GTK_OBJECT(tmp), "activate", GTK_SIGNAL_FUNC(gM_window_handle_menu), (char *) mHelp);
  gtk_widget_show(tmp);
  gtk_container_add(GTK_CONTAINER(tmpmenu), tmp);

/* showing befor the toolbar, to give the pixmaps a GdkWindow that is valid */

	gtk_widget_realize(main_window);

/*** TOOLBAR ***/
  toolbarhandle = gtk_handle_box_new();
  gtk_handle_box_set_handle_position(GTK_HANDLE_BOX(toolbarhandle), GTK_POS_LEFT);
  gtk_handle_box_set_snap_edge(GTK_HANDLE_BOX(toolbarhandle), GTK_POS_TOP);
  gtk_box_pack_start(GTK_BOX(mainvb), toolbarhandle, FALSE, TRUE, 0);

  toolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
  gtk_container_add(GTK_CONTAINER(toolbarhandle), toolbar);
  gtk_toolbar_set_button_relief(GTK_TOOLBAR(toolbar), GTK_RELIEF_NONE);
  gtk_container_set_border_width(GTK_CONTAINER(toolbar), 2);
  gtk_toolbar_set_space_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_SPACE_LINE);
  gtk_widget_show(toolbar);


  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      new_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
                          _("New"), _("Create a new world"), "Private", iconw, gM_window_handle_menu, (void *) mNew);
  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      open_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
                          _("Open"), _("Open a world"), "Private", iconw, gM_window_handle_menu, (void *) mOpen);
  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      exit_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
                          _("Exit"), _("Exit gMOO"), "Private", iconw, GTK_SIGNAL_FUNC(gM_window_destroy), NULL);

  gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      preferences_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
                          _("Preferences"), _("Open global preferences dialog"),
                          "Private", iconw, gM_window_handle_menu,
                          (void *) mPref);

  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      triggers_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
                          _("Highlites &\n Triggers"),
                          _("Open the highlites and triggers dialog"), 
                          "Private", iconw, gM_triggers_show_dialog, NULL);

  gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      connect_on_xpm);
  connect_on = gtk_pixmap_new(icon, mask);
  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      connect_off_xpm);
  connect_off = gtk_pixmap_new(icon, mask);
  gtk_widget_show(connect_off);
  tmp = gtk_hbox_new(TRUE, 0);
  gtk_widget_show(tmp);
  gtk_box_pack_start(GTK_BOX(tmp), connect_on, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(tmp), connect_off, FALSE, FALSE, 0);
  tb_discon = gtk_toolbar_append_element(GTK_TOOLBAR(toolbar),
                                            GTK_TOOLBAR_CHILD_TOGGLEBUTTON,
                                            NULL,
                                            _("Dis/Connect"),
                                            _("Dis/Connect"),
                                            "Private",
                                            tmp,
                                            GTK_SIGNAL_FUNC(handle_discon),
                                            NULL);
  gtk_signal_connect_after(GTK_OBJECT(tb_discon), "toggled", update_discon_pixmap, NULL);

  icon = gdk_pixmap_create_from_xpm_d(main_window->window, 
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      close_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_close = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
                          _("Close"), _("Close current world"), "Private", iconw, gM_window_handle_menu,
                          (void *) mClose);
  
  gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

  icon = gdk_pixmap_create_from_xpm_d(main_window->window, 
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      properties_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_prop = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
                          _("Properties"), _("Open world properties dialog"), "Private", iconw, gM_window_handle_menu,
                          (void *) mProp);

  gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      send_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_send = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
            _("Send"), _("Show `Send text to world' dialog"), "Private",
            iconw, gM_window_handle_menu, (void *) mSend);

  gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));

  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      userlist_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_userlist = gtk_toolbar_append_element(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_CHILD_TOGGLEBUTTON, NULL,
            _("Userlist"), _("Toggle the Userlist"), "Private", iconw, GTK_SIGNAL_FUNC(gM_userlist_toggle), NULL);

/* EDITORs TOOLBAR : */  
  edittoolbarhandle = gtk_handle_box_new();
  gtk_handle_box_set_handle_position(GTK_HANDLE_BOX(edittoolbarhandle), GTK_POS_LEFT);
  gtk_handle_box_set_snap_edge(GTK_HANDLE_BOX(edittoolbarhandle), GTK_POS_TOP);
  gtk_box_pack_start(GTK_BOX(mainvb), edittoolbarhandle, FALSE, TRUE, 0);

  edittoolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
  gtk_container_add(GTK_CONTAINER(edittoolbarhandle), edittoolbar);
  gtk_toolbar_set_button_relief(GTK_TOOLBAR(edittoolbar), GTK_RELIEF_NONE);
  gtk_container_set_border_width(GTK_CONTAINER(edittoolbar), 2);
  gtk_toolbar_set_space_style(GTK_TOOLBAR(edittoolbar), GTK_TOOLBAR_SPACE_LINE);
  gtk_widget_show(edittoolbar);

  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      editsend_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_editsend = gtk_toolbar_append_item(GTK_TOOLBAR(edittoolbar), 
                _("Send"), _("Send the text to the world"), "Private", iconw, GTK_SIGNAL_FUNC(gM_editor_handle_menu),
                (void *) SEND);
  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      editsendclose_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_editsendclose = gtk_toolbar_append_item(GTK_TOOLBAR(edittoolbar), 
                _("Send & Close"), _("Send the text to the world\nAnd close the editor"),
                "Private", iconw,GTK_SIGNAL_FUNC(gM_editor_handle_menu), (void *) SENDCLOSE);
  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      close_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_editclose = gtk_toolbar_append_item(GTK_TOOLBAR(edittoolbar),
                 _("Close"), _("Close the editor"), "Private", iconw, GTK_SIGNAL_FUNC(gM_editor_handle_menu),
                 (void *) CLOSE);

  gtk_toolbar_append_space(GTK_TOOLBAR(edittoolbar));

  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      open_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_editopen = gtk_toolbar_append_item(GTK_TOOLBAR(edittoolbar),
                _("Open"), _("Open a local file, and insert it"), 
                "Private", iconw, GTK_SIGNAL_FUNC(gM_editor_handle_menu),
                (void *) OPEN);
  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      editsave_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_editsave = gtk_toolbar_append_item(GTK_TOOLBAR(edittoolbar), 
                _("Save"), _("Save the text to a local file"),
                "Private", iconw, GTK_SIGNAL_FUNC(gM_editor_handle_menu),
                (void *) SAVE);

  gtk_toolbar_append_space(GTK_TOOLBAR(edittoolbar));

  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      editsyntax_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_editsyntax = gtk_toolbar_append_element(GTK_TOOLBAR(edittoolbar), GTK_TOOLBAR_CHILD_TOGGLEBUTTON, NULL,
                  _("Syntax HL"), _("Toggle the syntax highlighter"),
                  "Private", iconw, GTK_SIGNAL_FUNC(gM_editor_do_syntax),
                  NULL);
  
  gtk_toolbar_append_space(GTK_TOOLBAR(edittoolbar));

  icon = gdk_pixmap_create_from_xpm_d(main_window->window, &mask, &main_window->style->bg[GTK_STATE_NORMAL],
    		 editfind_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_editfind = gtk_toolbar_append_item(GTK_TOOLBAR(edittoolbar),
                _("Find"), _("Find text in editor"), "Private", iconw, GTK_SIGNAL_FUNC(gM_editor_handle_menu),
                (void *) FIND);
  icon = gdk_pixmap_create_from_xpm_d(main_window->window, &mask, &main_window->style->bg[GTK_STATE_NORMAL],
         editreplace_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_editreplace = gtk_toolbar_append_item(GTK_TOOLBAR(edittoolbar),
                   _("Replace"), _("Replace text in editor"),
                   "Private", iconw, GTK_SIGNAL_FUNC(gM_editor_handle_menu),
                   (void *) REPLACE);
  icon = gdk_pixmap_create_from_xpm_d(main_window->window, &mask, &main_window->style->bg[GTK_STATE_NORMAL],
    		 editagain_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_editagain = gtk_toolbar_append_item(GTK_TOOLBAR(edittoolbar),
                _("Again"), _("Again\nRedo last find/replace"),
                "Private", iconw, GTK_SIGNAL_FUNC(gM_editor_handle_menu),
                (void *) AGAIN);
  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask, &main_window->style->bg[GTK_STATE_NORMAL],
                                      editgoto_xpm);
  iconw = gtk_pixmap_new(icon, mask);
  tb_editgoto = gtk_toolbar_append_item(GTK_TOOLBAR(edittoolbar),
                _("Goto"), _("Goto line number"), "Private", iconw, GTK_SIGNAL_FUNC(gM_editor_handle_menu),
                (void *) GOTOLINE);


  tb_linenum = gtk_label_new(_("Line : 1"));
  gtk_toolbar_append_element(GTK_TOOLBAR(edittoolbar),
                                            GTK_TOOLBAR_CHILD_WIDGET,
                                            tb_linenum,
                                            "Dis/Connect",
                                            "Dis/Connect",
                                            "Private",
                                            NULL,
                                            NULL,
                                            NULL);


  gtk_widget_show(tb_prop);
  gtk_widget_show(tb_discon);
  gtk_widget_show(tb_close);
  gtk_widget_show(tb_send);
  gtk_widget_show(tb_userlist);

  gtk_widget_set_sensitive(tb_prop, FALSE);
  gtk_widget_set_sensitive(tb_discon, FALSE);
  gtk_widget_set_sensitive(tb_close, FALSE);
  gtk_widget_set_sensitive(tb_send, FALSE);
  gtk_widget_set_sensitive(tb_userlist, FALSE);

  gtk_widget_show(tb_editsend);
  gtk_widget_show(tb_editsendclose);
  gtk_widget_show(tb_editclose);
  gtk_widget_show(tb_editsave);
  gtk_widget_show(tb_editopen);
  gtk_widget_show(tb_editsyntax);
  gtk_widget_show(tb_editfind);
  gtk_widget_show(tb_editreplace);
  gtk_widget_show(tb_editagain);
  gtk_widget_show(tb_editgoto);
  gtk_widget_show(tb_linenum);

/* Hide it by default:
  gtk_widget_show(edittoolbarhandle); */

  if(show_toolbar > 0)
    gtk_widget_show(toolbarhandle);

/*** END TOOLBAR ***/

  notebook = gtk_notebook_new();
  gtk_notebook_set_scrollable(GTK_NOTEBOOK(notebook), TRUE);
  gtk_widget_show(notebook);
  gtk_box_pack_start(GTK_BOX(mainvb), notebook, TRUE, TRUE, 0);
  gtk_notebook_set_show_border(GTK_NOTEBOOK(notebook), FALSE);
  gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), TRUE);
  gtk_notebook_set_homogeneous_tabs(GTK_NOTEBOOK(notebook), homo_tabs);
  gtk_notebook_set_tab_vborder(GTK_NOTEBOOK(notebook), 0);
  gtk_signal_connect_after(GTK_OBJECT(notebook), "switch_page", GTK_SIGNAL_FUNC(switch_page), NULL);

  /* Statusbar && clock */
  
  frame = gtk_frame_new(NULL);
  gtk_widget_show(frame);
  gtk_box_pack_start(GTK_BOX(mainvb), frame, FALSE, FALSE, 0);
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_OUT);

  hbox = gtk_hbox_new (FALSE, 5);
  gtk_widget_show(hbox);
  gtk_container_add (GTK_CONTAINER (frame), hbox);

  frame = gtk_frame_new(NULL);
  gtk_widget_show(frame);
  gtk_box_pack_start(GTK_BOX(hbox), frame, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);

  statusbar = gtk_label_new ("gMOO Started, have fun ! (The statusbar doesn't do anything else yet)");
  gtk_widget_show (statusbar);
  gtk_container_add (GTK_CONTAINER (frame), statusbar);
  gtk_misc_set_alignment (GTK_MISC (statusbar), 0, 0.5);
  gtk_misc_set_padding (GTK_MISC (statusbar), 5, 0);

  /* The nice little clock: */
  clockframe = gtk_frame_new (NULL);
  gtk_box_pack_start (GTK_BOX (hbox), clockframe, FALSE, TRUE, 0);
  gtk_frame_set_shadow_type (GTK_FRAME (clockframe), GTK_SHADOW_IN);

  clockbar = gtk_label_new ("");
  gtk_widget_show (clockbar);
  gtk_container_add (GTK_CONTAINER (clockframe), clockbar);
  gtk_misc_set_padding (GTK_MISC (clockbar), 5, 0);
  
  /* load fonts */

  normalfont = gdk_font_load(normalfontstr);
  boldfont   = gdk_font_load(boldfontstr);

/* create the 2 styles for on the tabs */
  tabnorm_style = gtk_style_new();
  tabhigh_style = gtk_style_new();
  memcpy(tabhigh_style->fg, &colors[RED], sizeof(GdkColor));


  /* Create the userlist pixmaps */
  userlistpix[UNKNOWN] = gdk_pixmap_create_from_xpm_d(
                main_window->window, &userlistmask[UNKNOWN],
                &main_window->style->bg[GTK_STATE_NORMAL],
                userunknown_xpm);
  userlistpix[IDLE] = gdk_pixmap_create_from_xpm_d(
                main_window->window, &userlistmask[IDLE],
                &main_window->style->bg[GTK_STATE_NORMAL],
                useridle_xpm);
  userlistpix[AWAY] = gdk_pixmap_create_from_xpm_d(
                main_window->window, &userlistmask[AWAY],
                &main_window->style->bg[GTK_STATE_NORMAL],
                useraway_xpm);
  userlistpix[IDLEAWAY] = gdk_pixmap_create_from_xpm_d(
                main_window->window, &userlistmask[IDLEAWAY],
                &main_window->style->bg[GTK_STATE_NORMAL],
                useridleaway_xpm);
#ifdef TCL
  /* Create the Tcl msgbox pixmaps */
  msgboxpix[ICON_ERROR] = gdk_pixmap_create_from_xpm_d(
                main_window->window, &msgboxmask[ICON_ERROR],
                &main_window->style->bg[GTK_STATE_NORMAL],
                icon_error_xpm);
  msgboxpix[ICON_QUESTION] = gdk_pixmap_create_from_xpm_d(
                main_window->window, &msgboxmask[ICON_QUESTION],
                &main_window->style->bg[GTK_STATE_NORMAL],
                icon_question_xpm);
  msgboxpix[ICON_WARNING] = gdk_pixmap_create_from_xpm_d(
                main_window->window, &msgboxmask[ICON_WARNING],
                &main_window->style->bg[GTK_STATE_NORMAL],
                icon_warning_xpm);
  msgboxpix[ICON_INFORMATION] = gdk_pixmap_create_from_xpm_d(
                main_window->window, &msgboxmask[ICON_INFORMATION],
                &main_window->style->bg[GTK_STATE_NORMAL],
                icon_information_xpm);
#endif /* TCL */

  if(save_window_size != NULL) {
    s = strchr(save_window_size, ',');
    h = atoi(s + 1);
    s[0] = 0;
    w = atoi(save_window_size);
    s[0] = ',';
  }

  if(save_window_pos != NULL) {
    s = strchr(save_window_pos, ',');
    y = atoi(s + 1);
    s[0] = 0;
    x = atoi(save_window_pos);
    s[0] = ',';
  }

  gtk_widget_show(main_window);

  if(x < 0 || y < 0)
    gdk_window_resize(main_window->window, w, h);
  else {
    gdk_window_move_resize(main_window->window, x, y, w, h);
  }

  /* Add the gMOO logo */
  icon = gdk_pixmap_create_from_xpm_d(main_window->window,
                                      &mask,
                                      &main_window->style->bg[GTK_STATE_NORMAL],
                                      gMOO_xpm);

  gdk_window_set_icon(main_window->window,
                      NULL,
                      icon,
                      mask);
  gdk_window_set_icon_name(main_window->window, "gMOO");

  /* done at last */
}

int gM_window_destroy() {
  int x,y;
  if(gM_worlds_close_all() > 0) {
    if(save_window_pos != NULL) {
      gdk_window_get_root_origin(main_window->window, &x, &y);
      save_window_pos = g_malloc(40);
      sprintf(save_window_pos, "%d,%d", x, y);
    }
    if(save_window_size != NULL) {
      save_window_size = g_malloc(40);
      sprintf(save_window_size, "%d,%d", main_window->allocation.width, main_window->allocation.height);
    }
    gtk_main_quit();
  }
  return FALSE;
}

int gM_window_delete(GtkWidget *window, GdkEvent  *event, void *data) {
  if(gM_worlds_close_all() > 0) {
    gtk_main_quit();
    return FALSE; /* just pro forma */
  } else
    return TRUE; /* keep the window */
}


void gM_window_handle_menu(GtkWidget *w, int i) {
  switch(i) {
  case mNew:
    gM_dialogs_new_world();
    break;
  case mOpen:
    gM_dialogs_open_world();
    break;
  case mPref:
    gM_preferences();
    break;

  case mConDia:
    gM_worlddialogs_connect(gM_window_get_current_world());
    break;
  case mDiscon:
    handle_discon();
    break;
  case mClose:
    gM_worlds_remove_world(gM_window_get_current_world());
    break;
  case mSend:
    gM_worlddialogs_send(gM_window_get_current_world());
    break;
  case mSendSel:
    gM_world_send_selection(gM_window_get_current_world());
    break;
  case mFind:
    gM_worlddialogs_find(gM_window_get_current_world());
    break;
  case mCScreen:
    gM_world_clear_screen(gM_window_get_current_world());
    break;
  case mCBuffer:
    gM_world_clear_buffer(gM_window_get_current_world());
    break;
  case mUpdateLine:
    gM_world_force_update_linelen(gM_window_get_current_world());
    break;
  case mProp:
    gM_dialogs_world_properties(gM_window_get_current_world());
    break;

  case mAbout:
    gM_help_show_about();
    break;
  case mAboutWorld:
    gM_help_show_about_world(gM_window_get_current_world());
    break;
  case mConfigInfo:
    gM_help_show_config_info();
    break;
  case mKeyboard:
    gM_help_show_keyboard();
    break;
  case mHelp:
    gM_help_show_index();
  }
}


void gM_window_add_world(world *w) {
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), w->main, w->label);
  w->page_num = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), w->main);
  /* swtich focus */
  gtk_notebook_set_page(GTK_NOTEBOOK(notebook), w->page_num);
}

void gM_window_add_editor(editor *e) {
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), e->main, e->label);
  e->page_num = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), e->main);
  /* swtich focus */
  gtk_notebook_set_page(GTK_NOTEBOOK(notebook), e->page_num);
}

world *gM_window_get_current_world() {
  return gM_worlds_get_by_main(gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook),
    			gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook))));
}

world *gM_window_get_current_editor_world() {
  int dummy;
  return gM_worlds_get_world_by_editor_main(
           gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook),
    			   gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook))),
           &dummy);
}

void switch_page(GtkWidget *nb, GtkWidget *nbp, int page_num, void *pointer) {
  GtkWidget *wid;
  world *w;
  int i;

/*printf("switch\n");*/
  wid = gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page_num);
  w = gM_worlds_get_by_main(wid);

  if(w == NULL) {
    w = gM_worlds_get_world_by_editor_main(wid, &i);
    if(w != NULL) {
/*printf("grabbing focus to editor %d\n", i);*/
      gtk_widget_grab_focus(w->edit[i]->text);

      gtk_signal_handler_block(GTK_OBJECT(syntaxmenu), syntaxhlsignal);
      gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(syntaxmenu), w->edit[i]->syntax);
      gtk_signal_handler_unblock(GTK_OBJECT(syntaxmenu), syntaxhlsignal);

      gtk_signal_handler_block_by_func(GTK_OBJECT(tb_editsyntax), gM_editor_do_syntax, NULL);
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tb_editsyntax), w->edit[i]->syntax);
      gtk_signal_handler_unblock_by_func(GTK_OBJECT(tb_editsyntax), gM_editor_do_syntax, NULL);

      gtk_widget_set_sensitive(tb_prop, FALSE);
      gtk_widget_set_sensitive(tb_discon, FALSE);
    	gtk_widget_set_sensitive(tb_close, FALSE);
    	gtk_widget_set_sensitive(tb_send, FALSE);
      gtk_widget_set_sensitive(tb_userlist, FALSE);
      
      if(show_toolbar)
        gtk_widget_show(edittoolbarhandle);

      gM_window_update_discon(NULL);
    }
  } else {
/*printf("grabbing focus to a world\n");*/
    gtk_widget_grab_focus(w->input);

    gtk_signal_handler_block(GTK_OBJECT(userlistmenu), userlistsignal);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(userlistmenu), w->userlist_on);
    gtk_signal_handler_unblock(GTK_OBJECT(userlistmenu), userlistsignal);

    gtk_signal_handler_block_by_func(GTK_OBJECT(tb_userlist), GTK_SIGNAL_FUNC(gM_userlist_toggle), NULL);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tb_userlist), w->userlist_on);
    gtk_signal_handler_unblock_by_func(GTK_OBJECT(tb_userlist), GTK_SIGNAL_FUNC(gM_userlist_toggle), NULL);

    gtk_widget_set_sensitive(tb_prop, TRUE);
    gtk_widget_set_sensitive(tb_discon, TRUE);
    gtk_widget_set_sensitive(tb_close, TRUE);
    gtk_widget_set_sensitive(tb_send, TRUE);
    gtk_widget_set_sensitive(tb_userlist, TRUE);
      
    gtk_widget_hide(edittoolbarhandle);

    gM_window_update_discon(w);
  }
#if 0 /* fixed ... */
  if(w == NULL) {/* wierd, it's a bug, but I cand find it ... this is just a bad hack .. oh well :) */
    return;
  }
#endif
  gtk_widget_set_style(w->label, tabnorm_style);
  gtk_label_set_text(GTK_LABEL(w->label), w->name);
  w->lines_pending = 0;
  
  /* Update the clock (if any) */
  clockfunc(NULL);
}


void handle_discon() {
  world *w;
  
  w = gM_window_get_current_world();
  if(w == NULL)
    return;
  
  if(w->sock > -1) /* world is connected */
    gM_world_disconnect(w, FALSE);
  else
    gM_world_connect(w);
}

void gM_window_update_discon(world *w) {
  int page_num;
  char *s[] = {_("Disconnect"), _("Connect")};
  int i;

/*printf("update discon\n");*/
  page_num = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
  if(w == NULL) {
    gtk_widget_hide(worldmenu);
    if(gM_worlds_get_world_by_editor_main(gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page_num), &i) != NULL) {
      gtk_widget_show(editmenu);
      gtk_widget_show(searchmenu);
    }
    return;
  }
  
  gtk_label_set_text(GTK_LABEL(GTK_BIN(disconmenu)->child), w->sock > 0 ? s[0] : s[1]);

  gtk_signal_handler_block_by_func(GTK_OBJECT(tb_discon), GTK_SIGNAL_FUNC(handle_discon), NULL);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tb_discon), w->sock > 0 ? TRUE : FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(tb_discon), GTK_SIGNAL_FUNC(handle_discon), NULL);

  gtk_label_set_text(GTK_LABEL(GTK_BIN(worldmenu)->child), w->name);
  gtk_widget_hide(searchmenu);
  gtk_widget_hide(editmenu);
  gtk_widget_show(worldmenu);
}

void update_discon_pixmap(GtkWidget *dc, void *pointer) {
  if(GTK_TOGGLE_BUTTON(tb_discon)->active > 0) {
    gtk_widget_hide(connect_off);
    gtk_widget_show(connect_on);
  } else {
    gtk_widget_hide(connect_on);
    gtk_widget_show(connect_off);
  }
}

void toggle_userlist(GtkWidget *menuitem, void *pointer) {
  gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(tb_userlist), GTK_CHECK_MENU_ITEM(menuitem)->active);
}



