/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>	/* typedefs */
#include <sys/socket.h>	/* connect() */
#include <netinet/in.h>	/* gethostbyname() */
#include <netdb.h>			/* htons() */
#include <unistd.h>			/* close() */
#include <string.h>
#include <time.h>       /* ctime() */
#include <fcntl.h>      /* open(), close() ao */
#include <errno.h>

#include "structs.h"
#include "world.h"
#include "config.h"
#include "window.h"
#include "history.h"
#include "handy.h"
#include "global_vars.h"
#include "mcp.h"
#include "userlist.h"
#include "editor.h"
#include "dialog.h"
#include "script.h"
#include "worlddialogs.h"
#include "triggers.h"

#if 0 /* obsolete, but might be used in the future */
#define HIDDEN_STR "#$#__gMOO__#$#\n"
#endif 

int  init_sockaddr(struct sockaddr_in *wname, char *whostname, int wport);
int  parse(world *w, char *s, int i, int maxlen);
void enter_callback(GtkWidget *input, world *w);
int  key_callback(GtkWidget *input, GdkEventKey *event, world *w);
void handle_out_of_band(world *w, char *s, int len);
void handle_external2(world *w, char *line, int i, int len);
int  editors_ok_for_discon(world *w);
void resize_output(GtkWidget *output, GtkAllocation  *allocation, world *w);

void add_10_if_needed(world *w);
void update_line_label(world *w);

/* logging : */
void write_log(world *w, char *s);
void open_log(world *w);
void close_log(world *w);
/* macros */
#if 0 /* macors */
char *parse_macro(world *w, char *mtext);
char *get_word(char *from, int num);
#endif

/* needed for signal blocking */
extern void handle_discon();

void do_keypad(world *w, int key);

/* used in gM_world_print */
void parse_print_trigger(world *w, char *line, int *i, int *napos, int len);

void input_size_allocate(GtkWidget *input, GtkAllocation *a);

world *gM_world_new(char *worldname) {
  world *w = NULL;

  w = gM_config_read_world(worldname);
  if(w == NULL) /* something went wrong reeading the config file */
    return w;
  w->next           = w->prev = NULL;
  w->sock 				  = w->gsock = -1;
  w->page_num       = -1;
  w->fgcolor        = DEFAULTFG;
  w->bgcolor        = DEFAULTBG;
  w->curfont        = normalfont;
  w->hist           = NULL;
  w->selection      = NULL;
  w->get_for_editor = -1;
  w->editcount      = 0;
  w->edit           = NULL;
  w->log            = -1;
#if 0 /* macros */
  w->mac            = NULL;
#endif
  w->locked         = 0;
  w->ignore         = 0;
  w->insert_10      = 0;
  w->session        = NULL;
  w->lines_pending  = 0;
#if 0 /* That damned GtkText widget is acting up again */
  w->buffer_len     = 0;
#endif
  w->senddialog     = w->sendfirst = w->sendtext = w->sendlast = NULL;
  w->finddialog     = NULL;
  w->conname        = w->conpasswd = NULL;
  w->outputsize[0]  = w->outputsize[1] = 0;
  w->ansibuffer     = NULL;
  w->force_no_echo  = 0;
  w->really_reconnect = 0;
#ifdef TCL
  w->interp         = NULL;
  w->tcl_name       = NULL;
  w->send_buffer    = NULL;
#endif
  w->highlites      = NULL;
  w->runcount       = 0;
  w->oob_buffer     = NULL; /* MCP Out Of Band buffering */
#if 0 /* macros */
  gM_macros_init_for(w);
#endif
  return w; /* succes */
}

void gM_world_close(world *w) {
  gM_world_disconnect(w, FALSE);
}

void gM_world_show(world *w) {
  char *name;

  if(multiline > 0)
    w->main = gtk_vpaned_new();
  else
    w->main = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(w->main);
  GTK_WIDGET_UNSET_FLAGS(w->main, GTK_CAN_FOCUS); 

  w->hpaned = gtk_hpaned_new();
/*  gtk_widget_show(w->hpaned); */
  if(multiline < 1)
    gtk_box_pack_start(GTK_BOX(w->main), w->hpaned, TRUE, TRUE, 0);
  else
    gtk_widget_ref(w->hpaned);
  GTK_WIDGET_UNSET_FLAGS(w->hpaned, GTK_CAN_FOCUS);

/* TEXT output: */
  w->output_scroll = gtk_scrolled_window_new(NULL, NULL);
  gtk_widget_show(w->output_scroll);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(w->output_scroll), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
  GTK_WIDGET_UNSET_FLAGS(w->output_scroll, GTK_CAN_FOCUS);
  if(multiline > 0)
    gtk_paned_pack1(GTK_PANED(w->main), w->output_scroll, TRUE, TRUE);
  else
    gtk_box_pack_start(GTK_BOX(w->main), w->output_scroll, TRUE, TRUE, 0);

  w->output = gtk_text_new(NULL, NULL);
  gtk_widget_show(w->output);
  gtk_widget_set_name(w->output, "output");
  gtk_container_add(GTK_CONTAINER(w->output_scroll), w->output);
  gtk_signal_connect(GTK_OBJECT(w->output), "size_allocate", GTK_SIGNAL_FUNC(resize_output), w);
  GTK_WIDGET_UNSET_FLAGS(w->output, GTK_CAN_FOCUS);
  gtk_text_set_word_wrap(GTK_TEXT(w->output), wordwrap);

/* END TEXT output */

  /* USER LIST, inserted when the MCP package is finished */
  gM_userlist_init(w);

  /* ENTRY and clock: */
  w->ihbox = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(w->ihbox);
  if(multiline > 0)
    gtk_paned_pack2(GTK_PANED(w->main), w->ihbox, FALSE, TRUE);
  else
    gtk_box_pack_end(GTK_BOX(w->main), w->ihbox, FALSE, TRUE, 0);
  GTK_WIDGET_UNSET_FLAGS(w->ihbox, GTK_CAN_FOCUS);

  if(multiline > 0) {
    w->input = gtk_text_new(NULL, NULL);
    gtk_editable_set_editable(GTK_EDITABLE(w->input), TRUE);
    gtk_widget_set_usize(w->input, -1, multilinesize);
    gtk_signal_connect(GTK_OBJECT(w->input), "size_allocate",
                       GTK_SIGNAL_FUNC(input_size_allocate), NULL);
  } else {
    w->input = gtk_entry_new ();
    gtk_signal_connect(GTK_OBJECT(w->input), "activate", GTK_SIGNAL_FUNC(enter_callback), (void *) w);
  }
  gtk_signal_connect(GTK_OBJECT(w->input), "key_press_event", GTK_SIGNAL_FUNC(key_callback), (void *) w);

  gtk_box_pack_start(GTK_BOX(w->ihbox), w->input, TRUE, TRUE, 0);
  gtk_widget_show(w->input);

  /* END ENTRY and clock */

  if(multiline > 0)
    w->seperator = NULL;
  else {
    w->seperator = gtk_hseparator_new();
    gtk_box_pack_end(GTK_BOX(w->main), w->seperator, FALSE, TRUE, 4);
    if(showsep > 0)
      gtk_widget_show(w->seperator);
  }

  /* for the notebook */
  name = g_malloc(21);
  limit_to_20(FOR_WORLD, w->name, &name);
  w->label = gtk_label_new(short_tab_names > 0 ? name : w->name);
  g_free(name);
  gtk_widget_show(w->label);
  GTK_WIDGET_UNSET_FLAGS(w->label, GTK_CAN_FOCUS);


  set_global_color(w->output);
  if(definputstyle == 0)
    set_global_color(w->input);
}



int init_sockaddr(struct sockaddr_in *wname, char *whostname, int wport) {
  struct hostent *whostinfo;
  
  wname->sin_family = AF_INET;
  wname->sin_port = htons(wport);
  whostinfo = gethostbyname(whostname);
  if(whostinfo == NULL) {
    return -1;
  }
  wname->sin_addr = *(struct in_addr *) whostinfo->h_addr;
  return 0; 
}

void gM_world_connect(world *w) {
  struct sockaddr_in servername;
  char s[10];

  if(w->sock > 0) {
    gM_world_print(w, _("\e[31m% MOO is already connected\e[0m\n"), -1);
    return;
  }
  /* create socket */
  w->sock = socket(PF_INET, SOCK_STREAM, 0);
  if(w->sock < 0) {
    gM_world_print(w, _("\e[31m% Error creating socket\e[0m\n"), -1);
    return;
  }
  /* connect: */
  if(init_sockaddr(&servername, w->host, w->port) < 0) {
    gM_world_print(w, _("\e[31m% Could not resolve hostname \""), -1);
    gM_world_print(w, w->host, -1);
    gM_world_print(w, _("\" - Is your network down ?\e[0m\n"), -1);
    w->sock = -1;
    gtk_signal_handler_block_by_func(GTK_OBJECT(tb_discon), GTK_SIGNAL_FUNC(handle_discon), NULL);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tb_discon), FALSE);
    gtk_signal_handler_unblock_by_func(GTK_OBJECT(tb_discon), GTK_SIGNAL_FUNC(handle_discon), NULL);

    return;
  }

  sprintf(s, "%d", w->port);

  if(connect(w->sock, (struct sockaddr *)&servername, sizeof(servername)) < 0) {
    gM_world_print(w, _("\e[31m% Could not connect to \""), -1);
    gM_world_print(w, w->host, -1);
    gM_world_print(w, _("\" at port "), -1);
    gM_world_print(w, s, -1);
    gM_world_print(w, _(" - Host not listening\e[0m\n"), -1);
/*    error("connecting to socket"); */
    w->sock = -1;
    gtk_signal_handler_block_by_func(GTK_OBJECT(tb_discon), GTK_SIGNAL_FUNC(handle_discon), NULL);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tb_discon), FALSE);
    gtk_signal_handler_unblock_by_func(GTK_OBJECT(tb_discon), GTK_SIGNAL_FUNC(handle_discon), NULL);
    return;
  }
  gM_window_update_discon(w);

  gM_world_print(w, _("\e[32m% Connected to "), -1);
  gM_world_print(w, w->host, -1);
  gM_world_print(w, ":", 1);
  gM_world_print(w, s, -1);
  gM_world_print(w, "\e[0m\n", -1);

  w->gsock = gdk_input_add(w->sock, GDK_INPUT_READ, (GdkInputFunction) gM_world_read, w);
  
  open_log(w);
  w->connect_time = time(NULL);
  
  if(w->autoconnect > 0) { /* don't auto login */
    gM_world_write(w, "connect ");
    gM_world_write(w, w->user);
    gM_world_write(w, " ");
    gM_world_write(w, w->password);
    gM_world_write(w, "\n");
  } else if(popupcon > 0) { /* show connect dialog box: */
    gM_worlddialogs_connect(w);
  }

  gM_config_lock_add(w->name);
}


void gM_world_disconnect(world *w, int recon) {
  if(!recon && !editors_ok_for_discon(w))
    return;

  if(w->sock < 0) {
    gM_world_print(w, "% ", -1);
    gM_world_print(w, w->name, -1);
    gM_world_print(w, _(" was already disconnected, or wasn't ever open\n"), -1);
    return;
  }
  gM_world_print(w, _("% Disconnected\n"), -1);
  gdk_input_remove(w->gsock);
  close(w->sock);
  w->sock = w->gsock = -1; /* show that there has been disconnected */

  gM_window_update_discon(w);
  
  close_log(w);
  if(recon == TRUE && reconnect > 0) {
    if(w->really_reconnect) {
      gM_world_print(w, _("% Connection to world "), -1);
      gM_world_print(w, w->name, -1);
      gM_world_print(w, _(" lost, reconnecting...\n"), -1);
      gM_world_connect(w);
    } else {
      gM_world_print(w, _("\e[31m% Reconnect required, "
                          "but you haven't entered any text, "
                          "not reconnecting ...\e[0m\n"), -1);
    }
  }

  gM_config_lock_remove(w->name);

  gM_mcp_close(w);
}

/* output stuff */
void gM_world_handle_output(world *w, char *s, int len) {
  char *line;
  int linelen, i;
#ifdef DEBUG
printf("\n*** Got new block of data\n");
#endif
  line = g_malloc(len + 1);
  linelen = 0;
  for(i = 0; i < len; i++) {
    if(s[i] != 13) {
      line[linelen] = s[i];
      linelen++;
      if(s[i] == '\n') {
        line[linelen] = 0;
        gM_world_handle_output_line(w, line, linelen);
        linelen = 0;
      }
    }
  }
  /* Is the last char wasn't a '\n' */
  if(linelen != 0) {
    line[linelen] = 0;
    gM_world_handle_output_line(w, line, linelen);
  }
  g_free(line);
}

void gM_world_handle_output_line(world *w, char *line, int len) {
  char *line2;
  int   len2;
  int   i;
  int free_line = 0;

#ifdef DEBUG
printf("*** Handling line (len %d) >> %s", len, line);
#endif
  if(w->get_for_editor > -1) {
    if(len == 2 && strcmp(line, ".\n") == 0) {
      gM_editor_data_received(w->edit[w->get_for_editor]);
      w->get_for_editor = -1;
    } else {
      gM_editor_add(w->edit[w->get_for_editor], line, len);
    }  
    return;
  }
 
  if(w->ansibuffer != NULL) {
    line = strcatnew(w->ansibuffer, line);
    len += strlen(w->ansibuffer);
    g_free(w->ansibuffer);
    w->ansibuffer = NULL;
  }

  line2 = strip_chars(line, len, &len2);
#ifdef DEBUG
printf("*** Stripping resulted in (len %d)>> %s\n", len2, line2);
#endif
  if(w->oob_buffer != NULL) {
    /* NOTE: reuse line string here, so freein() it if needed ! */
    if(free_line) g_free(line);
    if(line[len - 1] != 10) { /* Buffed, but have to rebuffer :( */
        line = w->oob_buffer;
        w->oob_buffer = strcatnew(line, line2);
        g_free(line); /* the old oob_buffer */
    } else { 
      line = strcatnew(w->oob_buffer, line2 + 3);
      handle_out_of_band(w, line, strlen(line));
      g_free(line);
      g_free(w->oob_buffer);
      w->oob_buffer = NULL;
    }
    g_free(line2);
    return;
  } else if(strncmp(line2, "#$#", 3) == 0) { /* the "#$#" will be cut off */
#ifdef DEBUG
printf("*** MCP OOB command found\n");
#endif
  	if(line[len - 1] != 10) /* Line not terminated with an \n, buffer it */
      /* Buffer _must_ be NULL (no other way, see if statement above) */
      /* The +3 kills the #$# form the line */
      w->oob_buffer = strnew(line2 + 3);
    else
      handle_out_of_band(w, line2 + 3, len2 - 3);
#ifdef DEBUG
printf("\n");
#endif
  	g_free(line2);
    if(free_line)
      g_free(line);
    return;
  } else if(strncmp(line2, "#@#EXTERNAL#@#", 14) == 0) {
#ifdef DEBUG
printf("*** Externa, found\n");
#endif
    handle_external2(w, line2, 15, len2);
#ifdef DEBUG
printf("\n");
#endif
  	g_free(line2);
    if(free_line)
      g_free(line);
    return;
  }

#ifdef DEBUG
printf("*** Checking triggers\n");
#endif
  i = gM_triggers_check(w, line2, len2);
  g_free(line2);

#ifdef DEBUG
printf("*** Printing to screen\n");
#endif
  gM_world_print(w, line, len);

  if(i) {
#ifdef TCL
    if(w->highlites != NULL) {
      g_free(w->highlites);
      w->highlites = NULL;
    }
#endif
    if(w->runcount) {
      for(i = 0; i < w->runcount; i++) {
#ifdef DEBUG
printf("*** Doing trigger run [%d/%d]\n", i, w->runcount);
#endif
        gM_triggers_run(w, w->runs[i]);
        g_free(w->runs[i]->line);
        g_free(w->runs[i]->matched);
        g_free(w->runs[i]);
      }
      g_free(w->runs);
      w->runcount = 0;
    }
  }
#ifdef DEBUG
printf("\n");
#endif
  if(free_line)
    g_free(line);
}

void add_10_if_needed(world *w) {
  char *tmp;
  char *name;
#if 0 /* That damned GtkText widget is acting up again */
  int i;
#endif
  GtkAdjustment *vadj = GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj);
  
  if(w->insert_10 == 1) {
    if(w->locked < 1 && (vadj->value != (vadj->upper - vadj->page_size))) {
    	gtk_text_freeze(GTK_TEXT(w->output));
      gtk_adjustment_set_value(vadj, vadj->upper - vadj->page_size);
      gtk_text_thaw(GTK_TEXT(w->output));
    }
    gtk_text_insert(GTK_TEXT(w->output), normalfont, NULL, NULL, "\n", 1);
    write_log(w, "\n");
    /* update the length of the buffer (in lines) */
#if 0 /* That damned GtkText widget is acting up again */
    w->buffer_len++;
    /* Remove old line from the buffer (if needed) */
    if(buffer_len > 0 && w->buffer_len > buffer_len) {
      tmp = gtk_editable_get_chars(GTK_EDITABLE(w->output), 0, gtk_text_get_length(GTK_TEXT(w->output)));
      for(i=0; tmp[i] != '\n' && tmp[i] != 0; i++);
      gtk_editable_delete_text(GTK_EDITABLE(w->output), 0, i);
      g_free(tmp);
    }
#endif
    /* Update the line(count) label if required */
    if(gM_window_get_current_world() != w) {
      w->lines_pending++;
      gtk_widget_set_style(w->label, tabhigh_style);
      if(w->lines_pending > 1) {
        tmp = g_malloc(strlen(w->name) + 30);
        name = g_malloc(21);
  			limit_to_20(FOR_WORLD, w->name, &name);
        sprintf(tmp, "%s [%d]", short_tab_names > 0 ? name : w->name, w->lines_pending);
        gtk_label_set_text(GTK_LABEL(w->label), tmp);
        g_free(name);
        g_free(tmp);
      }
    }
  }
  w->insert_10 = 0;
}

#define scroll_down()\
        if(w->locked < 1 && vadj->value != vadj->upper - vadj->page_size) {\
          gtk_text_freeze(GTK_TEXT(w->output));\
          gtk_adjustment_set_value(vadj, vadj->upper - vadj->page_size);\
          gtk_text_thaw(GTK_TEXT(w->output));\
        }
#define insert_current()\
        gtk_text_insert(GTK_TEXT(w->output),\
                        w->curfont,\
                        w->fgcolor != DEFAULTFG ? &colors[w->fgcolor] : NULL,\
                        w->bgcolor != DEFAULTBG ? &colors[w->bgcolor] : NULL,\
                        s, j);\
        write_log(w, s)
void gM_world_print_raw(world *w, char *line, int len) {
  GtkAdjustment *vadj = GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj);
  int i,j;
  char *s;

  if(w == NULL)
    return;

  if(len < 0)
    len = strlen(line);

  s = g_malloc(len + 1);
  j = 0;
  for(i = 0; i < len; i++) {
    switch(line[i]) {
    case 10:
      add_10_if_needed(w);
      if(j != 0) {
        s[j] = 0;
        scroll_down()
        insert_current();
        j = 0;
      }
      w->insert_10 = 1;
      break;
    default:
      s[j] = line[i];
      j++;
    }
  }
  if(j != 0) {
    s[j] = 0;
    add_10_if_needed(w);
    scroll_down()
    insert_current();
  }
  g_free(s);
}

void gM_world_print(world *w, char *line, int len) {
  int i,j;
  char *s;
  GtkAdjustment *vadj = GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj);
  int napos = 0;  /* non ansi position, used by the highliter */

  if(w == NULL)
    return;

  if(len < 0)
    len = strlen(line);

  s = g_malloc(len + 1);
  j = 0;
  
  for(i = 0; i < len; i++) {
    if(w->highlites != NULL) { 
#ifdef DEBUG
printf("*** Highlighting the text\n");
#endif
      if(j != 0) {
        s[j] = 0;
        add_10_if_needed(w);
        scroll_down()
        insert_current();
        j = 0;
      }
      while(napos < w->highlitecount && w->highlites[napos] != -1) {
        parse_print_trigger(w, line, &i, &napos, len);
      }
    }
    switch(line[i]) {
    case -1:
      if(i + 2 < len && line[i + 2] == 1) {
        if(line[i + 1] == -5) /* echo off */
          w->force_no_echo = 1;
        else if(line[i + 1] == -4) /* echo back on */
          w->force_no_echo = 0;
        i += 2;
      }
      break;
    case 7:
      if(beep_type == beep_type_speaker ||
         beep_type == beep_type_speaker + beep_type_visual)
        gdk_beep();
      if(beep_type == beep_type_visual ||
         beep_type == beep_type_speaker + beep_type_visual) {
        gtk_widget_hide(w->output);
        gtk_widget_show(w->output);
        gtk_widget_hide(w->output);
        gtk_widget_show(w->output);
        gtk_widget_hide(w->output);
        gtk_widget_show(w->output);
      }
      break;
    case 10:
      add_10_if_needed(w);
      if(j != 0) {
        s[j] = 0;
        scroll_down()
        insert_current();
        j = 0;
      }
      w->insert_10 = 1;
      break;
    case 27:
      if(j != 0) {
        s[j] = 0;
        add_10_if_needed(w);
        scroll_down()
        insert_current();
        j = 0;
      }
      i = parse(w, line, i, len);
      break;
    default:
      s[j] = line[i];
      j++;
      napos++;
    }
  }
  if(j != 0) {
    s[j] = 0;
    add_10_if_needed(w);
    scroll_down()
    insert_current();
  }
  g_free(s);
  if(w->locked < 1)
    gtk_adjustment_set_value(vadj, vadj->upper - vadj->page_size);
}

int parse(world *w, char *s, int i, int len) {
  char *tmp;
  int *numlist = NULL;
  int numcount = 0;
  int j, k;
  
  if(s[i + 1] != '[') /* no escape code found */
    return i;
  tmp = g_malloc(len - i + 1);
  i += 2;
  for(j = 0; 1; j++) {
    if(s[j + i] == 0) {/* end of line, broken escape code ? */
  		w->ansibuffer = strnew(s + i - 2); /* now pointing at the ESC char */
      /*return i;*/ return len;          /* skipping the ramaining chars */
    } else if(s[j + i] == ';') /* one number found */
      numcount++;
    else if(s[j + i] < '0' || s[j + i] > '9') { /* last code found */
      numcount++;
      break;
    }
  }
  if(numcount != 0) 
    numlist = g_malloc(sizeof(int) * numcount);
  for(j = 0; j < numcount; j++) {
    k = 0;
    while(s[k + i] >= '0' && s[k + i] <= '9') {
      tmp[k] = s[k + i];
      k++;
    }
    tmp[k] = 0;
    i += k + 1;
    numlist[j] = atoi(tmp);
  }
  i--;
/*printf("parsing complete\n");*/
  switch(s[i]) {
  case 'm': /* A display attribute */
    for(j = 0; j < numcount; j++) {
      switch(numlist[j]) {
      case  0: /* normal    */
        w->fgcolor = DEFAULTFG;
        w->bgcolor = DEFAULTBG;
        w->curfont = normalfont;
        break;
      case  1: /* bold      */
        w->curfont = boldfont;
        break;
      case  4: /* underline  *** NOT SUPPORTED *** */
        break;
      case  5: /* blink      *** NOT SUPPORTED *** */
        break;
      case  7: /* reverse    *** NOT SUPPORTED ****/
        break;
      case  8: /* invisible */
        w->fgcolor = w->bgcolor;
        break;
      case 30: /* black     */
        w->fgcolor = BLACK;
        break;
      case 31: /* red       */
        w->fgcolor = RED;
        break;
      case 32: /* green     */
        w->fgcolor = GREEN;
        break;
      case 33: /* yellow    */
        w->fgcolor = YELLOW;
        break;
      case 34: /* blue      */
        w->fgcolor = BLUE;
        break;
      case 35: /* magenta   */
        w->fgcolor = MAGENTA;
        break;
      case 36: /* cyan      */
        w->fgcolor = CYAN;
        break;
      case 37: /* white     */
        w->fgcolor = WHITE;
        break;
      case 40: /* black     */
        w->bgcolor = BLACK;
        break;
      case 41: /* red       */
        w->bgcolor = RED;
        break;
      case 42: /* green     */
        w->bgcolor = GREEN;
        break;
      case 43: /* yellow    */
        w->bgcolor = YELLOW;
        break;
      case 44: /* blue      */
        w->bgcolor = BLUE;
        break;
      case 45: /* magenta   */
        w->bgcolor = MAGENTA;
        break;
      case 46: /* cyan      */
        w->bgcolor = CYAN;
        break;
      case 47: /* white     */
        w->bgcolor = WHITE;
        break;
      }
    }
    break;
  }
  g_free(numlist);
  g_free(tmp);
  return i;
}



/* Read from the world */
void gM_world_read(world *w, int source, GdkInputCondition con) {
#define BUF_SIZE 2048 /* 2 Kb buffer (enough for most cases) */
  int len;
  int mult = 1;
  char *s = NULL;

  while(1) {
    s = g_realloc(s, BUF_SIZE * mult + 1);
    len = read(w->sock, s + BUF_SIZE * (mult - 1), BUF_SIZE);
    if(len == 0) {
      g_free(s);
      gM_world_disconnect(w, TRUE);
      return;
    }
    s[len + BUF_SIZE * (mult - 1)] = 0; /* add terminating nul char ... */
    if(len != BUF_SIZE)	/* Break if buffer isn't completely filled */
      break;
    mult++;
  }
  gM_world_handle_output(w, s, len + BUF_SIZE * (mult - 1));
  g_free(s);
}

/* Write strings to the world */
void gM_world_write(world *w, char *s) {
  if(w->sock < 0) {
    gM_world_print(w, _("\e[31m% MOO not connected, text not send\n"), -1);
    return;
  }
  /* Maby check if writing is possible ... never had an error though */
/*printf(">>%s<<", s);*/
  write(w->sock, s, strlen(s));
}                      

/* write it, and also set the prefix and suffix strings */
#if 0  /* This is now obsolete, byt might be used in a future time ... */
void gM_world_write_hidden(world *w, char *s) {
  if(w->sock < 0) {
    error_arg(_("MOO not conneted, \"%s\" not send\n"), s);
    return;
  }
  write(w->sock, "SUFFIX ", 7);
  write(w->sock, HIDDEN_STR, strlen(HIDDEN_STR));
  write(w->sock, "PREFIX ", 7);
  write(w->sock, HIDDEN_STR, strlen(HIDDEN_STR));
  write(w->sock, s, strlen(s));
}                      
#endif


void enter_callback(GtkWidget *input, world *w) {
  char *t;
  char *s;
  int i = 0;

/*printf("Enter Press\n");*/
  if(multiline > 0)
    t = gtk_editable_get_chars(GTK_EDITABLE(input), 0,
          gtk_text_get_length(GTK_TEXT(input)));
  else
    t  = gtk_entry_get_text(GTK_ENTRY(input));

  if(t[0] != '/') {
    s = strcpy(g_malloc(strlen(t) + 2), t); /* reserve a space for the \n char */
    if(multiline > 0)
      g_free(t);
#ifdef TCL
  } else if(tcl_debug > 0 && t[1] == '/') { /* Special TCL_eval macro */
    gM_history_add(w);
    if(multiline > 0)
      gtk_editable_delete_text(GTK_EDITABLE(input), 0,
        gtk_text_get_length(GTK_TEXT(input)));
    else
      gtk_entry_set_text(GTK_ENTRY(input), "");
    gM_script_do_eval(w, t + 2);
    if(multiline > 0)
      g_free(t);
    return;
#endif /* TCL */
  } else { /* run a macro */
    gM_history_add(w);
    if(multiline > 0)
      gtk_editable_delete_text(GTK_EDITABLE(input), 0,
        gtk_text_get_length(GTK_TEXT(input)));
    else
      gtk_entry_set_text(GTK_ENTRY(input), "");
#ifdef TCL
    gM_script_do_macro(w, t);
#else
  	gM_world_print(w, _("\e[31m% Error in macro: No tcl support :(\e[0m\n"), -1);
#endif
    if(multiline > 0)
      g_free(t);
    return;
  }

  /* if MCP is enabled, lines starting with #$# and #$" will be prefixed with #$" */
  if(w->session != NULL &&
     (strncmp(s, "#$#", 3) == 0 || strncmp(s, "#$\"", 3) == 0)) {
    gM_world_write(w, "#$\"");
  }
  
  if(show_own > -1 && w->force_no_echo < 1) {
    t = g_malloc(strlen(s) + 15);
    switch(show_own) {
    case BLACK:   i = 30; break;
    case WHITE:   i = 37; break;
    case RED:     i = 31; break;
    case GREEN:   i = 32; break;
    case BLUE:    i = 34; break;
    case CYAN:    i = 36; break;
    case MAGENTA: i = 35; break;
    case YELLOW:  i = 33; break;
    }
    sprintf(t, "\e[%dm%s\e[0m\n", i, s);
    gM_world_print(w, t, -1);
    g_free(t);
  }

  /* do one write() instead of 2 */
  strcat(s, "\n");
  gM_world_write(w, s);

  gM_history_add(w);

  if(logtype & loguse && logtype & logsaveinput) { /* check if logging of user input is required */
    if(strncasecmp(s, "co ", 3) == 0 || strncasecmp(s, "connect ", 8) == 0) {
      gM_world_print(w, _("\e[32m% Password detected, NOT appended to the logfile\e[0m\n"), -1);
      /* linength may be updated after an connect */
      gM_config_lock_add(w->name);
    } else {
      /* here we must first write the \n for the world output is
       * wroted with an trailing \n too */
      s[strlen(s) - 1] = 0;
      write_log(w, "\n");
      write_log(w, s);
    }
  }
  g_free(s);
  if(multiline > 0)
    gtk_editable_delete_text(GTK_EDITABLE(input), 0,
      gtk_text_get_length(GTK_TEXT(input)));
  else
    gtk_entry_set_text(GTK_ENTRY(input), "");
  
  /* You have antered text, so we really want to reconnect */
  w->really_reconnect = 1;
}

int key_callback(GtkWidget *input, GdkEventKey *event, world *w) {
  int num;

/*printf("Keypress\n");*/
  gtk_widget_grab_focus(input);
  if(event->keyval == GDK_Tab || event->keyval == GDK_KP_Tab || event->keyval == GDK_ISO_Left_Tab) {
    num = g_list_length(GTK_NOTEBOOK(notebook)->children);
    if(event->state & 4) { /* CTR-tab */
      if(w->page_num != num - 1)
        gtk_notebook_next_page(GTK_NOTEBOOK(notebook));
      else
        gtk_notebook_set_page(GTK_NOTEBOOK(notebook), 0);
    } else if(event->state & 1) { /* SHIFT-tab */
      if(w->page_num != 0)
        gtk_notebook_prev_page(GTK_NOTEBOOK(notebook));
      else
        gtk_notebook_set_page(GTK_NOTEBOOK(notebook), num - 1);
    }
  } else if(event->keyval == GDK_Left  && event->state & 4) {
    num = g_list_length(GTK_NOTEBOOK(notebook)->children);
    if(w->page_num != 0)
      gtk_notebook_prev_page(GTK_NOTEBOOK(notebook));
    else
      gtk_notebook_set_page(GTK_NOTEBOOK(notebook), num - 1);
  } else if(event->keyval == GDK_Right && event->state & 4) {
    num = g_list_length(GTK_NOTEBOOK(notebook)->children);
    if(w->page_num != num - 1)
      gtk_notebook_next_page(GTK_NOTEBOOK(notebook));
    else
      gtk_notebook_set_page(GTK_NOTEBOOK(notebook), 0);
  } else if( (event->keyval == GDK_p && event->state & 4) ||
             (multiline < 1 && event->keyval == GDK_Up) ||
             (event->keyval == GDK_Up && event->state & 4)) {
    gM_history_up(w);
  } else if( (event->keyval == GDK_n && event->state & 4) ||
             (multiline < 1 && event->keyval == GDK_Down) ||
             (event->keyval == GDK_Down && event->state & 4)) {
    gM_history_down(w);
  } else if(event->keyval == GDK_Page_Up) {
    gtk_adjustment_set_value(GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj),
      											 GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj)->value -
                             GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj)->page_size);
  } else if(event->keyval == GDK_Page_Down) {
    if(GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj)->upper - 
       GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj)->value <=
       GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj)->page_size) {
      gtk_adjustment_set_value(GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj),
                               GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj)->upper -
                               GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj)->page_size);
    } else {
      gtk_adjustment_set_value(GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj),
                               GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj)->value +
                               GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj)->page_size);
    }
  } else if(event->keyval == GDK_Scroll_Lock) {
    if(event->state & 128) {
      w->locked = 0;
  	  gtk_text_thaw(GTK_TEXT(w->output));
      gtk_adjustment_set_value(GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj),
                               GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj)->upper -
                               GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj)->page_size);
    } else {
  	  gtk_text_freeze(GTK_TEXT(w->output));
      w->locked = 1;
    }
  } else if(event->keyval == GDK_l && event->state & 4) {
    gM_world_clear_screen(w);
  } else if(event->keyval == GDK_r && event->state & 4) {
    gM_history_purge(w);
  } else if(event->keyval == GDK_f && event->state & 4) {
    gtk_menu_item_activate(GTK_MENU_ITEM(findmenu));
  } else if(event->state == 4 && event->keyval >= GDK_1 && event->keyval <= GDK_9) {
    gtk_notebook_set_page(GTK_NOTEBOOK(notebook), event->keyval - GDK_1);
  } else if(event->state == 0 && event->keyval >= GDK_KP_Enter && event->keyval <= GDK_KP_9) {
    do_keypad(w, event->keyval);
  } else if((multiline > 0 &&
               event->keyval == GDK_Return && 
               event->state != 4) ||
            (event->state == 16 && event->keyval == GDK_KP_Enter)) {
  	if(multiline > 0)
      enter_callback(input, w);
    else
      gtk_signal_emit_by_name(GTK_OBJECT(input), "activate");
  } else if(multiline > 0 && event->keyval == GDK_Return && event->state == 4) {
  	gtk_text_insert(GTK_TEXT(input), NULL, NULL, NULL, "\n", 1);
  } else
    return FALSE;
  gtk_signal_emit_stop_by_name(GTK_OBJECT(input), "key_press_event");

  return TRUE;
}


void handle_out_of_band(world *w, char *s, int len) {
  char *tmp;
  char *name;
  char *upload;

  /* begin checking for the #$# edit non-MCP thningie */
  if(strncasecmp(s, " edit", 5) == 0) { /* found edit */
    s += 11;

    tmp = strstr(s, " upload: ");
    name = g_malloc(tmp - s);
    strncpy(name, s + 1, tmp - s - 1);
    name[tmp - s - 1] = 0; /* strncpy doesn't do this ?? */

    s = tmp + 9;
    upload = strnew(s);

    w->get_for_editor = gM_editor_new(w, name, upload);
    /* s is unreliable now */
  } else {
    /* Trailing \n is cut of in the handle_output_line .. */
#ifdef DEBUG
    printf("\n\n\n**** Handling MCP Out Of Band (green):\n>>\e[32m%s\e[0m<<\n\n\n", s);
#endif
    gM_mcp_do_out_of_band(w, s, len);
  }
}


void write_log(world *w, char *s) {
  if(!(logtype & loguse))
    return;

  if(w->log < 1)
    open_log(w);

  write(w->log, s, strlen(s));
}

void open_log(world *w) {
  char *s;

  if( !(logtype & loguse) || w->log > 0)
    return;

  s = g_malloc(sizeof(char) * (strlen(config_dir) + strlen(w->name) + 50));
  strcpy(s, config_dir);
  strcat(s, "/Logs/");
  strcat(s, w->name);
/*if(logtype & logappend) { Don't do anything } */
  if(logtype & logday) {
    strcat(s, " - ");
  	strcat(s, get_date());
  } else if(logtype & logtime) {
    strcat(s, " - ");
  	strcat(s, get_date());
  	strcat(s, " ");
  	strcat(s, get_time());
  }
  strcat(s, ".log");

  w->log = open(s, O_WRONLY | O_CREAT | O_APPEND, 00644);
  if(w->log < 0) { /* an error occured */
    gM_world_print(w, _("% Error opening logfile (\""), -1);
    gM_world_print(w, s, -1);
    gM_world_print(w, "\"): ", -1);
    gM_world_print(w, strerror(errno), -1);
    gM_world_print(w, _("\n% Output wil NOT be added to the log.\n"), -1);
  }
  g_free(s);
}

void close_log(world *w) {
  if(w->log < 0) {
    return;
  }
  close(w->log);
  w->log = -1;
}
#if 0 /* macros */
char *parse_macro(world *w, char *mtext) {
  macro *m;
  char *mname;
  int len;
  char *ret;
  int i;
  char *tmp;
  
  if(w->mac == NULL) {
/*printf("No macros found: ");*/
    return NULL;
  }

  mname = strchr(mtext, ' ');
  if(mname == NULL) {
/*printf("No args were found\n");*/
    mname = strnew(mtext + 1);
  } else {
/*printf("One arg was found\n");*/
    len = mname - mtext - 1;
    mname = g_malloc(len + 2);
    strncpy(mname, mtext + 1, len);
    mname[len] = 0; /* ensurance */
  }
/*printf("mtext == \"%s\",  ", mtext);*/
/*printf("mname == \"%s\"\n",mname);*/
  for(m = w->mac; m != NULL; m = m->next) {
    if(strcmp(m->name, mname) == 0)
      break;
  }

  g_free(mname);

  if(m == NULL) {/* nothing was found */
/*printf("No matching macro found: ");*/
    return NULL;
  }


  if(strchr(m->what, '$') == NULL) {
    return m->what;
  }
/*printf("actual parsing must be done\n");*/
  
  ret = g_malloc((strlen(mname) + strlen(m->what)) * 2);
  ret[0] = 0;
  for(i = 0; i <= strlen(m->what); i++) {
    if(m->what[i] != '$') {
/*printf("char found: '%c' (%d) at pos %d\n", m->what[i], m->what[i], i);*/
      sprintf(ret, "%s%c", ret, m->what[i]);
    } else {
/*printf("$ sign found at pos %d\n", i);*/
      i++;
      if(m->what[i] == '$') {
/*printf("another one found, adding it\n");*/
        sprintf(ret, "%s$", ret);
      } else if(m->what[i] >= '1' && m->what[i] <= '9') {
/*printf("argument detexted\n");*/
        tmp = get_word(mtext, m->what[i] - 48);
        if(tmp == NULL) {
/*          printf("Error, arg not found: ");*/
          g_free(ret);
          return NULL;
        }
        sprintf(ret, "%s%s", ret, tmp);
        g_free(tmp);
      } else {
/*        printf("Error in $ assignment: ");*/
        g_free(ret);
        return NULL;
      }
    }
  }

  return ret;
}

char *get_word(char *from, int num) {
  int len = strlen(from);
  char *ret = g_malloc(len);
  int i;
  int j = 0;

/*printf("Request for word %d received\n", num);*/

  for(i = 0; i <= len; i++) {
    if(from[i] == ' ') {
      while(from[i + 1] == ' ')
        i++;
      j++;
      if(j == num)
        break;
    }
  }
  if(j != num) /* not so many words have been found */
    return NULL;

  j=0; /* reuse the var */
  i++;
  for(; i <= len; i++) {
    if(from[i] == ' ')
      ret[j] = 0;
    else
      ret[j] = from[i];
    if(ret[j] == 0)
      break;
    j++;
  }
  return ret;
}
#endif

void handle_external2(world *w, char *line, int i, int len) {
  if(strncmp(line + i, "mailreader",10) == 0) {
    if(ext_mail == NULL)
      gM_world_print(w, _("\e[31m% Error: No mail reader specified\e[0m\n"), -1);
    else
      system(ext_mail);
  } else if(strncmp(line + i, "newsreader",10) == 0) {
    if(ext_news == NULL)
      gM_world_print(w, _("\e[31m% Error: No news reader specified\e[0m\n"), -1);
    else
      system(ext_news);
  }
}

int editors_ok_for_discon(world *w) {
  int i;
  int ok_count = 0;
  char s[80];
  
  if(w->editcount < 1)
    return 1;
    
  for(i = 0; i < w->editcount; i++) {
    if(w->edit[i]->has_changed < 1)
      ok_count++;
  }
  if(ok_count == w->editcount)
    return 1;
  
  i = w->editcount - ok_count;
  sprintf(s, _("There %s %d %s open, that %s send its text ...\nClose %s first."),
             i == 1 ? _("is")  : _("are"),
             i,
             i == 1 ? _("editor") : _("editors"),
             i == 1 ? _("hasn't") : _("haven't"),
             i == 1 ? _("it")  : _("them"));
  gM_dialog_show(_("Question:"), s, 1);

  return 0;
}

void set_global_color(GtkWidget *w) {
#if 0
  static GtkStyle style;

  memcpy(&style, w->style, sizeof(GtkStyle));

  memcpy(&style.fg[GTK_STATE_NORMAL],   &colors[DEFAULTFG], sizeof(GdkColor));
  memcpy(&style.text[GTK_STATE_NORMAL], &colors[DEFAULTFG], sizeof(GdkColor));

  memcpy(&style.bg[GTK_STATE_NORMAL],   &colors[DEFAULTBG], sizeof(GdkColor));
  memcpy(&style.base[GTK_STATE_NORMAL], &colors[DEFAULTBG], sizeof(GdkColor));

  style.font = normalfont;
  /* The next line doesn't work, how do I get rid of the bg pixmap ? */
  style.bg_pixmap[GTK_STATE_NORMAL] = NULL;

  gtk_widget_set_style(w, &style);
#endif
   GtkStyle  *style;

   style = gtk_style_copy(gtk_widget_get_style(w));

   memcpy(&style->fg[GTK_STATE_NORMAL],   &colors[DEFAULTFG], sizeof(GdkColor));
   memcpy(&style->text[GTK_STATE_NORMAL], &colors[DEFAULTFG], sizeof(GdkColor));

   memcpy(&style->bg[GTK_STATE_NORMAL],   &colors[DEFAULTBG], sizeof(GdkColor));
   memcpy(&style->base[GTK_STATE_NORMAL], &colors[DEFAULTBG], sizeof(GdkColor));

   style->font = normalfont;
   gdk_font_ref(normalfont);
   /* The next line doesn't work, how do I get rid of the bg pixmap ? */
   style->bg_pixmap[GTK_STATE_NORMAL] = NULL;

   gtk_widget_set_style(w, style);
   
   gtk_style_unref(style);
}

void resize_output(GtkWidget *output, GtkAllocation *allocation, world *w) {
  if(w->session != NULL && w->sock > 0)
    gM_mcp_vgmoo_client_screensize(w->session);
}

void gM_world_force_update_linelen(world *w) {
  char s[80];
  if(w != NULL && w->sock > 0) {
    sprintf(s, "@linelength %d\n",
               (w->output->allocation.width - 15) /
               gdk_char_width(normalfont, 'm'));
    gM_world_write(w,s);
  }
}

void do_keypad(world *w, int key) {
  int i = -1, j = 0;
  char *s;
  
  switch(key) {
  case GDK_KP_Home:      i = HOME; break;
  case GDK_KP_End:       i = END; break;
  case GDK_KP_Page_Up:   i = PGUP; break;
  case GDK_KP_Page_Down: i = PGDN; break;

  case GDK_KP_Left:      i = LEFT; break;
  case GDK_KP_Right:     i = RIGHT; break;
  case GDK_KP_Up:        i = UP; break;
  case GDK_KP_Down:      i = DOWN; break;
  case GDK_KP_Begin:     i = MIDDLE; break;

  case GDK_KP_Insert:    i = INS; break;
  case GDK_KP_Delete:    i = DEL; break;

  case GDK_KP_Divide:    i = DIV; break;
  case GDK_KP_Multiply:  i = MUL; break;
  case GDK_KP_Subtract:  i = MINUS; break;
  case GDK_KP_Add:       i = PLUS; break;
  case GDK_KP_Enter:     i = ENTER; break;
  }
  if(i != -1 && keypad_keys[i] != NULL) {
    /* first, echo the text */
    if(show_own > -1 && w->force_no_echo < 1) {
      s = g_malloc(strlen(keypad_keys[i]) + 15);
      switch(show_own) {
      case BLACK:   j = 30; break;
      case WHITE:   j = 37; break;
      case RED:     j = 31; break;
      case GREEN:   j = 32; break;
      case BLUE:    j = 34; break;
      case CYAN:    j = 36; break;
      case MAGENTA: j = 35; break;
      case YELLOW:  j = 33; break;
      }
      sprintf(s, "\e[%dm%s\e[0m\n", j, keypad_keys[i]);
      gM_world_handle_output_line(w, s, strlen(s));
      g_free(s);
    }

    gM_world_write(w, keypad_keys[i]);
    gM_world_write(w, "\n");
  }
}

void gM_world_clear_screen(world *w) {
  int i,num;

  if(w == NULL)
    return;

  num = GTK_WIDGET(w->output)->allocation.height / gdk_char_height(normalfont, '\n');
  for(i = 0; i < num; i++)
    gtk_text_insert(GTK_TEXT(w->output), normalfont, NULL, NULL, "\n", 1);
}

void gM_world_clear_buffer(world *w) {
  if(w == NULL)
    return;

  gtk_text_freeze(GTK_TEXT(w->output));
  gtk_editable_delete_text(GTK_EDITABLE(w->output),
                           0,
                           gtk_text_get_length(GTK_TEXT(w->output)));
  gtk_text_thaw(GTK_TEXT(w->output));
  w->insert_10 = 0;
}

void gM_world_send_selection(world *w) {
printf(_("This doens't work yet :(\n"));
}

void parse_print_trigger(world *w, char *line, int *i, int *napos, int len) {
  int j;
  char *s;
  GtkAdjustment *vadj = GTK_ADJUSTMENT(GTK_TEXT(w->output)->vadj);
  int trig = w->highlites[*napos];
  int flags = 0;
  GdkColor fg;
  GdkColor bg;
  gM_triggers_get_color(trig,
                        &fg.red, &fg.green, &fg.blue,
                        &bg.red, &bg.green, &bg.blue,
                        &flags);
  fg.pixel = bg.pixel = 0;

  s = g_malloc(w->highlitecount + 1);
  j = 0;
  
  for(; *i < len; (*i)++) {
    if(w->highlites[*napos] != trig)
      break;
    switch(line[*i]) {
    case -1:
      if(*i + 2 < len && line[*i + 2] == 1) {
        if(line[*i + 1] == -5) /* echo off */
          w->force_no_echo = 1;
        else if(line[*i + 1] == -4) /* echo back on */
          w->force_no_echo = 0;
        *i += 2;
      }
      break;
    case 7:
      if(beep_type == beep_type_speaker ||
         beep_type == beep_type_speaker + beep_type_visual)
        gdk_beep();
      if(beep_type == beep_type_visual ||
         beep_type == beep_type_speaker + beep_type_visual) {
        gtk_widget_hide(w->output);
        gtk_widget_show(w->output);
        gtk_widget_hide(w->output);
        gtk_widget_show(w->output);
        gtk_widget_hide(w->output);
        gtk_widget_show(w->output);
      }
      break;
/*    case 10: No char 10 _can_ appear here */
    case 27:
      if(j != 0) {
        s[j] = 0;
        add_10_if_needed(w);
        if(w->locked < 1 && vadj->value != vadj->upper - vadj->page_size) {
          gtk_text_freeze(GTK_TEXT(w->output));
          gtk_adjustment_set_value(vadj, vadj->upper - vadj->page_size);
          gtk_text_thaw(GTK_TEXT(w->output));
        }
        gtk_text_insert(GTK_TEXT(w->output),
                        w->curfont,
                        flags & T_F_FG ? &fg : w->fgcolor != DEFAULTFG ? &colors[w->fgcolor] : NULL,
                        flags & T_F_BG ? &bg : w->bgcolor != DEFAULTBG ? &colors[w->bgcolor] : NULL,
                        s, j);
        write_log(w, s);
      }
      j = 0;
      *i = parse(w, line, *i, len);
      break;
    default:
      s[j] = line[*i];
      j++;
      (*napos)++;
    }
  }
  if(j != 0) {
    s[j] = 0;
    add_10_if_needed(w);
    if(w->locked < 1 && vadj->value != vadj->upper - vadj->page_size) {
    	gtk_text_freeze(GTK_TEXT(w->output));
      gtk_adjustment_set_value(vadj, vadj->upper - vadj->page_size);
      gtk_text_thaw(GTK_TEXT(w->output));
    }
    gtk_text_insert(GTK_TEXT(w->output),
                    w->curfont,
                    flags & T_F_FG ? &fg : w->fgcolor != DEFAULTFG ? &colors[w->fgcolor] : NULL,
                    flags & T_F_BG ? &bg : w->bgcolor != DEFAULTBG ? &colors[w->bgcolor] : NULL,
                    s, j);
    write_log(w, s);
  }

  g_free(s);
}

void input_size_allocate(GtkWidget *input, GtkAllocation *a) {
  multilinesize = a->height;
}

