/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
 
#include "structs.h"
#include "help.h"
#include "world.h"
#include "dialog.h"
#include "handy.h"
#include "config.h"

void create_new_send_dialog(world *w);

void but_cb_clear(GtkWidget *but, world *w);
void but_cb_ok(GtkWidget *but, world *w);
void but_cb_cancel(GtkWidget *but, world *w);
void but_cb_help(GtkWidget *but, world *w);

void find_cb_find(GtkWidget *but, world *w);
void find_cb_close(GtkWidget *but, world *w);
void find_changed(GtkWidget *but, world *w);

void connect_ok_cb(GtkWidget *but, world *w);
void connect_cancel_cb(GtkWidget *but, world *w);
void enter_cb(GtkWidget *ent, world *w);

void gM_worlddialogs_send(world *w) {
  if(w == NULL)
    return;
  else if(w->sock < 1) {
    gM_world_print(w, _("\e[31m% Error, this world isn't connected, sending isn't possible\e[0m\n"), -1);
    return;
  }
  if(w->senddialog == NULL)
    create_new_send_dialog(w);

  if(savesend < 1) {
    gtk_entry_set_text (GTK_ENTRY (w->sendfirst), "@paste");
    gtk_entry_set_text (GTK_ENTRY (w->sendlast), ".");
  }

  gtk_editable_delete_text(GTK_EDITABLE(w->sendtext), 0,
                           gtk_text_get_length(GTK_TEXT(w->sendtext)));

  gtk_widget_show(w->senddialog);
}

void create_new_send_dialog(world *w) {
  GtkWidget *vbox;
  GtkWidget *vbox2;
  GtkWidget *hbox;
  GtkWidget *hbox2;
  GtkWidget *hbox3;
  GtkWidget *tmp;
  GtkWidget *hbb;
  char *title = g_malloc(strlen(w->name) + 21);
  sprintf(title, _("Send text to world %s"), w->name);

  w->senddialog = gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_widget_set_usize(w->senddialog, 320, 200);
  gtk_window_set_title(GTK_WINDOW(w->senddialog), title);
  gtk_window_set_policy(GTK_WINDOW(w->senddialog), FALSE, FALSE, FALSE);
  gtk_window_position(GTK_WINDOW(w->senddialog), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(w->senddialog), "delete_event", GTK_SIGNAL_FUNC(but_cb_cancel), w);
  
  g_free(title);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER(w->senddialog), vbox);

  vbox2 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox2);
  gtk_box_pack_start (GTK_BOX (vbox), vbox2, TRUE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (vbox2), 5);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox);
  gtk_box_pack_start (GTK_BOX (vbox2), hbox, FALSE, TRUE, 0);

  tmp = gtk_label_new (_("Prefix: "));
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (hbox), tmp, FALSE, TRUE, 0);
  gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);

  w->sendfirst = gtk_entry_new ();
  gtk_widget_show (w->sendfirst);
  gtk_box_pack_start (GTK_BOX (hbox), w->sendfirst, TRUE, TRUE, 0);
  gtk_entry_set_text (GTK_ENTRY (w->sendfirst), "@paste");

  hbox2 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox2);
  gtk_box_pack_start (GTK_BOX (vbox2), hbox2, TRUE, TRUE, 0);

  w->sendtext = gtk_text_new(NULL, NULL);
  gtk_widget_show (w->sendtext);
  gtk_box_pack_start (GTK_BOX (hbox2), w->sendtext, TRUE, TRUE, 0);
  gtk_text_set_editable (GTK_TEXT (w->sendtext), TRUE);    

  tmp = gtk_vscrollbar_new(GTK_TEXT(w->sendtext)->vadj);
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (hbox2), tmp, FALSE, TRUE, 0);

  hbox3 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox3);
  gtk_box_pack_start (GTK_BOX (vbox2), hbox3, FALSE, TRUE, 0);

  tmp = gtk_label_new (_("Suffix: "));
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (hbox3), tmp, FALSE, TRUE, 0);

  w->sendlast = gtk_entry_new ();
  gtk_widget_show (w->sendlast);
  gtk_box_pack_start (GTK_BOX (hbox3), w->sendlast, TRUE, TRUE, 0);
  gtk_entry_set_text (GTK_ENTRY (w->sendlast), ".");

  tmp = gtk_hseparator_new ();
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (vbox), tmp, FALSE, TRUE, 5);
                                                             

  hbb = gtk_hbutton_box_new ();
  gtk_widget_show (hbb);
  gtk_box_pack_start (GTK_BOX (vbox), hbb, FALSE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbb), 5);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbb), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbb), 10);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbb), 0, 0);

  tmp = gtk_button_new_with_label (_("Clear"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb_clear), w);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);

  tmp = gtk_button_new_with_label (_("Ok"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb_ok), w);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);

  tmp = gtk_button_new_with_label (_("Cancel"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb_cancel), w);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);
  
  tmp = gtk_button_new_with_label (_("Help"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(but_cb_help), w);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);
  gtk_widget_grab_focus (tmp);
  gtk_widget_grab_default (tmp);
  
  gtk_widget_grab_focus(w->sendtext);
}


void but_cb_clear(GtkWidget *but, world *w) {
  gtk_entry_set_text(GTK_ENTRY(w->sendfirst), "");
  gtk_editable_delete_text(GTK_EDITABLE(w->sendtext), 0, gtk_text_get_length(GTK_TEXT(w->sendtext)));
  gtk_entry_set_text(GTK_ENTRY(w->sendlast), "");
}

void but_cb_ok(GtkWidget *but, world *w) {
  char *s;
  int len;

  s = gtk_entry_get_text(GTK_ENTRY(w->sendfirst));
  if(s[0] != 0)
    gM_world_write(w, s);
  if(s[strlen(s) - 1] != 10)
    gM_world_write(w, "\n");

  len = gtk_text_get_length(GTK_TEXT(w->sendtext));
  s   = gtk_editable_get_chars(GTK_EDITABLE(w->sendtext), 0, len);
  if(len) {
    gM_world_write(w, s);
    if(s[len - 1] != 10)
      gM_world_write(w, "\n");
  }
  g_free(s);

  s = gtk_entry_get_text(GTK_ENTRY(w->sendlast));
  if(s[0] != 0)
    gM_world_write(w, s);
  if(s[strlen(s) - 1] != 10)
    gM_world_write(w, "\n");

  gtk_widget_hide(w->senddialog);
}

void but_cb_cancel(GtkWidget *but, world *w) {
  gtk_widget_hide(w->senddialog);
}

void but_cb_help(GtkWidget *but, world *w) {
  gM_help_show_send();
}


void gM_worlddialogs_find(world *w) {
  GtkWidget *vbox;
  GtkWidget *vbox2;
  GtkWidget *hbox;
  GtkWidget *hbb;
  GtkWidget *tmp;
  
  if(w == NULL || w->finddialog != NULL)
    return;

  w->finddialog = gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_window_set_title(GTK_WINDOW(w->finddialog), _("Find"));
  gtk_window_set_policy(GTK_WINDOW(w->finddialog), FALSE, FALSE, FALSE);
  gtk_window_position(GTK_WINDOW(w->finddialog), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(w->finddialog), "delete_event", GTK_SIGNAL_FUNC(find_cb_close), w);
  
  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER(w->finddialog), vbox);

  vbox2 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox2);
  gtk_box_pack_start (GTK_BOX (vbox), vbox2, TRUE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (vbox2), 5);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(vbox2), hbox, TRUE, TRUE, 0);

  tmp = gtk_label_new (_("Find text: "));
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (hbox), tmp, FALSE, TRUE, 0);
  gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);

  w->findentry = gtk_entry_new();
  gtk_widget_show(w->findentry);
  gtk_box_pack_start(GTK_BOX(hbox), w->findentry, TRUE, TRUE, 0);

  w->findcase = gtk_check_button_new_with_label(_("Case sensitive"));
  gtk_widget_show(w->findcase);
  gtk_box_pack_start(GTK_BOX(vbox2), w->findcase, TRUE, FALSE, 0);
  
  tmp = gtk_hseparator_new ();
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (vbox), tmp, FALSE, TRUE, 5);

  hbb = gtk_hbutton_box_new ();
  gtk_widget_show (hbb);
  gtk_box_pack_start (GTK_BOX (vbox), hbb, FALSE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbb), 5);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbb), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbb), 10);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbb), 0, 0);

  w->findok = gtk_button_new_with_label(_("Find"));
  gtk_widget_show(w->findok);
  gtk_container_add(GTK_CONTAINER(hbb), w->findok);
  gtk_signal_connect(GTK_OBJECT(w->findok), "clicked", GTK_SIGNAL_FUNC(find_cb_find), w);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);

  tmp = gtk_button_new_with_label(_("Close"));
  gtk_widget_show(tmp);
  gtk_container_add (GTK_CONTAINER(hbb), tmp);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);
  gtk_widget_grab_focus (tmp);
  gtk_widget_grab_default (tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(find_cb_close), w);
  
  gtk_signal_connect(GTK_OBJECT(w->findentry), "activate", GTK_SIGNAL_FUNC(find_cb_find), w);

  gtk_signal_connect(GTK_OBJECT(w->findcase),  "clicked", GTK_SIGNAL_FUNC(find_changed), w);
  gtk_signal_connect(GTK_OBJECT(w->findentry), "changed", GTK_SIGNAL_FUNC(find_changed), w);
  w->findpos = -1;

  gtk_widget_grab_focus(w->findentry);
  
  gtk_widget_show(w->finddialog);
}

void find_cb_close(GtkWidget *but, world *w) {
  gtk_widget_destroy(w->finddialog);
  w->finddialog = NULL;
  w->findpos = -1;
}

void find_cb_find(GtkWidget *but, world *w) {
  char *find;
  char *text;
  int len;
  int findlen;
  char *result;
  
  find = gtk_entry_get_text(GTK_ENTRY(w->findentry));
  findlen = strlen(find);
  if(findlen < 1)
    return;
  if(w->findpos == -2) {
/*    printf("ALWEER NIET gevonden\n");*/
    return;
  } else
    len = w->findpos < 0 ? gtk_text_get_length(GTK_TEXT(w->output)) : w->findpos;
  if(len < 1)
    return;
  text = gtk_editable_get_chars(GTK_EDITABLE(w->output), 0, len);
  while(1) {
    result = strrcasechr(text, find[0]);
    if(result == NULL)
      break;
  	if(GTK_TOGGLE_BUTTON(w->findcase)->active > 0) {
      if(strncmp(result, find, findlen) == 0)
        break;
      else
        result[0] = 0;
    } else {
      if(strncasecmp(result, find, findlen) == 0)
        break;
      else
        result[0] = 0;
    }

  }
  if(result == NULL) {
  	gM_dialog_show(_("Message"), _("Not found"), 1);
    w->findpos = -2;
    gtk_widget_set_sensitive(w->findok, FALSE);
  } else {
    text[(int)result - (int)text + findlen] = 0;
  gtk_editable_set_position(GTK_EDITABLE(w->output), (int)result - (int)text);

    gtk_editable_select_region(GTK_EDITABLE(w->output),
                               (int)result - (int)text,
                               (int)result - (int)text + findlen);
    w->findpos = (int)result - (int)text - 1;
  }
  g_free(text);
}

void find_changed(GtkWidget *but, world *w) {
  w->findpos = -1;
  gtk_widget_set_sensitive(w->findok, TRUE);
}

void gM_worlddialogs_connect(world *w) {
  GtkWidget *window;
  GtkWidget *vbox;
  GtkWidget *table;
  GtkWidget *hbb;
  GtkWidget *tmp;

  char *name;

  if(w == NULL || w->conname != NULL)
    return;

  name = strcatnew(_("Connect to world "), w->name);
  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (window), name);
  gtk_window_set_policy (GTK_WINDOW (window), TRUE, TRUE, FALSE);
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event", connect_cancel_cb, w);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  table = gtk_table_new (2, 2, FALSE);
  gtk_widget_show (table);
  gtk_box_pack_start (GTK_BOX (vbox), table, TRUE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (table), 5);

  tmp = gtk_label_new (_("Login name: "));
  gtk_widget_show (tmp);
  gtk_table_attach_defaults(GTK_TABLE (table), tmp, 0, 1, 0, 1);
  gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);

  tmp = gtk_label_new (_("Password: "));
  gtk_widget_show (tmp);
  gtk_table_attach_defaults(GTK_TABLE (table), tmp, 0, 1, 1, 2);
  gtk_misc_set_alignment (GTK_MISC (tmp), 1, 0.5);

  w->conname = gtk_entry_new ();
  gtk_widget_show (w->conname);
  gtk_table_attach_defaults(GTK_TABLE (table), w->conname, 1, 2, 0, 1);
  if(w->user != NULL)
    gtk_entry_set_text(GTK_ENTRY (w->conname), w->user);
  gtk_signal_connect(GTK_OBJECT(w->conname), "activate", GTK_SIGNAL_FUNC(enter_cb), w);

  w->conpasswd = gtk_entry_new ();
  gtk_widget_show (w->conpasswd);
  gtk_table_attach_defaults(GTK_TABLE (table), w->conpasswd, 1, 2, 1, 2);
  gtk_entry_set_visibility (GTK_ENTRY (w->conpasswd), FALSE);
  if(w->password != NULL)
    gtk_entry_set_text(GTK_ENTRY (w->conpasswd), w->password);
  gtk_signal_connect(GTK_OBJECT(w->conpasswd), "activate", GTK_SIGNAL_FUNC(enter_cb), w);

  tmp = gtk_hseparator_new ();
  gtk_widget_show (tmp);
  gtk_box_pack_start (GTK_BOX (vbox), tmp, FALSE, FALSE, 5);

  hbb = gtk_hbutton_box_new ();
  gtk_widget_show (hbb);
  gtk_box_pack_start (GTK_BOX (vbox), hbb, FALSE, TRUE, 0);
  gtk_container_border_width (GTK_CONTAINER (hbb), 5);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbb), GTK_BUTTONBOX_END);
  gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbb), 10);
  gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbb), 0, 0);

  tmp = gtk_button_new_with_label (_("Ok"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", connect_ok_cb, w);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);

  tmp = gtk_button_new_with_label (_("Cancel"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", connect_cancel_cb, w);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);

  tmp = gtk_button_new_with_label (_("Help"));
  gtk_widget_show (tmp);
  gtk_container_add (GTK_CONTAINER (hbb), tmp);
  gtk_signal_connect(GTK_OBJECT(tmp), "clicked", GTK_SIGNAL_FUNC(gM_help_show_connect), NULL);
  GTK_WIDGET_SET_FLAGS (tmp, GTK_CAN_DEFAULT);
  gtk_widget_grab_focus (tmp);
  gtk_widget_grab_default (tmp);

  gtk_widget_grab_focus(w->conname);

  gtk_widget_show(window);
}

void connect_ok_cb(GtkWidget *but, world *w) {
  if(w->user != NULL) g_free(w->user);
  w->user = strnew(gtk_entry_get_text(GTK_ENTRY(w->conname)));

  if(w->password != NULL) g_free(w->password);
  w->password = strnew(gtk_entry_get_text(GTK_ENTRY(w->conpasswd)));

  if(w->user[0] == 0) { g_free(w->user); w->user = NULL; }
  if(w->password[0] == 0) { g_free(w->password); w->password = NULL; }

  if(w->user != NULL) {
    connect_cancel_cb(but, w); /* destroy window and stuff */
    gM_world_write(w, "connect ");
    gM_world_write(w, w->user);
    if(w->password != NULL) {
      gM_world_write(w, " ");
      gM_world_write(w, w->password);
    }
    gM_world_write(w, "\n");
    gM_config_save_world(w); /* update the userstuff */
  } else { /* no username filled in ... */
    gM_dialog_show(_("Error:"), _("You have to fill in a username\n"), 1);
  }
}

void connect_cancel_cb(GtkWidget *but, world *w) {
  gtk_widget_destroy(gtk_widget_get_toplevel(but));
  w->conname = NULL;
  w->conpasswd = NULL;
}

void enter_cb(GtkWidget *ent, world *w) {
  if(ent == w->conname)
    gtk_widget_grab_focus(w->conpasswd);
  else 
    connect_ok_cb(ent, w);
}

