/*  gMOO
 *  Copyright (C) 1999  Gert Scholten
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>

#include "global_vars.h"
#include "structs.h"
#include "world.h"
#include "worlds.h"
#include "window.h"
#include "config.h"
#include "editor.h"
#include "dialog.h"
#include "script.h"
#include "handy.h"

void cleanup_lockfile();
      
void gM_worlds_init() {
  firstworld = NULL;
  clock_timeout = gtk_timeout_add(clocktype == clocksecond ? 1000 : 60000, (GtkFunction) clockfunc, NULL);
}

int gM_worlds_close_all() {
  world *w;
  
  if(firstworld == NULL)
    return 1;
  
  for(w = firstworld; w != NULL; w = w->next) {
    if(gM_worlds_remove_world(w) < 0)
      return -1;
/*    gM_editor_close(w);
    gM_world_close(w);*/
  }
  return 1;
}

void gM_worlds_add_world(char *worldname) {
  world *nw;
  world *w;
  int err = 0;

  nw = gM_world_new(worldname);
  if(nw == NULL) /* error */
    return;


    for(w = firstworld; w != NULL; w = w->next)
      if(strcmp(w->name, nw->name) == 0) {
        gM_dialog_show(_("Error:"), _("World has already been opened in this gMOO."), 1);
        err = 1;
        break;
      }
    if(err != 1 && gM_config_lock_check(nw->name) > -1) {
      if(gM_dialog_show_question(_("Warning:"),_("Warning ! This world has already been opened by\n"
                                            "an other gMOO... If reconnect on disconnect is\n"
                                            "enabled, a connect/disconnect loop will occure\n"
                                            "\n"
                                            "It is also possible that the lock file is\n"
                                            "corrupt, as the result of a kill of gMOO\n"
                                            "\n"
                                            "  Do you still want to open this world ?")
                                            , bYes | bNo) == bNo)
        err = 1;
      else if(gM_dialog_show_question(_("Question:"),_("Because you still want to open this word,\n"
                                                  "can I assume that this is the only gMOO\n"
                                                  "currently open, and that the lockfile is\n"
                                                  "corrupt?\n"
                                                  "If you select yes here, the lockfile will\n"
                                                  "be cleaned up, this will add all connected\n"
                                                  "worlds in this gMOO to the lockfile.\n"
                                                  "If this is not the only open gMOO or this\n"
                                                  "world is also open in another gMOO, many\n"
                                                  "wierd things may happen ...\n"
                                                  "\n"
                                                  "Continue ?")
                                                 , bYes | bNo) == bYes)
        cleanup_lockfile();
    }
    if(err == 1) {
      if(nw->name != NULL)     g_free(nw->name);
      if(nw->host != NULL)     g_free(nw->host);
      if(nw->user != NULL)     g_free(nw->user);
      if(nw->password != NULL) g_free(nw->password);
      g_free(nw);
      return;
    }

  if(firstworld == NULL)
    firstworld = nw;
  else {
    for(w = firstworld; w->next != NULL; w = w->next);
    w->next = nw;
    nw->prev = w;
  }
  gM_world_show(nw);
  gM_window_add_world(nw);
#ifdef TCL
  gM_script_init_for(nw);
#endif
  gM_world_connect(nw);
}

int gM_worlds_remove_world(world *w) {
  world *t;
  int i;
  
  if(w == NULL)
    return 1;
  t = w;

  for(t = firstworld; t != NULL; t = t->next) {
    if(t == w)
      break;
  }
  if(t == NULL) /* major error, sould not be happening */
    printf(_("gMOO: Error: World %s not fount in list of active worlds\n"), w->name);
 /* first close open editors */
  for(i = 0; i < t->editcount; i++)
    if(gM_editor_close(t->edit[i]) < 0)
      return -1;
  gtk_widget_hide(edittoolbarhandle);
      
  gtk_widget_set_sensitive(tb_prop, FALSE);
  gtk_widget_set_sensitive(tb_discon, FALSE);
  gtk_widget_set_sensitive(tb_close, FALSE);
  gtk_widget_set_sensitive(tb_send, FALSE);
  gtk_widget_set_sensitive(tb_userlist, FALSE);

  gM_world_close(t); /* disconnect and hide it */

  /* remove it from the chain */
  if(t->prev != NULL)
    t->prev->next = t->next;
  else
    firstworld = t->next;
  if(t->next != NULL)
    t->next->prev = t->prev;

  gtk_widget_destroy(w->main);
  if(w->senddialog != NULL)
    gtk_widget_destroy(w->senddialog);

  if(t->ansibuffer != NULL)
    g_free(t->ansibuffer);

  if(t->name != NULL)
    g_free(t->name);
  if(t->host != NULL)
    g_free(t->host);
  if(t->user != NULL)
    g_free(t->user);
  if(t->password != NULL)
    g_free(t->password);

  g_free(t);

  /* reorder the pages: */
  gM_worlds_reorder_pages();

  if(firstworld == NULL) {
    gtk_widget_hide(worldmenu);
  }
#ifdef TCL
  if(w->interp != NULL) {
    Tcl_Release(w->interp);
    Tcl_DeleteInterp(w->interp);
  }
  if(w->tcl_name != NULL)
    g_free(w->tcl_name);
#endif
  return 1;
}


world *gM_worlds_get_by_main(GtkWidget *main) {
  world *t;

  if(firstworld == NULL)
    return NULL;
    
  for(t = firstworld; t != NULL; t = t->next)
    if(t->main == main)
      return t;
  return NULL;
}

world *gM_worlds_get_world_by_editor_main(GtkWidget *main, int *num) {
  world *w;

  if(firstworld == NULL)
    return NULL;

  for(w = firstworld; w != NULL; w = w->next) {
/*printf("Checking world named \"%s\", it has %d editors\n", w->name, w->editcount);*/
    for(*num = 0; *num < w->editcount; (*num)++) {
/*printf("  Checking editor that is editing: \"%s\" (%d)\n", w->edit[*num]->verbname, *num);*/
      if(w->edit[*num]->main == main) {
/*printf("Editor found, returning\n");*/
        return w;
      }
    }
  }
/*printf("Editor NOT found !\n");*/
  return NULL;
}

void gM_worlds_reorder_pages() {
  world *w;
  int i;

  for(w = firstworld; w != NULL; w = w->next) {
    w->page_num = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), w->main);
    for(i = 0; i < w->editcount; i++) {
      w->edit[i]->page_num = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), w->edit[i]->main);
    }
  }
}



void gM_worlds_update_clock() {
  if(clock_timeout > -1)
    gtk_timeout_remove(clock_timeout);
  clock_timeout = gtk_timeout_add(clocktype == clocksecond ? 1000 : 60000, (GtkFunction) clockfunc, NULL);
  clockfunc(NULL);
}

void gM_worlds_update_default_color() {
  world *w;
  
  for(w = firstworld; w != NULL; w = w->next)
    set_global_color(w->output);
}

void cleanup_lockfile() {
  world *w;
  
  remove(lock_file_name);
  gM_config_lock_open();
  for(w = firstworld; w != NULL; w = w->next)
    gM_config_lock_add(w->name);
}

#ifdef TCL
world *gM_worlds_get_by_interp(Tcl_Interp *interp) {
  world *w;
  
  for(w = firstworld; w != NULL; w = w->next) {
    if(w->interp == interp)
      break;
  }
  return w;
}

world *gM_worlds_get_by_tcl_id(char *id) {
  world *w;
  
  for(w = firstworld; w != NULL; w = w->next) {
    if(strcmp(w->tcl_name, id) == 0)
      break;
  }
  return w;
}
#endif

int clockfunc(void *pointer) {
  time_t tim;
  char str[15];
  world *w = gM_window_get_current_world();
  int h,m,s;

  if(w == NULL || clocktype == 0) { /* No clock */
    gtk_widget_hide(clockframe);
  } else {
    if(w->sock > -1) {
      gtk_widget_show(clockframe);
      tim = time(NULL) - w->connect_time;

      h = (int) (tim % (216000)) / (3600);
      m = (int) (tim % (3600)) / 60;
      s = (int) tim % 60;
      sprintf(str, "%s%d:%s%d:%s%d",
                   h > 9 ? "" : "0", h,
                   m > 9 ? "" : "0", m,
                   s > 9 ? "" : "0", s);

      gtk_label_set_text(GTK_LABEL(clockbar), str);
    } else {
      gtk_widget_hide(clockframe);
    }
  }
  return TRUE;
}

void gM_worlds_set_wordwrap(int on) {
  world *w;

  for(w = firstworld; w != NULL; w = w->next) {
    gtk_text_set_word_wrap(GTK_TEXT(w->output), on);
  }
}

void gM_worlds_set_input_color() {
  world *w;

  for(w = firstworld; w != NULL; w = w->next) {
    if(definputstyle == 0)
      set_global_color(w->input);
    else
      gtk_widget_set_style(w->input, gtk_style_copy(gtk_widget_get_style(main_window)));
  }
}

void gM_worlds_show_sep() {
  world *w;

  for(w = firstworld; w != NULL; w = w->next) {
    if(w->seperator != NULL) {
      if(showsep > 0)
        gtk_widget_show(w->seperator);
      else
        gtk_widget_hide(w->seperator);
    }
  }
}


void gM_worlds_set_tab_size() {
  world *w;
  int i;
  char *name = g_malloc(21);

  for(w = firstworld; w != NULL; w = w->next) {
    limit_to_20(FOR_WORLD, w->name, &name);
    gtk_label_set_text(GTK_LABEL(w->label),
      short_tab_names > 0 ? name : w->name);
    for(i = 0; i < w->editcount; i++) {
  		if(w->edit[i]->verbname != NULL) { /* none MCP editor */
        limit_to_20(FOR_EDITOR, w->edit[i]->verbname, &name);
        gtk_label_set_text(GTK_LABEL(w->edit[i]->label),
          short_tab_names > 0 ? name : w->edit[i]->verbname);
      } else { /* MCP editor */
        limit_to_20(FOR_EDITOR, w->edit[i]->simpleedit->name, &name);
        gtk_label_set_text(GTK_LABEL(w->edit[i]->label),
          short_tab_names > 0 ? name : w->edit[i]->simpleedit->name);
      }
    }
  }
  g_free(name);
}


